<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/lib/alertmanager.class.php - manage alerts for changes in areas/nodes
 *
 * This file defines a class for managing alerts (add, edit, delete, view)
 * and underlying rules (add, delete).
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wascore
 * @version $Id: alertmanager.class.php,v 1.1 2016/04/26 09:33:19 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }

define('ALERTMANAGER_CHORE_VIEW',       'view');
define('ALERTMANAGER_CHORE_ADD',        'add');
define('ALERTMANAGER_CHORE_EDIT',       'edit');
define('ALERTMANAGER_CHORE_DELETE',     'delete');
define('ALERTMANAGER_CHORE_SAVE',       'save');
define('ALERTMANAGER_CHORE_RULE_ADD',   'addrule');
define('ALERTMANAGER_CHORE_RULE_EDIT',  'editrule');
define('ALERTMANAGER_CHORE_RULE_DELETE','deleterule');
define('ALERTMANAGER_CHORE_RULE_SAVE',  'saverule');

/** Methods to access properties of an alert
 *
 * This class is used to manage alerts. The following functions are supplied
 *
 *  - add a new alert
 *  - edit properties of an exising alert
 *  - delete an alert
 *  - view list of currently existing alerts
 *  - add a rule to an alert
 *  - edit a rule
 *  - delete a rule from an alert
 *
 * The default action is to show a list of existing alerts.
 *
 */
class AlertManager {
    /** @var object|null $output collects the html output */
    var $output = NULL;

    /** @var bool $show_parent_menu if TRUE caller is allowed to use menu area (show config mgr menu) */
    var $show_parent_menu = FALSE;

    /** @var array $intervals quick translation between cron interval and human readable values */
    var $intervals = array();


    /** construct an AlertManager object
     *
     * This initialises the AlertManager.
     *
     * @param object &$output collects the html output
     * @uses $CFG
     */
    function AlertManager(&$output) {
        global $CFG;
        $this->output = &$output;
        $this->output->set_helptopic('alerts');
        $this->intervals = array(60 => t('alerts_hourly','admin'),
                                 120 => t('alerts_2hours','admin'),
                                 240 => t('alerts_4hours','admin'),
                                 480 => t('alerts_8hours','admin'),
                                 960 => t('alerts_16hours','admin'),
                                 1440 => t('alerts_daily','admin'),
                                 2880 => t('alerts_2days','admin'),
                                 5760 => t('alerts_4days','admin'),
                                 10080 => t('alerts_weekly','admin')
                                 );
    } // AlertManager()


    /** entry point
     *
     * @return void output written to browser via $output
     */
    function run() {
        $this->show_parent_menu = TRUE;
        $chore = get_parameter_string('chore',ALERTMANAGER_CHORE_VIEW);
        switch($chore) {
        case ALERTMANAGER_CHORE_VIEW:
            $this->alerts_overview();
            break;

        case ALERTMANAGER_CHORE_ADD:
            $this->alert_edit(0);
            break;

        case ALERTMANAGER_CHORE_EDIT:
            $alert_id = get_parameter_int('alert',0);
            $this->alert_edit($alert_id);
            break;

        case ALERTMANAGER_CHORE_DELETE:
            $alert_id = get_parameter_int('alert',0);
            $this->alert_delete($alert_id);
            break;

        case ALERTMANAGER_CHORE_SAVE:
            $alert_id = get_parameter_int('alert',0);
            $this->alert_save($alert_id);
            break;

        case ALERTMANAGER_CHORE_RULE_ADD:
            $alert_id = get_parameter_int('alert',0);
            $this->alert_rule_edit($alert_id,0);
            break;

        case ALERTMANAGER_CHORE_RULE_EDIT:
            $alert_id = get_parameter_int('alert',0);
            $rule_id = get_parameter_int('rule',0);
            $this->alert_rule_edit($alert_id,$rule_id);
            break;

        case ALERTMANAGER_CHORE_RULE_DELETE:
            $alert_id = get_parameter_int('alert',0);
            $rule_id = get_parameter_int('rule',0);
            $this->alert_rule_delete($alert_id,$rule_id);
            break;

        case ALERTMANAGER_CHORE_RULE_SAVE:
            $alert_id = get_parameter_int('alert',0);
            $rule_id = get_parameter_int('rule',0);
            $this->alert_rule_save($alert_id,$rule_id);
            break;

        default:
            $s = (utf8_strlen($chore) <= 50) ? $chore : utf8_substr($chore,0,44).' (...)';
            $message = t('chore_unknown','admin',array('{CHORE}' => htmlspecialchars($s)));
            $this->output->add_message($message);
            logger('alertmanager: unknown chore: '.htmlspecialchars($s));
            $this->alerts_overview();
            break;
        }
    } // run()


    /** allow the caller to use the menu area (or not)
     *
     * this routine tells the caller if it is OK to use
     * the menu area (TRUE returned) or not (FALSE returned).
     (
     * @return bool TRUE if menu area is available, FALSE otherwise
     */
    function show_parent_menu() {
        return $this->show_parent_menu;
    } // show_parent_menu()


    // ==================================================================
    // ============================= ALERTS =============================
    // ==================================================================


    /** show link to 'add an alert' followed by a list of existing alerts
     *
     * this constructs an overview dialogue with options add, edit and
     * delete alerts. Records are ordered by 'is_active' and 'full name'.
     *
     * @return void output via $this->output
     */
    function alerts_overview() {
        global $WAS_SCRIPT_NAME;

        // 0 -- initialise
        $this->output->add_content('<h2>'.t('menu_alerts','admin').'</h2>');
        $this->output->add_content('<ul>');

        // 1 -- show link to 'add an alert'
        $this->output->add_content('  <li class="list">');
        if (!$this->output->text_only) {
            $this->output->add_content('    '.$this->output->skin->get_icon('blank'));
        } // else don't clutter text interface
        $attr = array('title'=> t('alerts_alert_add_title','admin'));
        $a_param = $this->a_param(ALERTMANAGER_CHORE_ADD);
        $anchor = t('alerts_alert_add','admin');
        $this->output->add_content('    '.html_a($WAS_SCRIPT_NAME,$a_param,$attr,$anchor));

        // 2 -- step through existing alerts if any
        $table = 'alerts';
        $fields = '*';
        $order = array('CASE WHEN (is_active) THEN 0 ELSE 1 END','full_name','alert_id');
        if (($records = db_select_all_records($table,$fields,'',$order,'alert_id')) !== FALSE) {
            foreach($records as $alert_id => $alert) {
                $this->output->add_content('  <li class="list">');
                $t = intval($alert['cron_interval']);
                $interval = (isset($this->intervals[$t])) ? $this->intervals[$t] : strval($t);
                $params = array('{ALERT_ID}' => $alert_id,
                                '{FULL_NAME}' => htmlspecialchars($alert['full_name']),
                                '{EMAIL}' => htmlspecialchars($alert['email']),
                                '{INTERVAL}' => $interval);

                // Delete icon
                $title = t('icon_alert_delete','admin');
                $alt = t('icon_alert_delete_alt','admin');
                $text =  t('icon_alert_delete_text','admin');
                $anchor = $this->output->skin->get_icon('delete', $title, $alt, $text);
                $a_params = $this->a_param(ALERTMANAGER_CHORE_DELETE,$alert_id);
                $attr = array('title' => $title);
                $this->output->add_content('    '.html_a($WAS_SCRIPT_NAME,$a_params,$attr,$anchor));

                // Edit link                                           
                $a_param = $this->a_param(ALERTMANAGER_CHORE_EDIT,$alert_id);
                $attr = array('title' => t('alerts_alert_title','admin'));
                $anchor = t('alerts_alert','admin',$params);
                $active = (db_bool_is(TRUE,$alert['is_active'])) ? '' : ' ('.t('inactive','admin').')';
                $this->output->add_content('    '.html_a($WAS_SCRIPT_NAME,$a_param,$attr,$anchor).$active);
                $this->output->add_content('  </li>');
            }
        }
        $this->output->add_content('</ul>');
    } // alerts_overview()


    /** display dialogue for an alert (new or existing)
     *
     * if $alert_id is 0 {@link show_alert()} presents the
     * add alert dialogue, otherwise the edit dialogue.
     *
     * @return void dialogue written via $output
     */
    function alert_edit($alert_id=0) {
        $dialogdef = $this->get_dialogdef_alert($alert_id);
        $this->show_alert($dialogdef,$alert_id);
    } // alert_edit()


    /** show a dialog to add a new or edit an existing alert
     *
     * if $alert_id is 0 we show the add new alert dialogue,
     * with just three fields: Name, E-mail and Frequency (is_active
     * is set to TRUE upon saving data). There are only two buttons:
     * [Save] or [Cancel]. The former takes the user to the edit
     * dialogue after adding the initial alert, the latter takes the user
     * back to the overview.
     *
     * If $alert_id is non-0, we show the same dialogue but this
     * time including is_active flag. This time there is also a [Done]
     * button: this saves the changes and returns to the overview.
     * [Save] stores the data but restarts the edit.
     *
     * Also, if we are not adding but editing the current list of
     * alert rules is displayed in a table, beneath the Save/Done/Cancel
     * buttons. There is also an Add rule link.
     * 
     * @param in $alert_id indicates which alert to show, 0 means add new dialogue
     * @return void output via $output
     */
    function show_alert(&$dialogdef,$alert_id=0) {
        global $WAS_SCRIPT_NAME,$DB;
        // 1 -- dialogue for alert properties
        if ($alert_id <= 0) {
            $this->output->add_content('<h2>'.t('alerts_alert_add','admin').'</h2>');
        } else {
            $this->output->add_content('<h2>'.t('alerts_alert_edit','admin').'</h2>');
        }
        $href = href($WAS_SCRIPT_NAME,$this->a_param(ALERTMANAGER_CHORE_SAVE,$alert_id));
        $this->output->add_content(dialog_quickform($href,$dialogdef));
        if ($alert_id <= 0) {
            return;
        }
        // 2 -- list of existing alert rules (only when editing existing alert)

        // 2A -- Prepare an 'add rule' link
        $a_attr = array('title'=> t('alerts_rule_add_title','admin'));
        $a_param = $this->a_param(ALERTMANAGER_CHORE_RULE_ADD,$alert_id);
        $anchor = t('alerts_rule_add','admin');
        $add_rule = html_a($WAS_SCRIPT_NAME,$a_param,$a_attr,$anchor);

        // 2B -- retrieve a list of rules associated with this alert
        $sql = sprintf('SELECT an.alert_area_node_id, an.alert_id, '.
                              'an.area_id, a.title area_title, '.
                              'an.node_id, n.title node_title, n.link_text node_link_text, '.
                              'an.flag, '.
                              'n.is_page, n.area_id node_area_id '.
                       'FROM %salerts_areas_nodes an '.
                       'LEFT JOIN %sareas a USING (area_id) '.
                       'LEFT JOIN %snodes n ON an.node_id = n.node_id '.
                       'WHERE alert_id = %d '.
                       'ORDER BY an.area_id, an.node_id ',
                       $DB->prefix,$DB->prefix,$DB->prefix,$alert_id);
        if (($DBResult = $DB->query($sql)) === FALSE) {
            logger(sprintf('%s(): cannot retrieve alert rules: %s',__FUNCTION__,db_errormessage()));
        } else {
            if ($DBResult->num_rows > 0) {
                $this->output->add_content('<h3>'.t('alerts_rules','admin').'</h3>');
                //$this->output->add_content('<p>'.$add_rule.'</p>');
                $class = 'header';
                $this->output->add_content('<p>');
                $this->output->add_content(html_table());
                $this->output->add_content('  '.html_table_row(array('class' => $class)));
                $this->output->add_content('    '.html_table_head(NULL,t('alerts_rules_nr','admin')));
                $this->output->add_content('    '.html_table_head(NULL,t('alerts_rules_delete','admin')));
                $this->output->add_content('    '.html_table_head(NULL,t('alerts_rules_area','admin')));
                $this->output->add_content('    '.html_table_head(NULL,t('alerts_rules_node','admin')));
                $this->output->add_content('  '.html_table_row_close());
                $index = 0;
                while (($record = $DBResult->fetch_row_assoc()) !== FALSE) {
                    ++$index;
                    $class = ($class == 'odd') ? 'even' : 'odd';
                    $rule_id = intval($record['alert_area_node_id']);
                    $area_id = intval($record['area_id']);
                    $node_id = intval($record['node_id']);
                    $node_area_id = intval($record['node_area_id']);
                    
                    if ($area_id <= 0) {
                        $area = t('alerts_rule_any_area','admin');
                    } else {
                        $area = htmlspecialchars($record['area_title']);
                    }
                    if ($node_id <= 0) {
                        $node = t('alerts_rule_any_node','admin');
                    } elseif (!empty($record['node_link_text'])) {
                        $node = htmlspecialchars($record['node_link_text']);
                    } else {
                        $node = htmlspecialchars($record['node_title']);
                    }
                    $params = array('{INDEX}' => $index,
                                    '{NODE}' => $node_id,
                                    '{NODE_FULL_NAME}' => $node,
                                    '{AREA}' => $area_id,
                                    '{AREA_FULL_NAME}' => $area);
                    if (($area_id != 0) && ($node_id != 0) && ($area_id != $node_area_id)) {
                        $class .= ' error';
                        $this->output->add_message(t('alerts_rule_warning','admin',$params));
                    }
                    // Delete icon
                    $title = t('icon_alert_rule_delete','admin');
                    $alt = t('icon_alert_rule_delete_alt','admin');
                    $text =  t('icon_alert_rule_delete_text','admin');
                    $anchor = $this->output->skin->get_icon('delete', $title, $alt, $text);
                    $a_params = $this->a_param(ALERTMANAGER_CHORE_RULE_DELETE,$alert_id,$rule_id);
                    $a_attr = array('title' => $title);
                    $icon = html_a($WAS_SCRIPT_NAME,$a_params,$a_attr,$anchor);
                    // Edit link
                    $a_attr = array('title'=> t('alerts_rule_edit_title','admin'));
                    $a_param = $this->a_param(ALERTMANAGER_CHORE_RULE_EDIT,$alert_id,$rule_id);
                    $anchor = t('alerts_rule_add','admin');
                    $number = html_a($WAS_SCRIPT_NAME,$a_param,$a_attr,strval($index));
                    // Area & Node
                    if ($area_id > 0) { $area = t('alerts_rule_area','admin',$params); }
                    if ($node_id > 0) { $node = t('alerts_rule_node','admin',$params); }
                    
                    $attributes = array('class' => $class);
                    $this->output->add_content('  '.html_table_row($attributes));
                    $attributes['align'] = 'right';
                    $this->output->add_content('    '.html_table_cell($attributes,$number));
                    $attributes['align'] = 'center';
                    $this->output->add_content('    '.html_table_cell($attributes,$icon));
                    unset($attributes['align']);
                    $this->output->add_content('    '.html_table_cell($attributes,$area));
                    $this->output->add_content('    '.html_table_cell($attributes,$node));
                    $this->output->add_content('  '.html_table_row_close());
                }
                $this->output->add_content(html_table_close());
                $this->output->add_content('</p>');
            }
            $this->output->add_content('<p>'.$add_rule.'</p>');
            $DBResult->close();
        }
    } // show_alert()


    /** handle confirmation and actual delete of an alert
     *
     * this single routine handles
     *
     *  - the display of a confirmation message,
     *  - handling of the actual delete, and
     *  - handling of the bail out/cancelbutton
     *
     * @param int $alert_id
     * @return void output written via $output
     */
    function alert_delete($alert_id) {
        global $WAS_SCRIPT_NAME,$DB;

        // 0 -- initialise
        $dialogdef = array(
            'csrftoken' => dialog_csrftoken(),
            'button_delete' => dialog_buttondef(BUTTON_DELETE),
            'buton_cancel' => dialog_buttondef(BUTTON_CANCEL)
            );

        // 1 -- bail out?
        if (isset($_POST['button_cancel'])) {
            $this->output->add_message(t('cancelled','admin'));
            $this->alerts_overview();
            return;
        }

        // 2 -- already in confirmation phase?
        if (isset($_POST['button_delete'])) {
            if (!dialog_validate($dialogdef)) {
                foreach($dialogdef as $k => $item) {
                    if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                        $this->output->add_message($item['error_messages']);
                    }
                }
            } else {
                $table = 'alerts_areas_nodes';
                $where = array('alert_id' => intval($alert_id));
                if (($retval = db_delete($table,$where)) === FALSE) {
                    logger(sprintf('%s(): errors deleting alert rule: %s',__FUNCTION__,db_errormessage()));
                } else {
                    $table = 'alerts';
                    if (($retval = db_delete($table,$where)) === FALSE) {
                        logger(sprintf('%s(): errors deleting alert: %s',__FUNCTION__,db_errormessage()));
                    }
                }
                if ($retval) {
                    $this->output->add_message(t('success_saving_data','admin'));
                } else { 
                    $this->output->add_message(t('errors_saving_data','admin',array('{ERRORS}'=>'1')));
                }
            }
            $this->alerts_overview();
            return;
        }

        // 3 -- show existing data in confirmation dialogue
        $alert = '';
        $table = 'alerts';
        $fields = '*';
        $where = array('alert_id' => intval($alert_id));
        if (($record = db_select_single_record($table,$fields,$where)) !== FALSE) {
            $t = intval($record['cron_interval']);
            $interval = (isset($this->intervals[$t])) ? $this->intervals[$t] : strval($t);
            $params = array('{ALERT_ID}' => $alert_id,
                            '{FULL_NAME}' => htmlspecialchars($record['full_name']),
                            '{EMAIL}' => htmlspecialchars($record['email']),
                            '{INTERVAL}' => $interval);
            $alert = t('alerts_alert','admin',$params);
        }
        $this->output->add_content('<h2>'.t('alerts_alert_delete','admin').'</h2>');
        $this->output->add_content('<p>'.t('alerts_alert_delete_confirm','admin').'</p>');
        $this->output->add_content('<p><strong>'.$alert.'</strong></p>');
        $this->output->add_content('<p>'.t('delete_are_you_sure','admin').'</p>');
        $href = href($WAS_SCRIPT_NAME,$this->a_param(ALERTMANAGER_CHORE_DELETE,$alert_id));
        $this->output->add_content(dialog_quickform($href,$dialogdef));
    } // alert_delete()


    /** store a new or modified alert in the database
     *
     * if the user pressed cancel (or somehow no button was pressed)
     * she is returned to the alerts overview.
     * after that the data is validated. On error the user is returned
     * to the same dialogue to correct errors.
     * after adding or updating the alert the user is taken to either
     *
     *  - the alerts overview (after pressing [Done] 
     *  - the add rule dialogue (in case of a new alert), or
     *  - the alert edit dialogue
     *
     * This attempts to seduce the user to actually add at least one rule to
     * the alert.
     *
     * @param int $alert_id
     * @return void data saved and dialogue displayed via $output
     */
    function alert_save($alert_id) {
        global $WAS_SCRIPT_NAME;

        // 1 -- cancelled?
        if ((isset($_POST['button_cancel'])) || 
            ((!isset($_POST['button_save']))  && (!isset($_POST['button_done'])))) {
            $this->output->add_message(t('cancelled','admin'));
            $this->alerts_overview();
            return;
        }

        // 2 -- errors?
        $dialogdef = $this->get_dialogdef_alert($alert_id);
        if (!$this->dialog_validate_alert($dialogdef)) {
            foreach($dialogdef as $k => $item) {
                if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                    $this->output->add_message($item['error_messages']);
                }
            }
            $this->show_parent_menu = FALSE; // let user focus on errors in dialog
            return $this->show_alert($dialogdef,$alert_id);
        }

        // 3 -- no validation errors, go save data
        $add_new = ($alert_id <= 0);
        $table = 'alerts';                        
        $t = intval($dialogdef['cron_interval']['value']);
        if ($add_new) {
            // 3A -- add new
            $now = strftime("%Y-%m-%d %T");
            $interval = (isset($this->intervals[$t])) ? $this->intervals[$t] : strval($t);
            $params = array('{FULL_NAME}' => $dialogdef['full_name']['value'],
                            '{EMAIL}' => $dialogdef['email']['value'],
                            '{INTERVAL}' => $interval);
            $fields = array('full_name' => $dialogdef['full_name']['value'],
                            'email' => $dialogdef['email']['value'],
                            'cron_interval' => intval($dialogdef['cron_interval']['value']),
                            'cron_next' => $now,
                            'messages' => 1,
                            'message_buffer' => $now."\n".t('alerts_created','admin',$params)."\n",
                            'is_active' => TRUE
                            );
            if (($retval = db_insert_into_and_get_id($table,$fields,'alert_id')) !== FALSE) {
                $alert_id = intval($retval);
            }
        } else {
            // 3B -- update existing
            $fields = array('full_name' => $dialogdef['full_name']['value'],
                            'email' =>$dialogdef['email']['value'],
                            'cron_interval' => $t,
                            'is_active' => ($dialogdef['is_active']['value']) ? TRUE : FALSE
                            );
            $where = array('alert_id' => intval($alert_id));
            $retval = db_update($table,$fields,$where);
        }
        // 3C -- something wrong?
        if ($retval === FALSE) {
            logger(sprintf('%s(): error saving alert %d: %s',__FUNCTION__,$alert_id,db_errormessage()));
            $this->output->add_message(t('errors_saving_data','admin',array('{ERRORS}'=>'1')));
            $this->show_alert($dialogdef,$alert_id);
            return;
        }

        // 4 -- ready: 'done' -> overview, add alert -> add rule too, otherwise -> stay in dialogue
        $this->output->add_message(t('success_saving_data','admin'));
        if (isset($_POST['button_done'])) {
            $this->alerts_overview();
        } elseif ($add_new) {
            $this->alert_rule_edit($alert_id,0);
        } else {
            $this->alert_edit($alert_id);
        }
    } // alert_save()


    // ==================================================================
    // ============================== RULES =============================
    // ==================================================================


    /** display dialogue for a rule  (new or existing)
     *
     * if $rule_id is 0 {@link show_rule()} presents the
     * add rule dialogue, otherwise the edit rule dialogue.
     *
     * @param int $alert_id
     * @param int $rule_id
     * @return void dialogue written via $output
     */
    function alert_rule_edit($alert_id,$rule_id=0) {
        $dialogdef = $this->get_dialogdef_rule($alert_id,$rule_id);
        $this->show_rule($dialogdef,$alert_id,$rule_id);
    } // alert_rule_edit()


    /** show a dialogue where a rule can be edited
     *
     * we expect the caller to fill dialogdef with the
     * correct values etc. This allows for reusing this
     * routine to easily display errors,
     *
     * @param array &$dialogdef
     * @param int $alert_id
     * @param int $rule_id
     * @return void dialogue sent to user via $output
     */
    function show_rule(&$dialogdef,$alert_id,$rule_id=0) {
        global $WAS_SCRIPT_NAME,$DB;
        // 1 -- dialogue for alert properties
        if ($rule_id <= 0) {
            $this->output->add_content('<h2>'.t('alerts_rule_add','admin').'</h2>');
        } else {
            $this->output->add_content('<h2>'.t('alerts_rule_edit','admin').'</h2>');
        }
        $href = href($WAS_SCRIPT_NAME,$this->a_param(ALERTMANAGER_CHORE_RULE_SAVE,$alert_id,$rule_id));
        $this->output->add_content(dialog_quickform($href,$dialogdef));
    } // show_rule()


    /** handle confirmation and actual delete of a rule
     *
     * this single routine handles
     *
     *  - the display of a confirmation message,
     *  - handling of the actual delete, and
     *  - handling of the bail out/cancelbutton
     *
     * @param int $alert_id
     * @param int $rule_id
     * @return void output written via $output
     */
    function alert_rule_delete($alert_id,$rule_id) {
        global $WAS_SCRIPT_NAME,$DB;

        // 0 -- initialise
        $dialogdef = array(
            'csrftoken' => dialog_csrftoken(),
            'button_delete' => dialog_buttondef(BUTTON_DELETE),
            'buton_cancel' => dialog_buttondef(BUTTON_CANCEL)
            );

        // 1 -- bail out?
        if (isset($_POST['button_cancel'])) {
            $this->output->add_message(t('cancelled','admin'));
            $this->alert_edit($alert_id);
            return;
        }

        // 2 -- already in confirmation phase?
        if (isset($_POST['button_delete'])) {
            if (!dialog_validate($dialogdef)) {
                foreach($dialogdef as $k => $item) {
                    if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                        $this->output->add_message($item['error_messages']);
                    }
                }
            } else {
                $table = 'alerts_areas_nodes';
                $where = array('alert_id' => intval($alert_id),'alert_area_node_id' => intval($rule_id));
                if (db_delete($table,$where) === FALSE) {
                    $this->output->add_message(t('errors_saving_data','admin',array('{ERRORS}'=>'1')));
                    logger(sprintf('%s(): errors saving alert rule: %s',__FUNCTION__,db_errormessage()));
                } else {
                    $this->output->add_message(t('success_saving_data','admin'));
                }
            }
            $this->alert_edit($alert_id);
            return;
        }

        // 3 -- show confirmation dialogue
        $rule = '?';
        $sql = sprintf('SELECT an.area_id, a.title area_title, '.
                              'an.node_id, n.title node_title, n.link_text node_link_text, n.is_page '.
                       'FROM %salerts_areas_nodes an '.
                       'LEFT JOIN %sareas a USING (area_id) '.
                       'LEFT JOIN %snodes n ON an.node_id = n.node_id '.
                       'WHERE alert_id = %d AND alert_area_node_id = %d ',
                       $DB->prefix,$DB->prefix,$DB->prefix,$alert_id,$rule_id);
        if ((($DBResult = $DB->query($sql,1)) === FALSE) || 
            (($record = $DBResult->fetch_row_assoc()) === FALSE)) {
            logger(sprintf('%s(): cannot retrieve alert rule: %s',__FUNCTION__,db_errormessage()));
        } else {
            $DBResult->close();
            $area_id = intval($record['area_id']);
            $node_id = intval($record['node_id']);
            if ($node_id <= 0) {
                if ($area_id <= 0) {
                    $rule = t('alerts_rule_any_any','admin');
                } else {
                    $params = array('{AREA}' => strval($area_id),
                                    '{AREA_FULL_NAME}' => $record['area_title']);
                    $rule = t('alerts_rule_area_any','admin',$params);
                }
            } else {
                $link_text = $record['node_link_text'];
                if (empty($link_text)) {
                    $link_text = $record['node_title'];
                }
                $params = array('{AREA}' => strval($area_id),
                                '{AREA_FULL_NAME}' => $record['area_title'],
                                '{NODE}' => strval($node_id),
                                '{NODE_FULL_NAME}' => $link_text
                                );
                if (db_bool_is(TRUE,$record['is_page'])) {
                    $rule = t('alerts_rule_page','admin',$params);
                } else {
                    $rule = t('alerts_rule_section','admin',$params);
                }
            }
        }
        $this->output->add_content('<h2>'.t('alerts_rule_delete','admin').'</h2>');
        $this->output->add_content('<p>'.t('alerts_rule_delete_confirm','admin').'</p>');
        $this->output->add_content('<p><strong>'.$rule.'</strong></p>');
        $this->output->add_content('<p>'.t('delete_are_you_sure','admin').'</p>');
        $href = href($WAS_SCRIPT_NAME,$this->a_param(ALERTMANAGER_CHORE_RULE_DELETE,$alert_id,$rule_id));
        $this->output->add_content(dialog_quickform($href,$dialogdef));
    } // alert_rule_delete()


    /** store a new or modified alert rule in the database
     *
     * @param int $alert_id
     * @param int $rule_id
     * @return void data saved and dialogue displayed via $output
     */
    function alert_rule_save($alert_id,$rule_id) {
        global $WAS_SCRIPT_NAME;

        // 1 -- cancelled?
        if ((isset($_POST['button_cancel'])) || 
            ((!isset($_POST['button_save']))  && (!isset($_POST['button_done'])))) {
            $this->output->add_message(t('cancelled','admin'));
            $this->alert_edit($alert_id);
            return;
        }
        // 2 -- errors?
        $dialogdef = $this->get_dialogdef_rule($alert_id,$rule_id);
        if (!$this->dialog_validate_rule($dialogdef,$alert_id,$rule_id)) {
            foreach($dialogdef as $k => $item) {
                if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                    $this->output->add_message($item['error_messages']);
                }
            }
            $this->show_parent_menu = FALSE; // let user focus on errors in dialog
            return $this->show_rule($dialogdef,$alert_id,$rule_id);
        }

        // 3 -- no errors, actually save data
        list($area_id,$node_id) = explode(':',$dialogdef['area_node']['value']);
        $table = 'alerts_areas_nodes';
        $fields = array('alert_id' => intval($alert_id),
                        'area_id' => intval($area_id),
                        'node_id' => intval($node_id),
                        'flag' => TRUE);
        if ($rule_id <= 0) {
            $retval = db_insert_into($table,$fields);
        } else {
            $where = array('alert_area_node_id' => intval($rule_id));
            $retval = db_update($table,$fields,$where);
        }
        if ($retval === FALSE) {
            $this->output->add_message(t('errors_saving_data','admin',array('{ERRORS}'=>'1')));
            logger(sprintf('%s(): errors saving alert rule %d: %s',__FUNCTION__,$rule_id,db_errormessage()));
        } else {
            $this->output->add_message(t('success_saving_data','admin'));
        }
        // 4 -- return to alert dialogue
        $this->alert_edit($alert_id);
    } // alert_rule_save()


    // ==================================================================
    // =========================== WORKHORSES ===========================
    // ==================================================================


    /** shorthand for the anchor parameters that lead to the alert manager
     *
     * @param string $chore the next chore that could be done
     * @param int|null $alert_id the alert of interest or NULL if none
     * @param int|null $rule_id the alert rule of interest or NULL if none
     * @return array ready-to-use array with parameters for constructing a-tag
     */
    function a_param($chore,$alert_id=NULL,$rule_id=NULL) {
        $parameters = array('job' => JOB_CONFIGURATIONMANAGER,'task' => TASK_ALERTS,'chore' => $chore);
        if ((!is_null($alert_id)) && ($alert_id > 0)) {
            $parameters['alert'] = strval($alert_id);
        }
        if ((!is_null($rule_id)) && ($rule_id > 0)) {
            $parameters['rule'] = strval($rule_id);
        }
        return $parameters;
    } // a_param()


    /** construct a dialogue defintion for adding/editing selected alert properties
     *
     * @param int $alert_id
     * @return array with dialogue definition
     */
    function get_dialogdef_alert($alert_id=0) {
        $alert = array('full_name' => '',
                       'email' => '',
                       'cron_interval' => 1440,
                       'is_active' => SQL_TRUE);
        if ($alert_id > 0) {
            $table = 'alerts';
            $fields = array_keys($alert);
            $where = array('alert_id' => intval($alert_id));
            if (($record = db_select_single_record($table,$fields,$where)) !== FALSE) {
                $alert = $record;
            }
        }
        $dialogdef = array(
            'full_name' => array(
                'type' => F_ALPHANUMERIC,
                'name' => 'full_name',
                'minlength' => 1,
                'maxlength' => 255,
                'columns' => 50,
                'label' => t('alerts_full_name_label','admin'),
                'title' => t('alerts_full_name_title','admin'),
                'value' => $alert['full_name']
                ),
            'email' => array(
                'type' => F_ALPHANUMERIC,
                'name' => 'email',
                'minlength' => 3,
                'maxlength' => 255,
                'columns' => 50,
                'label' => t('alerts_email_label','admin'),
                'title' => t('alerts_email_title','admin'),
                'value' => $alert['email']
                ),
            'cron_interval' => array(
                'type' => F_LISTBOX,
                'name' => 'cron_interval',
                'options' => $this->intervals,
                'label' => t('alerts_cron_interval_label','admin'),
                'title' => t('alerts_cron_interval_title','admin'),
                'value' => $alert['cron_interval']
                ),
            'csrftoken' => dialog_csrftoken(),
            );
        if ($alert_id > 0) {
           $dialogdef['is_active'] = array(
                'type' => F_CHECKBOX,
                'name' => 'is_active',
                'options' => array(1 => t('alerts_is_active_check','admin')),
                'label' => t('alerts_is_active_label','admin'),
                'title' => t('alerts_is_active_title','admin'),
                'value' => (db_bool_is(TRUE,$alert['is_active'])) ? '1' : ''
                );
        }
        $dialogdef['button_save'] = dialog_buttondef(BUTTON_SAVE);
        if ($alert_id > 0) {
            $dialogdef['button_done'] = dialog_buttondef(BUTTON_DONE);
        }
        $dialogdef['button_cancel'] = dialog_buttondef(BUTTON_CANCEL);
        return $dialogdef;
    } // get_dialogdef_alert()


    /** construct a dialogdef for adding/editing a rule for alert $alert_id
     *
     * @param int $alert_id
     * @param int $rule_id
     * @return array with dialog definition
     */
    function get_dialogdef_rule($alert_id=0,$rule_id=0) {
        $rule = array('area_id' => 0,'node_id' => 0);
        if ($rule_id > 0) {
            $table = 'alerts_areas_nodes';
            $fields = array_keys($rule);
            $where = array('alert_id' => intval($alert_id), 'alert_area_node_id' => intval($rule_id));
            if (($record = db_select_single_record($table,$fields,$where)) !== FALSE) {
                $rule = $record;
            }
        }
        $options = array('0:0' => array('option' => t('alerts_rule_any_any','admin'),
                                        'class' => 'level0'));
        if (($areas = get_area_records()) !== FALSE) {
            foreach($areas as $area_id => $area) {
                $area_title = $area['title'];
                $params = array('{AREA}' => strval($area_id),
                                '{AREA_FULL_NAME}' => $area_title);
                $options[$area_id.':0'] = array('option' => t('alerts_rule_area_any','admin',$params),
                                                'class' => 'level0');
                $tree = tree_build($area_id);
                $this->tree_walk($tree,$options,$area_id,$area_title,$tree[0]['first_child_id']);
            }
        }
        $dialogdef = array(
            'area_node' => array(
                'type' => F_LISTBOX,
                'name' => 'area_node',
                'options' => $options,
                'label' => t('alerts_rule_label','admin'),
                'title' => t('alerts_rule_title','admin'),
                'value' => $rule['area_id'].':'.$rule['node_id']
                ),
            'csrftoken' => dialog_csrftoken(),
            'button_done' => dialog_buttondef(BUTTON_DONE),
            'button_cancel' => dialog_buttondef(BUTTON_CANCEL)
            );
        return $dialogdef;
    } // get_dialogdef_rule()


    /** walk the tree and add all nodes to an options array
     *
     * @param array &$tree contains all nodes in $area_id
     * @param array &$options receives nodes in tree order
     * @param int &$area_id 
     * @param string &$area_title
     * @param int $node_id start of the (sub)tree
     * @return void and nodes added to $options
     * @uses tree_walk()
     */
    function tree_walk(&$tree,&$options,&$area_id,&$area_title,$node_id) {
        static $level = 1;
        $class = 'level'.intval($level);
        $anode = array('{AREA}' => $area_id,
                       '{AREA_FULL_NAME}' => $area_title);
        while ($node_id > 0) {
            $is_page = $tree[$node_id]['is_page'];
            $link_text = $tree[$node_id]['record']['link_text'];
            $title = $tree[$node_id]['record']['title'];
            if (empty($link_text)) {
                $link_text = $title;
            }
            $anode['{NODE}'] = strval($node_id);
            $anode['{NODE_FULL_NAME}'] = $link_text;
            $option = t(($is_page) ? 'alerts_rule_page' : 'alerts_rule_section','admin',$anode);
            $options[$area_id.':'.$node_id] = array('option'=>$option, 'title'=>$title, 'class'=>$class);
            if ((!$is_page) && (($subtree_id = $tree[$node_id]['first_child_id']) > 0)) {
                if ($level >= MAXIMUM_ITERATIONS) { // silently ignore but log it
                    logger(__FUNCTION__.'(): too many levels in node '.$node_id,WLOG_DEBUG);
                } else {
                    ++$level;
                    $this->tree_walk($tree,$options,$area_id,$area_title,$subtree_id);
                    --$level;
                }
            }
            $node_id = $tree[$node_id]['next_sibling_id'];
        }
        return;
    }

    /** validate add/edit alert dialog
     *
     * the usual validation and an additional check on the email address
     *
     * @param array &$dialogdef
     * @return bool TRUE if all submitted values are considered valid, FALSE otherwise
     */
    function dialog_validate_alert(&$dialogdef) {
        $retval = dialog_validate($dialogdef);
        // massage email: no trailing/leading spaces and always lowercase
        $dialogdef['email']['value'] = trim(utf8_strtolower($dialogdef['email']['value']));
        if (strpos($dialogdef['email']['value'],'@') === FALSE) { // no @-sign in mail address
            $retval = FALSE;
            ++$dialogdef['email']['errors'];
            $fname = str_replace('~','',$dialogdef['email']['label']);
            $dialogdef['email']['error_messages'][] = t('validate_invalid','',array('{FIELD}' => $fname));
        }
        return $retval;
    } // dialog_validate_alert()


    /** validate add/edit rule dialog
     *
     * the usual validation and an additional check on existing rule
     *
     * @param array &$dialogdef
     * @param int $alert_id
     * @param in $rule_id
     * @return bool TRUE if all submitted values are considered valid, FALSE otherwise
     */
    function dialog_validate_rule(&$dialogdef,$alert_id,$rule_id) {
        $retval = dialog_validate($dialogdef);
        // this rule already exists?
        list($area_id,$node_id) = explode(':',$dialogdef['area_node']['value']);
        $alert_id = intval($alert_id);
        $area_id = intval($area_id);
        $node_id = intval($node_id);
        $table = 'alerts_areas_nodes';
        $fields = 'alert_area_node_id';
        $where = array('alert_id' => $alert_id, 'area_id' => $area_id, 'node_id' => $node_id);
        if (($record = db_select_single_record($table,$fields,$where)) !== FALSE) {
            // this combination already exist, is it really us?
            if ($record['alert_area_node_id'] != $rule_id) {
                ++$dialogdef['area_node']['errors'];
                $dialogdef['area_node']['error_messages'][] = t('alerts_rule_exists','admin');
                $retval = FALSE;
            }
        }
        return $retval;
    } // dialog_validate_rule()

} // AlertManager

?>