<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/lib/email.class.php - wrapper for sending mail
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wascore
 * @version $Id: email.class.php,v 1.11 2016/03/23 09:28:34 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }

/** Email implements a simple interface to send mail
 *
 * This class can be used to send mail from Website@School,
 * e.g. alerts, new passwords, feedback to the project (with
 * attached translations), etc.
 *
 * Typical use:
 * <pre>
 * require_once('email.class.php');
 * $mailer = new Email;
 * $mailer->set_mailto($email,$name);
 * $mailer->set_subject($subject);
 * $mailer->set_message($message);
 * $mailer->add_attachment($data,$name);
 * $mailer->send();
 * </pre>
 *
 */
class Email {
    /** @var array $mailfrom contains addr and name for From: (single address) */
    var $mailfrom = array();

    /** @var array $mailreplyto contains addr and name for Reply-To: (single address) */
    var $mailreplyto = array();

    /** @var array $mailto contains addr and name for To: (single address)*/
    var $mailto = array();

    /** @var array $mailcc contains an array of arrays containing addr+name for Cc: (array of addresses) */
    var $mailcc = array();

    /** @var string $subject contains the message subject */
    var $subject = '';

    /** @var array $headers associative array with field names and field values of additional headers */
    var $headers = array();

    /** @var array $messages array of arrays with message properties: body, mimetype, charset, encoding */
    var $messages = array();

    /** @var array $attachments array of arrays with attachment properties: body, name, mimetype, etc. */
    var $attachments = array();

    /** @var bool $related TRUE indicates multipart/related mail rather than multipart/mixed */
    var $related = FALSE;

    /** @var string $eol end of line character(s), usually CR + LF */
    var $eol = "\r\n";

    /** @var string $charset default character set to use in display names and subject */
    var $charset = "UTF-8";

    /** @var bool $minimal default flag limiting the literal representation in {@link rfc2047_qchar()} */
    var $minimal = FALSE;

    /** @var int $max_length limit for line length */
    var $max_length = 76;

    /** constructor resets all variables to a known (default) state
     *
     * @return void
     */
    function Email() {
        $this->reset_all();
    } // Email()

    /** reset all variables to their default values
     *
     * @return void and variables set
     */
    function reset_all() {
        global $CFG;
        $this->set_mailfrom($CFG->website_from_address,$CFG->title);
        if (!empty($CFG->website_replyto_address)) {
            $this->set_mailreplyto($CFG->website_replyto_address,$CFG->title);
        } else {
            $this->mailreplyto = array();
        }
        $this->mailto = array();
        $this->mailcc = array();
        $this->subject = '';
        $this->headers = array();
        $this->messages = array();
        $this->attachments = array();
        $this->related = FALSE;
    } // reset_all()

    /** store the subject of the mail message
     *
     * Embedded CRs and LFs in $subject are removed
     * and the result is stored until message send time.
     *
     * @param string $subject
     * @return void;
     */
    function set_subject($subject) {
        $forbidden = array("\r","\n");
        $this->subject = str_replace($forbidden,'',$subject);
    } // set_subject()

    /** record the address and the name for the From: header
     *
     * Embedded CRs LFs "<" and "> are removed from $addr,
     * and the result is stored, together with $name.
     *
     * @param string $addr the address eg. 'webmaster@example.com'
     * @param string $name the name, eg. 'Exemplum Primary School'
     * @return void
     */
    function set_mailfrom($addr,$name='') {
        $forbidden = array("\r","\n","<",">");
        $this->mailfrom = array(
            'addr' => str_replace($forbidden,'',$addr),
            'name' => $name);
    } // set_mailfrom()

    /** record the address and the name for the Reply-To: header
     *
     * Embedded CRs LFs "<" and "> are removed from $addr,
     * and the result is stored, together with $name.
     *
     * @param string $addr the address eg. 'info@example.com'
     * @param string $name the name, eg. 'Exemplum Primary School'
     * @return void
     */
    function set_mailreplyto($addr,$name='') {
        $forbidden = array("\r","\n","<",">");
        $this->mailreplyto = array(
            'addr' => str_replace($forbidden,'',$addr),
            'name' => $name);
    } // set_mailreplyto()

    /** record the address and the name for the To: header
     *
     * Embedded CRs LFs "<" and ">" are removed from $addr,
     * and the result is stored, together with $name.
     *
     * @param string $addr the address eg. 'hparkh@example.com'
     * @param string $name the name, eg. 'Helen Parkhurst'
     * @return void
     */
    function set_mailto($addr,$name='') {
        $forbidden = array("\r","\n","<",">");
        $this->mailto = array(
            'addr' => str_replace($forbidden,'',$addr),
            'name' => $name);
    } // set_mailto()

    /** add an address and name for the Cc: header
     *
     * Embedded CRs LFs "<" and "> are removed from $addr,
     * and the result is stored, together with $name.
     * Note that this function can be called multiple times,
     * where each call adds an address to the list.
     *
     * @param string $addr the address eg. 'acackl@example.com'
     * @param string $name the name, eg. 'Amelia Cackle'
     * @return void
     */
    function add_mailcc($addr,$name='') {
        $forbidden = array("\r","\n","<",">");
        $this->mailcc[] = array(
            'addr' => str_replace($forbidden,'',$addr),
            'name' => $name);
    } // add_mailcc()

    /** set the (primary) message
     *
     * This simply stores the message primary body until it can be sent.
     * Many times there is just a single message (perhaps with attachments),
     * but it is possible to add an alternative message via {@link add_message()}.
     * In that case the array $this->messages[] holds more than 1 variation of
     * the message. This routine always stores 'the' message, i.e. if there
     * was a set of messages before, those will be removed and an array with
     * a single message remains.
     *
     * Note that according to RFC1341 the sender should place body parts in
     * increasing order of preference, ie. first text/plain, then text/html and
     * finally application/pdf. However, it is up to the caller to call
     * {@link set_message()} and {@link add_message()} in the correct order.
     *
     * @param string $message content to send
     * @param string $mimetype type of the content, usually 'text/plain'
     * @param string $charset character set to use
     * @param string $encoding the desired encoding (defaults to quoted-printable because we expect text)
     * @return void
     */
    function set_message($message,$mimetype='text/plain',$charset='UTF-8',$encoding='quoted-printable') {
        $this->messages = array(
            array('body' => $message,
                  'mimetype' => $mimetype,
                  'charset' => $charset,
                  'encoding' => $encoding)
            );
    } // set_message()

    /** add an (alternative version of) message
     *
     * This simply stores an alternative message body until it can be 
     * combined (to a multipart/alternative) and subsequently sent.
     *
     * Note that according to RFC1341 the sender should place body parts in
     * increasing order of preference, ie. first text/plain, then text/html and
     * finally application/pdf. However, it is up to the caller to call
     * {@link set_message()} and {@link add_message()} in the correct order.
     *
     * @param string $message content to send
     * @param string $mimetype type of the content, usually 'text/plain'
     * @param string $charset character set to use
     * @param string $encoding the desired encoding (defaults to quoted-printable because we expect text)
     * @return void
     */
    function add_message($message,$mimetype='text/plain',$charset='UTF-8',$encoding='quoted-printable') {
        $this->messages[] = array('body' => $message,
                                  'mimetype' => $mimetype,
                                  'charset' => $charset,
                                  'encoding' => $encoding);
    } // add_message()

    /** add an attachment
     *
     * This simply adds an attachment with associated properties.
     * Multiple attachments can be added by calling this routine multiple times.
     *
     * The attachments added via this routine are simply added to the current list
     * of attachments in $this->attachments. There is also another routine (@link add_related()}
     * which takes care of attachments in a multipart/related context.
     * Both routines do add to the same array.
     *
     * @param string $attachment presumably 8bit data to attach to the message
     * @param string $name the suggested filename to use when receiving the attachment
     * @param string $mimetype type of the content, usually the generic 'application/octet-stream'
     * @param string $charset character set to use (only applicable when $mimetype indicates 'text')
     * @param string $encoding the desired encoding (defaults to base64 because we expect binary data)
     * @param string $description optional description of the attachment
     * @param string $disposition either 'inline' or 'attachment'
     * @return void
     */
    function add_attachment($attachment,$name,$mimetype='application/octet-stream',$charset='UTF-8',$encoding='base64',$description='',$disposition='attachment') {
        $this->attachments[] = array(
            'body' => $attachment,
            'name' => $name,
            'mimetype' => $mimetype,
            'charset' => $charset,
            'encoding' => $encoding,
            'description' => $description,
            'disposition' => $disposition,
            'content_id' => ''); // only set in a multipart/related context
    } // add_attachment()

    /** add a related attachment
     *
     * This adds a related attachment with associated properties.
     * Multiple related attachments can be added by calling this
     * routine multiple times.
     *
     * The attachments added via this routine are added to the current
     * list of attachments in $this->attachments, but they are handled
     * in a slightly different way once the mail body is constructed. 
     * There is another routine (@link add_attachment()} which takes care 
     * of 'plain' attachments (which usually leads to a multipart/mixed
     * message). Both routines add to the same array.
     *
     * Note 1: all related attachments are assigned a (hopefully globally)
     * unique Content-ID. This ID can be used to refer to this message part,
     * e.g. via HTML img-tags src="cid:$content_id". See also RFC2111.
     * The Content-ID adheres to the same rules as a Message-ID, hence
     * we use the same routine for both.
     *
     * Note 2: once a related attachment is added via this routine, the flag
     * $this->related is set tot TRUE, making sure we eventually generate a
     * multipart/related rather than a multipart/mixed message.
     *
     * @param string $attachment presumably 8bit data that relates to the message
     * @param string $name the suggested filename to use when receiving the attachment
     * @param string $mimetype type of the content, usually the generic 'application/octet-stream'
     * @param string $charset character set to use (only applicable when $mimetype indicates 'text')
     * @param string $encoding the desired encoding (defaults to base64 because we expect binary data)
     * @param string $description optional description of the attachment
     * @param string $disposition either 'inline' or 'attachment'
     * @return string the generated unique content-id
     * @uses rfc5322_message_id()
     */
    function add_related($attachment,$name,$mimetype='application/octet-stream',$charset='UTF-8',$encoding='base64',$description='',$disposition='inline') {
        $content_id = $this->rfc5322_message_id();
        $this->attachments[] = array(
            'body' => $attachment,
            'name' => $name,
            'mimetype' => $mimetype,
            'charset' => $charset,
            'encoding' => $encoding,
            'description' => $description,
            'disposition' => $disposition,
            'content_id' => $content_id);
        $this->related = TRUE;
        return $content_id;
    } // add_related()

    /** manually add a header to the mail message
     *
     * This adds a header to the message headers. Possible candidates are
     *  - 'Priority' with possible values (from RFC2156): "normal" | "non-urgent" | "urgent"
     *  - 'Importance' with possible values (from RFC2156): "low" | "normal" | "high"
     *
     * Note that the following headers may be overwritten in the course
     * of constructing the message to send (see {@link send()}:
     *  - To: (depending on how mail() handles the first parameter internally)
     *  - Subject: (depending on how mail() handles the second parameter internally)
     *  - From:
     *  - Reply-To:
     *  - Cc:
     *  - X-Mailer:
     *  - Message-ID:
     *  - MIME-Version:
     *  - Content-Type:
     *  - Content-Transfer-Encoding:
     *
     * It is possible to use tricks such as using a different capitalisation to defeat this.
     * (I said it was a simple class, didn't I?)
     *      
     * @param string $name is the name of the header field 
     * @param string $value is the contents of the header field
     * @return void
     * @todo should we bring the Capi-Tali-Sation of $name in line with the default capitalisation in the list above?
     */
    function set_header($name,$value='') {
        $this->headers[$name] = $value;
    } // set_header()

    /** send the message using the prepared information (To:, Subject:, the message and attachments etc.)
     *
     * this routine prepares the raw message + necessary headers and subsequently sends the message.
     *
     * @return bool TRUE on success, FALSE on failure (actually the return value of mail())
     * @uses is_7bit()
     * @uses rfc5322_address()
     * @uses rfc2047_qstring()
     * @uses rfc5322_message_id()
     */
    function send() {
        $body = '';
        $headers = array();
        $this->prepare_body($body,$headers);
        return $this->send_body($body,$headers);
    } // send()


    /** construct the full mail body and the necessary top-level mail header fields
     *
     * this routine examines the messages that were added to Email (in $this->messages):
     * that array could contain 0, 1 or more versions of the message to send. Also this
     * routine looks at the 0, 1 or more plain or related attachments that
     * were added to $this->attachments.
     *
     * If there are multiple entries in $this->messages, those are wrapped in
     * a multipart/alternative body part. If there is only a single message
     * it becomes one of the body parts in its own right. If there are any
     * attachments, both the message (which might be multipart/alternative)
     * and the attachments are combined using either multipart/mixed (there are
     * only plain attachments) or multipart/related (there was at least
     * one related attachment).
     *
     * The resulting mail body is returned to the caller via &$body. Hopefully
     * this saves some memory moving that possibly big string around. The
     * headers that need to go in the mail headers are returned via &$headers.
     * It is the responsabilty of the caller to make sure that these headers
     * end up in the correct place.
     *
     * @param string &$body receives the complete mail body
     * @param array &$headers receives the necessary mail header fields
     * @return bool TRUE on success
     */
    function prepare_body(&$body,&$headers) {
        $body = '';
        $headers = array('Message-ID' => '<'.$this->rfc5322_message_id().'>',
                         'MIME-Version' => '1.0',
                         'Content-Type' => 'text/plain; charset="US-ASCII"');
        $num_attachments = sizeof($this->attachments);
        $num_messages = sizeof($this->messages);
        $retval = TRUE; // assume success

        if ($num_attachments <= 0) {
            if ($num_messages > 0) {
                // at least a single message
                $retval = $this->encode_message($body,$headers,TRUE);
            }
        } elseif (($num_attachments == 1) & ($num_messages == 0))  {
            // no message and a single attachment
            $retval = $this->encode_part($this->attachments[0],$body,$headers,TRUE);
        } else {
            // multiple bodyparts (at least 2)
            $boundary = $this->boundary();
            $headers['Content-Type'] = sprintf('multipart/%s; boundary="%s"',
                                               ($this->related) ? 'related' : 'mixed', $boundary);
            $body .= sprintf('%s%s%s',$this->eol,t('mime_preamble'),$this->eol);
            if ($num_messages > 0) {
                $body .= sprintf('%s--%s%s',$this->eol,$boundary,$this->eol);
                if ($this->encode_message($body,$headers) == FALSE) {
                    $retval = FALSE;
                }
            }
            for ($i=0; $i<$num_attachments; ++$i) {
                $body .= sprintf('%s--%s%s',$this->eol,$boundary,$this->eol);
                if ($this->encode_part($this->attachments[$i],$body,$headers) == FALSE) {
                    $retval = FALSE;
                }
            }
            $body .= sprintf('%s--%s--%s',$this->eol,$boundary,$this->eol);
        }
        return $retval;
    } // prepare_body()
           

    /** encode the main message, optionally including 1 or more alternative versions
     *
     * this encodes the main message as a single body part. If
     * there are alternative versions of the message, all variants
     * are wrapped in a multipart/alternative body part.
     *
     * Depending on the $toplevel flag the associated headers are
     * either appended to the $body ($toplevel==FALSE) OR added to
     * the $headers array ($toplevel==TRUE).
     *
     * The actual encoding is done in {@link encode_part()}.
     *
     * @uses encode_part()
     * @param string &$body receives the email body
     * @param array &$headers receives necessary header fields (if $toplevel == TRUE)
     * @param bool $toplevel determines where the headers are stored: in $body or $headers
     * @return bool TRUE on success, FALSE otherhwise
     */
    function encode_message(&$body,&$headers,$toplevel=FALSE) {
        $retval = TRUE;
        $num_messages  = sizeof($this->messages);
        if ($num_messages == 1) {
            $retval = $this->encode_part($this->messages[0],$body,$headers,$toplevel);
        } else {
            $boundary = $this->boundary();
            $content_type = sprintf('multipart/alternative; boundary="%s"',$boundary);
            if ($toplevel) {
                $headers['Content-Type'] = $content_type;
            } else {
                $body .= sprintf('Content-Type: %s%s',$content_type,$this->eol);
            }
            for ($i=0; $i<$num_messages; ++$i) {
                $body .= sprintf('%s--%s%s',$this->eol,$boundary,$this->eol);
                if ($this->encode_part($this->messages[$i],$body,$headers) == FALSE) {
                    $retval = FALSE;
                }
            }
            $body .= sprintf('%s--%s--%s',$this->eol,$boundary,$this->eol);
        }
        return $retval;
    } // encode_message()


    /** encode an email body part with headers and all
     *
     * this encodes the message or attachment in $source with the correct
     * mime type and other headers. Depending on the $toplevel flag,  the
     * headers end up in the $body (as part of a multipart mime message) OR
     * in the $headers array (if the caller has decided that there is only
     * a single body part in this email message).
     *
     * The necessary headers are computed in a separate routine {@link encode_part_headers()}.
     *
     * @param array &$source holds either a message or an attachment + associated properties
     * @param string &$body receives the email body
     * @param array &$headers receives necessary header fields (if $toplevel == TRUE)
     * @param bool $toplevel determines where the headers are stored: in $body or $headers
     * @return bool TRUE on success, FALSE otherhwise
     * @uses encode_part_headers()
     */
    function encode_part(&$source,&$body,&$headers,$toplevel=FALSE) {
        $part_headers = $this->encode_part_headers($source);

        if ($toplevel) {
            $headers = array_merge($headers,$part_headers);
        } else {
            foreach($part_headers as $k => $v) {
                $body .= $k.': '.$v.$this->eol;
            }
            $body .= $this->eol;
        }
        switch ($part_headers['Content-Transfer-Encoding']) {
        case '7bit':
            $body .= str_replace("\n",$this->eol,
                                 str_replace(array("\r\n","\n\r","\r"),"\n",$source['body']));
            break;

        case 'quoted-printable':
            $body .= quoted_printable($source['body']);
            break;

        case 'base64':
            $body .= wordwrap(base64_encode($source['body']),$this->max_length,$this->eol,TRUE);
            break;
        }
        return TRUE;
    } // encode_part()


    /** construct necessary headers for a MIME body part
     *
     * this routine constructs the following headers:
     *
     *  - Content-Type
     *  - Content-Transfert-Encoding
     *  - Content-Description (optional)
     *  - Content-Disposition (optional)
     *  - Content-ID (optional)
     *
     * based on the information in $source. Basically this is copying
     * and combining the information of the elements in $source. However,
     * there is 1 exception: if $source appears to be a 7-bit ASCII
     * message with lines shorter than 78 characters, the characterset in
     * Content-Type type is overruled and set to US-ASCII and the Content-
     * Transfer-Encoding is set to 7bit, all in an attempt to keep simple
     * messages as readable as possible.
     *
     * @param array &$source points to a message or an attachment
     * @return array contains the calculated headers
     */
    function encode_part_headers(&$source) {
        $crlf = array("\r\n","\n\r","\r", "\n");

        // 1 -- Content-Type and Content-Transfer-Encoding
        $mimetype = $source['mimetype'];
        $encoding = (strcasecmp($source['encoding'],'quoted-printable')==0) ? 'quoted-printable':'base64';
        if (strncasecmp($mimetype,'text/',5) == 0) {
            if (($this->is_7bit($source['body'])) && (wordwrap($source['body'],78) == $source['body'])) {
                $charset = 'US-ASCII';
                $encoding = '7bit';
            } else {
                $charset = $source['charset'];
            }
            $mimetype .= sprintf('; charset="%s"',$charset);
        }
        if (isset($source['name'])) {
            $mimetype .= sprintf('; name="%s"',str_replace($crlf,' ', $source['name']));
        }
        $headers = array('Content-Type' => $mimetype, 'Content-Transfer-Encoding' => $encoding);

        // 2 -- Process remaining headers (if any) mostly for attachments
        if ((isset($source['description'])) && (!empty($source['description']))) {
            $headers['Content-Description'] = str_replace($crlf,' ',$source['description']);
        }
        if (isset($source['disposition'])) {
            $disposition = (strcasecmp($source['disposition'],'inline') == 0) ? 'inline' : 'attachment';
            if (isset($source['name'])) {
                $disposition .= sprintf('; filename="%s"',str_replace($crlf,' ', $source['name']));
            }
            $headers['Content-Disposition'] = $disposition;
        }
        if ((isset($source['content_id'])) && (!empty($source['content_id']))) {
            $headers['Content-ID'] = sprintf('<%s>',$source['content_id']);
        }
        return $headers;
    } // encode_part_headers()

    /** actually send a prepared mail body (with headers) to recipient(s)
     *
     * this routine constructs the necessary header fields ('To', 'Subject',
     * 'From', 'Reply-To', 'Cc', 'X-Mailer', etc. and subsequently sends
     * $message to the recipient(s).
     *
     * Note that there are several sources for headers: we start with $this->headers,
     * followed by the list of headers mentioned above and ending with the headers
     * in $message_headers. This means that $message_headers has the last word
     * in headers; existing headers in $headers will be overwritten with headers
     * present in $message_headers. As a rule at most the following headers
     * are set in $message_headers: 'MIME-Version', 'Content-Type',
     * 'Content-Transfer-Encoding' and 'Message-ID:'.
     *
     * @param string &$message contains the completely prepared message (including multiparts etc.)
     * @param array &$message_headers contains headers that need to be added the message's headers
     * @return bool TRUE on success, FALSE on failure
     */
    function send_body(&$message, &$message_headers) {
        $headers = $this->headers; // working copy of $this->headers; keep original 'as-is'

        // 1 - To:
        $remaining = $this->max_length - 4; // strlen('To: ') = 4
        // quirks in mail(): the (single) to: address MUST be on one line!
        $mailto = $this->rfc5322_address($this->mailto['addr'],$remaining,
                                         $this->mailto['name'],TRUE, // TRUE=use legacy: addr "(" name ")"
                                         $this->charset,$this->minimal,$this->max_length,''); // no CR,LF
        // 2 - Subject:
        $remaining = $this->max_length - 9; // strlen('Subject: ') = 9
        $subject = $this->rfc2047_qstring($this->subject,$remaining,
                                          $this->charset,$this->minimal,
                                          $this->max_length,$this->eol);

        // 3 - From:
        $remaining = $this->max_length - 6; // strlen('From: ') = 6
        $headers['From'] = $this->rfc5322_address($this->mailfrom['addr'],$remaining,
                                                  $this->mailfrom['name'],FALSE, // FALSE=no legacy mode
                                                  $this->charset,$this->minimal,
                                                  $this->max_length,$this->eol);
        // 4 - Reply-To:
        if (strcasecmp($this->mailfrom['addr'],$this->mailreplyto['addr']) != 0) {
            $remaining = $this->max_length - 10; // strlen('Reply-To: ') = 10
            $headers['Reply-To'] = $this->rfc5322_address($this->mailreplyto['addr'],$remaining,
                                                          $this->mailreplyto['name'],FALSE,
                                                          $this->charset,$this->minimal,
                                                          $this->max_length,$this->eol);
        }

        // 5 - Cc:
        if (sizeof($this->mailcc) > 0) {
            $remaining = $this->max_length - 5; // strlen('Cc: '.$glue) = 5
            $headers['Cc'] = '';
            $glue = '';
            foreach($this->mailcc as $mailcc) {
                $headers['Cc'] .= $glue.$this->rfc5322_address($mailcc['addr'],$remaining,
                                                               $mailcc['name'],FALSE,
                                                               $this->charset,$this->minimal,
                                                               $this->max_length - 1, // -1 for glue
                                                               $this->eol);
                if ($glue == '') { $glue = ','; }
            }
        }
        // 6 - X-Mailer: and necessary (MIME-)headers
        $headers['X-Mailer'] = sprintf('Website@School Mailer %s (%s, v%s)',
                                       WAS_RELEASE,WAS_RELEASE_DATE,WAS_VERSION);
        $headers = array_merge($headers,$message_headers); // the latter headers prevail

        // 7 - Join all header fields to a single header
        $headertext = '';
        $glue = '';
        foreach($headers as $k => $v) {
            $headertext .= $glue.$k.': '.$v;
            if ($glue == '') { $glue = $this->eol; }
        }
        
        // 8 - Finally send the message
        $additional = '-f'.$this->mailfrom['addr'];
        $retval = mail($mailto,$subject,$message,$headertext,$additional);
        return $retval;
    } // send_body()

    /** encode a string according to RFC2047 (Message Header Extensions for Non-ASCII Text)
     *
     * This routine encodes $source according to RFC2047 (Message Header Extensions for Non-ASCII Text)
     * using the 'Q'-encoding (somewhat comparable to quoted_printable). However, if the $source uses
     * only harmless 7bit characters and falls within the limit of $remaining characters it is returned
     * unchanged and the number of $remaining characters is updated accordingly.
     *
     * In all other cases ($source contains bytes > 127, $source is longer than $remaining, etc.) this
     * encodes the string into 'encoded-word's of max 75 chars. These 'encoded-word's look like this:
     * <pre>
     * "=?" charset "?" encoding "?" encoded-text "?="
     * </pre>
     * with 'encoding' always equal to "Q" (similar to quoted printable).
     * The actual encoding of characters is done in {@link rfc2047_qchar()}. The boolean flag $minimal
     * can be used to limit the literal representation to only digits and letters, using generic 8bit
     * encoding by setting it to TRUE.
     *
     * If multiple 'encoded-word's are necessary, they are separated from each other by a folding space,
     * i.e. newline (using $eol) followed by a normal space (ASCII 32). The end result never ends with
     * such a folding space; the returned value always ends with the "?=" of the last 'encoded-word'.
     *
     * Note 1:
     * I found it quite hard to read the combination of RFC5322 and RFC2047 because I had some trouble
     * distinguishing the rules for RFC5322-type headers. I finally settled for this simplified set
     *
     *  - From:, To:, Cc: and Reply-To: are all of type 'mailbox' to me (KISS, no 'group's and stuff)
     *
     *  - A 'mailbox' can be either written as [ 'display-name' ] "<" addr-spec ">" OR as
     *    addr-spec "(" ctext ")", which both allow for FWS (folding white space)
     *
     *  - A Subject: field is simply 'unstructured', which also allows for FWS
     *
     * Furthermore, this routine simply encodes non-ASCII text and therefore makes no assumptions
     * about the contents of $source; it is the caller's responsability to make sure that
     * the 'ctext' or 'unstructured' or 'display-name' conforms to RFC5322.
     *
     * Note 2:
     * I have not implemented fancy and streamlined code to minimise the amount of encoded characters,
     * ie. leaving pure ASCII-words unencoded and encoding only non-ASCII words or words containing "=?"
     * because I could not invest that amount of time. Maybe in a later version... (famous last words).
     * I did optimise for short and pure and simple ASCII strings because I expect that generated 
     * Subject: headers and other headers will be ASCII most of the time. We'll see how that works out.
     *
     * Note 3:
     * Quirk: if initially there is not enough space for the shortest possible 'encoded-word', we insert
     * a FWS even if $source has no WSP at that point. Basically it means that we add a character to the
     * result. This may or may not be a problem for the caller. OTOH: the caller should provide enough
     * space in the first place, so there.
     *
     * References: see {@link http://www.ietf.org/rfc/rfc2047.txt} and {@link http://www.ietf.org/rfc/rfc5322.txt}.
     *
     * @param string $source the string to encode (could be 7bit)
     * @param int &$remaining the number of bytes remanining on the current output line (not counting any CR+LF)
     * @param string $charset indicates character set used in $source
     * @param bool $minimal if TRUE, rfc2047_qchar() limits literal encoding to digits and letters
     * @param int $max_length limit on output line length (and indirect of the 'encoded-word' length) to max 76 (75)
     * @param string $eol the end of line character(s), default as per RFC5322 (RFC822) is CR chr(13) + LF chr(10)
     * @return string possibly encoded $source with possibly embedded folding whitespace 
     * @todo maybe optimise this routine to let pure ASCII-words through unencoded (in a later version)
     * @uses rfc2047_qchar()
     * @uses is_7bit()
     */
    function rfc2047_qstring($source,&$remaining,$charset="UTF-8",$minimal=FALSE,$max_length=76,$eol="\r\n") {
        //
        // 1 -- shortcut for the easy cases (simple and short 7-bit strings)
        //
        $n = strlen($source);
        if (($this->is_7bit($source)) &&                    // maybe not US-ASCII but 7 bits nevertheless...
            (strpos($source,"=?") === FALSE) &&             // ...and no spurious begin of an 'encoded-word'...
            ($n <= $remaining)) {                           // ...and room to move: we're done!
            $remaining -= $n;
            return $source;
        }

        //
        // 2 -- prepare for actual rfc2047 encoding (always of type 'Q')
        //
        $encode_head = '=?'.$charset.'?Q?';
        $encode_tail = '?=';
        $overhead = strlen($encode_head) + strlen($encode_tail);

        // 3 -- quick and dirty: simply encode everything including plain ASCII
        $target = '';                                       // collects output
        $i = 0;                                             // index to step through $source
        $max_length = max(min(76,$max_length),$overhead+4); // +1 for folding, +3 for 'worst case' "=XX" encoded_char
        $remaining = min($max_length - 1,$remaining);       // keep first 'encoded-word' within 75 bytes too

        //
        // 3A -- get started with special case (no space left on the current output line)
        //
        $c = ord($source{0});
        $required_len = 1;                                  // create parameter to allow return of value
        $encoded_char = $this->rfc2047_qchar($c,$required_len,$minimal);
        if ($remaining < $overhead + $required_len) {       // Alas, we MUST fold at this point... :-(
            if (($c == 9) || ($c == 32)) {
                $target .= $eol.chr($c).$encode_head; 
                $remaining = $max_length - 1 - $overhead;   // already take length of $encode_tail into account
            } else {                                        // See note above: alas, we HAVE to insert an extra WSP here
                $target .= $eol.chr(32).$encode_head.$encoded_char;
                $remaining = $max_length - 1 - $overhead - strlen($encoded_char);
            }
            ++$i;                                           // Remember we already processed the first character
        } else {                                            // unconditionally start with the first 'encoded-word'
            $target .= $encode_head;
            $remaining -= $overhead;
        }

        //
        // 3B -- step through string and convert into 'encoded-word's of at most $max_length - 1 bytes
        //
        while ($i < $n) {
            $c = ord($source{$i++});
            $encoded_char = $this->rfc2047_qchar($c,$required_len,$minimal);
            if ($remaining < $required_len) {
                $target .= $encode_tail.$eol.chr(32).$encode_head;
                $remaining = $max_length - 1 - $overhead;
            }
            $target .= $encoded_char;
            $remaining -= strlen($encoded_char);
        }
        $target .= $encode_tail;
        return $target;
    } // rfc2047_qstring()

    /** encode an 8-bit byte according to Q-encoding in RFC2047
     *
     * This routine encodes a single integer ASCII code into either
     *  - literal representaion
     *  - generic 8bit representation, ie. "=" followed by 2 (uppercas)e hexdigits
     *  - an underscore character
     *
     * If $minimal is FALSE, all printable ASCII characters from 33 "!" to 126 "~"
     * except 61 "=", 63 "?" and 95 "_" use literal representation. Character 32 " "
     * is represented as an underscore (for improved readabilitu/deciphering).
     *
     * If $minimal is TRUE, only digits "0" - "9" and letters "A" - "Z" and "a" - "z"
     * use literal representation and character 32 " " uses generic 8bit encoding "=20".
     *
     * The latter case yields only digits, letters, equal-sign and question mark, which
     * should travel undisturbed through any mail transdfer agent.
     *
     * There is a special situation when encoding UTF8 where characters can span multiple
     * octets. The length of such a sequence can be determined by the number of most significant
     * 1's in a row in the first octet. If $c is the first octet of a UTF8-sequence, we tell
     * the caller the total length of the encoded sequence, not just the length of the encoded
     * 1st octet (which would always be 3). This forces the caller to start a new 'encoded-word'
     * with enough room for the complete sequence if necessary, preventing a multi-octet sequence
     * to span two 'encoded-words'. Note that characters in a UTF8-tail yield length 3, even when
     * more UTF8-tail octets follow. That is OK because the first character already 'reserved' the
     * space when the first octet was processed.
     *
     * Here is a small truth table for sequence lengths (see also RFC3629).
     * bit pattern   range len  comments   
     * 0xxx.xxxx     0-127   3  ASCII
     * 10xx.xxxx   128-191   3  octet is part of UTF8-tail
     * 110x.xxxx   192-223   6  UTF8-2, beginning of a sequence of 2 octets
     * 1110.xxxx   224-239   9  UTF8-3, beginning of a sequence of 3 octets
     * 1111.0xxx   240-247  12  UTF8-4, beginning of a sequence of 4 octets
     * 1111.10xx   248-251   3  sequence of 5 characters not defined in RFC3929, settle for length 3
     * 1111.110x   252-253   3  sequence of 6 characters not defined in RFC3929, settle for length 3
     * 1111.111x   254-255   3  no sequence at all, settle for length 3
     *
     * Note that if $c is NOT UTF8 but say ISO-5988-1, the worst that can happen is that a perfectly
     * valid single octet character in the range 192-247 would indicat a length of more than the
     * necessary 3, pushing up to 4 characters to the next 'encoded-word'. Oh well, I can live with that.
     *
     * References: {@link http://www.ietf.org/rfc/rfc2047.txt}, {@link http://www.ietf.org/rfc/rfc3629.txt}.
     *
     * @param int $c the character to encode
     * @param int &$required_len returns the space required for this encoded char (UTF8-aware)
     * @param bool $minimal if TRUE, only [0-9A-Za-z] use literal representation, otherwise encoding is more relaxed
     * @return string the encoded character either as literal or as generic 8bit in the form "=XX"; $required_len set
     */
    function rfc2047_qchar($c,&$required_len,$minimal=FALSE) {
        if ($c == 32) {
            $encoded_char = ($minimal) ? '=20' : '_';
            $required_len = ($minimal) ? 3 : 1;
        } elseif (($minimal) && (((48 <= $c) && ($c <= 57)) ||      // '0' - '9'
                                 ((65 <= $c) && ($c <= 90)) ||      // 'A' - 'Z'
                                 ((97 <= $c) && ($c <= 122)))) {    // 'a' - 'z'

            $encoded_char = chr($c);
            $required_len = 1;
        } elseif ((!$minimal) && ((33 <= $c) && ($c <= 126) &&               // printable ASCII
                                  ($c != 61) && ($c != 63) && ($c != 95))) { // and not '=', '?', or '_'
            $encoded_char = chr($c);
            $required_len = 1;
        } else {
            $encoded_char = sprintf('=%02X',$c);
            if     ($c < 192) { $required_len =  3; } // ASCII/UTF8-1 or UTF8-tail (see RFC3629 section 4) 
            elseif ($c < 224) { $required_len =  6; } // UTF8-2 (including overlong sequences)
            elseif ($c < 240) { $required_len =  9; } // UTF8-3 (including surrogate pairs)
            elseif ($c < 248) { $required_len = 12; } // UTF8-4
            else              { $required_len =  3; } //longer/unrecognised sequences
        }
        return $encoded_char;
    } // rfc2047_qchar()

    /** a small utility routine to determine if a string has only 7bit characters
     *
     * @param string &$source the text to examine
     * @return bool TRUE if no 8-bit characters were found, FALSE otherwise
     */
    function is_7bit(&$source) {
        $n = strlen($source);
        for ($i=0; $i < $n; ++$i) {
            if (ord($source{$i}) & 128) {
               return FALSE;
            }
        }
        return TRUE;
    } // is_7bit()

    /** construct an address field according to RFC5322 (RFC822)
     *
     * This routine constructs an address according to (simplified) rules in RFC5322 section 3.4.
     * Depending on the $legacy flag, the parameters are used to construct either an 'angle-addr'
     * <pre>
     * DQUOT $display_name DQUOT SPACE "<" $addr_spec ">"
     * </pre>
     * or an address with the display-name "hidden" in a comment
     * <pre>
     * $addr_spec SPACE "(" $display_name ")"
     * </pre>
     *
     * Basically the $addr_spec is not modified; it is assumed that this string obeys the rules
     * for 'addr-spec' in RFC5322. Specifically we do not encode this information
     * (with {@link rfc2047_qstring()} or otherwise). However, we DO strip any CR and/or LF characters
     * because these might cause problems lateron (eg. an unwanted extra blank line in the mail headers).
     * Also, we definately do not want to have angle brackets, so we remove those too, just to be sure.
     *
     * The $display_name can be modified. Reading RFC5322 yields the following (simplified) rules.
     * 'display-name' => 'phrase' => 1*'word'; 'word' => 'atom' | 'quoted-string'. In other words:
     * it is allowed to use a quoted string, ie. 
     *  - a DQUOTE, followed by
     *  - a string with printable ASCII characters not being DQUOTE or the quote character backslash, followed by
     *  - a DQOTE.
     *
     * Note that FWS (folding white space) is allowed between the two DQUOTEs.
     * This leads an easy way out of to stripping DQUOTEs and backslashes before the
     * 'display-name' is encoded and/or folded.
     *
     * If the legacy-flag is set, we use the parameter $display_name to construct a 
     * (simplified) comment, ie.
     *  - a "(", followed by
     *  - a string of ctext characters not containing "(", ")" or a backslash, followed by
     *  - a ")".
     * Here, too, folding whitespace is allowed between the opening "(" and closing ")".
     * This variant leads the easy way out of stripping parentheses and backslashes
     * before the 'display-name' is encoded and/or folded as a comment.
     *
     * Note:
     * All this cleaning up of CR, LF, DQUOTE, etc. does NOT weed out other CTLs (ASCII 0...ASCII 31).
     *
     * @param string $addr_spec is an address of the form 'local-part' "@" 'domain' (RFC5322 section 3.4.1)
     * @param int &$remaining indicates how much space is left on the current output line
     * @param string $display_name is the human readable name associated with $addr_spec
     * @param bool $legacy if TRUE, output is in the legacy format 'addr-spec' "(" 'ctext' ")"
     * @param string $charset indicates character set used in $display_name
     * @param bool $minimal if TRUE, eventually rfc2047_qchar() limits literal encoding to digits and letters
     * @param int $max_length limit on output line length (and indirect of the 'encoded-word' length) to max 76 (75)
     * @param string $eol the end of line character(s), default as per RFC5322 (RFC822) is CR chr(13) + LF chr(10)
     * @return string ready-to-use header, possibly encoded and/or folded; $remaining updated
     * @uses rfc2047_qstring()
     */
    function rfc5322_address($addr_spec,&$remaining,$display_name='',$legacy=FALSE,
                                                    $charset="UTF-8",$minimal=FALSE,$max_length=76,$eol="\r\n") {
        $address = ''; // collects output
        $forbidden = array(chr(10),chr(13),'<','>');
        $addr_spec = str_replace($forbidden,'',$addr_spec); // make sure addr_spec is clean
        if ($legacy) {
            $required = strlen($addr_spec);
            if ($remaining < $required) {
                $address .= $eol.chr(32);
                $remaining = $max_length - 1;       // -1 for the folding space
            }
            $address .= $addr_spec;
            $remaining -= $required;
            $forbidden = array(chr(10),chr(13),'(',')','\\'); // easy way out: strip unwanted characters from input
            $display_name = str_replace($forbidden,'',$display_name);
            if (!empty($display_name)) {
                if ($remaining < 2) {
                    $address .= $eol;
                    $remaining = $max_length;;
                }
                $address .= " (";
                $remaining -= 2;
                $address .= $this->rfc2047_qstring($display_name,$remaining,$charset,$minimal,$max_length,$eol);
                if ($remaining < 1) {
                    $address .= $eol.chr(32);
                    $remaining = $max_length - 1;       // -1 for the folding space
                }
                $address .= ")";
                --$remaining;
            }
        } else {
            $forbidden = array(chr(10),chr(13),chr(34),'\\'); // easy way out: strip unwanted characters from input
            $display_name = str_replace($forbidden,'',$display_name);
            $angle_addr = '<'.$addr_spec.'>';
            if (!empty($display_name)) {
                if ($remaining < 1) {
                    $address .= $eol.chr(32);
                    $remaining = $max_length - 1;       // -1 for the folding space
                }
                $address .= chr(34);                    // DQUOTE opening 'quoted-string'
                --$remaining;
                $address .= $this->rfc2047_qstring($display_name,$remaining,$charset,$minimal,$max_length,$eol);
                if ($remaining < 1) {
                    $address .= $eol.chr(32);
                    $remaining = $max_length - 1;       // -1 for the folding space
                }
                $address .= chr(34);                    // DQUOTE closing 'quoted-string'
                --$remaining;
                if (strlen($angle_addr) < $remaining) { // SP + angle_addr will fit on remaining space
                    $address .= chr(32);                // so we won't have folding problems after adding
                    --$remaining;                       // this simple SP
                }
            }
            $required = strlen($angle_addr);
            if ($remaining < $required) {
                $address .= $eol.chr(32);
                $remaining = $max_length - 1;           // -1 for the folding space
            }
            $address .= $angle_addr;
            $remaining -= $required;
        }
        return $address;
    } // rfc5322_address()

    /** construct a message-id conforming to RFC5322 (RFC2822, RFC822)
     *
     * This constructs a message id according to specifications in section
     * 3.6.4 in RFC5322 Internet Message Format (October 2008),
     * see {@link http://www.ietf.org/rfc/rfc5322.txt}.
     *
     * Note that the 'id-left' in the 'msg-id' also contains the remote
     * IP-address. This could be an IPv4 address in the usual dotted-decimal
     * form but it could also be an IPv6 address like '::1' (3 characters) or
     * '[FEDC:BA98:7654:3210:FEDC:BA98:7654:3210]' (41 characters).
     * The total maximum-length of 'id-left' may add up to
     * 11 (32-bit signed pid) + 1 (dot) + 41 (full IPv6 w/ brackets) + 1 (hash) +
     * 11 (32-bit signed remote port) + 1 (dot) + 14 (date/time) + 11 (signed
     * unique number) = 91 characters. This is longer than the recommended
     * linelength of 78 characters, but the absolute maximum of 998 characters
     * will probably NOT be reached. OTOH: we do not actually check for
     * huge domain names and/or server names. Oh well.
     *
     * Note that we massage the IPv6 address by replacing any ':', '[' and ']'
     * with '!', '{' and '}' respectively because the former are not allowed in
     * a dot-atom-text. As a matter of fact we translate most 'specials' to
     * 'atext' (RFC5322 3.2.3). Notable exception: the dot stays.
     *
     * @return string message-id according to RFC5322 section 3.6.4
     * @todo how about UTF-8 hostnames? Mmmm...
     */
    function rfc5322_message_id() {
        global $CFG;

        // 1 -- construct id-right according to RFC5322
        $id_right = '';

        // 1A -- try bare host/domain name from this site's URL
        if (($url = parse_url($CFG->www)) !== FALSE) {
            // labels [RFC1035] are dot-atom-text [RFC5322] hence the preg_replace()
            $id_right = (isset($url['host'])) ? trim(preg_replace('/[^a-zA-Z0-9.-]/','',$url['host']),'.') : '';
        }
        // 1B -- if no joy, try the server name
        if ((empty($id_right)) && (isset($_SERVER['SERVER_NAME']))) {
            $id_right = trim(preg_replace('/[^a-zA-Z0-9.-]/','',$_SERVER['SERVER_NAME']),'.');
        }
        // 1C -- last resort: localhost
        if (empty($id_right)) {
            $id_right = 'localhost.localdomain'; // last resort
        }
        // 2 -- construct msg-id with a hopefully unique id-left
        $msg_id = sprintf('%d.%s#%d.%s.%d@%s',
                          intval(getmypid()),
                          (isset($_SERVER['REMOTE_ADDR'])) ? strtr($_SERVER['REMOTE_ADDR'],
                                                                   ':[]()<>;@,\\/',
                                                                   '!{}{}{}!*-==' ) : '127.0.0.1',
                          (isset($_SERVER['REMOTE_PORT'])) ? intval($_SERVER['REMOTE_PORT']) : 42,
                          strftime('%Y%m%d%H%M%S'),
                          get_unique_number(),
                          $id_right);
        return $msg_id;
    } // rfc5322_message_id()


    /** construct a unique boundary for use within this mail message
     *
     * the combination of our pid, the current date up to a second and a unique number
     * within this run should provide enough uniqueness within this mail message.
     * The maximum length of the resulting string is at most 44 characters (assuming
     * signed integers take at most 11 positions in decimal form). However, as a rule
     * get_unique_number() starts with a single '1' so a practical max length of 34
     * is far more likely.
     *
     * @return string unique boundary value
     */
    function boundary() {
        $b = sprintf('----=_%s_%d_%d',strftime('%Y%m%d%H%M%S'),intval(getmypid()),get_unique_number());
        return $b;
    } // boundary()

} // Email

?>