<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/lib/filelib.php - utilities for manipulating files
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wascore
 * @version $Id: filelib.php,v 1.11 2016/06/03 13:44:46 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }

/** determine the mimetype of a file
 *
 * Try to discover the mimetype of a file.
 * First we see if getimagesize() yields results; if so we're done (0).
 * Very efficient if there are many image files and not so many other files.
 *
 * If not an image we try finfo-method (1), mime_content_type()-method (2)
 * and finally the shell-to-file(1)-method (3) to determine the mime type.
 *
 * Note that in step 3 we shell out and try to execute the file(1) command.
 * The results are checked against a pattern to assert that we are
 * really dealing with a mime type. The pattern is described in RFC2616
 * (see sections 3.7 and 2.2) (see also {@link get_mediatype()}.:
 *
 *<pre>
 *      media-type     = type "/" subtype *( ";" parameter )
 *      type           = token
 *      subtype        = token
 *      token          = 1*&lt;any CHAR except CTLs or separators&gt;
 *      separators     = "(" | ")" | "&lt;" | "&gt;" | "@"
 *                     | "," | ";" | ":" | "\" | &lt;"&gt;
 *                     | "/" | "[" | "]" | "?" | "="
 *                     | "{" | "}" | SP | HT
 *      CHAR           = &lt;any US-ASCII character (octets 0 - 127)&gt;
 *      CTL            = &lt;any US-ASCII control character
 *                       (octets 0 - 31) and DEL (127)&gt;
 *      SP             = &lt;US-ASCII SP, space (32)&gt;
 *      HT             = &lt;US-ASCII HT, horizontal-tab (9)&gt;
 *      &lt;"&gt;            = &lt;US-ASCII double-quote mark (34)&gt;
 *</pre>
 *
 * This description means we should look for two tokens containing
 * letters a-z or A-Z, digits 0-9 and these special characters:
 * ! # $ % & ' * + - . ^ _ ` | or ~. That's it.
 *
 * Note that some methods may return a mime type with additional parameters.
 * e.g. 'text/plain; charset=US-ASCII'. This fits the pattern, because
 * it starts with a token, a slash and another token.
 *
 * The optional parameter $name is used to perform special handling of
 * the text/css type: some methods (notably old versions of file(1))
 * return 'text/plain' instead of 'text/css'. Workaround: if we see
 * a mimetype starting with 'text/' AND the extension is .css we assume
 * text/css.
 *
 * Note that we trust the user at least partially($name eventually
 * originales in the $_FILES[] array). The path to the actual file
 * to check is in $path. If $name is not specified, we extract the
 * extension from $path (as in the case of {@link send_file_from_datadir()}.
 *
 * @param string $path fully qualified path to the file to test
 * @param string $name name of the file, possibly different from $path
 * @return string mimetype of the file $path or FALSE on error/not found
 * @uses get_mediatype()
 */
function get_mimetype($path,$name='') {
    $mime = FALSE;
    $method = 0; // DEBUG: determine which attempt succeeds

    // 0 -- getimagesize (quick check, works for only for (some) image/*)
    if ((($imgsiz = @getimagesize($path)) !== FALSE) && (is_array($imgsiz)) && (isset($imgsiz['mime']))) {
        return $imgsiz['mime'];
    }

    // 1 -- finfo (since 5.3.0)
    if ($mime === FALSE) {
        if ((function_exists('finfo_open')) &&
            (function_exists('finfo_file')) &&
            (function_exists('finfo_close')) &&
            (defined('FILEINFO_MIME'))) {
            $finfo = finfo_open(FILEINFO_MIME);
            if ($finfo !== FALSE) {
                $mime = finfo_file($finfo,$path);
                finfo_close($finfo);
                $method = 1;
            }
        }
    }

    // 2 -- mime_content_type() (since 4.3.0 but now deprecated)
    if (($mime === FALSE) && (function_exists('mime_content_type'))) {
        $mime = mime_content_type($path);
        $method = 2;
    }

    // 3 -- shell out for the file(1) command
    if ($mime === FALSE) {
        $command = sprintf('file -b -i %s',escapeshellarg($path)); // -b = brief, -i = mime type
        $dummy = array();
        $retval = 0;
        $output = exec($command,$dummy,$retval);
        if ($retval == 0) { // assert the result looks like a mimetype and not an error message
            if (get_mediatype($output) !== FALSE) {
                $mime = $output;
                $method = 3;
            }
        }
    }

    // 4 -- deal with known issue (see note)
    if (($ext = strtolower(pathinfo((empty($name)) ? $path : $name,PATHINFO_EXTENSION))) == 'css') {
        if (($typ = get_mediatype($mime)) !== FALSE) {
            if (substr($typ,0,5) == 'text/') {
                $mime = 'text/css';
            }
        }
    }
    // logger(sprintf('%s(): method=%d path=%s name=%s mime=%s',__FUNCTION__,$method,$path,$name,$mime));
    return $mime;
} // get_mimetype()


/** extract the mediatype and -subtype from a full mimetype
 *
 * this extracts the mediatype and -subtype from a full mimetype, i.e.
 * 'text/plain' from 'text/plain; charset=US-ASCII' (see also {@link get_mimetype()} and RFC2616).
 * If $mimetype doesn't look like a mimetype, we return FALSE.
 *
 * @param string $mimetype the full mimetype to examine, possibly with parameters
 * @return bool|string FALSE on invalid mimetype, the lowercase extracted mediatype and -subtype otherwise
 */
function get_mediatype($mimetype) {
    $pattern = "/^[a-zA-Z0-9!#\$%&'\*+\-.\^_`|~]+\/[a-zA-Z0-9!#\$%&'\*+\-.\^_`|~]+/";
    $matches = array();
    if (preg_match($pattern,$mimetype,$matches) == 1) {
        return strtolower($matches[0]);
    } else {
        return FALSE;
    }
} // get_mediatype()


/** check availability of requested file for the current user
 *
 * This checks the $filename for permissions and existence.
 * Returns FALSE if the current user is not allowed to see the file
 * or if the file does not exist. We remember failures in the logger.
 *
 * @param string $filename is the requested file
 * @return bool|string FALSE if access denied, rewritten path if access granted etc.
 */
function file_available($filename) {
    global $USER,$CFG;
    $parts = explode('/',trim(strtr($filename,'\\','/'),'/'));
    $path = '/'.implode('/',$parts);

    // 1 -- always disallow attempts to escape from tree via parent directory tricks
    if (in_array('..',$parts)) {
        logger(sprintf("%s(): access denied for file '%s': no tricks with '/../'",
                       __FUNCTION__,$path),WLOG_DEBUG);
        return FALSE;
    }

    // 2 -- check the 1st (areas/users/groups) and 2nd (subdir) component of the requested file
    $table = $parts[0];
    $where = array('is_active' => TRUE, 'path' => (isset($parts[1])) ? $parts[1] : '');
    switch ($table) {
    case 'areas':
        $fields = array('area_id','is_private');
        if (($record = db_select_single_record($table,$fields,$where)) === FALSE) {
            logger(sprintf("%s(): access denied for file '%s': non-existing or inactive area",
                           __FUNCTION__,$path),WLOG_DEBUG);
            return FALSE;
        }
        $area_id = intval($record['area_id']);
        if ((db_bool_is(TRUE,$record['is_private'])) &&
            (!$USER->has_intranet_permissions(ACL_ROLE_INTRANET_ACCESS,$area_id))) {
            logger(sprintf("%s(): access denied for file '%s' in private area '%d'",
                            __FUNCTION__,$path,$area_id),WLOG_DEBUG);
            return FALSE;
        }
        break;

    case 'users':
        $fields = array('user_id');
        if (($record = db_select_single_record($table,$fields,$where)) === FALSE) {
            logger(sprintf("%s(): access denied for file '%s': non-existing or inactive user",
                           __FUNCTION__,$path),WLOG_DEBUG);
            return FALSE;
        }
        break;

    case 'groups':
        $fields = array('group_id');
        if (($record = db_select_single_record($table,$fields,$where)) === FALSE) {
            logger(sprintf("%s(): access denied for file '%s': non-existing or inactive group",
                           __FUNCTION__,$path),WLOG_DEBUG);
            return FALSE;
        }
        break;

    default:
        logger(sprintf("%s(): access denied for file '%s': subdirectory '%s' not recognised",
                        __FUNCTION__,$path,$parts[0]),WLOG_DEBUG);
        return FALSE;
        break;
    }

    // 3 -- still here? 1st and 2nd components are good but does the file exist?
    if (!is_file($CFG->datadir.$path)) {
        logger(sprintf("%s(): access denied for file '%s': file does not exist",
                       __FUNCTION__,$path),WLOG_DEBUG);
        return FALSE;
    }
    return $path;
} // file_available()


/** verify the combination of filename extension and mimetype
 *
 * this looks up the mimetypes associated with filename extension 
 * $ext in the global table $MIME_EXT_TYP. If any of those mimetypes
 * matches the requested mimetype we confirm that ext and typ match.
 *
 * Note 1:
 * There is another table: $MIME_TYP_EXT that works the other way around.
 * We could have used either table; I choose the former.
 *
 * Note 2:
 * The two tables are generated with a development tool (see /devel/tools)
 * {@link mimemerge.php}. This tool reads the existing table in this file
 * filelib.php and optionally a newer version of the Apache mime.types file
 * creating updated tables.
 *
 * @param string $typ mimetype
 * @param string $ext filename extension
 * @return bool TRUE if both parameters match, FALSE otherwise
 * @uses $MIME_EXT_TYP
 */
function mime_typ_ext_match($typ,$ext) {
    global $MIME_EXT_TYP;
    if (isset($MIME_EXT_TYP[$ext])) {
        $typs = $MIME_EXT_TYP[$ext];
        if (is_array($typs)) {
            return in_array($typ,$typs);
        } else {
            return ($typ == $typs);
        }
    } else {
        return FALSE;
    }
} // mime_typ_ext_match()


/** lookup extension in table, using mimetype as key
 *
 * return the first extension associated with mimetype $typ
 * 
 * @param string $typ mime type
 * @return bool|string FALSE on error/not found, extension otherwise
 * @uses $MIME_TYP_EXT
 */
function mime_ext_by_typ($typ) {
    global $MIME_TYP_EXT;
    if (isset($MIME_TYP_EXT[$typ])) {
        $exts = $MIME_TYP_EXT[$typ];
        return (is_array($exts)) ? $exts[0] : $exts;
    } else {
        return FALSE;
    }
} // mime_ext_by_typ()


/** lookup mimetype in table, using extension as key
 *
 * return the first mimetype associated with filename extension $ext
 *
 * @param string $ext filename extension
 * @return bool|string FALSE on error/not found, mimetype otherwise
 * @uses $MIME_EXT_TYP
 */
function mime_typ_by_ext($ext) {
    global $MIME_EXT_TYP;
    if (isset($MIME_EXT_TYP[$ext])) {
        $typs = $MIME_EXT_TYP[$ext];
        return (is_array($typs)) ? $typs[0] : $typs;
    } else {
        return FALSE;
    }
} // mime_ext_by_typ()


global $MIME_EXT_TYP;
if (empty($MIME_EXT_TYP)) { $MIME_EXT_TYP=array(
 ''=>'application/octet-stream',
 '3dml'=>'text/vnd.in3d.3dml',
 '3ds'=>'image/x-3ds',
 '3g2'=>'video/3gpp2',
 '3gp'=>'video/3gpp',
 '7z'=>'application/x-7z-compressed',
 'aab'=>'application/x-authorware-bin',
 'aac'=>'audio/x-aac',
 'aam'=>'application/x-authorware-map',
 'aas'=>'application/x-authorware-seg',
 'abw'=>'application/x-abiword',
 'ac'=>'application/pkix-attr-cert',
 'acc'=>'application/vnd.americandynamics.acc',
 'ace'=>'application/x-ace-compressed',
 'acu'=>'application/vnd.acucobol',
 'acutc'=>'application/vnd.acucorp',
 'adp'=>'audio/adpcm',
 'aep'=>'application/vnd.audiograph',
 'afm'=>'application/x-font-type1',
 'afp'=>'application/vnd.ibm.modcap',
 'ahead'=>'application/vnd.ahead.space',
 'ai'=>'application/postscript',
 'aif'=>'audio/x-aiff',
 'aifc'=>'audio/x-aiff',
 'aiff'=>'audio/x-aiff',
 'air'=>'application/vnd.adobe.air-application-installer-package+zip',
 'ait'=>'application/vnd.dvb.ait',
 'ami'=>'application/vnd.amiga.ami',
 'apk'=>'application/vnd.android.package-archive',
 'appcache'=>'text/cache-manifest',
 'application'=>'application/x-ms-application',
 'apr'=>'application/vnd.lotus-approach',
 'arc'=>'application/x-freearc',
 'asc'=>array('text/plain','application/pgp-signature'),
 'asf'=>'video/x-ms-asf',
 'asm'=>'text/x-asm',
 'aso'=>'application/vnd.accpac.simply.aso',
 'asx'=>'video/x-ms-asf',
 'atc'=>'application/vnd.acucorp',
 'atom'=>'application/atom+xml',
 'atomcat'=>'application/atomcat+xml',
 'atomsvc'=>'application/atomsvc+xml',
 'atx'=>'application/vnd.antix.game-component',
 'au'=>'audio/basic',
 'avi'=>'video/x-msvideo',
 'aw'=>'application/applixware',
 'azf'=>'application/vnd.airzip.filesecure.azf',
 'azs'=>'application/vnd.airzip.filesecure.azs',
 'azw'=>'application/vnd.amazon.ebook',
 'bat'=>'application/x-msdownload',
 'bcpio'=>'application/x-bcpio',
 'bdf'=>'application/x-font-bdf',
 'bdm'=>'application/vnd.syncml.dm+wbxml',
 'bed'=>'application/vnd.realvnc.bed',
 'bh2'=>'application/vnd.fujitsu.oasysprs',
 'bin'=>'application/octet-stream',
 'blb'=>'application/x-blorb',
 'blorb'=>'application/x-blorb',
 'bmi'=>'application/vnd.bmi',
 'bmp'=>array('image/bmp','image/x-ms-bmp'),
 'book'=>'application/vnd.framemaker',
 'box'=>'application/vnd.previewsystems.box',
 'boz'=>'application/x-bzip2',
 'bpk'=>'application/octet-stream',
 'btif'=>'image/prs.btif',
 'bz'=>'application/x-bzip',
 'bz2'=>'application/x-bzip2',
 'c'=>'text/x-c',
 'c11amc'=>'application/vnd.cluetrust.cartomobile-config',
 'c11amz'=>'application/vnd.cluetrust.cartomobile-config-pkg',
 'c4d'=>'application/vnd.clonk.c4group',
 'c4f'=>'application/vnd.clonk.c4group',
 'c4g'=>'application/vnd.clonk.c4group',
 'c4p'=>'application/vnd.clonk.c4group',
 'c4u'=>'application/vnd.clonk.c4group',
 'cab'=>'application/vnd.ms-cab-compressed',
 'caf'=>'audio/x-caf',
 'cap'=>'application/vnd.tcpdump.pcap',
 'car'=>'application/vnd.curl.car',
 'cat'=>'application/vnd.ms-pki.seccat',
 'cb7'=>'application/x-cbr',
 'cba'=>'application/x-cbr',
 'cbr'=>'application/x-cbr',
 'cbt'=>'application/x-cbr',
 'cbz'=>'application/x-cbr',
 'cc'=>'text/x-c',
 'cct'=>'application/x-director',
 'ccxml'=>'application/ccxml+xml',
 'cdbcmsg'=>'application/vnd.contact.cmsg',
 'cdf'=>'application/x-netcdf',
 'cdkey'=>'application/vnd.mediastation.cdkey',
 'cdmia'=>'application/cdmi-capability',
 'cdmic'=>'application/cdmi-container',
 'cdmid'=>'application/cdmi-domain',
 'cdmio'=>'application/cdmi-object',
 'cdmiq'=>'application/cdmi-queue',
 'cdx'=>'chemical/x-cdx',
 'cdxml'=>'application/vnd.chemdraw+xml',
 'cdy'=>'application/vnd.cinderella',
 'cer'=>'application/pkix-cert',
 'cfs'=>'application/x-cfs-compressed',
 'cgm'=>'image/cgm',
 'chat'=>'application/x-chat',
 'chm'=>'application/vnd.ms-htmlhelp',
 'chrt'=>'application/vnd.kde.kchart',
 'cif'=>'chemical/x-cif',
 'cii'=>'application/vnd.anser-web-certificate-issue-initiation',
 'cil'=>'application/vnd.ms-artgalry',
 'cla'=>'application/vnd.claymore',
 'class'=>array('application/octet-stream','application/java-vm'),
 'clkk'=>'application/vnd.crick.clicker.keyboard',
 'clkp'=>'application/vnd.crick.clicker.palette',
 'clkt'=>'application/vnd.crick.clicker.template',
 'clkw'=>'application/vnd.crick.clicker.wordbank',
 'clkx'=>'application/vnd.crick.clicker',
 'clp'=>'application/x-msclip',
 'cmc'=>'application/vnd.cosmocaller',
 'cmdf'=>'chemical/x-cmdf',
 'cml'=>'chemical/x-cml',
 'cmp'=>'application/vnd.yellowriver-custom-menu',
 'cmx'=>'image/x-cmx',
 'cod'=>'application/vnd.rim.cod',
 'com'=>'application/x-msdownload',
 'conf'=>'text/plain',
 'cpio'=>'application/x-cpio',
 'cpp'=>'text/x-c',
 'cpt'=>'application/mac-compactpro',
 'crd'=>'application/x-mscardfile',
 'crl'=>'application/pkix-crl',
 'crt'=>'application/x-x509-ca-cert',
 'cryptonote'=>'application/vnd.rig.cryptonote',
 'csh'=>'application/x-csh',
 'csml'=>'chemical/x-csml',
 'csp'=>'application/vnd.commonspace',
 'css'=>array('text/css','text/plain'),
 'cst'=>'application/x-director',
 'csv'=>array('text/csv','text/plain'),
 'cu'=>'application/cu-seeme',
 'curl'=>'text/vnd.curl',
 'cww'=>'application/prs.cww',
 'cxt'=>'application/x-director',
 'cxx'=>'text/x-c',
 'dae'=>'model/vnd.collada+xml',
 'daf'=>'application/vnd.mobius.daf',
 'dart'=>'application/vnd.dart',
 'dataless'=>'application/vnd.fdsn.seed',
 'davmount'=>'application/davmount+xml',
 'dbk'=>'application/docbook+xml',
 'dcr'=>'application/x-director',
 'dcurl'=>'text/vnd.curl.dcurl',
 'dd2'=>'application/vnd.oma.dd2+xml',
 'ddd'=>'application/vnd.fujixerox.ddd',
 'deb'=>'application/x-debian-package',
 'def'=>'text/plain',
 'deploy'=>'application/octet-stream',
 'der'=>'application/x-x509-ca-cert',
 'dfac'=>'application/vnd.dreamfactory',
 'dgc'=>'application/x-dgc-compressed',
 'dic'=>'text/x-c',
 'dir'=>'application/x-director',
 'dis'=>'application/vnd.mobius.dis',
 'dist'=>'application/octet-stream',
 'distz'=>'application/octet-stream',
 'djv'=>'image/vnd.djvu',
 'djvu'=>'image/vnd.djvu',
 'dll'=>array('application/octet-stream','application/x-msdownload'),
 'dmg'=>array('application/octet-stream','application/x-apple-diskimage'),
 'dmp'=>'application/vnd.tcpdump.pcap',
 'dms'=>'application/octet-stream',
 'dna'=>'application/vnd.dna',
 'doc'=>'application/msword',
 'docm'=>'application/vnd.ms-word.document.macroenabled.12',
 'docx'=>'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
 'dot'=>'application/msword',
 'dotm'=>'application/vnd.ms-word.template.macroenabled.12',
 'dotx'=>'application/vnd.openxmlformats-officedocument.wordprocessingml.template',
 'dp'=>'application/vnd.osgi.dp',
 'dpg'=>'application/vnd.dpgraph',
 'dra'=>'audio/vnd.dra',
 'dsc'=>'text/prs.lines.tag',
 'dssc'=>'application/dssc+der',
 'dtb'=>'application/x-dtbook+xml',
 'dtd'=>'application/xml-dtd',
 'dts'=>'audio/vnd.dts',
 'dtshd'=>'audio/vnd.dts.hd',
 'dump'=>'application/octet-stream',
 'dvb'=>'video/vnd.dvb.file',
 'dvi'=>'application/x-dvi',
 'dwf'=>'model/vnd.dwf',
 'dwg'=>'image/vnd.dwg',
 'dxf'=>'image/vnd.dxf',
 'dxp'=>'application/vnd.spotfire.dxp',
 'dxr'=>'application/x-director',
 'ecelp4800'=>'audio/vnd.nuera.ecelp4800',
 'ecelp7470'=>'audio/vnd.nuera.ecelp7470',
 'ecelp9600'=>'audio/vnd.nuera.ecelp9600',
 'ecma'=>'application/ecmascript',
 'edm'=>'application/vnd.novadigm.edm',
 'edx'=>'application/vnd.novadigm.edx',
 'efif'=>'application/vnd.picsel',
 'ei6'=>'application/vnd.pg.osasli',
 'elc'=>'application/octet-stream',
 'emf'=>'application/x-msmetafile',
 'eml'=>'message/rfc822',
 'emma'=>'application/emma+xml',
 'emz'=>'application/x-msmetafile',
 'eol'=>'audio/vnd.digital-winds',
 'eot'=>'application/vnd.ms-fontobject',
 'eps'=>'application/postscript',
 'epub'=>'application/epub+zip',
 'es3'=>'application/vnd.eszigno3+xml',
 'esa'=>'application/vnd.osgi.subsystem',
 'esf'=>'application/vnd.epson.esf',
 'et3'=>'application/vnd.eszigno3+xml',
 'etx'=>'text/x-setext',
 'eva'=>'application/x-eva',
 'evy'=>'application/x-envoy',
 'exe'=>array('application/octet-stream','application/x-msdownload','application/x-dosexec'),
 'exi'=>'application/exi',
 'ext'=>'application/vnd.novadigm.ext',
 'ez'=>'application/andrew-inset',
 'ez2'=>'application/vnd.ezpix-album',
 'ez3'=>'application/vnd.ezpix-package',
 'f'=>'text/x-fortran',
 'f4v'=>'video/x-f4v',
 'f77'=>'text/x-fortran',
 'f90'=>'text/x-fortran',
 'fbs'=>'image/vnd.fastbidsheet',
 'fcdt'=>'application/vnd.adobe.formscentral.fcdt',
 'fcs'=>'application/vnd.isac.fcs',
 'fdf'=>'application/vnd.fdf',
 'fe_launch'=>'application/vnd.denovo.fcselayout-link',
 'fg5'=>'application/vnd.fujitsu.oasysgp',
 'fgd'=>'application/x-director',
 'fh'=>'image/x-freehand',
 'fh4'=>'image/x-freehand',
 'fh5'=>'image/x-freehand',
 'fh7'=>'image/x-freehand',
 'fhc'=>'image/x-freehand',
 'fig'=>'application/x-xfig',
 'flac'=>'audio/x-flac',
 'fli'=>'video/x-fli',
 'flo'=>'application/vnd.micrografx.flo',
 'flv'=>'video/x-flv',
 'flw'=>'application/vnd.kde.kivio',
 'flx'=>'text/vnd.fmi.flexstor',
 'fly'=>'text/vnd.fly',
 'fm'=>'application/vnd.framemaker',
 'fnc'=>'application/vnd.frogans.fnc',
 'for'=>'text/x-fortran',
 'fpx'=>'image/vnd.fpx',
 'frame'=>'application/vnd.framemaker',
 'fsc'=>'application/vnd.fsc.weblaunch',
 'fst'=>'image/vnd.fst',
 'ftc'=>'application/vnd.fluxtime.clip',
 'fti'=>'application/vnd.anser-web-funds-transfer-initiation',
 'fvt'=>'video/vnd.fvt',
 'fxp'=>'application/vnd.adobe.fxp',
 'fxpl'=>'application/vnd.adobe.fxp',
 'fzs'=>'application/vnd.fuzzysheet',
 'g2w'=>'application/vnd.geoplan',
 'g3'=>'image/g3fax',
 'g3w'=>'application/vnd.geospace',
 'gac'=>'application/vnd.groove-account',
 'gam'=>'application/x-tads',
 'gbr'=>'application/rpki-ghostbusters',
 'gca'=>'application/x-gca-compressed',
 'gdl'=>'model/vnd.gdl',
 'geo'=>'application/vnd.dynageo',
 'gex'=>'application/vnd.geometry-explorer',
 'ggb'=>'application/vnd.geogebra.file',
 'ggt'=>'application/vnd.geogebra.tool',
 'ghf'=>'application/vnd.groove-help',
 'gif'=>'image/gif',
 'gim'=>'application/vnd.groove-identity-message',
 'gml'=>'application/gml+xml',
 'gmx'=>'application/vnd.gmx',
 'gnumeric'=>'application/x-gnumeric',
 'gph'=>'application/vnd.flographit',
 'gpx'=>'application/gpx+xml',
 'gqf'=>'application/vnd.grafeq',
 'gqs'=>'application/vnd.grafeq',
 'gram'=>'application/srgs',
 'gramps'=>'application/x-gramps-xml',
 'gre'=>'application/vnd.geometry-explorer',
 'grv'=>'application/vnd.groove-injector',
 'grxml'=>'application/srgs+xml',
 'gsf'=>'application/x-font-ghostscript',
 'gtar'=>'application/x-gtar',
 'gtm'=>'application/vnd.groove-tool-message',
 'gtw'=>'model/vnd.gtw',
 'gv'=>'text/vnd.graphviz',
 'gxf'=>'application/gxf',
 'gxt'=>'application/vnd.geonext',
 'h'=>'text/x-c',
 'h261'=>'video/h261',
 'h263'=>'video/h263',
 'h264'=>'video/h264',
 'hal'=>'application/vnd.hal+xml',
 'hbci'=>'application/vnd.hbci',
 'hdf'=>'application/x-hdf',
 'hh'=>'text/x-c',
 'hlp'=>'application/winhlp',
 'hpgl'=>'application/vnd.hp-hpgl',
 'hpid'=>'application/vnd.hp-hpid',
 'hps'=>'application/vnd.hp-hps',
 'hqx'=>'application/mac-binhex40',
 'htke'=>'application/vnd.kenameaapp',
 'htm'=>'text/html',
 'html'=>'text/html',
 'hvd'=>'application/vnd.yamaha.hv-dic',
 'hvp'=>'application/vnd.yamaha.hv-voice',
 'hvs'=>'application/vnd.yamaha.hv-script',
 'i2g'=>'application/vnd.intergeo',
 'icc'=>'application/vnd.iccprofile',
 'ice'=>'x-conference/x-cooltalk',
 'icm'=>'application/vnd.iccprofile',
 'ico'=>'image/x-icon',
 'ics'=>'text/calendar',
 'ief'=>'image/ief',
 'ifb'=>'text/calendar',
 'ifm'=>'application/vnd.shana.informed.formdata',
 'iges'=>'model/iges',
 'igl'=>'application/vnd.igloader',
 'igm'=>'application/vnd.insors.igm',
 'igs'=>'model/iges',
 'igx'=>'application/vnd.micrografx.igx',
 'iif'=>'application/vnd.shana.informed.interchange',
 'imp'=>'application/vnd.accpac.simply.imp',
 'ims'=>'application/vnd.ms-ims',
 'in'=>'text/plain',
 'ink'=>'application/inkml+xml',
 'inkml'=>'application/inkml+xml',
 'install'=>'application/x-install-instructions',
 'iota'=>'application/vnd.astraea-software.iota',
 'ipfix'=>'application/ipfix',
 'ipk'=>'application/vnd.shana.informed.package',
 'irm'=>'application/vnd.ibm.rights-management',
 'irp'=>'application/vnd.irepository.package+xml',
 'iso'=>'application/x-iso9660-image',
 'itp'=>'application/vnd.shana.informed.formtemplate',
 'ivp'=>'application/vnd.immervision-ivp',
 'ivu'=>'application/vnd.immervision-ivu',
 'jad'=>'text/vnd.sun.j2me.app-descriptor',
 'jam'=>'application/vnd.jam',
 'jar'=>'application/java-archive',
 'java'=>'text/x-java-source',
 'jisp'=>'application/vnd.jisp',
 'jlt'=>'application/vnd.hp-jlyt',
 'jnlp'=>'application/x-java-jnlp-file',
 'joda'=>'application/vnd.joost.joda-archive',
 'jpe'=>'image/jpeg',
 'jpeg'=>'image/jpeg',
 'jpg'=>'image/jpeg',
 'jpgm'=>'video/jpm',
 'jpgv'=>'video/jpeg',
 'jpm'=>'video/jpm',
 'js'=>array('application/x-javascript','application/javascript'),
 'json'=>'application/json',
 'jsonml'=>'application/jsonml+json',
 'kar'=>'audio/midi',
 'karbon'=>'application/vnd.kde.karbon',
 'kfo'=>'application/vnd.kde.kformula',
 'kia'=>'application/vnd.kidspiration',
 'kml'=>'application/vnd.google-earth.kml+xml',
 'kmz'=>'application/vnd.google-earth.kmz',
 'kne'=>'application/vnd.kinar',
 'knp'=>'application/vnd.kinar',
 'kon'=>'application/vnd.kde.kontour',
 'kpr'=>'application/vnd.kde.kpresenter',
 'kpt'=>'application/vnd.kde.kpresenter',
 'kpxx'=>'application/vnd.ds-keypoint',
 'ksp'=>'application/vnd.kde.kspread',
 'ktr'=>'application/vnd.kahootz',
 'ktx'=>'image/ktx',
 'ktz'=>'application/vnd.kahootz',
 'kwd'=>'application/vnd.kde.kword',
 'kwt'=>'application/vnd.kde.kword',
 'lasxml'=>'application/vnd.las.las+xml',
 'latex'=>'application/x-latex',
 'lbd'=>'application/vnd.llamagraphics.life-balance.desktop',
 'lbe'=>'application/vnd.llamagraphics.life-balance.exchange+xml',
 'les'=>'application/vnd.hhe.lesson-player',
 'lha'=>array('application/octet-stream','application/x-lzh-compressed'),
 'link66'=>'application/vnd.route66.link66+xml',
 'list'=>'text/plain',
 'list3820'=>'application/vnd.ibm.modcap',
 'listafp'=>'application/vnd.ibm.modcap',
 'lnk'=>'application/x-ms-shortcut',
 'log'=>'text/plain',
 'lostxml'=>'application/lost+xml',
 'lrf'=>'application/octet-stream',
 'lrm'=>'application/vnd.ms-lrm',
 'ltf'=>'application/vnd.frogans.ltf',
 'lvp'=>'audio/vnd.lucent.voice',
 'lwp'=>'application/vnd.lotus-wordpro',
 'lzh'=>array('application/octet-stream','application/x-lzh-compressed'),
 'm13'=>'application/x-msmediaview',
 'm14'=>'application/x-msmediaview',
 'm1v'=>'video/mpeg',
 'm21'=>'application/mp21',
 'm2a'=>'audio/mpeg',
 'm2v'=>'video/mpeg',
 'm3a'=>'audio/mpeg',
 'm3u'=>'audio/x-mpegurl',
 'm3u8'=>'application/vnd.apple.mpegurl',
 'm4a'=>'audio/mp4',
 'm4u'=>'video/vnd.mpegurl',
 'm4v'=>'video/x-m4v',
 'ma'=>'application/mathematica',
 'mads'=>'application/mads+xml',
 'mag'=>'application/vnd.ecowin.chart',
 'maker'=>'application/vnd.framemaker',
 'man'=>array('application/x-troff-man','text/troff'),
 'mar'=>'application/octet-stream',
 'mathml'=>'application/mathml+xml',
 'mb'=>'application/mathematica',
 'mbk'=>'application/vnd.mobius.mbk',
 'mbox'=>'application/mbox',
 'mc1'=>'application/vnd.medcalcdata',
 'mcd'=>'application/vnd.mcd',
 'mcurl'=>'text/vnd.curl.mcurl',
 'mdb'=>'application/x-msaccess',
 'mdi'=>'image/vnd.ms-modi',
 'me'=>array('application/x-troff-me','text/troff'),
 'mesh'=>'model/mesh',
 'meta4'=>'application/metalink4+xml',
 'metalink'=>'application/metalink+xml',
 'mets'=>'application/mets+xml',
 'mfm'=>'application/vnd.mfmp',
 'mft'=>'application/rpki-manifest',
 'mgp'=>'application/vnd.osgeo.mapguide.package',
 'mgz'=>'application/vnd.proteus.magazine',
 'mid'=>'audio/midi',
 'midi'=>'audio/midi',
 'mie'=>'application/x-mie',
 'mif'=>'application/vnd.mif',
 'mime'=>'message/rfc822',
 'mj2'=>'video/mj2',
 'mjp2'=>'video/mj2',
 'mk3d'=>'video/x-matroska',
 'mka'=>'audio/x-matroska',
 'mks'=>'video/x-matroska',
 'mkv'=>'video/x-matroska',
 'mlp'=>'application/vnd.dolby.mlp',
 'mmd'=>'application/vnd.chipnuts.karaoke-mmd',
 'mmf'=>'application/vnd.smaf',
 'mmr'=>'image/vnd.fujixerox.edmics-mmr',
 'mng'=>'video/x-mng',
 'mny'=>'application/x-msmoney',
 'mobi'=>'application/x-mobipocket-ebook',
 'mods'=>'application/mods+xml',
 'mov'=>'video/quicktime',
 'movie'=>'video/x-sgi-movie',
 'mp2'=>'audio/mpeg',
 'mp21'=>'application/mp21',
 'mp2a'=>'audio/mpeg',
 'mp3'=>'audio/mpeg',
 'mp4'=>'video/mp4',
 'mp4a'=>'audio/mp4',
 'mp4s'=>'application/mp4',
 'mp4v'=>'video/mp4',
 'mpc'=>'application/vnd.mophun.certificate',
 'mpe'=>'video/mpeg',
 'mpeg'=>'video/mpeg',
 'mpg'=>'video/mpeg',
 'mpg4'=>'video/mp4',
 'mpga'=>'audio/mpeg',
 'mpkg'=>'application/vnd.apple.installer+xml',
 'mpm'=>'application/vnd.blueice.multipass',
 'mpn'=>'application/vnd.mophun.application',
 'mpp'=>'application/vnd.ms-project',
 'mpt'=>'application/vnd.ms-project',
 'mpy'=>'application/vnd.ibm.minipay',
 'mqy'=>'application/vnd.mobius.mqy',
 'mrc'=>'application/marc',
 'mrcx'=>'application/marcxml+xml',
 'ms'=>array('application/x-troff-ms','text/troff'),
 'mscml'=>'application/mediaservercontrol+xml',
 'mseed'=>'application/vnd.fdsn.mseed',
 'mseq'=>'application/vnd.mseq',
 'msf'=>'application/vnd.epson.msf',
 'msh'=>'model/mesh',
 'msi'=>'application/x-msdownload',
 'msl'=>'application/vnd.mobius.msl',
 'msty'=>'application/vnd.muvee.style',
 'mts'=>'model/vnd.mts',
 'mus'=>'application/vnd.musician',
 'musicxml'=>'application/vnd.recordare.musicxml+xml',
 'mvb'=>'application/x-msmediaview',
 'mwf'=>'application/vnd.mfer',
 'mxf'=>'application/mxf',
 'mxl'=>'application/vnd.recordare.musicxml',
 'mxml'=>'application/xv+xml',
 'mxs'=>'application/vnd.triscape.mxs',
 'mxu'=>'video/vnd.mpegurl',
 'n-gage'=>'application/vnd.nokia.n-gage.symbian.install',
 'n3'=>'text/n3',
 'nb'=>'application/mathematica',
 'nbp'=>'application/vnd.wolfram.player',
 'nc'=>'application/x-netcdf',
 'ncx'=>'application/x-dtbncx+xml',
 'nfo'=>'text/x-nfo',
 'ngdat'=>'application/vnd.nokia.n-gage.data',
 'nitf'=>'application/vnd.nitf',
 'nlu'=>'application/vnd.neurolanguage.nlu',
 'nml'=>'application/vnd.enliven',
 'nnd'=>'application/vnd.noblenet-directory',
 'nns'=>'application/vnd.noblenet-sealer',
 'nnw'=>'application/vnd.noblenet-web',
 'npx'=>'image/vnd.net-fpx',
 'nsc'=>'application/x-conference',
 'nsf'=>'application/vnd.lotus-notes',
 'ntf'=>'application/vnd.nitf',
 'nzb'=>'application/x-nzb',
 'oa2'=>'application/vnd.fujitsu.oasys2',
 'oa3'=>'application/vnd.fujitsu.oasys3',
 'oas'=>'application/vnd.fujitsu.oasys',
 'obd'=>'application/x-msbinder',
 'obj'=>'application/x-tgif',
 'oda'=>'application/oda',
 'odb'=>'application/vnd.oasis.opendocument.database',
 'odc'=>'application/vnd.oasis.opendocument.chart',
 'odf'=>'application/vnd.oasis.opendocument.formula',
 'odft'=>'application/vnd.oasis.opendocument.formula-template',
 'odg'=>'application/vnd.oasis.opendocument.graphics',
 'odi'=>'application/vnd.oasis.opendocument.image',
 'odm'=>'application/vnd.oasis.opendocument.text-master',
 'odp'=>'application/vnd.oasis.opendocument.presentation',
 'ods'=>'application/vnd.oasis.opendocument.spreadsheet',
 'odt'=>'application/vnd.oasis.opendocument.text',
 'oga'=>'audio/ogg',
 'ogg'=>array('application/ogg','audio/ogg'),
 'ogv'=>'video/ogg',
 'ogx'=>'application/ogg',
 'omdoc'=>'application/omdoc+xml',
 'onepkg'=>'application/onenote',
 'onetmp'=>'application/onenote',
 'onetoc'=>'application/onenote',
 'onetoc2'=>'application/onenote',
 'opf'=>'application/oebps-package+xml',
 'opml'=>'text/x-opml',
 'oprc'=>'application/vnd.palm',
 'org'=>'application/vnd.lotus-organizer',
 'osf'=>'application/vnd.yamaha.openscoreformat',
 'osfpvg'=>'application/vnd.yamaha.openscoreformat.osfpvg+xml',
 'otc'=>'application/vnd.oasis.opendocument.chart-template',
 'otf'=>'application/x-font-otf',
 'otg'=>'application/vnd.oasis.opendocument.graphics-template',
 'oth'=>'application/vnd.oasis.opendocument.text-web',
 'oti'=>'application/vnd.oasis.opendocument.image-template',
 'otp'=>'application/vnd.oasis.opendocument.presentation-template',
 'ots'=>'application/vnd.oasis.opendocument.spreadsheet-template',
 'ott'=>'application/vnd.oasis.opendocument.text-template',
 'oxps'=>'application/oxps',
 'oxt'=>'application/vnd.openofficeorg.extension',
 'p'=>'text/x-pascal',
 'p10'=>'application/pkcs10',
 'p12'=>'application/x-pkcs12',
 'p7b'=>'application/x-pkcs7-certificates',
 'p7c'=>'application/pkcs7-mime',
 'p7m'=>'application/pkcs7-mime',
 'p7r'=>'application/x-pkcs7-certreqresp',
 'p7s'=>'application/pkcs7-signature',
 'p8'=>'application/pkcs8',
 'pas'=>'text/x-pascal',
 'paw'=>'application/vnd.pawaafile',
 'pbd'=>'application/vnd.powerbuilder6',
 'pbm'=>'image/x-portable-bitmap',
 'pcap'=>'application/vnd.tcpdump.pcap',
 'pcf'=>'application/x-font-pcf',
 'pcl'=>'application/vnd.hp-pcl',
 'pclxl'=>'application/vnd.hp-pclxl',
 'pct'=>'image/x-pict',
 'pcurl'=>'application/vnd.curl.pcurl',
 'pcx'=>'image/x-pcx',
 'pdb'=>array('chemical/x-pdb','application/vnd.palm'),
 'pdf'=>'application/pdf',
 'pfa'=>'application/x-font-type1',
 'pfb'=>'application/x-font-type1',
 'pfm'=>'application/x-font-type1',
 'pfr'=>'application/font-tdpfr',
 'pfx'=>'application/x-pkcs12',
 'pgm'=>'image/x-portable-graymap',
 'pgn'=>'application/x-chess-pgn',
 'pgp'=>'application/pgp-encrypted',
 'pic'=>'image/x-pict',
 'pkg'=>'application/octet-stream',
 'pki'=>'application/pkixcmp',
 'pkipath'=>'application/pkix-pkipath',
 'plb'=>'application/vnd.3gpp.pic-bw-large',
 'plc'=>'application/vnd.mobius.plc',
 'plf'=>'application/vnd.pocketlearn',
 'pls'=>'application/pls+xml',
 'pml'=>'application/vnd.ctc-posml',
 'png'=>'image/png',
 'pnm'=>'image/x-portable-anymap',
 'portpkg'=>'application/vnd.macports.portpkg',
 'pot'=>'application/vnd.ms-powerpoint',
 'potm'=>'application/vnd.ms-powerpoint.template.macroenabled.12',
 'potx'=>'application/vnd.openxmlformats-officedocument.presentationml.template',
 'ppam'=>'application/vnd.ms-powerpoint.addin.macroenabled.12',
 'ppd'=>'application/vnd.cups-ppd',
 'ppm'=>'image/x-portable-pixmap',
 'pps'=>'application/vnd.ms-powerpoint',
 'ppsm'=>'application/vnd.ms-powerpoint.slideshow.macroenabled.12',
 'ppsx'=>'application/vnd.openxmlformats-officedocument.presentationml.slideshow',
 'ppt'=>'application/vnd.ms-powerpoint',
 'pptm'=>'application/vnd.ms-powerpoint.presentation.macroenabled.12',
 'pptx'=>'application/vnd.openxmlformats-officedocument.presentationml.presentation',
 'pqa'=>'application/vnd.palm',
 'prc'=>'application/x-mobipocket-ebook',
 'pre'=>'application/vnd.lotus-freelance',
 'prf'=>'application/pics-rules',
 'ps'=>'application/postscript',
 'psb'=>'application/vnd.3gpp.pic-bw-small',
 'psd'=>'image/vnd.adobe.photoshop',
 'psf'=>'application/x-font-linux-psf',
 'pskcxml'=>'application/pskc+xml',
 'ptid'=>'application/vnd.pvi.ptid1',
 'pub'=>'application/x-mspublisher',
 'pvb'=>'application/vnd.3gpp.pic-bw-var',
 'pwn'=>'application/vnd.3m.post-it-notes',
 'pya'=>'audio/vnd.ms-playready.media.pya',
 'pyv'=>'video/vnd.ms-playready.media.pyv',
 'qam'=>'application/vnd.epson.quickanime',
 'qbo'=>'application/vnd.intu.qbo',
 'qfx'=>'application/vnd.intu.qfx',
 'qps'=>'application/vnd.publishare-delta-tree',
 'qt'=>'video/quicktime',
 'qwd'=>'application/vnd.quark.quarkxpress',
 'qwt'=>'application/vnd.quark.quarkxpress',
 'qxb'=>'application/vnd.quark.quarkxpress',
 'qxd'=>'application/vnd.quark.quarkxpress',
 'qxl'=>'application/vnd.quark.quarkxpress',
 'qxt'=>'application/vnd.quark.quarkxpress',
 'ra'=>'audio/x-pn-realaudio',
 'ram'=>'audio/x-pn-realaudio',
 'rar'=>'application/x-rar-compressed',
 'ras'=>'image/x-cmu-raster',
 'rcprofile'=>'application/vnd.ipunplugged.rcprofile',
 'rdf'=>'application/rdf+xml',
 'rdz'=>'application/vnd.data-vision.rdz',
 'rep'=>'application/vnd.businessobjects',
 'res'=>'application/x-dtbresource+xml',
 'rgb'=>'image/x-rgb',
 'rif'=>'application/reginfo+xml',
 'rip'=>'audio/vnd.rip',
 'ris'=>'application/x-research-info-systems',
 'rl'=>'application/resource-lists+xml',
 'rlc'=>'image/vnd.fujixerox.edmics-rlc',
 'rld'=>'application/resource-lists-diff+xml',
 'rm'=>'application/vnd.rn-realmedia',
 'rmi'=>'audio/midi',
 'rmp'=>'audio/x-pn-realaudio-plugin',
 'rms'=>'application/vnd.jcp.javame.midlet-rms',
 'rmvb'=>'application/vnd.rn-realmedia-vbr',
 'rnc'=>'application/relax-ng-compact-syntax',
 'roa'=>'application/rpki-roa',
 'roff'=>array('application/x-troff','text/troff'),
 'rp9'=>'application/vnd.cloanto.rp9',
 'rpss'=>'application/vnd.nokia.radio-presets',
 'rpst'=>'application/vnd.nokia.radio-preset',
 'rq'=>'application/sparql-query',
 'rs'=>'application/rls-services+xml',
 'rsd'=>'application/rsd+xml',
 'rss'=>'application/rss+xml',
 'rtf'=>array('text/rtf','application/rtf'),
 'rtx'=>'text/richtext',
 's'=>'text/x-asm',
 's3m'=>'audio/s3m',
 'saf'=>'application/vnd.yamaha.smaf-audio',
 'sbml'=>'application/sbml+xml',
 'sc'=>'application/vnd.ibm.secure-container',
 'scd'=>'application/x-msschedule',
 'scm'=>'application/vnd.lotus-screencam',
 'scq'=>'application/scvp-cv-request',
 'scs'=>'application/scvp-cv-response',
 'scurl'=>'text/vnd.curl.scurl',
 'sda'=>'application/vnd.stardivision.draw',
 'sdc'=>'application/vnd.stardivision.calc',
 'sdd'=>'application/vnd.stardivision.impress',
 'sdkd'=>'application/vnd.solent.sdkm+xml',
 'sdkm'=>'application/vnd.solent.sdkm+xml',
 'sdp'=>'application/sdp',
 'sdw'=>'application/vnd.stardivision.writer',
 'see'=>'application/vnd.seemail',
 'seed'=>'application/vnd.fdsn.seed',
 'sema'=>'application/vnd.sema',
 'semd'=>'application/vnd.semd',
 'semf'=>'application/vnd.semf',
 'ser'=>'application/java-serialized-object',
 'setpay'=>'application/set-payment-initiation',
 'setreg'=>'application/set-registration-initiation',
 'sfd-hdstx'=>'application/vnd.hydrostatix.sof-data',
 'sfs'=>'application/vnd.spotfire.sfs',
 'sfv'=>'text/x-sfv',
 'sgi'=>'image/sgi',
 'sgl'=>'application/vnd.stardivision.writer-global',
 'sgm'=>'text/sgml',
 'sgml'=>'text/sgml',
 'sh'=>'application/x-sh',
 'shar'=>'application/x-shar',
 'shf'=>'application/shf+xml',
 'sid'=>'image/x-mrsid-image',
 'sig'=>'application/pgp-signature',
 'sil'=>'audio/silk',
 'silo'=>'model/mesh',
 'sis'=>'application/vnd.symbian.install',
 'sisx'=>'application/vnd.symbian.install',
 'sit'=>'application/x-stuffit',
 'sitx'=>'application/x-stuffitx',
 'skd'=>array('application/x-koan','application/vnd.koan'),
 'skm'=>array('application/x-koan','application/vnd.koan'),
 'skp'=>array('application/x-koan','application/vnd.koan'),
 'skt'=>array('application/x-koan','application/vnd.koan'),
 'sldm'=>'application/vnd.ms-powerpoint.slide.macroenabled.12',
 'sldx'=>'application/vnd.openxmlformats-officedocument.presentationml.slide',
 'slt'=>'application/vnd.epson.salt',
 'sm'=>'application/vnd.stepmania.stepchart',
 'smf'=>'application/vnd.stardivision.math',
 'smi'=>array('application/smil','application/smil+xml'),
 'smil'=>array('application/smil','application/smil+xml'),
 'smv'=>'video/x-smv',
 'smzip'=>'application/vnd.stepmania.package',
 'snd'=>'audio/basic',
 'snf'=>'application/x-font-snf',
 'so'=>'application/octet-stream',
 'spc'=>'application/x-pkcs7-certificates',
 'spf'=>'application/vnd.yamaha.smaf-phrase',
 'spl'=>'application/x-futuresplash',
 'spot'=>'text/vnd.in3d.spot',
 'spp'=>'application/scvp-vp-response',
 'spq'=>'application/scvp-vp-request',
 'spx'=>'audio/ogg',
 'sql'=>'application/x-sql',
 'src'=>'application/x-wais-source',
 'srt'=>'application/x-subrip',
 'sru'=>'application/sru+xml',
 'srx'=>'application/sparql-results+xml',
 'ssdl'=>'application/ssdl+xml',
 'sse'=>'application/vnd.kodak-descriptor',
 'ssf'=>'application/vnd.epson.ssf',
 'ssml'=>'application/ssml+xml',
 'st'=>'application/vnd.sailingtracker.track',
 'stc'=>'application/vnd.sun.xml.calc.template',
 'std'=>'application/vnd.sun.xml.draw.template',
 'stf'=>'application/vnd.wt.stf',
 'sti'=>'application/vnd.sun.xml.impress.template',
 'stk'=>'application/hyperstudio',
 'stl'=>'application/vnd.ms-pki.stl',
 'str'=>'application/vnd.pg.format',
 'stw'=>'application/vnd.sun.xml.writer.template',
 'sub'=>array('image/vnd.dvb.subtitle','text/vnd.dvb.subtitle'),
 'sus'=>'application/vnd.sus-calendar',
 'susp'=>'application/vnd.sus-calendar',
 'sv4cpio'=>'application/x-sv4cpio',
 'sv4crc'=>'application/x-sv4crc',
 'svc'=>'application/vnd.dvb.service',
 'svd'=>'application/vnd.svd',
 'svg'=>'image/svg+xml',
 'svgz'=>'image/svg+xml',
 'swa'=>'application/x-director',
 'swf'=>'application/x-shockwave-flash',
 'swi'=>'application/vnd.aristanetworks.swi',
 'sxc'=>'application/vnd.sun.xml.calc',
 'sxd'=>'application/vnd.sun.xml.draw',
 'sxg'=>'application/vnd.sun.xml.writer.global',
 'sxi'=>'application/vnd.sun.xml.impress',
 'sxm'=>'application/vnd.sun.xml.math',
 'sxw'=>'application/vnd.sun.xml.writer',
 't'=>array('application/x-troff','text/troff'),
 't3'=>'application/x-t3vm-image',
 'taglet'=>'application/vnd.mynfc',
 'tao'=>'application/vnd.tao.intent-module-archive',
 'tar'=>'application/x-tar',
 'tcap'=>'application/vnd.3gpp2.tcap',
 'tcl'=>'application/x-tcl',
 'teacher'=>'application/vnd.smart.teacher',
 'tei'=>'application/tei+xml',
 'teicorpus'=>'application/tei+xml',
 'tex'=>'application/x-tex',
 'texi'=>'application/x-texinfo',
 'texinfo'=>'application/x-texinfo',
 'text'=>'text/plain',
 'tfi'=>'application/thraud+xml',
 'tfm'=>'application/x-tex-tfm',
 'tga'=>'image/x-tga',
 'thmx'=>'application/vnd.ms-officetheme',
 'tif'=>'image/tiff',
 'tiff'=>'image/tiff',
 'tmo'=>'application/vnd.tmobile-livetv',
 'torrent'=>'application/x-bittorrent',
 'tpl'=>'application/vnd.groove-tool-template',
 'tpt'=>'application/vnd.trid.tpt',
 'tr'=>array('application/x-troff','text/troff'),
 'tra'=>'application/vnd.trueapp',
 'trm'=>'application/x-msterminal',
 'tsd'=>'application/timestamped-data',
 'tsv'=>'text/tab-separated-values',
 'ttc'=>'application/x-font-ttf',
 'ttf'=>'application/x-font-ttf',
 'ttl'=>'text/turtle',
 'twd'=>'application/vnd.simtech-mindmapper',
 'twds'=>'application/vnd.simtech-mindmapper',
 'txd'=>'application/vnd.genomatix.tuxedo',
 'txf'=>'application/vnd.mobius.txf',
 'txt'=>'text/plain',
 'u32'=>'application/x-authorware-bin',
 'udeb'=>'application/x-debian-package',
 'ufd'=>'application/vnd.ufdl',
 'ufdl'=>'application/vnd.ufdl',
 'ulx'=>'application/x-glulx',
 'umj'=>'application/vnd.umajin',
 'unityweb'=>'application/vnd.unity',
 'uoml'=>'application/vnd.uoml+xml',
 'uri'=>'text/uri-list',
 'uris'=>'text/uri-list',
 'urls'=>'text/uri-list',
 'ustar'=>'application/x-ustar',
 'utz'=>'application/vnd.uiq.theme',
 'uu'=>'text/x-uuencode',
 'uva'=>'audio/vnd.dece.audio',
 'uvd'=>'application/vnd.dece.data',
 'uvf'=>'application/vnd.dece.data',
 'uvg'=>'image/vnd.dece.graphic',
 'uvh'=>'video/vnd.dece.hd',
 'uvi'=>'image/vnd.dece.graphic',
 'uvm'=>'video/vnd.dece.mobile',
 'uvp'=>'video/vnd.dece.pd',
 'uvs'=>'video/vnd.dece.sd',
 'uvt'=>'application/vnd.dece.ttml+xml',
 'uvu'=>'video/vnd.uvvu.mp4',
 'uvv'=>'video/vnd.dece.video',
 'uvva'=>'audio/vnd.dece.audio',
 'uvvd'=>'application/vnd.dece.data',
 'uvvf'=>'application/vnd.dece.data',
 'uvvg'=>'image/vnd.dece.graphic',
 'uvvh'=>'video/vnd.dece.hd',
 'uvvi'=>'image/vnd.dece.graphic',
 'uvvm'=>'video/vnd.dece.mobile',
 'uvvp'=>'video/vnd.dece.pd',
 'uvvs'=>'video/vnd.dece.sd',
 'uvvt'=>'application/vnd.dece.ttml+xml',
 'uvvu'=>'video/vnd.uvvu.mp4',
 'uvvv'=>'video/vnd.dece.video',
 'uvvx'=>'application/vnd.dece.unspecified',
 'uvvz'=>'application/vnd.dece.zip',
 'uvx'=>'application/vnd.dece.unspecified',
 'uvz'=>'application/vnd.dece.zip',
 'vcard'=>'text/vcard',
 'vcd'=>'application/x-cdlink',
 'vcf'=>'text/x-vcard',
 'vcg'=>'application/vnd.groove-vcard',
 'vcs'=>'text/x-vcalendar',
 'vcx'=>'application/vnd.vcx',
 'vis'=>'application/vnd.visionary',
 'viv'=>'video/vnd.vivo',
 'vob'=>'video/x-ms-vob',
 'vor'=>'application/vnd.stardivision.writer',
 'vox'=>'application/x-authorware-bin',
 'vrml'=>'model/vrml',
 'vsd'=>'application/vnd.visio',
 'vsf'=>'application/vnd.vsf',
 'vss'=>'application/vnd.visio',
 'vst'=>'application/vnd.visio',
 'vsw'=>'application/vnd.visio',
 'vtu'=>'model/vnd.vtu',
 'vxml'=>'application/voicexml+xml',
 'w3d'=>'application/x-director',
 'wad'=>'application/x-doom',
 'wav'=>'audio/x-wav',
 'wax'=>'audio/x-ms-wax',
 'wbmp'=>'image/vnd.wap.wbmp',
 'wbs'=>'application/vnd.criticaltools.wbs+xml',
 'wbxml'=>'application/vnd.wap.wbxml',
 'wcm'=>'application/vnd.ms-works',
 'wdb'=>'application/vnd.ms-works',
 'wdp'=>'image/vnd.ms-photo',
 'weba'=>'audio/webm',
 'webm'=>'video/webm',
 'webp'=>'image/webp',
 'wg'=>'application/vnd.pmi.widget',
 'wgt'=>'application/widget',
 'wks'=>'application/vnd.ms-works',
 'wm'=>'video/x-ms-wm',
 'wma'=>'audio/x-ms-wma',
 'wmd'=>'application/x-ms-wmd',
 'wmf'=>'application/x-msmetafile',
 'wml'=>'text/vnd.wap.wml',
 'wmlc'=>'application/vnd.wap.wmlc',
 'wmls'=>'text/vnd.wap.wmlscript',
 'wmlsc'=>'application/vnd.wap.wmlscriptc',
 'wmv'=>'video/x-ms-wmv',
 'wmx'=>'video/x-ms-wmx',
 'wmz'=>array('application/x-ms-wmz','application/x-msmetafile'),
 'woff'=>array('application/font-woff','application/x-font-woff'),
 'wpd'=>'application/vnd.wordperfect',
 'wpl'=>'application/vnd.ms-wpl',
 'wps'=>'application/vnd.ms-works',
 'wqd'=>'application/vnd.wqd',
 'wri'=>'application/x-mswrite',
 'wrl'=>'model/vrml',
 'wsdl'=>'application/wsdl+xml',
 'wspolicy'=>'application/wspolicy+xml',
 'wtb'=>'application/vnd.webturbo',
 'wvx'=>'video/x-ms-wvx',
 'x32'=>'application/x-authorware-bin',
 'x3d'=>'model/x3d+xml',
 'x3db'=>'model/x3d+binary',
 'x3dbz'=>'model/x3d+binary',
 'x3dv'=>'model/x3d+vrml',
 'x3dvz'=>'model/x3d+vrml',
 'x3dz'=>'model/x3d+xml',
 'xaml'=>'application/xaml+xml',
 'xap'=>'application/x-silverlight-app',
 'xar'=>'application/vnd.xara',
 'xbap'=>'application/x-ms-xbap',
 'xbd'=>'application/vnd.fujixerox.docuworks.binder',
 'xbm'=>'image/x-xbitmap',
 'xdf'=>'application/xcap-diff+xml',
 'xdm'=>'application/vnd.syncml.dm+xml',
 'xdp'=>'application/vnd.adobe.xdp+xml',
 'xdssc'=>'application/dssc+xml',
 'xdw'=>'application/vnd.fujixerox.docuworks',
 'xenc'=>'application/xenc+xml',
 'xer'=>'application/patch-ops-error+xml',
 'xfdf'=>'application/vnd.adobe.xfdf',
 'xfdl'=>'application/vnd.xfdl',
 'xht'=>'application/xhtml+xml',
 'xhtml'=>'application/xhtml+xml',
 'xhvml'=>'application/xv+xml',
 'xif'=>'image/vnd.xiff',
 'xla'=>'application/vnd.ms-excel',
 'xlam'=>'application/vnd.ms-excel.addin.macroenabled.12',
 'xlc'=>'application/vnd.ms-excel',
 'xlf'=>'application/x-xliff+xml',
 'xlm'=>'application/vnd.ms-excel',
 'xls'=>array('application/vnd.ms-excel','application/vnd.ms-office'),
 'xlsb'=>'application/vnd.ms-excel.sheet.binary.macroenabled.12',
 'xlsm'=>'application/vnd.ms-excel.sheet.macroenabled.12',
 'xlsx'=>'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
 'xlt'=>'application/vnd.ms-excel',
 'xltm'=>'application/vnd.ms-excel.template.macroenabled.12',
 'xltx'=>'application/vnd.openxmlformats-officedocument.spreadsheetml.template',
 'xlw'=>'application/vnd.ms-excel',
 'xm'=>'audio/xm',
 'xml'=>'application/xml',
 'xo'=>'application/vnd.olpc-sugar',
 'xop'=>'application/xop+xml',
 'xpi'=>'application/x-xpinstall',
 'xpl'=>'application/xproc+xml',
 'xpm'=>'image/x-xpixmap',
 'xpr'=>'application/vnd.is-xpr',
 'xps'=>'application/vnd.ms-xpsdocument',
 'xpw'=>'application/vnd.intercon.formnet',
 'xpx'=>'application/vnd.intercon.formnet',
 'xsl'=>'application/xml',
 'xslt'=>'application/xslt+xml',
 'xsm'=>'application/vnd.syncml+xml',
 'xspf'=>'application/xspf+xml',
 'xul'=>'application/vnd.mozilla.xul+xml',
 'xvm'=>'application/xv+xml',
 'xvml'=>'application/xv+xml',
 'xwd'=>'image/x-xwindowdump',
 'xyz'=>'chemical/x-xyz',
 'xz'=>'application/x-xz',
 'yang'=>'application/yang',
 'yin'=>'application/yin+xml',
 'z1'=>'application/x-zmachine',
 'z2'=>'application/x-zmachine',
 'z3'=>'application/x-zmachine',
 'z4'=>'application/x-zmachine',
 'z5'=>'application/x-zmachine',
 'z6'=>'application/x-zmachine',
 'z7'=>'application/x-zmachine',
 'z8'=>'application/x-zmachine',
 'zaz'=>'application/vnd.zzazz.deck+xml',
 'zip'=>array('application/zip','application/x-zip'),
 'zir'=>'application/vnd.zul',
 'zirz'=>'application/vnd.zul',
 'zmm'=>'application/vnd.handheld-entertainment+xml',
 '123'=>'application/vnd.lotus-1-2-3'
 );
}
global $MIME_TYP_EXT;
if (empty($MIME_TYP_EXT)) { $MIME_TYP_EXT=array(
 'application/andrew-inset'=>'ez',
 'application/applixware'=>'aw',
 'application/atom+xml'=>'atom',
 'application/atomcat+xml'=>'atomcat',
 'application/atomsvc+xml'=>'atomsvc',
 'application/ccxml+xml'=>'ccxml',
 'application/cdmi-capability'=>'cdmia',
 'application/cdmi-container'=>'cdmic',
 'application/cdmi-domain'=>'cdmid',
 'application/cdmi-object'=>'cdmio',
 'application/cdmi-queue'=>'cdmiq',
 'application/cu-seeme'=>'cu',
 'application/davmount+xml'=>'davmount',
 'application/docbook+xml'=>'dbk',
 'application/dssc+der'=>'dssc',
 'application/dssc+xml'=>'xdssc',
 'application/ecmascript'=>'ecma',
 'application/emma+xml'=>'emma',
 'application/epub+zip'=>'epub',
 'application/exi'=>'exi',
 'application/font-tdpfr'=>'pfr',
 'application/font-woff'=>'woff',
 'application/gml+xml'=>'gml',
 'application/gpx+xml'=>'gpx',
 'application/gxf'=>'gxf',
 'application/hyperstudio'=>'stk',
 'application/inkml+xml'=>array('ink','inkml'),
 'application/ipfix'=>'ipfix',
 'application/java-archive'=>'jar',
 'application/java-serialized-object'=>'ser',
 'application/java-vm'=>'class',
 'application/javascript'=>'js',
 'application/json'=>'json',
 'application/jsonml+json'=>'jsonml',
 'application/lost+xml'=>'lostxml',
 'application/mac-binhex40'=>'hqx',
 'application/mac-compactpro'=>'cpt',
 'application/mads+xml'=>'mads',
 'application/marc'=>'mrc',
 'application/marcxml+xml'=>'mrcx',
 'application/mathematica'=>array('ma','nb','mb'),
 'application/mathml+xml'=>'mathml',
 'application/mbox'=>'mbox',
 'application/mediaservercontrol+xml'=>'mscml',
 'application/metalink+xml'=>'metalink',
 'application/metalink4+xml'=>'meta4',
 'application/mets+xml'=>'mets',
 'application/mods+xml'=>'mods',
 'application/mp21'=>array('m21','mp21'),
 'application/mp4'=>'mp4s',
 'application/msword'=>array('doc','dot'),
 'application/mxf'=>'mxf',
 'application/octet-stream'=>array('','bin','class','dll','dmg','dms','exe','lha','lzh','so','lrf','mar','dist','distz','pkg','bpk','dump','elc','deploy'),
 'application/oda'=>'oda',
 'application/oebps-package+xml'=>'opf',
 'application/ogg'=>array('ogg','ogx'),
 'application/omdoc+xml'=>'omdoc',
 'application/onenote'=>array('onetoc','onetoc2','onetmp','onepkg'),
 'application/oxps'=>'oxps',
 'application/patch-ops-error+xml'=>'xer',
 'application/pdf'=>'pdf',
 'application/pgp-encrypted'=>'pgp',
 'application/pgp-signature'=>array('asc','sig'),
 'application/pics-rules'=>'prf',
 'application/pkcs10'=>'p10',
 'application/pkcs7-mime'=>array('p7m','p7c'),
 'application/pkcs7-signature'=>'p7s',
 'application/pkcs8'=>'p8',
 'application/pkix-attr-cert'=>'ac',
 'application/pkix-cert'=>'cer',
 'application/pkix-crl'=>'crl',
 'application/pkix-pkipath'=>'pkipath',
 'application/pkixcmp'=>'pki',
 'application/pls+xml'=>'pls',
 'application/postscript'=>array('ps','ai','eps'),
 'application/prs.cww'=>'cww',
 'application/pskc+xml'=>'pskcxml',
 'application/rdf+xml'=>'rdf',
 'application/reginfo+xml'=>'rif',
 'application/relax-ng-compact-syntax'=>'rnc',
 'application/resource-lists+xml'=>'rl',
 'application/resource-lists-diff+xml'=>'rld',
 'application/rls-services+xml'=>'rs',
 'application/rpki-ghostbusters'=>'gbr',
 'application/rpki-manifest'=>'mft',
 'application/rpki-roa'=>'roa',
 'application/rsd+xml'=>'rsd',
 'application/rss+xml'=>'rss',
 'application/rtf'=>'rtf',
 'application/sbml+xml'=>'sbml',
 'application/scvp-cv-request'=>'scq',
 'application/scvp-cv-response'=>'scs',
 'application/scvp-vp-request'=>'spq',
 'application/scvp-vp-response'=>'spp',
 'application/sdp'=>'sdp',
 'application/set-payment-initiation'=>'setpay',
 'application/set-registration-initiation'=>'setreg',
 'application/shf+xml'=>'shf',
 'application/smil'=>array('smi','smil'),
 'application/smil+xml'=>array('smi','smil'),
 'application/sparql-query'=>'rq',
 'application/sparql-results+xml'=>'srx',
 'application/srgs'=>'gram',
 'application/srgs+xml'=>'grxml',
 'application/sru+xml'=>'sru',
 'application/ssdl+xml'=>'ssdl',
 'application/ssml+xml'=>'ssml',
 'application/tei+xml'=>array('tei','teicorpus'),
 'application/thraud+xml'=>'tfi',
 'application/timestamped-data'=>'tsd',
 'application/vnd.3gpp.pic-bw-large'=>'plb',
 'application/vnd.3gpp.pic-bw-small'=>'psb',
 'application/vnd.3gpp.pic-bw-var'=>'pvb',
 'application/vnd.3gpp2.tcap'=>'tcap',
 'application/vnd.3m.post-it-notes'=>'pwn',
 'application/vnd.accpac.simply.aso'=>'aso',
 'application/vnd.accpac.simply.imp'=>'imp',
 'application/vnd.acucobol'=>'acu',
 'application/vnd.acucorp'=>array('atc','acutc'),
 'application/vnd.adobe.air-application-installer-package+zip'=>'air',
 'application/vnd.adobe.formscentral.fcdt'=>'fcdt',
 'application/vnd.adobe.fxp'=>array('fxp','fxpl'),
 'application/vnd.adobe.xdp+xml'=>'xdp',
 'application/vnd.adobe.xfdf'=>'xfdf',
 'application/vnd.ahead.space'=>'ahead',
 'application/vnd.airzip.filesecure.azf'=>'azf',
 'application/vnd.airzip.filesecure.azs'=>'azs',
 'application/vnd.amazon.ebook'=>'azw',
 'application/vnd.americandynamics.acc'=>'acc',
 'application/vnd.amiga.ami'=>'ami',
 'application/vnd.android.package-archive'=>'apk',
 'application/vnd.anser-web-certificate-issue-initiation'=>'cii',
 'application/vnd.anser-web-funds-transfer-initiation'=>'fti',
 'application/vnd.antix.game-component'=>'atx',
 'application/vnd.apple.installer+xml'=>'mpkg',
 'application/vnd.apple.mpegurl'=>'m3u8',
 'application/vnd.aristanetworks.swi'=>'swi',
 'application/vnd.astraea-software.iota'=>'iota',
 'application/vnd.audiograph'=>'aep',
 'application/vnd.blueice.multipass'=>'mpm',
 'application/vnd.bmi'=>'bmi',
 'application/vnd.businessobjects'=>'rep',
 'application/vnd.chemdraw+xml'=>'cdxml',
 'application/vnd.chipnuts.karaoke-mmd'=>'mmd',
 'application/vnd.cinderella'=>'cdy',
 'application/vnd.claymore'=>'cla',
 'application/vnd.cloanto.rp9'=>'rp9',
 'application/vnd.clonk.c4group'=>array('c4g','c4d','c4f','c4p','c4u'),
 'application/vnd.cluetrust.cartomobile-config'=>'c11amc',
 'application/vnd.cluetrust.cartomobile-config-pkg'=>'c11amz',
 'application/vnd.commonspace'=>'csp',
 'application/vnd.contact.cmsg'=>'cdbcmsg',
 'application/vnd.cosmocaller'=>'cmc',
 'application/vnd.crick.clicker'=>'clkx',
 'application/vnd.crick.clicker.keyboard'=>'clkk',
 'application/vnd.crick.clicker.palette'=>'clkp',
 'application/vnd.crick.clicker.template'=>'clkt',
 'application/vnd.crick.clicker.wordbank'=>'clkw',
 'application/vnd.criticaltools.wbs+xml'=>'wbs',
 'application/vnd.ctc-posml'=>'pml',
 'application/vnd.cups-ppd'=>'ppd',
 'application/vnd.curl.car'=>'car',
 'application/vnd.curl.pcurl'=>'pcurl',
 'application/vnd.dart'=>'dart',
 'application/vnd.data-vision.rdz'=>'rdz',
 'application/vnd.dece.data'=>array('uvf','uvvf','uvd','uvvd'),
 'application/vnd.dece.ttml+xml'=>array('uvt','uvvt'),
 'application/vnd.dece.unspecified'=>array('uvx','uvvx'),
 'application/vnd.dece.zip'=>array('uvz','uvvz'),
 'application/vnd.denovo.fcselayout-link'=>'fe_launch',
 'application/vnd.dna'=>'dna',
 'application/vnd.dolby.mlp'=>'mlp',
 'application/vnd.dpgraph'=>'dpg',
 'application/vnd.dreamfactory'=>'dfac',
 'application/vnd.ds-keypoint'=>'kpxx',
 'application/vnd.dvb.ait'=>'ait',
 'application/vnd.dvb.service'=>'svc',
 'application/vnd.dynageo'=>'geo',
 'application/vnd.ecowin.chart'=>'mag',
 'application/vnd.enliven'=>'nml',
 'application/vnd.epson.esf'=>'esf',
 'application/vnd.epson.msf'=>'msf',
 'application/vnd.epson.quickanime'=>'qam',
 'application/vnd.epson.salt'=>'slt',
 'application/vnd.epson.ssf'=>'ssf',
 'application/vnd.eszigno3+xml'=>array('es3','et3'),
 'application/vnd.ezpix-album'=>'ez2',
 'application/vnd.ezpix-package'=>'ez3',
 'application/vnd.fdf'=>'fdf',
 'application/vnd.fdsn.mseed'=>'mseed',
 'application/vnd.fdsn.seed'=>array('seed','dataless'),
 'application/vnd.flographit'=>'gph',
 'application/vnd.fluxtime.clip'=>'ftc',
 'application/vnd.framemaker'=>array('fm','frame','maker','book'),
 'application/vnd.frogans.fnc'=>'fnc',
 'application/vnd.frogans.ltf'=>'ltf',
 'application/vnd.fsc.weblaunch'=>'fsc',
 'application/vnd.fujitsu.oasys'=>'oas',
 'application/vnd.fujitsu.oasys2'=>'oa2',
 'application/vnd.fujitsu.oasys3'=>'oa3',
 'application/vnd.fujitsu.oasysgp'=>'fg5',
 'application/vnd.fujitsu.oasysprs'=>'bh2',
 'application/vnd.fujixerox.ddd'=>'ddd',
 'application/vnd.fujixerox.docuworks'=>'xdw',
 'application/vnd.fujixerox.docuworks.binder'=>'xbd',
 'application/vnd.fuzzysheet'=>'fzs',
 'application/vnd.genomatix.tuxedo'=>'txd',
 'application/vnd.geogebra.file'=>'ggb',
 'application/vnd.geogebra.tool'=>'ggt',
 'application/vnd.geometry-explorer'=>array('gex','gre'),
 'application/vnd.geonext'=>'gxt',
 'application/vnd.geoplan'=>'g2w',
 'application/vnd.geospace'=>'g3w',
 'application/vnd.gmx'=>'gmx',
 'application/vnd.google-earth.kml+xml'=>'kml',
 'application/vnd.google-earth.kmz'=>'kmz',
 'application/vnd.grafeq'=>array('gqf','gqs'),
 'application/vnd.groove-account'=>'gac',
 'application/vnd.groove-help'=>'ghf',
 'application/vnd.groove-identity-message'=>'gim',
 'application/vnd.groove-injector'=>'grv',
 'application/vnd.groove-tool-message'=>'gtm',
 'application/vnd.groove-tool-template'=>'tpl',
 'application/vnd.groove-vcard'=>'vcg',
 'application/vnd.hal+xml'=>'hal',
 'application/vnd.handheld-entertainment+xml'=>'zmm',
 'application/vnd.hbci'=>'hbci',
 'application/vnd.hhe.lesson-player'=>'les',
 'application/vnd.hp-hpgl'=>'hpgl',
 'application/vnd.hp-hpid'=>'hpid',
 'application/vnd.hp-hps'=>'hps',
 'application/vnd.hp-jlyt'=>'jlt',
 'application/vnd.hp-pcl'=>'pcl',
 'application/vnd.hp-pclxl'=>'pclxl',
 'application/vnd.hydrostatix.sof-data'=>'sfd-hdstx',
 'application/vnd.ibm.minipay'=>'mpy',
 'application/vnd.ibm.modcap'=>array('afp','listafp','list3820'),
 'application/vnd.ibm.rights-management'=>'irm',
 'application/vnd.ibm.secure-container'=>'sc',
 'application/vnd.iccprofile'=>array('icc','icm'),
 'application/vnd.igloader'=>'igl',
 'application/vnd.immervision-ivp'=>'ivp',
 'application/vnd.immervision-ivu'=>'ivu',
 'application/vnd.insors.igm'=>'igm',
 'application/vnd.intercon.formnet'=>array('xpw','xpx'),
 'application/vnd.intergeo'=>'i2g',
 'application/vnd.intu.qbo'=>'qbo',
 'application/vnd.intu.qfx'=>'qfx',
 'application/vnd.ipunplugged.rcprofile'=>'rcprofile',
 'application/vnd.irepository.package+xml'=>'irp',
 'application/vnd.is-xpr'=>'xpr',
 'application/vnd.isac.fcs'=>'fcs',
 'application/vnd.jam'=>'jam',
 'application/vnd.jcp.javame.midlet-rms'=>'rms',
 'application/vnd.jisp'=>'jisp',
 'application/vnd.joost.joda-archive'=>'joda',
 'application/vnd.kahootz'=>array('ktz','ktr'),
 'application/vnd.kde.karbon'=>'karbon',
 'application/vnd.kde.kchart'=>'chrt',
 'application/vnd.kde.kformula'=>'kfo',
 'application/vnd.kde.kivio'=>'flw',
 'application/vnd.kde.kontour'=>'kon',
 'application/vnd.kde.kpresenter'=>array('kpr','kpt'),
 'application/vnd.kde.kspread'=>'ksp',
 'application/vnd.kde.kword'=>array('kwd','kwt'),
 'application/vnd.kenameaapp'=>'htke',
 'application/vnd.kidspiration'=>'kia',
 'application/vnd.kinar'=>array('kne','knp'),
 'application/vnd.koan'=>array('skp','skd','skt','skm'),
 'application/vnd.kodak-descriptor'=>'sse',
 'application/vnd.las.las+xml'=>'lasxml',
 'application/vnd.llamagraphics.life-balance.desktop'=>'lbd',
 'application/vnd.llamagraphics.life-balance.exchange+xml'=>'lbe',
 'application/vnd.lotus-1-2-3'=>'123',
 'application/vnd.lotus-approach'=>'apr',
 'application/vnd.lotus-freelance'=>'pre',
 'application/vnd.lotus-notes'=>'nsf',
 'application/vnd.lotus-organizer'=>'org',
 'application/vnd.lotus-screencam'=>'scm',
 'application/vnd.lotus-wordpro'=>'lwp',
 'application/vnd.macports.portpkg'=>'portpkg',
 'application/vnd.mcd'=>'mcd',
 'application/vnd.medcalcdata'=>'mc1',
 'application/vnd.mediastation.cdkey'=>'cdkey',
 'application/vnd.mfer'=>'mwf',
 'application/vnd.mfmp'=>'mfm',
 'application/vnd.micrografx.flo'=>'flo',
 'application/vnd.micrografx.igx'=>'igx',
 'application/vnd.mif'=>'mif',
 'application/vnd.mobius.daf'=>'daf',
 'application/vnd.mobius.dis'=>'dis',
 'application/vnd.mobius.mbk'=>'mbk',
 'application/vnd.mobius.mqy'=>'mqy',
 'application/vnd.mobius.msl'=>'msl',
 'application/vnd.mobius.plc'=>'plc',
 'application/vnd.mobius.txf'=>'txf',
 'application/vnd.mophun.application'=>'mpn',
 'application/vnd.mophun.certificate'=>'mpc',
 'application/vnd.mozilla.xul+xml'=>'xul',
 'application/vnd.ms-artgalry'=>'cil',
 'application/vnd.ms-cab-compressed'=>'cab',
 'application/vnd.ms-excel'=>array('xls','xlm','xla','xlc','xlt','xlw'),
 'application/vnd.ms-excel.addin.macroenabled.12'=>'xlam',
 'application/vnd.ms-excel.sheet.binary.macroenabled.12'=>'xlsb',
 'application/vnd.ms-excel.sheet.macroenabled.12'=>'xlsm',
 'application/vnd.ms-excel.template.macroenabled.12'=>'xltm',
 'application/vnd.ms-fontobject'=>'eot',
 'application/vnd.ms-htmlhelp'=>'chm',
 'application/vnd.ms-ims'=>'ims',
 'application/vnd.ms-lrm'=>'lrm',
 'application/vnd.ms-office'=>'xls',
 'application/vnd.ms-officetheme'=>'thmx',
 'application/vnd.ms-pki.seccat'=>'cat',
 'application/vnd.ms-pki.stl'=>'stl',
 'application/vnd.ms-powerpoint'=>array('ppt','pps','pot'),
 'application/vnd.ms-powerpoint.addin.macroenabled.12'=>'ppam',
 'application/vnd.ms-powerpoint.presentation.macroenabled.12'=>'pptm',
 'application/vnd.ms-powerpoint.slide.macroenabled.12'=>'sldm',
 'application/vnd.ms-powerpoint.slideshow.macroenabled.12'=>'ppsm',
 'application/vnd.ms-powerpoint.template.macroenabled.12'=>'potm',
 'application/vnd.ms-project'=>array('mpp','mpt'),
 'application/vnd.ms-word.document.macroenabled.12'=>'docm',
 'application/vnd.ms-word.template.macroenabled.12'=>'dotm',
 'application/vnd.ms-works'=>array('wps','wks','wcm','wdb'),
 'application/vnd.ms-wpl'=>'wpl',
 'application/vnd.ms-xpsdocument'=>'xps',
 'application/vnd.mseq'=>'mseq',
 'application/vnd.musician'=>'mus',
 'application/vnd.muvee.style'=>'msty',
 'application/vnd.mynfc'=>'taglet',
 'application/vnd.neurolanguage.nlu'=>'nlu',
 'application/vnd.nitf'=>array('ntf','nitf'),
 'application/vnd.noblenet-directory'=>'nnd',
 'application/vnd.noblenet-sealer'=>'nns',
 'application/vnd.noblenet-web'=>'nnw',
 'application/vnd.nokia.n-gage.data'=>'ngdat',
 'application/vnd.nokia.n-gage.symbian.install'=>'n-gage',
 'application/vnd.nokia.radio-preset'=>'rpst',
 'application/vnd.nokia.radio-presets'=>'rpss',
 'application/vnd.novadigm.edm'=>'edm',
 'application/vnd.novadigm.edx'=>'edx',
 'application/vnd.novadigm.ext'=>'ext',
 'application/vnd.oasis.opendocument.chart'=>'odc',
 'application/vnd.oasis.opendocument.chart-template'=>'otc',
 'application/vnd.oasis.opendocument.database'=>'odb',
 'application/vnd.oasis.opendocument.formula'=>'odf',
 'application/vnd.oasis.opendocument.formula-template'=>'odft',
 'application/vnd.oasis.opendocument.graphics'=>'odg',
 'application/vnd.oasis.opendocument.graphics-template'=>'otg',
 'application/vnd.oasis.opendocument.image'=>'odi',
 'application/vnd.oasis.opendocument.image-template'=>'oti',
 'application/vnd.oasis.opendocument.presentation'=>'odp',
 'application/vnd.oasis.opendocument.presentation-template'=>'otp',
 'application/vnd.oasis.opendocument.spreadsheet'=>'ods',
 'application/vnd.oasis.opendocument.spreadsheet-template'=>'ots',
 'application/vnd.oasis.opendocument.text'=>'odt',
 'application/vnd.oasis.opendocument.text-master'=>'odm',
 'application/vnd.oasis.opendocument.text-template'=>'ott',
 'application/vnd.oasis.opendocument.text-web'=>'oth',
 'application/vnd.olpc-sugar'=>'xo',
 'application/vnd.oma.dd2+xml'=>'dd2',
 'application/vnd.openofficeorg.extension'=>'oxt',
 'application/vnd.openxmlformats-officedocument.presentationml.presentation'=>'pptx',
 'application/vnd.openxmlformats-officedocument.presentationml.slide'=>'sldx',
 'application/vnd.openxmlformats-officedocument.presentationml.slideshow'=>'ppsx',
 'application/vnd.openxmlformats-officedocument.presentationml.template'=>'potx',
 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'=>'xlsx',
 'application/vnd.openxmlformats-officedocument.spreadsheetml.template'=>'xltx',
 'application/vnd.openxmlformats-officedocument.wordprocessingml.document'=>'docx',
 'application/vnd.openxmlformats-officedocument.wordprocessingml.template'=>'dotx',
 'application/vnd.osgeo.mapguide.package'=>'mgp',
 'application/vnd.osgi.dp'=>'dp',
 'application/vnd.osgi.subsystem'=>'esa',
 'application/vnd.palm'=>array('pdb','pqa','oprc'),
 'application/vnd.pawaafile'=>'paw',
 'application/vnd.pg.format'=>'str',
 'application/vnd.pg.osasli'=>'ei6',
 'application/vnd.picsel'=>'efif',
 'application/vnd.pmi.widget'=>'wg',
 'application/vnd.pocketlearn'=>'plf',
 'application/vnd.powerbuilder6'=>'pbd',
 'application/vnd.previewsystems.box'=>'box',
 'application/vnd.proteus.magazine'=>'mgz',
 'application/vnd.publishare-delta-tree'=>'qps',
 'application/vnd.pvi.ptid1'=>'ptid',
 'application/vnd.quark.quarkxpress'=>array('qxd','qxt','qwd','qwt','qxl','qxb'),
 'application/vnd.realvnc.bed'=>'bed',
 'application/vnd.recordare.musicxml'=>'mxl',
 'application/vnd.recordare.musicxml+xml'=>'musicxml',
 'application/vnd.rig.cryptonote'=>'cryptonote',
 'application/vnd.rim.cod'=>'cod',
 'application/vnd.rn-realmedia'=>'rm',
 'application/vnd.rn-realmedia-vbr'=>'rmvb',
 'application/vnd.route66.link66+xml'=>'link66',
 'application/vnd.sailingtracker.track'=>'st',
 'application/vnd.seemail'=>'see',
 'application/vnd.sema'=>'sema',
 'application/vnd.semd'=>'semd',
 'application/vnd.semf'=>'semf',
 'application/vnd.shana.informed.formdata'=>'ifm',
 'application/vnd.shana.informed.formtemplate'=>'itp',
 'application/vnd.shana.informed.interchange'=>'iif',
 'application/vnd.shana.informed.package'=>'ipk',
 'application/vnd.simtech-mindmapper'=>array('twd','twds'),
 'application/vnd.smaf'=>'mmf',
 'application/vnd.smart.teacher'=>'teacher',
 'application/vnd.solent.sdkm+xml'=>array('sdkm','sdkd'),
 'application/vnd.spotfire.dxp'=>'dxp',
 'application/vnd.spotfire.sfs'=>'sfs',
 'application/vnd.stardivision.calc'=>'sdc',
 'application/vnd.stardivision.draw'=>'sda',
 'application/vnd.stardivision.impress'=>'sdd',
 'application/vnd.stardivision.math'=>'smf',
 'application/vnd.stardivision.writer'=>array('sdw','vor'),
 'application/vnd.stardivision.writer-global'=>'sgl',
 'application/vnd.stepmania.package'=>'smzip',
 'application/vnd.stepmania.stepchart'=>'sm',
 'application/vnd.sun.xml.calc'=>'sxc',
 'application/vnd.sun.xml.calc.template'=>'stc',
 'application/vnd.sun.xml.draw'=>'sxd',
 'application/vnd.sun.xml.draw.template'=>'std',
 'application/vnd.sun.xml.impress'=>'sxi',
 'application/vnd.sun.xml.impress.template'=>'sti',
 'application/vnd.sun.xml.math'=>'sxm',
 'application/vnd.sun.xml.writer'=>'sxw',
 'application/vnd.sun.xml.writer.global'=>'sxg',
 'application/vnd.sun.xml.writer.template'=>'stw',
 'application/vnd.sus-calendar'=>array('sus','susp'),
 'application/vnd.svd'=>'svd',
 'application/vnd.symbian.install'=>array('sis','sisx'),
 'application/vnd.syncml+xml'=>'xsm',
 'application/vnd.syncml.dm+wbxml'=>'bdm',
 'application/vnd.syncml.dm+xml'=>'xdm',
 'application/vnd.tao.intent-module-archive'=>'tao',
 'application/vnd.tcpdump.pcap'=>array('pcap','cap','dmp'),
 'application/vnd.tmobile-livetv'=>'tmo',
 'application/vnd.trid.tpt'=>'tpt',
 'application/vnd.triscape.mxs'=>'mxs',
 'application/vnd.trueapp'=>'tra',
 'application/vnd.ufdl'=>array('ufd','ufdl'),
 'application/vnd.uiq.theme'=>'utz',
 'application/vnd.umajin'=>'umj',
 'application/vnd.unity'=>'unityweb',
 'application/vnd.uoml+xml'=>'uoml',
 'application/vnd.vcx'=>'vcx',
 'application/vnd.visio'=>array('vsd','vst','vss','vsw'),
 'application/vnd.visionary'=>'vis',
 'application/vnd.vsf'=>'vsf',
 'application/vnd.wap.wbxml'=>'wbxml',
 'application/vnd.wap.wmlc'=>'wmlc',
 'application/vnd.wap.wmlscriptc'=>'wmlsc',
 'application/vnd.webturbo'=>'wtb',
 'application/vnd.wolfram.player'=>'nbp',
 'application/vnd.wordperfect'=>'wpd',
 'application/vnd.wqd'=>'wqd',
 'application/vnd.wt.stf'=>'stf',
 'application/vnd.xara'=>'xar',
 'application/vnd.xfdl'=>'xfdl',
 'application/vnd.yamaha.hv-dic'=>'hvd',
 'application/vnd.yamaha.hv-script'=>'hvs',
 'application/vnd.yamaha.hv-voice'=>'hvp',
 'application/vnd.yamaha.openscoreformat'=>'osf',
 'application/vnd.yamaha.openscoreformat.osfpvg+xml'=>'osfpvg',
 'application/vnd.yamaha.smaf-audio'=>'saf',
 'application/vnd.yamaha.smaf-phrase'=>'spf',
 'application/vnd.yellowriver-custom-menu'=>'cmp',
 'application/vnd.zul'=>array('zir','zirz'),
 'application/vnd.zzazz.deck+xml'=>'zaz',
 'application/voicexml+xml'=>'vxml',
 'application/widget'=>'wgt',
 'application/winhlp'=>'hlp',
 'application/wsdl+xml'=>'wsdl',
 'application/wspolicy+xml'=>'wspolicy',
 'application/x-7z-compressed'=>'7z',
 'application/x-abiword'=>'abw',
 'application/x-ace-compressed'=>'ace',
 'application/x-apple-diskimage'=>'dmg',
 'application/x-authorware-bin'=>array('aab','x32','u32','vox'),
 'application/x-authorware-map'=>'aam',
 'application/x-authorware-seg'=>'aas',
 'application/x-bcpio'=>'bcpio',
 'application/x-bittorrent'=>'torrent',
 'application/x-blorb'=>array('blb','blorb'),
 'application/x-bzip'=>'bz',
 'application/x-bzip2'=>array('bz2','boz'),
 'application/x-cbr'=>array('cbr','cba','cbt','cbz','cb7'),
 'application/x-cdlink'=>'vcd',
 'application/x-cfs-compressed'=>'cfs',
 'application/x-chat'=>'chat',
 'application/x-chess-pgn'=>'pgn',
 'application/x-conference'=>'nsc',
 'application/x-cpio'=>'cpio',
 'application/x-csh'=>'csh',
 'application/x-debian-package'=>array('deb','udeb'),
 'application/x-dgc-compressed'=>'dgc',
 'application/x-director'=>array('dcr','dir','dxr','cst','cct','cxt','w3d','fgd','swa'),
 'application/x-doom'=>'wad',
 'application/x-dosexec'=>'exe',
 'application/x-dtbncx+xml'=>'ncx',
 'application/x-dtbook+xml'=>'dtb',
 'application/x-dtbresource+xml'=>'res',
 'application/x-dvi'=>'dvi',
 'application/x-envoy'=>'evy',
 'application/x-eva'=>'eva',
 'application/x-font-bdf'=>'bdf',
 'application/x-font-ghostscript'=>'gsf',
 'application/x-font-linux-psf'=>'psf',
 'application/x-font-otf'=>'otf',
 'application/x-font-pcf'=>'pcf',
 'application/x-font-snf'=>'snf',
 'application/x-font-ttf'=>array('ttf','ttc'),
 'application/x-font-type1'=>array('pfa','pfb','pfm','afm'),
 'application/x-font-woff'=>'woff',
 'application/x-freearc'=>'arc',
 'application/x-futuresplash'=>'spl',
 'application/x-gca-compressed'=>'gca',
 'application/x-glulx'=>'ulx',
 'application/x-gnumeric'=>'gnumeric',
 'application/x-gramps-xml'=>'gramps',
 'application/x-gtar'=>'gtar',
 'application/x-hdf'=>'hdf',
 'application/x-install-instructions'=>'install',
 'application/x-iso9660-image'=>'iso',
 'application/x-java-jnlp-file'=>'jnlp',
 'application/x-javascript'=>'js',
 'application/x-koan'=>array('skp','skd','skm','skt'),
 'application/x-latex'=>'latex',
 'application/x-lzh-compressed'=>array('lzh','lha'),
 'application/x-mie'=>'mie',
 'application/x-mobipocket-ebook'=>array('prc','mobi'),
 'application/x-ms-application'=>'application',
 'application/x-ms-shortcut'=>'lnk',
 'application/x-ms-wmd'=>'wmd',
 'application/x-ms-wmz'=>'wmz',
 'application/x-ms-xbap'=>'xbap',
 'application/x-msaccess'=>'mdb',
 'application/x-msbinder'=>'obd',
 'application/x-mscardfile'=>'crd',
 'application/x-msclip'=>'clp',
 'application/x-msdownload'=>array('exe','dll','com','bat','msi'),
 'application/x-msmediaview'=>array('mvb','m13','m14'),
 'application/x-msmetafile'=>array('wmf','wmz','emf','emz'),
 'application/x-msmoney'=>'mny',
 'application/x-mspublisher'=>'pub',
 'application/x-msschedule'=>'scd',
 'application/x-msterminal'=>'trm',
 'application/x-mswrite'=>'wri',
 'application/x-netcdf'=>array('nc','cdf'),
 'application/x-nzb'=>'nzb',
 'application/x-pkcs12'=>array('p12','pfx'),
 'application/x-pkcs7-certificates'=>array('p7b','spc'),
 'application/x-pkcs7-certreqresp'=>'p7r',
 'application/x-rar-compressed'=>'rar',
 'application/x-research-info-systems'=>'ris',
 'application/x-sh'=>'sh',
 'application/x-shar'=>'shar',
 'application/x-shockwave-flash'=>'swf',
 'application/x-silverlight-app'=>'xap',
 'application/x-sql'=>'sql',
 'application/x-stuffit'=>'sit',
 'application/x-stuffitx'=>'sitx',
 'application/x-subrip'=>'srt',
 'application/x-sv4cpio'=>'sv4cpio',
 'application/x-sv4crc'=>'sv4crc',
 'application/x-t3vm-image'=>'t3',
 'application/x-tads'=>'gam',
 'application/x-tar'=>'tar',
 'application/x-tcl'=>'tcl',
 'application/x-tex'=>'tex',
 'application/x-tex-tfm'=>'tfm',
 'application/x-texinfo'=>array('texinfo','texi'),
 'application/x-tgif'=>'obj',
 'application/x-troff'=>array('t','roff','tr'),
 'application/x-troff-man'=>'man',
 'application/x-troff-me'=>'me',
 'application/x-troff-ms'=>'ms',
 'application/x-ustar'=>'ustar',
 'application/x-wais-source'=>'src',
 'application/x-x509-ca-cert'=>array('der','crt'),
 'application/x-xfig'=>'fig',
 'application/x-xliff+xml'=>'xlf',
 'application/x-xpinstall'=>'xpi',
 'application/x-xz'=>'xz',
 'application/x-zip'=>'zip',
 'application/x-zmachine'=>array('z1','z2','z3','z4','z5','z6','z7','z8'),
 'application/xaml+xml'=>'xaml',
 'application/xcap-diff+xml'=>'xdf',
 'application/xenc+xml'=>'xenc',
 'application/xhtml+xml'=>array('xhtml','xht'),
 'application/xml'=>array('xml','xsl'),
 'application/xml-dtd'=>'dtd',
 'application/xop+xml'=>'xop',
 'application/xproc+xml'=>'xpl',
 'application/xslt+xml'=>'xslt',
 'application/xspf+xml'=>'xspf',
 'application/xv+xml'=>array('mxml','xhvml','xvml','xvm'),
 'application/yang'=>'yang',
 'application/yin+xml'=>'yin',
 'application/zip'=>'zip',
 'audio/adpcm'=>'adp',
 'audio/basic'=>array('au','snd'),
 'audio/midi'=>array('mid','kar','midi','rmi'),
 'audio/mp4'=>array('m4a','mp4a'),
 'audio/mpeg'=>array('mp3','mp2','mpga','mp2a','m2a','m3a'),
 'audio/ogg'=>array('oga','ogg','spx'),
 'audio/s3m'=>'s3m',
 'audio/silk'=>'sil',
 'audio/vnd.dece.audio'=>array('uva','uvva'),
 'audio/vnd.digital-winds'=>'eol',
 'audio/vnd.dra'=>'dra',
 'audio/vnd.dts'=>'dts',
 'audio/vnd.dts.hd'=>'dtshd',
 'audio/vnd.lucent.voice'=>'lvp',
 'audio/vnd.ms-playready.media.pya'=>'pya',
 'audio/vnd.nuera.ecelp4800'=>'ecelp4800',
 'audio/vnd.nuera.ecelp7470'=>'ecelp7470',
 'audio/vnd.nuera.ecelp9600'=>'ecelp9600',
 'audio/vnd.rip'=>'rip',
 'audio/webm'=>'weba',
 'audio/x-aac'=>'aac',
 'audio/x-aiff'=>array('aif','aifc','aiff'),
 'audio/x-caf'=>'caf',
 'audio/x-flac'=>'flac',
 'audio/x-matroska'=>'mka',
 'audio/x-mpegurl'=>'m3u',
 'audio/x-ms-wax'=>'wax',
 'audio/x-ms-wma'=>'wma',
 'audio/x-pn-realaudio'=>array('ram','ra'),
 'audio/x-pn-realaudio-plugin'=>'rmp',
 'audio/x-wav'=>'wav',
 'audio/xm'=>'xm',
 'chemical/x-cdx'=>'cdx',
 'chemical/x-cif'=>'cif',
 'chemical/x-cmdf'=>'cmdf',
 'chemical/x-cml'=>'cml',
 'chemical/x-csml'=>'csml',
 'chemical/x-pdb'=>'pdb',
 'chemical/x-xyz'=>'xyz',
 'image/bmp'=>'bmp',
 'image/cgm'=>'cgm',
 'image/g3fax'=>'g3',
 'image/gif'=>'gif',
 'image/ief'=>'ief',
 'image/jpeg'=>array('jpg','jpeg','jpe'),
 'image/ktx'=>'ktx',
 'image/png'=>'png',
 'image/prs.btif'=>'btif',
 'image/sgi'=>'sgi',
 'image/svg+xml'=>array('svg','svgz'),
 'image/tiff'=>array('tiff','tif'),
 'image/vnd.adobe.photoshop'=>'psd',
 'image/vnd.dece.graphic'=>array('uvi','uvvi','uvg','uvvg'),
 'image/vnd.djvu'=>array('djvu','djv'),
 'image/vnd.dvb.subtitle'=>'sub',
 'image/vnd.dwg'=>'dwg',
 'image/vnd.dxf'=>'dxf',
 'image/vnd.fastbidsheet'=>'fbs',
 'image/vnd.fpx'=>'fpx',
 'image/vnd.fst'=>'fst',
 'image/vnd.fujixerox.edmics-mmr'=>'mmr',
 'image/vnd.fujixerox.edmics-rlc'=>'rlc',
 'image/vnd.ms-modi'=>'mdi',
 'image/vnd.ms-photo'=>'wdp',
 'image/vnd.net-fpx'=>'npx',
 'image/vnd.wap.wbmp'=>'wbmp',
 'image/vnd.xiff'=>'xif',
 'image/webp'=>'webp',
 'image/x-3ds'=>'3ds',
 'image/x-cmu-raster'=>'ras',
 'image/x-cmx'=>'cmx',
 'image/x-freehand'=>array('fh','fhc','fh4','fh5','fh7'),
 'image/x-icon'=>'ico',
 'image/x-mrsid-image'=>'sid',
 'image/x-ms-bmp'=>'bmp',
 'image/x-pcx'=>'pcx',
 'image/x-pict'=>array('pic','pct'),
 'image/x-portable-anymap'=>'pnm',
 'image/x-portable-bitmap'=>'pbm',
 'image/x-portable-graymap'=>'pgm',
 'image/x-portable-pixmap'=>'ppm',
 'image/x-rgb'=>'rgb',
 'image/x-tga'=>'tga',
 'image/x-xbitmap'=>'xbm',
 'image/x-xpixmap'=>'xpm',
 'image/x-xwindowdump'=>'xwd',
 'message/rfc822'=>array('eml','mime'),
 'model/iges'=>array('igs','iges'),
 'model/mesh'=>array('msh','mesh','silo'),
 'model/vnd.collada+xml'=>'dae',
 'model/vnd.dwf'=>'dwf',
 'model/vnd.gdl'=>'gdl',
 'model/vnd.gtw'=>'gtw',
 'model/vnd.mts'=>'mts',
 'model/vnd.vtu'=>'vtu',
 'model/vrml'=>array('wrl','vrml'),
 'model/x3d+binary'=>array('x3db','x3dbz'),
 'model/x3d+vrml'=>array('x3dv','x3dvz'),
 'model/x3d+xml'=>array('x3d','x3dz'),
 'text/cache-manifest'=>'appcache',
 'text/calendar'=>array('ics','ifb'),
 'text/css'=>'css',
 'text/csv'=>'csv',
 'text/html'=>array('html','htm'),
 'text/n3'=>'n3',
 'text/plain'=>array('txt','asc','text','conf','def','list','log','in','csv','css'),
 'text/prs.lines.tag'=>'dsc',
 'text/richtext'=>'rtx',
 'text/rtf'=>'rtf',
 'text/sgml'=>array('sgml','sgm'),
 'text/tab-separated-values'=>'tsv',
 'text/troff'=>array('t','tr','roff','man','me','ms'),
 'text/turtle'=>'ttl',
 'text/uri-list'=>array('uri','uris','urls'),
 'text/vcard'=>'vcard',
 'text/vnd.curl'=>'curl',
 'text/vnd.curl.dcurl'=>'dcurl',
 'text/vnd.curl.mcurl'=>'mcurl',
 'text/vnd.curl.scurl'=>'scurl',
 'text/vnd.dvb.subtitle'=>'sub',
 'text/vnd.fly'=>'fly',
 'text/vnd.fmi.flexstor'=>'flx',
 'text/vnd.graphviz'=>'gv',
 'text/vnd.in3d.3dml'=>'3dml',
 'text/vnd.in3d.spot'=>'spot',
 'text/vnd.sun.j2me.app-descriptor'=>'jad',
 'text/vnd.wap.wml'=>'wml',
 'text/vnd.wap.wmlscript'=>'wmls',
 'text/x-asm'=>array('s','asm'),
 'text/x-c'=>array('c','cc','cxx','cpp','h','hh','dic'),
 'text/x-fortran'=>array('f','for','f77','f90'),
 'text/x-java-source'=>'java',
 'text/x-nfo'=>'nfo',
 'text/x-opml'=>'opml',
 'text/x-pascal'=>array('p','pas'),
 'text/x-setext'=>'etx',
 'text/x-sfv'=>'sfv',
 'text/x-uuencode'=>'uu',
 'text/x-vcalendar'=>'vcs',
 'text/x-vcard'=>'vcf',
 'video/3gpp'=>'3gp',
 'video/3gpp2'=>'3g2',
 'video/h261'=>'h261',
 'video/h263'=>'h263',
 'video/h264'=>'h264',
 'video/jpeg'=>'jpgv',
 'video/jpm'=>array('jpm','jpgm'),
 'video/mj2'=>array('mj2','mjp2'),
 'video/mp4'=>array('mp4','mp4v','mpg4'),
 'video/mpeg'=>array('mpeg','mpe','mpg','m1v','m2v'),
 'video/ogg'=>'ogv',
 'video/quicktime'=>array('mov','qt'),
 'video/vnd.dece.hd'=>array('uvh','uvvh'),
 'video/vnd.dece.mobile'=>array('uvm','uvvm'),
 'video/vnd.dece.pd'=>array('uvp','uvvp'),
 'video/vnd.dece.sd'=>array('uvs','uvvs'),
 'video/vnd.dece.video'=>array('uvv','uvvv'),
 'video/vnd.dvb.file'=>'dvb',
 'video/vnd.fvt'=>'fvt',
 'video/vnd.mpegurl'=>array('mxu','m4u'),
 'video/vnd.ms-playready.media.pyv'=>'pyv',
 'video/vnd.uvvu.mp4'=>array('uvu','uvvu'),
 'video/vnd.vivo'=>'viv',
 'video/webm'=>'webm',
 'video/x-f4v'=>'f4v',
 'video/x-fli'=>'fli',
 'video/x-flv'=>'flv',
 'video/x-m4v'=>'m4v',
 'video/x-matroska'=>array('mkv','mk3d','mks'),
 'video/x-mng'=>'mng',
 'video/x-ms-asf'=>array('asf','asx'),
 'video/x-ms-vob'=>'vob',
 'video/x-ms-wm'=>'wm',
 'video/x-ms-wmv'=>'wmv',
 'video/x-ms-wmx'=>'wmx',
 'video/x-ms-wvx'=>'wvx',
 'video/x-msvideo'=>'avi',
 'video/x-sgi-movie'=>'movie',
 'video/x-smv'=>'smv',
 'x-conference/x-cooltalk'=>'ice'
 );
}

?>