<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/lib/statisticslib.php - statistics
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wascore
 * @version $Id: statisticslib.php,v 1.9 2016/06/15 12:11:41 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }

/** main entry point for statistics (called from admin.php)
 *
 * Produce a simple ordered table of pages with # of page views
 * for one area or for all areas the user is allowed to access.
 *
 * Note:
 * Even though expired pages are no longer visible for visitors, the
 * # of views is still displayed. Pages under embargo are completely
 * suppressed.
 *
 * @param object &$output collects the html output
 * @return void results are returned as output in $output
 * @todo this routine is a little bit too long, it should be split into smaller chunks
 */
function job_statistics(&$output) {
    global $USER,$CFG,$WAS_SCRIPT_NAME,$DB;
    $output->set_helptopic('statistics');
    $header = t('description_statistics','admin');

    //
    // 1 -- any areas available for this user?
    //
    $areas = array();
    if (($records = get_area_records()) !== FALSE) {
        if (($USER->has_site_permissions(PERMISSION_SITE_ADD_AREA)) ||
            ($USER->has_site_permissions(PERMISSION_SITE_DROP_AREA))) {
            $areas = $records;
        } else {
            foreach($records as $area_id => $record) {
                if (($USER->has_area_permissions(PERMISSION_AREA_EDIT_AREA,$area_id)) ||
                    (db_bool_is(FALSE,$record['is_private'])) ||
                    ($USER->has_intranet_permissions(ACL_ROLE_INTRANET_ACCESS,$area_id))) {
                    $areas[$area_id] = $record;
                } // else suppress this area because user has no access whatsoever
            }
        }
    }
    if (sizeof($areas) <= 0) {
        $output->add_content('<h2>'.$header.'</h2>');
        $output->add_content(t('no_areas_available','admin'));
        return;
    }
    $area_id = get_parameter_int('area');
    if ((is_int($area_id)) && (!isset($areas[$area_id]))) {
        $output->add_content('<h2>'.$header.'</h2>');
        $output->add_content(t('access_denied','admin'));
        return;
    }
    statistics_show_area_menu($output,$areas,$area_id); // indicate available areas via selection menu

    //
    // 2 -- are there any pages in the selected area(s)?
    //
    if (($pages = statistics_get_pages($areas,$area_id)) === FALSE) {
        $output->add_message(t('statistics_error','admin'));
        $pages = array();
    }
    if (($count = sizeof($pages)) < 1) {
        $output->add_content('<h2>'.$header.'</h2>');
        $output->add_content(t('no_pages_available','admin'));
        return;
    }

    //
    // 3 -- check for possible pagination, maybe modify $header too
    //
    $offset = max(0,min($count-1,get_parameter_int('offset',0))); // 0 <= offset < count
    $limit = max(1,min($count, get_parameter_int('limit',$CFG->pagination_height))); // 1 <= limit <= count
    $last = min($count,$offset+$limit);
    if (($limit < $count) || ($offset != 0)) {
        $parameters = array('job' => JOB_STATISTICS);
        if (is_int($area_id)) {
            $parameters['area'] = $area_id;
        }
        $output->add_pagination($WAS_SCRIPT_NAME,$parameters,$count,$limit,$offset,$CFG->pagination_width);
        $header .= ' '.t('pagination_count_of_total','admin',array('{FIRST}'=> $offset+1,
                                                                  '{LAST}' => $last,
                                                                  '{TOTAL}'=> $count));
    }

    //
    // 4 -- prepare translation of module_id -> readable module (nick)name
    //
    $modules = array();
    if (($records = get_module_records()) !== FALSE) {
        foreach ($records as $module_id => $module) {
            $m_name = 'm_'.$module['name'];
            $modules[$module_id] = array('nick' => t('nickname',$m_name),'title' => t('title',$m_name));
        }
    }

    //
    // 5 - show results via HTML table
    //
    $class = 'header';
    $attributes = array('class' => $class, 'align' => 'left');
    $attr_right = array('class' => $class, 'align' => 'right');

    // 5A -- table header
    $output->add_content('<h2>'.$header.'</h2>');
    $output->add_content('<p>');
    $output->add_content(html_table());
    $output->add_content('  '.html_table_row($attributes));
    $output->add_content('    '.html_table_head($attr_right,t('statistics_nr','admin')));
    $output->add_content('    '.html_table_head($attributes,t('statistics_area','admin')));
    $output->add_content('    '.html_table_head($attributes,t('statistics_page','admin')));
    $output->add_content('    '.html_table_head($attr_right,t('statistics_views','admin')));
    $output->add_content('  '.html_table_row_close());

    // 5B -- table entries
    for ($index=$offset; $index < $last; ) {
        $page = $pages[$index++];
        $class = ($class == 'odd') ? 'even' : 'odd';
        $attributes = array('class' => $class);
        $attr_right = array('class' => $class, 'align' => 'right');
        $id = intval($page['area_id']);
        $module_id = intval($page['module_id']);
        $nick = $name = '';
        if (isset($modules[$module_id])) {
            $nick = '/'.$modules[$module_id]['nick'];
            $name = '/'.$modules[$module_id]['title'];
        }
        $area = sprintf('%s (%d)',htmlspecialchars($areas[$id]['title']),$id);
        $node = sprintf('%s (%d%s)',htmlspecialchars($page['link_text']),$page['node_id'],$nick);
        $title = sprintf('%s (%d%s)',htmlspecialchars($page['title']),$page['node_id'],$name);
        $output->add_content('  '.html_table_row(array('class' => $class,'title' => $title)));
        $output->add_content('    '.html_table_cell($attr_right,strval($index)));
        $output->add_content('    '.html_table_cell($attributes,$area));
        $output->add_content('    '.html_table_cell($attributes,$node));
        $output->add_content('    '.html_table_cell($attr_right,$page['view_count']));
        $output->add_content('  '.html_table_row_close());
        }
    // 5C -- done
    $output->add_content(html_table_close());
    $output->add_content('<p>');
} // job_statistics()


/** show a menu of available areas
 *
 * output a menu, starting with an entry to show _all_ available areas,
 * followed by links to individual areas. If there is only a single area,
 * the 'all areas' link is suppressed
 *
 * @param object &$output collects the html output
 * @param array &$areas list of all areas available for this user
 * @param int|null $current_area_id the current area
 * @return void results are returned as output in $this->output
 */
function statistics_show_area_menu(&$output,&$areas,$current_area_id=NULL) {
    global $USER,$WAS_SCRIPT_NAME;
    $output->add_menu('<h2>'.t('menu','admin').'</h2>');
    $output->add_menu('<ul>');
    $a_params = array('job' => JOB_STATISTICS);

    // suppress 'all areas' if there's only one
    if (sizeof($areas) > 1) {
        $title = t('all_areas_title','admin');
        $anchor = t('all_areas','admin');
        $attributes = array('title' => $title);
        if (is_null($current_area_id)) {
            $attributes['class'] = 'current';
        }
        $output->add_menu('  <li>'.html_a($WAS_SCRIPT_NAME,$a_params,$attributes,$anchor));
    }

    // step through all available areas
    foreach($areas as $area_id => $area) {
        $a_params['area'] = strval($area_id);
        $title = sprintf("%s %d%s",
                         t((db_bool_is(TRUE,$area['is_private'])) ? 'private_area':'public_area','admin'),
                         $area_id,
                         (db_bool_is(TRUE,$area['is_active'])) ? '':' ('.t('inactive','admin').')');
        $anchor = htmlspecialchars($area['title']);
        $attributes = array('title' => $title);
        if ($current_area_id == $area_id) {
            $attributes['class'] = 'current';
        }
        $output->add_menu('  <li>'.html_a($WAS_SCRIPT_NAME,$a_params,$attributes,$anchor));
    }
    $output->add_menu('</ul>');
} // statistics_show_area_menu()


/** check ancestors for embargo
 *
 * ascend the tree via $parent_id to see if there is an embargo.
 * the quest ends when we reach a top level section (indicated by
 * $id == $parent_id) or a section that was examined before (indicated
 * through the existence of the flag element in the $section).
 * before we start recursing we set that flag just to be sure we don't
 * end up in an endless loop if somehow the tree structure is not
 * sane (contains circular references).
 *
 * @param array &$sections holds pertinent information about sections in selected areas
 * @param string &$now current time to check agains the embargo time
 * @param int $id section id
 * @return bool TRUE if under embargo, FALSE otherwise
 */
function statistics_embargo(&$sections,&$now,$id) {
    if (!isset($sections[$id])) {
        return TRUE; // caller is orphan? go with embargo
    } elseif (!isset($sections[$id]['flag'])) {
        if ($now < $sections[$id]['embargo']) {
            $sections[$id]['flag'] = TRUE;
        } elseif (($parent_id = $sections[$id]['parent_id']) == $id) {
            $sections[$id]['flag'] = FALSE;
        } else {
            $sections[$id]['flag'] = TRUE; // prevent endless loop in case of circular reference
            $sections[$id]['flag'] = statistics_embargo($sections,$now,$parent_id);
        }
    }
    return $sections[$id]['flag'];
} // statistics_embargo


/**  construct an ordered list of pages to show in a statistics report
 *
 * strategy:
 *  - start with all pages in selected area(s) NOT under embargo
 *  - retrieve  all sections in selected areas(s)
 *  - for all pages: if none of the ancestors is under embargo then keep page in $pages[]
 *
 * If anything goes wrong, return FALSE, otherwise an array with the selected pages
 *
 * It would be nice if we could simply select $limit records from the database at $offset.
 * Unfortunately we cannot be sure if a page is elegible for display until we have
 * examined the ancestors so we have to retrieve them and perhaps later discard them.
 * oh well, there's room for improvement.
 *
 * @param array &$areas list of areas available to the current user
 * @param int|null $area_id area of interest or NULL indicating all areas in $areas[]
 * @return bool|array FALSE on error, array with pages otherwise (could be empty)
 */
function statistics_get_pages(&$areas,$area_id) {

    // 0 -- initialise
    $table = 'nodes';
    $now = strftime('%Y-%m-%d %T'); // needed for embargo
    $key = 'node_id';
    $pages = array();

    // 1 -- limit results to selected area(s) via where clause
    if (is_int($area_id)) {
        $where_areas = sprintf('area_id = %d',$area_id);
    } else {
        $a = array();
        foreach($areas as $k => $v) {
            $a[] = sprintf('(area_id=%d)',$k);
        }
        $where_areas = implode(' OR ',$a);
    }

    // 2 -- get raw page data in $nodes[], ordered by # of views (descending)
    $fields = array($key,'area_id','parent_id','title','link_text','module_id','view_count');
    $where = sprintf('(is_page) AND (embargo < %s) AND (%s)',db_escape_and_quote($now),$where_areas);
    $order = array('view_count DESC','area_id',$key);
    if (($nodes = db_select_all_records($table,$fields,$where,$order,$key)) === FALSE) {
        logger(sprintf('%s(): cannot retrieve raw page data: %s',__FUNCTION__,db_errormessage()));
        return FALSE;
    } elseif (sizeof($nodes) < 1) {
        logger(sprintf('%s(): no pages to show',__FUNCTION__),WLOG_DEBUG);
        return $pages;
    }

    // 3 -- get all sections in $sections[], as a helper to find pages in embargo'ed sections
    $fields = array($key,'parent_id','embargo');
    $where = sprintf('(not is_page) AND (%s)',$where_areas);
    $order = array($key);
    if (($sections = db_select_all_records($table,$fields,$where,$order,$key)) === FALSE) {
        logger(sprintf('%s(): cannot retrieve raw section data: %s',__FUNCTION__,db_errormessage()));
        return FALSE;
    }

    // 4 -- collect pages that are not under embargo in $pages[]
    foreach($nodes as $node_id => $node) {
        if (($node_id == $node['parent_id']) || (!statistics_embargo($sections,$now,$node['parent_id']))) {
            $pages[] = $node;
        }
    }
    if (sizeof($pages) < 1) {
        logger(sprintf('%s(): no entries to show',__FUNCTION__),WLOG_DEBUG);
    }
    return $pages;
} // statistics_get_pages()

?>