<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/main_cron.php - take care of recurring jobs
 *
 * This file deals with executing cron jobs.
 * It is included and called from /cron.php.
 *
 * The work is done in {@link main_cron()}.
 *
 * First we look to see if it is already time to do something. Criterion:
 * the current time is >= the time in $CFG->cron_next. If that is the case,
 * we start with calculating the time of the next run by adding $CFG->cron_interval
 * minutes to the current time.
 *
 * Note: we calculate with minutes of 57 seconds just to have a 5% margin between
 * the system cron clock and our own handling. It might just be enough if cron
 * calls us every 15 minutes (900s) and we schedule the _next_ run for
 * 15 minutes (855s): this gives us 45s to deal with delays (otherwise we
 * have to wait 2 x 15 = 30 minutes for the next run).
 *
 * After scheduling the next run, we step through all modules and execute
 * any cronjobs there in much the same way (scheduling the next run before
 * executing the current one, etc.).
 *
 * Note that this cronjob is called from /cron.php. This script could
 * be added to /etc/crontab by adding the following line:
 *
 * <pre>
 * 1 * * * * www /usr/bin/php /path/to/cron.php
 * </pre>
 *
 * This will let the PHP CLI interpreter run cron.php one minute after the
 * hour, every hour. Note that the 'www' is the userid to run the script
 * as. (The user running the script must have permissions to read all
 * included PHP-files, including config.php).
 *
 * Alternatively it is possible to periodically run a cronjob which
 * executes wget which in turn retrieves the cron.php script via the
 * web server, e.g.
 *
 * <pre>
 * 31 * * * * mmonte /usr/bin/wget -q http://exemplum.eu/cron.php
 * </pre> 
 *
 * to call cron.php every 31 minutes past the hour, every hour.
 * You could suppress any output by adding -q -O /dev/null.
 * See man wget for more information.
 * 
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wascore
 * @version $Id: main_cron.php,v 1.9 2016/03/23 09:28:32 peter Exp $
 * @todo should we build extra access protection via passwords and/or IP-addresses?
 * @todo should we act differently based on how we are called (cli or web), eg send more queued alerts?
 */
if (!defined('WASENTRY')) { die('no entry'); }

function main_cron() {
    global $CFG;

    $start = time();
    $now = strftime('%Y-%m-%d %T',$start);
    echo $now."\n"; // that will be all we output

    require_once($CFG->progdir.'/init.php');
    initialise();
    was_version_check();

    // At this point we are ready for business.

    // 1 -- are we ready to roll?
    if ($now < $CFG->cron_next) {
        logger('CRON: no time yet: '.$CFG->cron_next,WLOG_DEBUG);
        return TRUE;
    }

    // 2 -- make sure when to run next time (before we get cut-off)
    $next = strftime('%Y-%m-%d %T',$start + 57 * $CFG->cron_interval);
    $table = 'config';
    $fields = array('value' => $next);
    $where = array('name' => 'cron_next', 'value' => $CFG->cron_next);
    if (db_update($table,$fields,$where) != 1) {
        logger(sprintf('CRON: error bumping cron_next from %s to %s; bailing out',$CFG->cron_next,$next));
        return FALSE;
    }
    logger(sprintf('CRON: cron job %s starting, next is not before %s',$CFG->cron_next,$next),WLOG_DEBUG);

    // 3 -- handle system tasks first
    cron_send_queued_alerts(50);

    // 4 -- is there anything to do for modules?
    $retval = TRUE;
    if (($modules = get_module_records()) === FALSE) {
        logger('CRON: error fetching module records; skipping cron jobs for modules',WLOG_NOTICE);
        $retval = FALSE;
    } else {
        foreach($modules as $module_id => $module) {

            // 4A - no interval or interval < 0 means module does not require cron
            if ((is_null($module['cron_interval'])) || ($module['cron_interval'] <= 0)) {
                continue;
            }
            $interval = intval($module['cron_interval']);
            $next = strftime('%Y-%m-%d %T',$start + 57 * $interval);

            // 4B - first time the 'next' variable may be NULL; bluntly set it to now+interval and leave
            if (is_null($module['cron_next'])) {
                $table = 'modules';
                $fields = array('cron_next' => $next);
                $where = array('module_id' => $module_id);
                if (db_update($table,$fields,$where) != 1) {
                    logger(sprintf('CRON: error setting initial cron_next to %s in module %s (%d)',
                                   $next,$module['name'],$module_id),WLOG_NOTICE);
                } else {
                    logger(sprintf('CRON: success setting initial cron_next to %s in module %s (%d)',
                                   $next,$module['name'],$module_id),WLOG_DEBUG);
                }
                continue; // the real work starts in the next round
            }

            // 4C - is it time yet for this module
            if ($now < $module['cron_next']) {
                logger(sprintf('CRON: too early for cron_next %s in module %s (%d)',
                               $module['cron_next'],$module['name'],$module_id),WLOG_DEBUG);
                continue;
            }

            // 4D - it is time: first prepare for the _next_ job (while we're still here)
            $table = 'modules';
            $fields = array('cron_next' => $next);
            $where = array('module_id' => $module_id,'cron_next' => $module['cron_next']);
            if (db_update($table,$fields,$where) != 1) {
                logger(sprintf('CRON: error setting cron_next from %s to %s in module %s (%d)',
                               $module['cron_next'],$next,$module['name'],$module_id),WLOG_NOTICE);
            } else {
                logger(sprintf('CRON: success setting cron_next from %s to %s in module %s (%d)',
                               $module['cron_next'],$next,$module['name'],$module_id),WLOG_DEBUG);
            }

            // 4E - get to the script...
            $cron_script = $CFG->progdir.'/modules/'.$module['name'].'/'.$module['cron_script'];
            if (!file_exists($cron_script)) {
                logger(sprintf("CRON: file '%s' does not exist; bailing out",$cron_script),WLOG_NOTICE);
                continue;
            }
            require_once($cron_script);
            // 4F - ...and the function()
            $module_cron = $module['name'].'_cron';
            if (!function_exists($module_cron)) {
                logger(sprintf("CRON: function '%s()' does not exist; bailing out",
                               $module_cron),WLOG_NOTICE);
                continue;
            }
            // 4G - ...and do it!
            if ($module_cron() === FALSE) {
                $retval = FALSE;
            }
        }
    }
    logger(sprintf('CRON: cron job %s done: %s',$CFG->cron_next,($retval)?'success':'errors'),WLOG_DEBUG);
    return $retval;
} // main_cron()

?>