<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/althing/althing_admin.php - management interface for althing-module
 *
 * This file defines the administrative interface to this module.
 * The interface consists of the following four functions.
 *
 * <code>
 * althing_disconnect(&$output,$area_id,$node_id,$module)
 * althing_connect(&$output,$area_id,$node_id,$module)
 * althing_show_edit(&$output,$area_id,$node_id,$module,$viewonly,$edit_again,$href,$option)
 * althing_save(&$output,$area_id,$node_id,$module,$viewonly,&$edit_again,$option)
 * </code>
 *
 * These functions are called from pagemanagerlib.php whenever necessary.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_althing
 * @version $Id: althing_admin.php,v 1.18 2016/05/26 13:22:27 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }

require_once(dirname(__FILE__).'/althing_common.php'); // common code + constants

/** disconnect this module from a node
 *
 * this breaks the link between the node $node_id in area $area_id and
 * this module. For now we simply delete the records in all tables
 * 'althings*' using node_id == althing_id. Note the order in which
 * the tables are emptied (to satisfy the FK constraints).
 * 
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node from which we need to disconnect
 * @param array $module the module record straight from the database
 * @return bool TRUE on success, FALSE otherwise
 */
function althing_disconnect(&$output, $area_id, $node_id, $module) {
    $where = array('althing_id' => intval($node_id));
    $tables = array(
        'althing_log_messages',
        'althing_relations',
        'althing_posts',
        'althings');
    $retval = TRUE; // assume success
    foreach($tables as $table) {
        if (db_delete($table,$where) === FALSE) {
            logger(sprintf("%s(): cannot disconnect table '%s' from node '%d': %s",
                           __FUNCTION__,$table,$node_id,db_errormessage()));
            $retval = FALSE;
        }
    }
    return $retval;
} // althing_disconnect()


/** connect this module to a node
 *
 * this makes the link between the node $node_id in area $area_id and this module.
 * In this case we create a single 'althings' record linked to node_id via althing_id.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which we need to connect
 * @param array $module the module record straight from the database
 * @return bool TRUE on success, FALSE otherwise
 */
function althing_connect(&$output, $area_id, $node_id, $module) {
    global $USER;
    $now = strftime('%Y-%m-%d %T');
    $fields = array(
        'althing_id' => intval($node_id),
        'header' => '',
        'introduction' => '',
        'question' => '',
        'conclusion' => '',
        'footer' => '',
        'status' => ALTHING_STATUS_OPENED,
        'default_visibility' => ALTHING_DEFAULT_VISIBILITY,
        'email_moderators' => $USER->email,
        'email_subscribers' => '',
        'ctime' => $now,
        'cuser_id' => $USER->user_id,
        'mtime' => $now,
        'muser_id' => $USER->user_id);
    $retval = db_insert_into('althings',$fields);
    if ($retval !== 1) {
        logger(sprintf("%s(): cannot connect to node '%d': %s",__FUNCTION__,$node_id,db_errormessage()));
        $retval = FALSE;
    } else {
        $retval = TRUE;
    }
    return $retval;
} // althing_connect()


/** dispatcher for showing various dialog screens
 *
 * there are a few dialogs that are accessed via this routine:
 * - overview: a simple introduction to this module (via $option==NULL)
 * - configuration: all properties of the althing (main althings record) ($option=configuration)
 * - moderation: a list of postings OR an individual post ($option=moderation)
 * - logmessages: a list of all moderation messages for this althing ($option=viewlog)
 *
 * This routine only handles the main overview dialog, the rest is done in subroutines.
 * 
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @param string $option the selected submenu option or NULL if user used the 'Content' option to get here
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function althing_show_edit(&$output, $area_id, $node_id, $module, $viewonly, $edit_again, $href, $option) {

    if ($option == 'configuration') {
        return althing_show_edit_configuration($output, $node_id, $module, $viewonly,$edit_again, $href);
    } elseif ($option == 'moderation') {
        $output->add_stylesheet(sprintf('program/modules/%s/%s.css',$module['name'],$module['name']));
        return althing_show_edit_moderation($output,$area_id,$node_id,$module,$viewonly,$edit_again,$href);
    } elseif ($option == 'viewlog') {
        return althing_show_edit_viewlog($output,$node_id,$module,$href);
    } elseif ($option == 'report') {
        $output->add_stylesheet(sprintf('program/modules/%s/%s.css',$module['name'],$module['name']));
        return althing_show_edit_report($output,$node_id,$module,$edit_again,$href);
    } else { // generic call to Content ($option == NULL)
        althing_show_content($output,$node_id,$module);
    }
    return TRUE;
} // althing_show_edit()


/** show the main content screen (in fact documentation only)
 *
 * this displays the simple overview screen with documentation.
 * It ends with a cancel button that ends the whole edit operation for
 * this node.
 * 
 * @param object &$output collects the html output (if any)
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function althing_show_content(&$output,$node_id,$module) {
    global $WAS_SCRIPT_NAME;
    $params = array(
        'job' => JOB_PAGEMANAGER,
        'task' => TASK_NODE_EDIT_CONTENT,
        'node' => strval($node_id)
        );
    $aparams = array('{NODE}' => strval($node_id));
    if (!empty($module['options'])) {
        $options = explode(',',$module['options']);
        if (!empty($options)) {
            foreach($options as $option) {
                $key = sprintf('{%s}',strtoupper($option));
                $title = t($option.'_title','m_althing',$aparams);
                $params[PARAM_SUBMENU_OPTION] = $option;
                $a_attr = array('title' => $title);
                $anchor = t($option.'_label','m_althing',$aparams);
                $aparams[$key] = html_a($WAS_SCRIPT_NAME,$params,$a_attr,$anchor);
            }
        }
    }
    $output->add_content('<h2>'.t('althing_content_header','m_althing',$aparams).'</h2>');
    $output->add_content(t('althing_content_explanation','m_althing',$aparams));
    $params = array(
        'job'    => JOB_PAGEMANAGER,
        'task'   => TASK_SAVE_CONTENT,
        'node'   => intval($node_id)
        );
    $href = href($WAS_SCRIPT_NAME, $params);
    $dialogdef = array('button_cancel' => dialog_buttondef(BUTTON_CANCEL));
    $output->add_content(dialog_quickform($href,$dialogdef));
    return TRUE;
} // althing_show_content()


/** dispatcher for save routines
 *
 * We use different subroutines for clarity.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which the content is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing and hence saving is not allowed
 * @param bool &$edit_again set to TRUE if we need to edit the content again, FALSE otherwise
 * @param string $option the selected submenu option or NULL if user used the 'Content' option to get here
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function althing_save(&$output, $area_id, $node_id, $module, $viewonly, &$edit_again, $option) {
    if ($option == 'configuration') {
        return althing_save_configuration($output,$node_id,$viewonly,$edit_again);
    } elseif ($option == 'moderation') {
        return althing_save_moderation($output,$area_id, $node_id,$viewonly,$edit_again);
    } elseif ($option == 'viewlog') {
        return althing_save_viewlog($output,$viewonly,$edit_again);
    } elseif ($option == 'report') {
        return althing_save_report($output,$viewonly,$edit_again);
    } else { // save NULL or invalid option: indicate [Cancel]
        $edit_again = FALSE;
        return FALSE;
    }
    $edit_again = TRUE;
    return TRUE;
} // althing_save()


// ==================================================================
// =================== ALTHING MAIN CONFIGURATION ===================
// ==================================================================


/** show a simple edit screen for the althing configuration
 *
 *
 * @param object &$output collects the html output (if any)
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function althing_show_edit_configuration(&$output, $node_id, $module, $viewonly, $edit_again, $href) {
    global $WAS_SCRIPT_NAME;
    // Cancel and Done from edit configuration should return to content screen
    if ((isset($_POST['button_cancel'])) || (isset($_POST['button_done']))) {
        return althing_show_content($output,$node_id,$module);
    }
    $output->add_content('<h2>'.t('althing_configuration_header','m_althing').'</h2>');
    $output->add_content(t('althing_configuration_explanation','m_althing'));
    $dialogdef = althing_get_dialogdef_configuration($output,$viewonly,$node_id);
    if ($edit_again) {
        althing_config_dialog_validate($dialogdef); // just for side effect of grabbing data from $_POST
    }
    $output->add_content(dialog_quickform($href,$dialogdef));
    return TRUE;
} // althing_show_edit_configuration()


/** validate and save the modified althing configuration linked to node $node_id
 *
 * this validates and saves the data that was submitted by the user.
 *
 * If validation fails, the flag $edit_again is set to TRUE and the return value
 * is FALSE.
 *
 * If the user has cancelled the operation, the flag $edit_again is set to FALSE
 * and the return value is also FALSE.
 *
 * If the modified data is stored successfully, the return value is TRUE.
 * The flag edit_again is set to TRUE if the user saved via the [Save] button,
 * and to FALSE if the user saved via the [Done] button.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which the content is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing and hence saving is not allowed
 * @param bool &$edit_again set to TRUE if we need to edit the content again, FALSE otherwise
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function althing_save_configuration(&$output, $node_id, $viewonly, &$edit_again) {
    global $USER;

    // 1 -- bail out if cancelled or viewonly
    if ((isset($_POST['button_cancel'])) || ($viewonly)) {
        $edit_again = TRUE; // trick to let [Cancel] return to the Content screen
        return FALSE;
    }

    // 2 -- redo if invalid data was submitted
    $dialogdef = althing_get_dialogdef_configuration($output,$viewonly,$node_id);
    if (!althing_config_dialog_validate($dialogdef)) {
        // there were errors, show them to the user and do it again
        foreach($dialogdef as $k => $item) {
            if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                $output->add_message($item['error_messages']);
            }
        }
        $edit_again = TRUE;
        return FALSE;
    }

    // 3 -- actually save the new setting
    $now = strftime('%Y-%m-%d %T');
    $table = 'althings';
    $fields = array(
        'header'             => $dialogdef['header']['value'],
        'introduction'       => $dialogdef['introduction']['value'],
        'question'           => $dialogdef['question']['value'],
        'conclusion'         => $dialogdef['conclusion']['value'],
        'footer'             => $dialogdef['footer']['value'],
        'status'             => intval($dialogdef['status']['value']),
        'default_visibility' => intval($dialogdef['default_visibility']['value']),
        'email_moderators'   => $dialogdef['email_moderators']['value'],
        'email_subscribers'  => $dialogdef['email_subscribers']['value'],
        'mtime'              => $now,
        'muser_id'           => $USER->user_id);
    $where = array('althing_id' => intval($node_id));
    if (db_update($table,$fields,$where) === FALSE) {
        logger(sprintf('%s(): error saving config value: %s',__FUNCTION__,db_errormessage()));
        $edit_again = TRUE;
        $retval = FALSE;
    } else {
        $edit_again = TRUE; // we _always_ want to return to an edit screen
        $retval = TRUE;
    }
    return $retval;
} // althing_save_configuration()

// ==================================================================
// ======================= ALTHING MODERATION =======================
// ==================================================================


/** present the user with a dialog for moderation of althing posts at node $node_id
 *
 * this is a complex routine that is responsible for constructing and showing
 * the dialogs associated with posts.
 *
 *The following dialogs exist here.
 *<pre>
 * #1. Posts overview
 *
 * Id Date             E V M Name     Subject
 * d1 yyyy-mm-dd hh:mm 1 V 0 Nickname Subject
 * d2 yyyy-mm-dd hh:mm 0 V 0 Nickname Subject
 * d3 yyyy-mm-dd hh:mm 1 H 0 Nickname Subject
 * d4 yyyy-mm-dd hh:mm 0 V 0 Nickname Subject
 * d5 yyyy-mm-dd hh:mm 0 V 0 Nickname Subject
 * d6 yyyy-mm-dd hh:mm 0 U 0 Nickname Subject
 *
 * [Cancel]
 *
 * E=Number of edits, V=Visibility:Visible/Hidden/Unpublished, M=Marbles
 *
 * #2. Moderation of an individual posting
 *
 * Prev Up Next
 *
 * ID: dd (ppp)
 * Date: yyyy-mm-dd hh:mm:ss
 * From: Nickname <email@address> (registered)
 * Subject: Subject
 * IP-address: 127.0.0.1
 * Xx xx xx xxxxx xx xxxx xx xxxxx xx xxxx
 *
 * Status: ( ) Unpublished (X) Published
 * Visible: ( ) Hidden (X) Visible
 * Marbles:  _____
 * Remarks:
 * ____________________
 * ____________________
 * ____________________
 * [Save] [Done] [Cancel]
 *
 * Prev Up Next
 *</pre>
 *
 * There are some complex interactions between these screens and the corresponding 
 * save routine  {@link althing_save_moderation()}. The effects we try to achieve
 * are as follows.
 *
 * The listing in dialog #1 shows all posts. Every post ID is a link to dialog #2.
 * The [Cancel] button in dialog #1 ends the editing of this node; ie. returns to
 * Page Manager.
 *
 * Dialog #2 can be used to change the visibility of a post or to modify the number
 * of marbles. The contents of the remarks field eventually ends up in the log file.
 * The  [Cancel] button here bring the user back to dialog #1. In effect this is a kind of
 * zoom function: you can zoom in on a post and go back to the list without leaving this
 * node, etc. etc. Note that the link Up also takes you back to the list in dialog #1,
 * be it that in that case we use afragment to arrive at the correct position in the list.
 * The [Done] button saves the modified post (and the remarks) and also return to dialog #1.
 * The [Save] button always has the meaning "save the data and continue editing".
 *
 * Note that saving an unchanged record does not yield a message in the log because that
 * would only add noise.
 *
 * As a service there are simple links Prev, Up and Next which allow for quickly stepping
 * through the posts in detail without having to go back to the list in dialog #1.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function althing_show_edit_moderation(&$output,$area_id,$node_id,$module,$viewonly,$edit_again,$href) {

    $post_id = get_parameter_int('post',NULL);
    if ((isset($_POST['button_cancel'])) && (is_null($post_id))) {
        return althing_show_content($output,$node_id,$module);
    }

    // 1 -- get the posts in an array in the desired sort order
    $sort_index = get_parameter_int('sort',1);
    if (($records = althing_get_post_records($node_id,$sort_index)) === FALSE) {
        $output->add_message(t('error_retrieving_data','admin'));
        return FALSE;
    }

    // 2 -- handle cancellation from dialog #2; return to dialog #1
    if (((isset($_POST['button_cancel'])) && (!is_null($post_id))) ||
        ((is_null($post_id)) || (!isset($records[$post_id])))) {
        $retval = althing_show_edit_moderation_list($output,$records,$node_id,$sort_index,$href);
    } else {
        // not cancelled but perhaps we need to edit again?
        if ($edit_again) {
            $record = $records[$post_id];
            $dialogdef = althing_get_dialogdef_moderation($record,$viewonly);
            if ((althing_moderation_dialog_validate($dialogdef)) && (isset($_POST['button_done']))) {
                // force new data because Done saves data (unlike Cancel above)
                if (($records = althing_get_post_records($node_id,$sort_index,TRUE)) === FALSE) {
                    $output->add_message(t('error_retrieving_data','admin'));
                    return FALSE;
                }
                $retval = althing_show_edit_moderation_list($output,$records,$node_id,$sort_index,$href);
                return TRUE;
            } // else fall through to re-edit post
        } // else fall through to fresh-edit post
        $retval = althing_show_edit_moderation_post($output,$records[$post_id],$node_id,$sort_index,$viewonly,$edit_again);
    }
    return $retval;
} // althing_show_edit_moderation()


/** show a tabular overview of all posts in this althing
 *
 * this shows the main moderation screen: a list of all posts in a user-defined sort order.
 * Re-ordering can be done by clicking the column header. Click twice to reverse the order.
 * Due to space constraints the contents of the moderations column, visibility column and 
 * marbles column are just a single character. The explanation of these ad hoc codes is
 * given in the title. Note that the name column also shows the email address via a mouse
 * over (again, to save valuable space).
 *
 * The list ends with a [Cancel] button which ends the complete edit of this node (i.e.
 * returns to Page Manager).
 *
 * @param object &$output collects the html output (if any)
 * @param array &$records is the ordered array with all posts in this althing
 * @param int $node_id the node to which this module is connected
 * @param int $sort_index indicates the currently selected sort order
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function althing_show_edit_moderation_list(&$output,&$records,$node_id,$sort_index,$href) {
    global $WAS_SCRIPT_NAME;
    $hdrs = array('1' => array('title' => t('moderation_list_id_title','m_althing'),
                               'label' => t('moderation_list_id','m_althing')),
                  '2' => array('title' => t('moderation_list_datim_title','m_althing'),
                               'label' => t('moderation_list_datim','m_althing')),
                  '3' => array('title' => t('moderation_list_moderations_title','m_althing'),
                               'label' => t('moderation_list_moderations','m_althing')),
                  '4' => array('title' => t('moderation_list_visibility_title','m_althing'),
                               'label' => t('moderation_list_visibility','m_althing')),
                  '5' => array('title' => t('moderation_list_marbles_title','m_althing'),
                               'label' => t('moderation_list_marbles','m_althing')),
                  '6' => array('title' => t('moderation_list_name_title','m_althing'),
                               'label' => t('moderation_list_name','m_althing')),
                  '7' => array('title' => t('moderation_list_subject_title','m_althing'),
                               'label' => t('moderation_list_subject','m_althing'))
                  );
    $params = array(
        'job'    => JOB_PAGEMANAGER,
        'task'   => TASK_NODE_EDIT_CONTENT,
        'node'   => intval($node_id),
        'option' => 'moderation'
        );

    // 1A -- start with good intentions + explanation...
    $output->add_content('<h2>'.t('moderation_list_header','m_althing').'</h2>');
    $output->add_content(t('moderation_list_explanation','m_althing'));

    // 1B -- ...but bail out if there's nothing there
    if (sizeof($records) <= 0) {
        $msg = t('moderation_no_posts','m_althing');
        $output->add_content('<p><strong>'.$msg.'</strong>');
        $output->add_message($msg);
        $dialogdef = array('button_cancel' => dialog_buttondef(BUTTON_CANCEL));
        $output->add_content(dialog_quickform($href,$dialogdef));
        return TRUE;
    }

    // 2 -- table headers
    $class = 'header';
    $output->add_content('<p>');
    $output->add_content(html_table());
    $output->add_content('  '.html_table_row(array('class' => $class)));
    foreach($hdrs as $i => $hdr) {
        $attr = array('title' => $hdr['title']);
        $params['sort'] = (abs($sort_index) == $i) ? -$sort_index : $i;
        $label = html_a($WAS_SCRIPT_NAME,$params,$attr,$hdr['label']);
        $output->add_content('    '.html_table_head(NULL,$label));
    }
    $output->add_content('  '.html_table_row_close());
    $params['sort'] = $sort_index; // remember current sort order for correct next/prev in dialog #2

    // 3 -- table rows
    foreach($records as $id => $record) {
        $class = ($class == 'odd') ? 'even' : 'odd';
        $attributes = array('class' => $class, 'valign' => 'top');
        $attr_right = array('class' => $class, 'valign' => 'top', 'align' => 'right');

        // 1st column contains the link to the post detail, showing display_id, and post_id in title
        $params['post'] = intval($record['post_id']);
        $id_name = sprintf('p%d',$record['post_id']);
        $id_attr = array('name' => $id_name, 'title' => $id_name);
        $id_value = html_a($WAS_SCRIPT_NAME,$params,$id_attr,strval($record['display_id']));

        // 2nd column shows the short datim with the longer datim in the title
        $datim = html_tag('span',array('title' => strval($record['datim'])),short_datim($record['datim']));

        // 3rd column holds a clickable (if > 0) number of moderations; links to viewlog
        if ($record['moderations'] > 0) {
            $viewlog = array(
                'job'    => JOB_PAGEMANAGER,
                'task'   => TASK_NODE_EDIT_CONTENT,
                'node'   => intval($node_id),
                'option' => 'viewlog',
                'post'   => $record['post_id'],
                'sort'   => intval($sort_index)
                );
            $fragment = sprintf('p%d',$record['post_id']);
            $mods = html_a($WAS_SCRIPT_NAME,$viewlog,'',strval($record['moderations']),$fragment);
        } else {
            $mods = '0';
        }
        // 4th column contains 1-letter visibility and requires explanation via title
        if (db_bool_is(TRUE,$record['visible'])) {
            $vhu = html_tag('span', array('title' => t('post_visible_title','m_althing')),
                            t('post_visible','m_althing'));
        } elseif (db_bool_is(TRUE,$record['published'])) {
            $vhu = html_tag('span', array('title' => t('post_hidden_title','m_althing')),
                            t('post_hidden','m_althing'));
        } else {
            $vhu = html_tag('span', array('title' => t('post_unpublished_title','m_althing')),
                            t('post_unpublished','m_althing'));
        }
        // 6th column holds the author name + email in the title
        $author = html_tag('span',array('title' => htmlspecialchars($record['email'])),
                           htmlspecialchars($record['author_full_name']));
        $output->add_content('  '.html_table_row(array('class' => $class)));
        $output->add_content('    '.html_table_cell($attr_right,$id_value));
        $output->add_content('    '.html_table_cell($attributes,$datim));
        $output->add_content('    '.html_table_cell($attr_right,$mods));
        $output->add_content('    '.html_table_cell($attributes,$vhu));
        $output->add_content('    '.html_table_cell($attr_right,strval($record['marbles'])));
        $output->add_content('    '.html_table_cell($attributes,$author));
        $output->add_content('    '.html_table_cell($attributes,htmlspecialchars($record['subject'])));
        $output->add_content('  '.html_table_row_close());
    }
    $output->add_content(html_table_close());
    $output->add_content('<p>');
    $dialogdef = array('button_cancel' => dialog_buttondef(BUTTON_CANCEL));
    $output->add_content(dialog_quickform($href,$dialogdef));
    return TRUE;
} // althing_show_edit_moderation_list()


/** show a single post and allow for moderation of this post
 *
 * this shows the current contents of the post and allows the moderator to
 * change the visibility and the number of marbles assigned to this post.
 * The moderator is encouraged to add a remark to every moderation action.

 * @param object &$output collects the html output (if any)
 * @param array &$records is the ordered array with all posts in this althing
 * @param int $node_id the node to which this module is connected
 * @param int $sort_index indicates the currently selected sort order
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function althing_show_edit_moderation_post(&$output,&$record,$node_id,$sort_index,$viewonly,$edit_again) {
    global $WAS_SCRIPT_NAME;
    $params = array(
        'job'    => JOB_PAGEMANAGER,
        'task'   => TASK_NODE_EDIT_CONTENT,
        'node'   => intval($node_id),
        'option' => 'moderation',
        'sort'   => intval($sort_index)
        );
    //
    // 1 -- construct a navigation bar to step through all postings in selected order
    //
    $fragment = sprintf('p%d',$record['post_id']);
    $up = html_a($WAS_SCRIPT_NAME,$params,'',t('nav_up','m_althing'),$fragment);
    $prev = t('nav_prev','m_althing');
    if (isset($record['prev_id'])) {
        $params['post'] = $record['prev_id'];
        $prev = html_a($WAS_SCRIPT_NAME,$params,'',$prev);
    }
    $next = t('nav_next','m_althing');
    if (isset($record['next_id'])) {
        $params['post'] = $record['next_id'];
        $next = html_a($WAS_SCRIPT_NAME,$params,'',$next);
    }
    $navbar = array($prev,'-',$up,'-',$next);
    $output->add_content($navbar);

    //
    // 2 -- show the current post with details
    //
    $content = bbcode2html($record['content']);
    $aparams = array(
        '{POST_ID}' => $record['post_id'],
        '{DISPLAY_ID}' => $record['display_id'],
        '{ALTHING_ID}' => $record['althing_id'],
        '{AUTHOR_ID}' => (empty($record['author_id'])) ? '' : strval($record['author_id']),
        '{REGISTERED}' => t((empty($record['author_id'])) ? 'anonymous':'registered','m_althing'),
        '{AUTHOR}' => htmlspecialchars($record['author_full_name']),
        '{EMAIL}' => htmlspecialchars($record['email']),
        '{SUBJECT}' => htmlspecialchars($record['subject']),
        '{DATIM}' => $record['datim'],
        '{CONTENT}' =>  $content,
        '{VISIBLE}' => t((db_bool_is(TRUE,$record['visible'])) ? 'flag_yes' : 'flag_no','m_althing'),
        '{PUBLISHED}' => t((db_bool_is(TRUE,$record['published'])) ? 'flag_yes' : 'flag_no','m_althing'),
        '{MARBLES}' => strval($record['marbles']),
        '{REMOTE_ADDR}' => $record['remote_addr'],
        '{MTIME}' => $record['mtime'],
        '{MODERATIONS}' => strval($record['moderations']),
        '{MUSER_ID}' => (empty($record['muser_id'])) ? '' : strval($record['muser_id']),
        '{MODERATOR_NAME}' => (empty($record['moderator_name'])) ? '' : strval($record['moderator_name'])
        );
    $output->add_content('<h2>'.t('moderation_post_header','m_althing',$aparams).'</h2>');
    $output->add_content('<hr>');
    $output->add_content('<div class="althing_header">');
    for ($i=1; $i < 7; ++$i) {
        $s = t('moderation_post_headerline'.$i,'m_althing',$aparams);
        if (!empty($s)) {
            $output->add_content('  '.$s.'<br>');
        }
    }
    $output->add_content('</div>');
    $output->add_content('<div class="althing_post">');
    $output->add_content(t('moderation_post_content','m_althing',$aparams));
    $output->add_content('</div>');
    $output->add_content('<hr>');

    //
    // 3 -- show a dialog where the user can manipulate this posting
    //
    $n = intval($record['moderations']);
    $moderations = t('moderation_post_moderations','m_althing',array('{MODERATIONS}' => strval($n)));
    if ($n > 0) {
        $viewlog = array(
            'job'    => JOB_PAGEMANAGER,
            'task'   => TASK_NODE_EDIT_CONTENT,
            'node'   => intval($node_id),
            'option' => 'viewlog',
            'post'   => $record['post_id'],
            'sort'   => intval($sort_index)
            );
        $fragment = sprintf('p%d',$record['post_id']);
        $moderations = html_a($WAS_SCRIPT_NAME,$viewlog,'',$moderations,$fragment);
    }
    $dialogdef = althing_get_dialogdef_moderation($record,$viewonly);
    if ($edit_again) {
        if (althing_moderation_dialog_validate($dialogdef)) {
            // editing a valid dialog again implies the user used [Save]
            // we don't want to keep the remark because it is already logged, so there...
            $dialogdef['remark']['value'] = '';
        }
    }
    $params['task'] = TASK_SAVE_CONTENT;
    $params['post'] = $record['post_id'];
    $href = href($WAS_SCRIPT_NAME, $params);
    $output->add_content(html_form($href));
    $output->add_content($moderations.' |'); // separator between moderations and marbles


    // 3A -- 'manually' render all elements to get radio buttons on a single line etc.
    $buttons_seen = FALSE;
    foreach($dialogdef as $item) {
        $label = dialog_get_label($item);
        if (!empty($label)) {
            if ($item['name'] == 'published') {
                $output->add_content('|'); // separator between 'visible' and 'published' radio buttons
            }
            $output->add_content($label);
        }
        $widget = dialog_get_widget($item);
        if (is_array($widget)) {
            $output->add_content($widget);
            if ($item['name'] != 'visible') {
                $output->add_content('<br>');
            }
        } else {
            $postfix = '';
            if ($item['type'] == F_SUBMIT) {
                if (!$buttons_seen) {
                    $buttons_seen = TRUE;
                    $output->add_content('<p>');
                }
            } elseif (!((isset($item['hidden'])) && ($item['hidden']))) {
                $postfix = '<br>';
            }
            $output->add_content($widget.$postfix);
        }
    }
    $output->add_content('<hr>');
    $output->add_content(html_form_close());

    //    $output->add_content(dialog_quickform($href,$dialogdef));
    $output->add_content($navbar);
    $output->add_content('<p>');
    return TRUE;
} // althing_show_edit_moderation_post()


/** validate and save the modified althing configuration linked to node $node_id
 *
 * this validates and saves the data that was submitted by the user.
 *
 * If validation fails, the flag $edit_again is set to TRUE and the return value
 * is FALSE.
 *
 * If the user has cancelled the operation, the flag $edit_again is set to FALSE
 * and the return value is also FALSE.
 *
 * If the modified data is stored successfully, the return value is TRUE.
 * The flag edit_again is set to TRUE if the user saved via the [Save] button,
 * and to FALSE if the user saved via the [Done] button.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which the content is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing and hence saving is not allowed
 * @param bool &$edit_again set to TRUE if we need to edit the content again, FALSE otherwise
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function althing_save_moderation(&$output,$area_id, $node_id,$viewonly,&$edit_again) {
    global $USER;
    $now = strftime('%Y-%m-%d %T');

    // 1 -- bail out if cancelled or viewonly
    $post_id = get_parameter_int('post',NULL);
    if ((isset($_POST['button_cancel'])) || ($viewonly)) {
        if (is_null($post_id)) { // [Cancel] button on overview screen...
            $edit_again = TRUE; // trick to let [Cancel] return to the Content screen
        } else {
            $output->add_message(t('cancelled','admin')); // feedback about this 'partial' Cancel...
            $edit_again = TRUE; // We want to see the list of posts again, do not exit completely
        }
        return FALSE;
    }

    // 2 -- setup relevant data in array in the desired sort order
    $sort_index = get_parameter_int('sort',1);
    if (($records = althing_get_post_records($node_id,$sort_index)) === FALSE) {
        $output->add_message(t('error_retrieving_data','admin'));
        return FALSE;
    }
    if ((is_null($post_id)) || (!isset($records[$post_id]))) {
        logger(sprintf("%s(): weird: no valid post_id (%s) to save om '%d', mmm...",__FUNCTION__,
                       (is_null($post_id)) ? 'NULL' : strval($post_id), $node_id));
        $edit_again = TRUE; // Make sure we return to show_edit
        return FALSE; // Dirty trick to return to the destinations dialog
    }

    // 3 -- now is the time to actually look at the data of the moderated post
    $record = $records[$post_id];
    $dialogdef = althing_get_dialogdef_moderation($record,$viewonly);
    if (!althing_moderation_dialog_validate($dialogdef)) { // show errors to the user + do it again
        foreach($dialogdef as $k => $item) {
            if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                $output->add_message($item['error_messages']);
            }
        }
        $edit_again = TRUE;
        return FALSE;
    }
    // 4 -- actually save the settings
    $table = 'althing_posts';

    $old_published = intval($dialogdef['published']['old_value']);
    $new_published = intval($dialogdef['published']['value']);
    $old_visible = intval($dialogdef['visible']['old_value']);
    $new_visible = intval($dialogdef['visible']['value']);
    $old_marbles = intval($dialogdef['marbles']['old_value']);
    $new_marbles = intval($dialogdef['marbles']['value']);
    $dirty = FALSE;
    $announce = FALSE;
    if ($old_published != $new_published) {
        if ($new_published == 0) {
            logger(sprintf("%s(): Huh? Cannot UNpublish post %d; bail out",__FUNCTION__,$post_id));
            $output->add_message(t('error_invalid_parameters','admin'));
            $edit_again = TRUE;
            return FALSE;
        } else {
            $dirty = TRUE;
            $announce = TRUE; // because published changes from '0' => '1'
            $new_visible = 1; // and that implies visibility '1'
        }
    }
    if ($old_visible != $new_visible) {
        $dirty = TRUE;
    }
    if ($old_marbles != $new_marbles) {
        $dirty = TRUE;
    }
    if (!$dirty) { // nothing to do
        $edit_again = TRUE; // trick to stay in case of [Done] button pressed
        $output->add_message(t('nothing_to_save','m_althing'));
        return FALSE;
    }
    // At this point we know we actually have to save some changed data, including a log message
    // Go prepare the $action message
    $aparams = array(
        '{OLD_MARBLES}' => strval($old_marbles),
        '{NEW_MARBLES}' => strval($new_marbles)
        );
    if ($old_visible == 1) {
        $aparams['{OLD_VISIBILITY}'] = t('moderation_action_visible','m_althing');
    } elseif ($old_published == 1) {
        $aparams['{OLD_VISIBILITY}'] = t('moderation_action_hidden','m_althing');
    } else {
        $aparams['{OLD_VISIBILITY}'] = t('moderation_action_unpublished','m_althing');
    }
    if ($new_visible == 1) {
        $aparams['{NEW_VISIBILITY}'] = t('moderation_action_visible','m_althing');
    } elseif ($new_published == 1) {
        $aparams['{NEW_VISIBILITY}'] = t('moderation_action_hidden','m_althing');
    } else {
        $aparams['{NEW_VISIBILITY}'] = t('moderation_action_unpublished','m_althing');
    }
    if ($old_visible != $new_visible) {
        if ($old_marbles != $new_marbles) {
            $action = t('moderation_action_both','m_althing',$aparams);
        } else {
            $action = t('moderation_action_visibility','m_althing',$aparams);
        }
    } else {
        $action = t('moderation_action_marbles','m_althing',$aparams);
    }
    // Now we can actually save the data to the posts table...
    $table = 'althing_posts';
    $fields = array(
        'visible'   => ($new_visible == 1) ? TRUE : FALSE,
        'published' => ($new_published == 1) ? TRUE : FALSE,
        'marbles'   => $new_marbles,
        'mtime'     => $now,
        'muser_id'  => $USER->user_id
        );
    $where = array('post_id' => $post_id);
    if (db_update($table,$fields,$where) === FALSE) {
        logger(sprintf("%s(): error updating post '%d' in althing '%d': %s",
                       __FUNCTION__, $post_id, $node_id,db_errormessage()));
        $output->add_message(t('error_saving_data','m_althing'));
        $edit_again = TRUE;
        return FALSE;
    }
    // ...and the log table
    $table = 'althing_log_messages';
    $fields = array(
        'post_id'    => $post_id,
        'althing_id' => intval($node_id),
        'action'     => $action,
        'remark'     => $dialogdef['remark']['value'],
        'mtime'      => $now,
        'muser_id'   => $USER->user_id
        );
    if (db_insert_into($table,$fields) === FALSE) {
        logger(sprintf("%s(): error storing log message for  post '%d' in althing '%d': %s",
                       __FUNCTION__, $post_id, $node_id,db_errormessage()));
        $output->add_message(t('error_saving_data','m_althing'));
        $edit_again = TRUE;
        return FALSE;
    }
    $output->add_message(t('success_saving_data','admin'));

    // There's only one thing left now: perhaps send an announcement to the
    // subscribers IFF the publication status changed from UNpublished to published.
    // This means that we
    //  - have to retrieve the list of subscriber addresses
    //  - determine if the area is public or private.
    // Here we go...
    if ($announce) {
        $table = 'althings';
        $field = 'email_subscribers';
        $where = array('althing_id' => intval($node_id));
        if (($config = db_select_single_record($table, $field, $where)) === FALSE) {
            logger(sprintf('%s(): error retrieving configuration; no addresses for announce: %s',
                           __FUNCTION__,db_errormessage()));
        } else {
            $subscribers = althing_get_emails($config[$field]);
            if (sizeof($subscribers) > 0) {
                $moderator = FALSE; // subscribers
                $flag = FALSE; // for a post that used to be unpublished until now: visible=FALSE
                $where = array('area_id' => intval($area_id));
                if (($area = db_select_single_record('areas','is_private', $where)) === FALSE) {
                    logger(sprintf("%s(): cannot retrieve is_private for area=%d; assuming 'public': %s",
                                   __FUNCTION__,$area_id,db_errormessage()));
                    $private = FALSE; // assume public area after error
                } else {
                    $private = (db_bool_is(TRUE,$area['is_private'])) ? TRUE : FALSE;
                }
                $record['published'] = SQL_TRUE;   // keep $record up to date for mail. Since we are here
                $record['visible'] = SQL_TRUE;     // announcing, published (and visible) MUST be true...
                $record['marbles'] = $new_marbles;
                $record['althing_id'] = intval($node_id);
                althing_send_alerts($subscribers,$moderator,$private,$flag,$area_id,$record);
            }
        }
    }
    $edit_again = TRUE;
    return TRUE;
} // althing_save_moderation()


// ==================================================================
// ======================== ALTHING VIEW LOG ========================
// ==================================================================


/** show a list of moderation messages to the user
 *
 * This is basicly a dump of the althing_log_messages table.
 * However, we have tried to jazz it up a little. 
 *
 * The features are as follows.
 * You can arrive here via de link in the submenu. In that case we simply show
 * the available records (all available records) in ascending order, grouped by post_id.
 * This dump consists of a single row in the HTML-table with essential information about
 * the posting to which the messages belong, That row is followed by two rows per
 * logmessage, containinf the date of moderation, the moderator name, the remark and
 * the action.
 *
 * An additional feature is the link in the header line (the 1st column: ID): clicking
 * that link reverses the order of the list.
 * Also every posting has their own link (via the ID). This link leads to the
 * moderation screen (dialog #2) where this posting can be edited.
 * In that individual post dialog there is a link back to here. However, in this case
 * the output of this routine is limited to the selected posting.
 *
 * There is also a link in the moderation list (in the moderations column) which
 * yields the messages for a single posting.
 *
 * The layout of the table is more or less as follows:
 * <code>
 * ID Date             Name               Subject/remark
 * 10 yyyy-mm-dd hh:mm Anonymous Coward 4 Re: Chocolate galore
 *    yyyy-mm-dd hh:mm Maria Montessori   Remark of Maria
 *    here-comes-the-action-text-spanning-three-columns
 * </code>
 *
 * @param object &$output collects the html output (if any)
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function althing_show_edit_viewlog(&$output,$node_id,$module,$href) {
    global $DB,$WAS_SCRIPT_NAME;

    // Cancel from edit viewlog should return to content screen
    if (isset($_POST['button_cancel'])) {
        return althing_show_content($output,$node_id,$module);
    }

    // 1A - prepare a few tweaks for the complicated SQL-statement below
    $post_id = get_parameter_int('post',NULL);
    $where_post_id = (is_null($post_id)) ? '' : sprintf('AND p.post_id = %d ',$post_id);
    $sort_index = get_parameter_int('sort',1);
    $desc = ($sort_index < 0) ? ' DESC' : '';

    // 1B -- grab all relevant data in a single trip to the DB
    $sql = sprintf('SELECT m.*, p.display_id,p.datim,p.author_full_name,p.subject,p.email, u.full_name '.
                   'FROM %salthing_log_messages m '.
                   'LEFT JOIN %salthing_posts p USING (post_id) '.
                   'LEFT JOIN %susers u ON m.muser_id = u.user_id '.
                   'WHERE m.althing_id = %d %s'.
                   'ORDER BY m.post_id%s, m.log_message_id%s',
                   $DB->prefix,      // althing_log_messages
                   $DB->prefix,      // althing_posts
                   $DB->prefix,      // users
                   intval($node_id), // althing_id
                   $where_post_id,   // 'AND p.post_id = nnn '
                   $desc,            // ' DESC' (for post_id)
                   $desc);           // ' DESC' (for log_message_id)
    if (($DBResult = $DB->query($sql)) === FALSE) {
        logger(sprintf("%s(): error retrieving moderation messages for althing '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        return FALSE;
    }
    $records = $DBResult->fetch_all_assoc('log_message_id');
    $DBResult->close();

    // 1C -- start with good intentions + explanation...
    $output->add_content('<h2>'.t('viewlog_header','m_althing').'</h2>');
    $output->add_content(t('viewlog_explanation','m_althing'));

    // 1D -- ...but bail out if there's nothing there
    if (sizeof($records) <= 0) {
        $msg = t('viewlog_no_messages','m_althing');
        $output->add_content('<p><strong>'.$msg.'</strong>');
        $output->add_message($msg);
        $dialogdef = array('button_cancel' => dialog_buttondef(BUTTON_CANCEL));
        $output->add_content(dialog_quickform($href,$dialogdef));
        return TRUE;
    }

    // 2A -- preparations, also for the HTML table header
    $params = array(
        'job'    => JOB_PAGEMANAGER,
        'task'   => TASK_NODE_EDIT_CONTENT,
        'node'   => intval($node_id),
        'option' => 'moderation',
        'sort'   => $sort_index // remember the current sort order if any
        );
    $viewlog = array(
        'job'    => JOB_PAGEMANAGER,
        'task'   => TASK_NODE_EDIT_CONTENT,
        'node'   => intval($node_id),
        'option' => 'viewlog',
        'sort'   => -$sort_index // reverse the sort order when user clicks ID-header
        );
    if (!is_null($post_id)) {
        $viewlog['post'] = strval($post_id);
    }
    $id_hdr = html_a($WAS_SCRIPT_NAME,$viewlog,'',t('viewlog_id','m_althing'));

    // 2B -- start with the page and open the HTML-table + column headers
    $class = 'header';
    $output->add_content('<p>');
    $output->add_content(html_table());
    $output->add_content('  '.html_table_row(array('class' => $class)));
    $output->add_content('    '.html_table_head(NULL,$id_hdr));
    $output->add_content('    '.html_table_head(NULL,t('viewlog_date','m_althing')));
    $output->add_content('    '.html_table_head(NULL,t('viewlog_name','m_althing')));
    $output->add_content('    '.html_table_head(NULL,t('viewlog_subject','m_althing')));
    $output->add_content('  '.html_table_row_close());

    // 3 -- step through the available records
    $last_post_id = 0;
    foreach($records as $log_message_id => $record) {
        $post_id = intval($record['post_id']);
        if ($last_post_id != $post_id) {
            // 3A -- start of a new post detected; insert a 'post header', sort of...
            $class = 'header';
            $attributes = array('class' => $class);
            $attr_right= array('class' => $class,'align' => 'right');
            $params['post'] = intval($record['post_id']);
            $id_name = sprintf('p%d',$record['post_id']); // a fragment to return exactly here...
            $id_attr = array('name' => $id_name, 'title' => $id_name); // ...and a link to the post too
            $id_value = html_a($WAS_SCRIPT_NAME,$params,$id_attr,strval($record['display_id']));
            $author = html_tag('span',array('title' => htmlspecialchars($record['email'])),
                               htmlspecialchars($record['author_full_name']));
            $datim = html_tag('span',
                              array('title' => strval($record['datim'])),
                              short_datim($record['datim']));
            $subject = htmlspecialchars($record['subject']);
            $output->add_content('  '.html_table_row($attributes));
            $output->add_content('    '.html_table_cell($attr_right,html_tag('strong','',$id_value)));
            $output->add_content('    '.html_table_cell($attributes,html_tag('strong','',$datim)));
            $output->add_content('    '.html_table_cell($attributes,html_tag('strong','',$author)));
            $output->add_content('    '.html_table_cell($attributes,html_tag('strong','',$subject)));
            $output->add_content('  '.html_table_row_close());
            $last_post_id = $post_id;
        }
        // 3B -- output the first row of this log message, includes: ID, Date, Moderator, Remark
        $class = ($class == 'odd') ? 'even' : 'odd';
        $attributes = array('class' => $class);
        $attr_right = array('class' => $class,'align' => 'right');
        $attr_colspan =  array('class' => $class,'colspan' => '3');
        $datim = html_tag('span',array('title' => strval($record['mtime'])),short_datim($record['mtime']));
        $moderator = htmlspecialchars($record['full_name']);
        $remark = htmlspecialchars($record['remark']);
        $action =  htmlspecialchars($record['action']);
        $output->add_content('  '.html_table_row($attributes));
        $output->add_content('    '.html_table_cell($attributes,'&nbsp;'));
        $output->add_content('    '.html_table_cell($attributes,$datim));
        $output->add_content('    '.html_table_cell($attributes,$moderator));
        $output->add_content('    '.html_table_cell($attributes,$remark));
        $output->add_content('  '.html_table_row_close());

        // 3C -- output the action field over three columns wide
        $class = ($class == 'odd') ? 'even' : 'odd';
        $attributes = array('class' => $class);
        $attr_colspan =  array('class' => $class,'colspan' => '3');
        $output->add_content('  '.html_table_row($attributes));
        $output->add_content('    '.html_table_cell($attributes,'&nbsp;'));
        $output->add_content('    '.html_table_cell($attr_colspan,$action));
        $output->add_content('  '.html_table_row_close());
    }
    $output->add_content(html_table_close());
    $output->add_content('<p>');

    // 4 -- done, close with a lone [Cancel] button
    $dialogdef = array('button_cancel' => dialog_buttondef(BUTTON_CANCEL));
    $output->add_content(dialog_quickform($href,$dialogdef));
    return TRUE;
} // althing_show_edit_viewlog()


/** almost a dummy routine but it must exist to handle the lone [Cancel] button
 *
 * @param object &$output collects the html output (if any)
 * @param int $node_id the node to which the content is connected
 * @param bool $viewonly if TRUE, editing and hence saving is not allowed
 * @param bool &$edit_again set to TRUE if we need to edit the content again, FALSE otherwise
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function althing_save_viewlog(&$output,$viewonly,&$edit_again) {
    // 1 -- bail out if cancelled or viewonly
    if ((isset($_POST['button_cancel'])) || ($viewonly)) {
        $edit_again = TRUE; // trick to let [Cancel] return to the Content screen
        return FALSE;
    }
    // we should never get here because we only have a single Cancel button
    $edit_again = TRUE;
    return FALSE;
} // althing_save_viewlog()


// ==================================================================
// ========================= ALTHING REPORT =========================
// ==================================================================

function althing_show_edit_report(&$output,$node_id,$module,$edit_again,$href) {

    // 1 -- Cancel from report returns to content screen
    if (isset($_POST['button_cancel'])) {
        return althing_show_content($output,$node_id,$module);
    }

    // 2 -- get raw data in $posts array keyed by display_id (not post_id)
    $althing_id = intval($node_id);
    $table = 'althing_posts';
    $fields = '*';
    $where = array('althing_id' => $althing_id, 'published' => TRUE);
    $order = array('display_id');
    if (($posts = db_select_all_records($table, $fields,$where,$order,'display_id')) === FALSE) {
        $output->add_message(t('error_retrieving_posts','m_althing'));
        logger(sprintf("%s(): cannot retrieve althing posts '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        return FALSE;
    }

    // 3 -- fetch data from POST array and render the selection dialog
    $dialogdef = althing_get_dialogdef_report();
    if ($edit_again) {
        althing_report_dialog_validate($dialogdef,$posts); // side effect: fetch data from POST
        foreach($dialogdef as $k => $item) {
            if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                $output->add_message($item['error_messages']);
            }
        }
    }
    $output->add_content('<h2>'.t('report_header','m_althing').'</h2>');
    $output->add_content(dialog_quickform($href,$dialogdef));

    // 4 -- actually output the requested data as specified in the dialogdef
    if (empty($dialogdef['selection']['value'])) {
        return TRUE;
    }
    // 4A -- what posts and fields do we show and in which format?
    $display_ids = explode(',',$dialogdef['selection']['value']);
    $format = intval($dialogdef['format']['value']);
    $show_message_id = ($dialogdef['message_id']['value'] == 1) ? TRUE : FALSE;
    $show_author = ($dialogdef['author']['value'] == 1) ? TRUE : FALSE;
    $show_subject = ($dialogdef['subject']['value'] == 1) ? TRUE : FALSE;
    $show_date = ($dialogdef['date']['value'] == 1) ? TRUE : FALSE;
    $show_body = ($dialogdef['body']['value'] == 1) ? TRUE : FALSE;

    // 4B -- add a header repeating the choiced made
    $newline = '<br>';
    $output->add_content('<hr>');
    // Posts: 1,2,3,4,...
    $output->add_content(sprintf('%s %s%s',str_replace('~','',t('report_selection_label','m_althing')),
                                 $dialogdef['selection']['value'],$newline));
    // Fields: Field1,Field2,...
    $f = array();
    if ($show_message_id) { $f[] = str_replace('~','',t('report_message_id_option','m_althing')); }
    if ($show_author) { $f[] = str_replace('~','',t('report_author_option','m_althing')); }
    if ($show_subject) { $f[] = str_replace('~','',t('report_subject_option','m_althing')); }
    if ($show_date) { $f[] = str_replace('~','',t('report_date_option','m_althing')); }
    if ($show_body) { $f[] = str_replace('~','',t('report_body_option','m_althing')); }
    $output->add_content(sprintf('%s %s%s',t('report_fields_label','m_althing'),implode(',',$f),$newline));
    // Format: HTML
    switch($format) {
    case 1: $s = str_replace('~','',t('report_format_htmlpage_label','m_althing')); break;
    case 2: $s = str_replace('~','',t('report_format_htmlcode_label','m_althing')); 
        $newline = htmlspecialchars(nl2br("\n")).'<br>';
        break;
    case 3: $s = str_replace('~','',t('report_format_bbcode_label','m_althing')); break;
    default: $s = ''; break;
    }
    $output->add_content(sprintf('%s %s<br><br>',t('report_format_label','m_althing'),$s));

    // 4C -- output all selected fields of selected posts in selected format
    foreach($display_ids as $id) {
        $id = intval($id);
        switch($format) {
        case 1:
            $content = bbcode2html($posts[$id]['content'],TRUE); // TRUE=fully qualified domain
            break;
        case 2: 
            $content = nl2br(htmlspecialchars(bbcode2html($posts[$id]['content'],TRUE)));
            break;
        case 3: 
            $content = nl2br(htmlspecialchars($posts[$id]['content']));
            break;
        default: 
            $content = '?';
            break;
        }
        $aparams = array(
            '{DISPLAY_ID}' => $posts[$id]['display_id'],
            '{POST_ID}' => $posts[$id]['post_id'],
            '{AUTHOR}' => htmlspecialchars($posts[$id]['author_full_name']),
            '{REGISTERED}' => t((empty($posts[$id]['author_id'])) ? 'anonymous':'registered','m_althing'),
            '{SUBJECT}' => htmlspecialchars($posts[$id]['subject']),
            '{DATIM}' => htmlspecialchars($posts[$id]['datim']),
            '{SHORT_DATIM}' => short_datim($posts[$id]['datim']),
            '{CONTENT}' => $content
            );
        if ($show_message_id) {
            $output->add_content(t('report_id','m_althing',$aparams).$newline);
        }
        if ($show_author) {
            $output->add_content(t('report_author','m_althing',$aparams).$newline);
        }
        if ($show_subject) {
            $output->add_content(t('report_subject','m_althing',$aparams).$newline);
        }
        if ($show_date) {
            $output->add_content(t('report_date','m_althing',$aparams).$newline);
        }
        if ($show_body) {
            $output->add_content(t('report_content','m_althing',$aparams).$newline);
        }
    }
    $output->add_content('<br>');
    return TRUE;
} // althing_show_edit_report()


/** a dummy routine that pretends to "save" the report selections
 *
 * this is just a dummy routine to tell the caller that we want
 * to return to {@link althing_show_edit_report()} which is where the
 * actual work is done. It is not even necessary to check for either
 * Cancel or OK because that too is done in althing_show_edit_report().
 * We do set the edit_again flag.
 *
 * @param object &$output collects the html output (if any)
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @return bool always FALSE to indicate we need to return to althing_show_edit_report(), always
 */
function  althing_save_report(&$output,$viewonly,&$edit_again) {
    $edit_again = TRUE; // stay here
    return FALSE;
} // althing_save_report()


// ==================================================================
// =========================== WORKHORSES ===========================
// ==================================================================


/** construct a dialog definition for the main althing configuration
 *
 * @param object &$output collects the html output (if any)
 * @param int $viewonly if TRUE the Save button is not displayed and config values cannot be changed
 * @param int $node_id identifies the current althing
 * @return array dialog definition
 * @uses $USER, $CFG
 */
function althing_get_dialogdef_configuration(&$output, $viewonly, $node_id) {
    global $USER,$CFG;
    //
    // 1 -- generic configuration parameters
    //
    $table = 'althings';
    $fields = array('header','introduction','question','conclusion','footer','status',
                    'default_visibility','email_moderators','email_subscribers'
                   );
    $where = array('althing_id' => intval($node_id));
    if (($record = db_select_single_record($table, $fields, $where)) === FALSE) {
        logger(sprintf('%s(): error retrieving configuration: %s',__FUNCTION__,db_errormessage()));
        $output->add_message(t('error_retrieving_data','admin'));
        $record = array(
            'header' => '',
            'introduction' => '',
            'question' => '',
            'conclusion' => '',
            'footer' => '',
            'status' => ALTHING_STATUS_OPENED,
            'default_visibility' => ALTHING_DEFAULT_VISIBILITY,
            'email_moderators' => $USER->email,
            'email_subscribers' => ''
            );
    }
    $status_options = array(
        strval(ALTHING_STATUS_OPENED) => array(
            'option' => t('status_opened_label','m_althing'),
            'title' => t('status_opened_title', 'm_althing')
            ),
        strval(ALTHING_STATUS_FROZEN) => array(
            'option' => t('status_frozen_label','m_althing'),
            'title' => t('status_frozen_title', 'm_althing')
            ),
        strval(ALTHING_STATUS_CLOSED) => array(
            'option' => t('status_closed_label','m_althing'),
            'title' => t('status_closed_title', 'm_althing')
            )
        );
    $visibility_options = array(
        strval(ALTHING_DEFAULT_VISIBILITY_HIDDEN) => array(
            'option' => t('default_visibility_hidden_label','m_althing'),
            'title' => t('default_visibility_hidden_title', 'm_althing')
            ),
        strval(ALTHING_DEFAULT_VISIBILITY_REGISTERED) => array(
            'option' => t('default_visibility_registered_label','m_althing'),
            'title' => t('default_visibility_registered_title', 'm_althing')
            ),
        strval(ALTHING_DEFAULT_VISIBILITY_VISIBLE) => array(
            'option' => t('default_visibility_visible_label','m_althing'),
            'title' => t('default_visibility_visible_title', 'm_althing')
            )
        );
    $dialogdef = array(
        'status' => array(
            'type' => F_RADIO,
            'name' => 'status',
            'value' => strval($record['status']),
            'old_value' => strval($record['status']),
            'options' => $status_options,
            'viewonly' => $viewonly,
            'title' => t('status_title','m_althing'),
            'label' => t('status_label','m_althing')
            ),
        'default_visibility' => array(
            'type' => F_RADIO,
            'name' => 'default_visibility',
            'value' => strval($record['default_visibility']),
            'old_value' => strval($record['default_visibility']),
            'options' => $visibility_options,
            'viewonly' => $viewonly,
            'title' => t('default_visibility_title','m_althing'),
            'label' => t('default_visibility_label','m_althing')
            ),
        'email_moderators' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'email_moderators',
            'minlength' => 0,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 50,
            'rows' => 10,
            'label' => t('email_moderators_label','m_althing'),
            'title' => t('email_moderators_title','m_althing'),
            'viewonly' => $viewonly,
            'value' => $record['email_moderators'],
            'old_value' => $record['email_moderators']
            ),
        'email_subscribers' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'email_subscribers',
            'minlength' => 0,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 50,
            'rows' => 10,
            'label' => t('email_subscribers_label','m_althing'),
            'title' => t('email_subscribers_title','m_althing'),
            'viewonly' => $viewonly,
            'value' => $record['email_subscribers'],
            'old_value' => $record['email_subscribers']
            ),
        'header' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'header',
            'minlength' => 0,
            'maxlength' => 240,
            'columns' => 40,
            'label' => t('header_label','m_althing'),
            'title' => t('header_title','m_althing'),
            'viewonly' => $viewonly,
            'value' => $record['header'],
            'old_value' => $record['header']
            ),
        'introduction' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'introduction',
            'minlength' => 0,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 50,
            'rows' => 10,
            'label' => t('introduction_label','m_althing'),
            'title' => t('introduction_title','m_althing'),
            'viewonly' => $viewonly,
            'value' => $record['introduction'],
            'old_value' => $record['introduction']
            ),
        'question' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'question',
            'minlength' => 0,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 50,
            'rows' => 10,
            'label' => t('question_label','m_althing'),
            'title' => t('question_title','m_althing'),
            'viewonly' => $viewonly,
            'value' => $record['question'],
            'old_value' => $record['question']
            ),
        'conclusion' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'conclusion',
            'minlength' => 0,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 50,
            'rows' => 10,
            'label' => t('conclusion_label','m_althing'),
            'title' => t('conclusion_title','m_althing'),
            'viewonly' => $viewonly,
            'value' => $record['conclusion'],
            'old_value' => $record['conclusion']
            ),
        'footer' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'footer',
            'minlength' => 0,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 50,
            'rows' => 10,
            'label' => t('footer_label','m_althing'),
            'title' => t('footer_title','m_althing'),
            'viewonly' => $viewonly,
            'value' => $record['footer'],
            'old_value' => $record['footer']
            ),
        'csrftoken' => dialog_csrftoken()
        );
    //
    // 2 -- maybe add a save button but always a cancel button
    //
    if (!$viewonly) {
        $dialogdef['button_save'] = dialog_buttondef(BUTTON_SAVE);
        $dialogdef['button_done'] = dialog_buttondef(BUTTON_DONE);
    }
    $dialogdef['button_cancel'] = dialog_buttondef(BUTTON_CANCEL);
    return $dialogdef;
} // althing_get_dialogdef_configuration()


/** construct the moderation dialog
 *
 * @param array &$record contains the information about this post
 * @param bool $viewonly if TRUE, editing and hence saving is not allowed
 * @return array dialog definition
 */
function althing_get_dialogdef_moderation(&$record,$viewonly) {
    $visible_options = array(
        '1' => array(
            'option' => t('moderation_post_visible_visible_label','m_althing'),
            'title' => t('moderation_post_visible_visible_title', 'm_althing')
            ),
        '0' => array(
            'option' => t('moderation_post_visible_hidden_label','m_althing'),
            'title' => t('moderation_post_visible_hidden_title', 'm_althing')
            )
        );
    $published_options = array(
        '1' => array(
            'option' => t('moderation_post_published_published_label','m_althing'),
            'title' => t('moderation_post_published_published_title', 'm_althing')
            ),
        '0' => array(
            'option' => t('moderation_post_published_unpublished_label','m_althing'),
            'title' => t('moderation_post_published_unpublished_title', 'm_althing')
            )
        );
    if (db_bool_is(TRUE,$record['visible'])) {
        $visible = '1';
        $published = '1';
    } else {
        $visible = '0';
        $published = (db_bool_is(TRUE,$record['published'])) ? '1' : '0';
    }
    $dialogdef = array(
        'marbles' => array(
            'type' => F_INTEGER,
            'name' => 'marbles',
            'columns' => 5,
            'maxlength' => 10,
            'minvalue' => 0,
            'maxvalue' => ALTHING_MARBLES_LIMIT,
            'viewonly' => $viewonly,
            'value' => intval($record['marbles']),
            'old_value' => intval($record['marbles']),
            'title' => t('moderation_post_marbles_title','m_althing'),
            'label' => t('moderation_post_marbles_label','m_althing')
            ),
        'visible' => array(
            'type' => F_RADIO,
            'name' => 'visible',
            'value' => $visible,
            'old_value' => $visible,
            'options' => $visible_options,
            'viewonly' => (($viewonly) || ($published == '0')),
            'title' => t('moderation_post_visible_title','m_althing'),
            'label' => t('moderation_post_visible_label','m_althing')
            ),
        'published' => array(
            'type' => F_RADIO,
            'name' => 'published',
            'value' => $published,
            'old_value' => $published,
            'options' => $published_options,
            'viewonly' => (($viewonly) || ($published == '1')),
            'title' => t('moderation_post_published_title','m_althing'),
            'label' => t('moderation_post_published_label','m_althing')
            ),
        'remark' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'remark',
            'minlength' => 0, // adding a message is enforced in
            'maxlength' => 8192, // arbitrary; 8 kB
            'columns' => 50,
            'rows' => 1,
            'label' => t('moderation_post_remark_label','m_althing'),
            'title' => t('moderation_post_remark_title','m_althing'),
            'viewonly' => $viewonly,
            'value' => '',
            'old_value' => ''
            ),
        'csrftoken' => dialog_csrftoken()
        );
    //
    // 2 -- maybe add a save button but always a cancel button
    //
    if (!$viewonly) {
        $dialogdef['button_save'] = dialog_buttondef(BUTTON_SAVE);
        $dialogdef['button_done'] = dialog_buttondef(BUTTON_DONE);
    }
    $dialogdef['button_cancel'] = dialog_buttondef(BUTTON_CANCEL);
    return $dialogdef;
} // althing_get_dialogdef_moderation()


/** validation of moderation dialog
 *
 * this validates the the moderation dialog. Special issue here is
 * the contents of the remark-field: if there are no changes in the
 * visibility and the marbles, a remark is not necessary and should
 * not trigger an error message. We already have the limit set to 0
 * in the dialogdef, but we _do_ check for the length in case of changes.
 *
 * @param array &$dialogdef the dialog to check
 * @return TRUE if valid, FALSE otherwise (+messages in dialogdef)
 */
function althing_moderation_dialog_validate(&$dialogdef) {
    // 1 -- first the basic fields must be OK. If not, return failed status to caller
    if (!dialog_validate($dialogdef)) {
        return FALSE;
    }
    // 2 -- check for changes between old and new, all is well if no changes
    if ((intval($dialogdef['published']['old_value']) == intval($dialogdef['published']['value'])) &&
        (intval($dialogdef['visible']['old_value']) == intval($dialogdef['visible']['value'])) &&
        (intval($dialogdef['marbles']['old_value']) == intval($dialogdef['marbles']['value']))) {
        return TRUE;
    }
    // 3 -- there is a change, so remark should be longer than 1 character
    if (strlen($dialogdef['remark']['value']) > 0) {
        return TRUE;
    }
    // 4 -- oops. Tell user to add a remark
    ++$dialogdef['remark']['errors'];
    $aparams = array('{FIELD}' => str_replace('~','',$dialogdef['remark']['label']), '{MIN}'=> '1');
    $dialogdef['remark']['error_messages'][] = t('validate_too_short','',$aparams);
    return FALSE;
} // althing_moderation_dialog_validate()


/** retrieve the information about all posts in an array
 *
 * this routine retrieves all relevant information of all posts in this
 * althing in a single assoc array. We do this in 1 go, which makes
 * for a crafted piece of SQL. However, basically we are simply
 * SELECTing all posts in this althing, in the order indicated
 * by $sort. As a service we also COUNT() the number of moderations
 * a post has had AND we fetch the full name of the moderator that
 * last changed the post.
 *
 * Note that the result is cached in a local static variable, this
 * might save some time after the user presses a button to move from
 * post-view to list-view.
 *
 * Sort order #3 is a little special: we sort on the # of moderations,
 * allowing the user to show the 'hottest' posts re: moderation to
 * show at the top of the list.
 *
 * Sort order #4 is an attempt to order the posts by visibility.
 * We need the CASE WHEN construct because it is hard to sort by boolean.
 *
 * @param int $node_id indicates the althing we are working with
 * @param int $sort is the desired sort order
 * @param bool $forced if TRUE the data is retrieved from DB even if it was cached
 * @return array|bool FALSE on error, array with data on success
 */
function althing_get_post_records($node_id, $sort, $forced=FALSE) {
    global $DB;
    static $records = NULL;

    // 0 -- maybe return a cached version of the records
    if ((!is_null($records)) && ($forced == FALSE)) {
        return $records;
    }

    // 1 -- fetch needed information about all available posts in $node_id in one go
    $desc = ($sort < 0) ? ' DESC' : '';
    switch (abs($sort)) {
    case 1:  $order = 'p.display_id'; break;
    case 2:  $order = 'p.datim'; break;
    case 3:  $order = 'moderations'; break;
    case 4:  $order = 'CASE WHEN visible THEN 0 WHEN published THEN 1 ELSE 2 END'; break;
    case 5:  $order = 'p.marbles'; break;
    case 6:  $order = 'p.author_full_name'; break;
    case 7:  $order = 'p.subject'; break;
    default: $order = 'p.display_id'; $sort = 1; break;
    }
    $sql = sprintf('SELECT p.*, COUNT(m.log_message_id) AS moderations, u.full_name AS moderator_name '.
                   'FROM %salthing_posts p '.
                   'LEFT JOIN %salthing_log_messages m USING (post_id) '.
                   'LEFT JOIN %susers u ON m.muser_id = u.user_id '.
                   'WHERE p.althing_id = %d '.
                   'GROUP BY p.post_id '.
                   'ORDER BY %s%s, p.post_id%s',
                   $DB->prefix,      // althing_posts
                   $DB->prefix,      // althing_log_messages
                   $DB->prefix,      // users
                   intval($node_id), // althing_id
                   $order,$desc,     // order + direction
                   $desc);           // direction for post_id
    if (($DBResult = $DB->query($sql)) === FALSE) {
        logger(sprintf("%s(): error retrieving posts in althing '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        return FALSE;
    }
    $records = $DBResult->fetch_all_assoc('post_id');
    $DBResult->close();

    // 2 -- massage the result: add previous and next
    $prev_id = NULL;
    foreach($records as $id => $post) {
        if (!is_null($prev_id)) {
            $records[$id]['prev_id'] = $prev_id;
            $records[$prev_id]['next_id'] = $id;
        }
        $prev_id = $id;
    }
    return $records;
} // althing_get_post_records()


/** validation of configuration dialog + rewriting the lists of email addresses as a side effect
 *
 * @param array &$dialogdef the dialog to check
 * @return TRUE if valid, FALSE otherwise (+messages in dialogdef)
 */
function althing_config_dialog_validate(&$dialogdef) {
    $retval = dialog_validate($dialogdef);
    $moderators  = trim(preg_replace("/[\s,]+/","\n",$dialogdef['email_moderators']['value']));
    $subscribers = trim(preg_replace("/[\s,]+/","\n",$dialogdef['email_subscribers']['value']));
    $dialogdef['email_moderators']['value'] = ltrim($moderators."\n");
    $dialogdef['email_subscribers']['value'] = ltrim($subscribers."\n");
    return $retval;
} // althing_config_dialog_validate()


/** construct the dialog for the reporting tool
 *
 * @return array ready to use dialog definition
 */
function althing_get_dialogdef_report() {

    $format_options = array(
        '1' => array(
            'option' => t('report_format_htmlpage_label','m_althing'),
            'title' => t('report_format_htmlpage_title', 'm_althing')
            ),
        '2' => array(
            'option' => t('report_format_htmlcode_label','m_althing'),
            'title' => t('report_format_htmlcode_title', 'm_althing')
            ),
        '3' => array(
            'option' => t('report_format_bbcode_label','m_althing'),
            'title' => t('report_format_bbcode_title', 'm_althing')
            )
        );

    $dialogdef = array(
        'selection' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'selection',
            'minlength' => 0,
            'maxlength' => 240,
            'columns' => 40,
            'label' => t('report_selection_label','m_althing'),
            'title' => t('report_selection_title','m_althing'),
            'value' => ''
            ),
        'message_id' => array(
            'type' => F_CHECKBOX,
            'name' => 'message_id',
            'options' => array('1' => array('option' => t('report_message_id_option','m_althing'),
                                            'title' => t('report_message_id_title','m_althing'))),
            'label' => t('report_fields_label','m_althing'),
            'title' => t('report_fields_title','m_althing'),
            'value' => '1'
            ),
        'author' => array(
            'type' => F_CHECKBOX,
            'name' => 'author',
            'options' => array('1' => array('option' => t('report_author_option','m_althing'),
                                            'title' => t('report_author_title','m_althing'))),
            'value' => '1'
            ),
        'subject' => array(
            'type' => F_CHECKBOX,
            'name' => 'subject',
            'options' => array('1' => array('option' => t('report_subject_option','m_althing'),
                                            'title' => t('report_subject_title','m_althing'))),
            'value' => '1'
            ),
        'date' => array(
            'type' => F_CHECKBOX,
            'name' => 'date',
            'options' => array('1' => array('option' => t('report_date_option','m_althing'),
                                            'title' => t('report_date_title','m_althing'))),
            'value' => '1'
            ),
        'body' => array(
            'type' => F_CHECKBOX,
            'name' => 'body',
            'options' => array('1' => array('option' => t('report_body_option','m_althing'),
                                            'title' => t('report_body_title','m_althing'))),
            'value' => '1'
            ),
        'format' => array(
            'type' => F_RADIO,
            'name' => 'format',
            'value' => '1',
            'options' => $format_options,
            'title' => t('report_format_title','m_althing'),
            'label' => t('report_format_label','m_althing')
            ),
        'csrftoken' => dialog_csrftoken(),
        'button_ok' => dialog_buttondef(BUTTON_OK),
        'button_cancel' => dialog_buttondef(BUTTON_CANCEL)
        );
    return $dialogdef;
} // althing_get_dialogdef_report()


/** validate the report selection dialog, correct the list of posts if necessary
 *
 * @param array &$dialogdef to check
 * @param array &$posts all posts keyed by display_id (not by post_id)
 * @return bool TRUE if all fields are valid, FALSE otherwise + messages in dialogdef
 */
function althing_report_dialog_validate(&$dialogdef,&$posts) {
    $retval = dialog_validate($dialogdef);
    if (!empty($dialogdef['selection']['value'])) {
        $a = explode(',',$dialogdef['selection']['value']);
        $b = array();
        foreach($a as $id) {
            $id = intval($id);
            if (!isset($posts[$id])) {
                $retval = FALSE;
                ++$dialogdef['selection']['errors'];
                $msg = t('error_no_such_display_id','m_althing',array('{DISPLAY_ID}' => strval($id)));
                $dialogdef['selection']['error_messages'][] = $msg;
            } else {
                $b[] = strval($id);
            }
        }
        $dialogdef['selection']['value'] = (sizeof($b) > 0) ? implode(',',$b) : '';
    }
    return $retval;
} // althing_report_dialog_validate()

?>