<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/althing/althing_common.php - code shared between admin and view
 *
 * This file defines various constants and subroutines used from both
 * {@link althing_admin.php} and {@link althing_view.php}.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_althing
 * @version $Id: althing_common.php,v 1.7 2016/03/23 09:28:21 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }

define('ALTHING_STATUS_OPENED',0); // visitors can post to the althing
define('ALTHING_STATUS_FROZEN',1); // adding no longer possible but accepted posts are visible
define('ALTHING_STATUS_CLOSED',2); // no adding anymore and nothing is visible, only the conclusion

define('ALTHING_DEFAULT_VISIBILITY_HIDDEN',0);     // New posts are hidden and unpublished until approved
define('ALTHING_DEFAULT_VISIBILITY_REGISTERED',1); // New posts from known users are published immediately
define('ALTHING_DEFAULT_VISIBILITY_VISIBLE',2);    // New posts are all published immediately

define('ALTHING_DEFAULT_VISIBILITY', ALTHING_DEFAULT_VISIBILITY_HIDDEN); // posts are moderated by default
define('ALTHING_HARVEST_LIMIT',10);// Max number of posts to reply to simultaneously
define('ALTHING_MARBLES_LIMIT',10);// Max number of marbles a post can receive

// ==================================================================
// ==================== ALTHING UTILITY ROUTINES ====================
// ==================================================================


/** create a neat array with email-addresses from a (possibly messy) text
 *
 * this routine converts a (messy) list of email addresses to an array with
 * a single address per array element. If there are no addresses, an empty
 * array is returned. The addresses are supposed to be comma-delimited and/or
 * newline and/or space delimited. Note that we expect only plain email
 * addresses, no readable names etc. (KISS)
 *
 * @param string $buffer holds the messy list of email addresses
 * @return array with one email address per element (or an empty array)
 */
function althing_get_emails($buffer) {
    $s = trim(preg_replace("/[\s,]+/","\n",$buffer));
    $a = (empty($s)) ? array() : explode("\n",$s);
    return $a;
} // althing_get_emails()


/** send an alert message to every email address in $emails
 *
 * this routine sends an individual message to each of the addresses in
 * the array $emails notifying the receiver about the new post. Relevant
 * information (headers) are taken from the $post-record. Note that we
 * attempt to leak as less information as is possible, eg. we don't show
 * email addresses of posters, or the contents, or other information
 * like the name of the page (which could be in a private area) or the
 * title of the althing.
 *
 * There are several combinations flags that yield different mails to be sent.
 * Here is a truth-table:
 *
 * <code>
 * Visible Moderator Private Message            Subject
 *  FALSE   FALSE    FALSE   public_published   [nn] Public post #xx is now published
 *  FALSE   FALSE     TRUE   private_published  [nn] Private post #xx is now published
 *  FALSE   TRUE     FALSE   moderation         [nn] Moderation request for post #xx
 *  FALSE   TRUE      TRUE   moderation         [nn] Moderation request for post #xx
 *   TRUE   FALSE    FALSE   public_visible     [nn] New public post #xx
 *   TRUE   FALSE     TRUE   private_visible    [nn] New private post #xx
 *   TRUE   TRUE     FALSE   public_visible     [nn] New public post #xx
 *   TRUE   TRUE      TRUE   private_visible    [nn] New private post #xx
 * </code>
 * 
 * Note that most messages contain a warning about the need to login first
 * (either for accessing the private area OR for accessing the admin interface for
 * moderation). In total there are 5 different messages. These can be adapted via
 * the Translate Tool.
 *
 * Note that the messages 'public_published' and 'private_published' are to be sent
 * once a moderator publishes an unpublished post (ie. makes the post visible for
 * the first time). This means that in that case we are called from althing_admin.php
 * at moderation time rather than from althing_view.php at submit-time.
 *
 * Note: the URL in the mail now uses the new parameter 'post' to select a single
 * message for display (see also {@link althing_show_overview()} in althing_view.php).
 * By showing only a single post, the user is not distracted by other messages that
 * were already there (July 2014).
 *
 * @param array $emails holds all email addresses to send an alert to
 * @param bool $moderator TRUE if emails are of moderators, FALSE means subscribers
 * @param bool $private TRUE if the althing lives in a private area
 * @param bool $visible TRUE if post already published, FALSE if still unpublished
 * @param int $area_id holds the page with this althing
 * @param array $post the information about the post
 * @return bool TRUE on success + messages mailed out, FALSE if an error occurred
 */
function althing_send_alerts($emails,$moderator,$private,$visible,$area_id,$post) {
    global $CFG;
    static $counter = 0;
    $overtime = max(intval(ini_get('max_execution_time')),30); // additional processing time in seconds
    include_once($CFG->progdir.'/lib/email.class.php');

    $node_id = $post['althing_id'];
    $post_id = $post['post_id'];
    $index_link = ($CFG->friendly_url)?sprintf('%s/index.php/%d/post/%d',$CFG->www,$node_id,$post_id) :
                                       sprintf('%s/index.php?node=%d&post=%d',$CFG->www,$node_id,$post_id);
    $param = array(
        'job' => 'pagemanager',
        'task' => 'editcontent',
        'area' => intval($area_id), // necessary for deep-linking if the current area is another one
        'node' => $node_id,
        'option' => 'moderation',
        'sort' => '-2',
        'post' => $post_id
        );
    $admin_link = href($CFG->www.'/admin.php',$param);
    $params = array(
        '{NODE}' => strval($node_id),
        '{AUTHOR}' => htmlspecialchars($post['author_full_name']),
        '{DATIM}' => $post['datim'],
        '{SUBJECT}' => htmlspecialchars($post['subject']),
        '{DISPLAY_ID}' => $post['display_id'],
        '{POST_ID}' => strval($post_id),
        '{REGISTERED}' => t((empty($post['author_id'])) ? 'anonymous':'registered','m_althing'),
        '{ADMIN_LINK}' => $admin_link,
        '{ADMIN_LOGIN}' => sprintf("%s/admin.php?login=1",$CFG->www),
        '{INDEX_LINK}' => $index_link,
        '{INDEX_LOGIN}' => sprintf("%s/index.php?login=1",$CFG->www)
        );
    $name = ($moderator) ? t('moderators_name','m_althing') : t('subscribers_name','m_althing');
    if ($private) {
        if ($visible) {
            $subj = t('alert_private_visible_subj','m_althing',$params);
            $body = t('alert_private_visible_body','m_althing',$params);
        } else { // hidden
            if ($moderator) {
                $subj = t('alert_moderation_subj','m_althing',$params);
                $body = t('alert_moderation_body','m_althing',$params);
            } else { // subscriber
                $subj = t('alert_private_published_subj','m_althing',$params);
                $body = t('alert_private_published_body','m_althing',$params);
            }
        }
    } else { // public
        if ($visible) {
            $subj = t('alert_public_visible_subj','m_althing',$params);
            $body = t('alert_public_visible_body','m_althing',$params);
        } else { // hidden
            if ($moderator) {
                $subj = t('alert_moderation_subj','m_althing',$params);
                $body = t('alert_moderation_body','m_althing',$params);
            } else { // subscriber
                $subj = t('alert_public_published_subj','m_althing',$params);
                $body = t('alert_public_published_body','m_althing',$params);
            }
        }
    }
    $retval = TRUE; // assume success
    $mailer = new Email;
    $mailer->set_subject($subj);
    $mailer->set_message($body);
    foreach ($emails as $email) {
        $mailer->set_mailto($email,$name);
        if (!$mailer->send()) {
            $retval = FALSE;
        }
        if (($counter++ % 16) == 0) {
            @set_time_limit($overtime); // try to get additional work time after every 16 mails
            logger(sprintf("%s(): working overtime after %d mails (t=%ds)",
                           __FUNCTION__,$counter,$overtime),WLOG_DEBUG);
        }
    }
    return $retval;
} // althing_send_alerts()


// ==================================================================
// ===================== ALTHING BBCODE HELPERS =====================
// ==================================================================


/** convert valid BBCode to valid HTML
 *
 * this routine more or less replaces BBCode tags with the HTML equivalent.
 * We assume that the BBCode is validated before. If you are unsure, you
 * can use {@link bbcode_validate()} to double check (at a cost of processing
 * time).
 *
 * Strategy here is to
 *  - get rid of any superfluous CR's and CR+LF's
 *  - defuse any HTML-tricks via htmlspecialchars() (note that this also
 *    converts the " in [url="path"]...[/url])
 *  - convert the 'simple' tags via a cheap routine str_replace()
 *  - convert the tags with parameters (path, name) via the more expensive preg_replace
 *
 * Since we probably need to use this routine over and over again in a single
 * call, we buffer the preg search and replace arrays on the first call, speeding up
 * subsequent calls, at a cost of a single if (is_null()). The plain search and replace
 * arrays are so static that we simply declare them static from the start.
 *
 * Note 1:
 * We do get rid of a single newline after [quote] or [quote=...] because
 * that would otherwise ruin the display with an extra blank line in the quote div.
 * If you really want to have a blank line, use two newlines in a row.
 *
 * Note 2:
 * We do take into consideration the friendly URL feature and also an option for
 * fully qualified domain names.
 *
 * @param string $bbcode the BBCode-string to translate
 * @param bool $q TRUE means any URLs are fully qualified, FALSE uses the short variety
 * @return string ready-to-use HTML
 * @uses bbcode_callback()
 * @uses bbcode_callback_q()
 */
function bbcode2html($bbcode,$q=FALSE) {
    static $s=array('[b]','[/b]','[u]','[/u]','[i]','[/i]','[s]','[/s]','[code]','[/code]','[quote]','[/quote]');
    static $r=array('<b>','</b>','<u>','</u>','<i>','</i>','<s>','</s>','<pre>','</pre>','<div class="quote">','</div>');
    static $regex = array(
        '/\[(url=)(&quot;)?(.*?)(&quot;)?\](.*?)\[\/url\]/s', // [url=path] and [url="path"]... in 1 go
        '/\[(url)\](.*?)\[\/url\]/s',                         // [url]path[/url]
        '/\[(img=)(&quot;)?(.*?)(&quot;)?\]/s',               // [imgl=path] and [img="path"]... in 1 go
        '/\[(img)\](.*?)\[\/img\]/s',                         // [url]path[/url]
        '/\[(quote=)(&quot;)?(.*?)(&quot;)?\](\<br \/\>)?/s', // [quote=author] and [quote="author"]...
        );

    // 1--lose the superfluous CR and CR+LF and make remaining LF -> <br>
    $t = str_replace("\r\n","\n",$bbcode);
    $t = str_replace("\n\r","\n",$t);
    $t = str_replace("\r","\n",$t);
    $t = str_replace("[quote]\n","[quote]",$t);
    $t = str_replace("\n[/quote]","[/quote]",$t);
    $t = htmlspecialchars($t);
    $t = nl2br($t);
    // 2--simple replacements
    $t = str_replace($s,$r,$t);
    // 3--RE-replacements, maybe with qualified domain names
    $t = preg_replace_callback($regex,($q) ? 'bbcode_callback_q' : 'bbcode_callback',$t);
    return $t;
} // bbcode2html()


/** one-line wrapper to allow for fully qualified URLs
 * 
 * @param array $m holds complete matched code + all individual sub-matches
 * @return string the replacement string including everything
 * @uses bbcode_callback()
 */
function bbcode_callback_q($m) { return bbcode_callback($m,TRUE); }


/** handle advanced BBCode search/replace specifically for encoding filenames
 *
 * we use this callback routine to perform the regex-based search/replace
 * so we get a chance to encode the URLs and also deal with the procy-friendly
 * style of servind files via file.php.
 *
 * Note that we are transparent about 'hidden' URLs: if the user uses the format
 * [url=path]anchor[/url] we show the path in plain text too, like so:
 * <a href="path">anchor</a> (anchor)
 * No surprises with deceiving links..
 *
 * @param array $m holds complete matched code + all individual sub-matches
 * @param bool $q TRUE means any URLs are fully qualified, FALSE uses the short variety
 * @return string the replacement string including everything
 */
function bbcode_callback($m,$q=FALSE) {
    global $CFG;
    static $s = array('%2F','%2f'); // replace encoded slashes, make result more readable 
    static $r = array('/','/');
    if ($m[1] == 'quote=') {
        return '<div class="quote"><div class="quoting">'.$m[3].':</div>';
    }
    $file_php = (($q) ? $CFG->www : $CFG->www_short).'/file.php'.(($CFG->friendly_url) ? '' : '?file=');
    switch($m[1]) {
    case 'url=': 
        return sprintf('<a target="_blank" href="%s%s">%s</a> (%s)',
                       $file_php,str_replace($s,$r,rawurlencode($m[3])),$m[5],$m[3]);
        break;
    case 'url':
        return sprintf('<a target="_blank" href="%s%s">%s</a>',
                       $file_php,str_replace($s,$r,rawurlencode($m[2])),$m[2]);
        break;
    case 'img=':
        return sprintf('<img src="%s%s">',$file_php,str_replace($s,$r,rawurlencode($m[3])));
        break;
    case 'img':
        return sprintf('<img src="%s%s">',$file_php,str_replace($s,$r,rawurlencode($m[2])));
        break;
    }
    return '#';
} // bbcode_callback()


/** validate the BBCode in the dialogdef item
 *  *
 * Simple tags always should match in pairs, without overlap:
 * [b] [/b]
 * [u] [/u]
 * [i] [/i]
 * [s] [/s]
 * [code] [/code]
 *
 * The more complicated constructs with opening/closing tags are:
 * [url=path]anchor[/url]
 * [url]path[/url]
 * [img]path[/img]
 * [quote]text[/quote]
 * [quote=author]text[/quote]
 * [quote="author"]text[/quote]
 *
 * There is one single tag, ie without closing tag:
 *
 * [img=path]
 *
 * Path should be a path relative to the data root, e.g.
 * '/areas/exemplum/snapshots/cyanara.jpg'. This limits
 * the URLs and the IMGs to files on this server.
 *
 * Note that we also validate the specified path via
 * {@link bbcode_valid_file()} which also checks permissions
 * of the current user. The prevents leaking of files
 * via pupils from the teachers private intranet.
 *
 * @param array &$item an element of a dialogdef that supposedly holds BBCode
 * @return bool TRUE if no validation errors, FALSE otherwise + errors in $item[]
 */
function bbcode_validate(&$item) {
    $stack = array(0=>array('i'=>0,'t'=>'?')); // sentinel for stack underrun);
    $sp = 0;              // stack pointer always points to the last valid opening tag (0 means none)
    $s = $item['value'];  // our working copy of the text to analyze
    $i = 0;               // index into $s; keeps track of where we are in the string
    while (($i = strpos($s,'[',$i)) !== FALSE) {
        // 1 -- check for [b] [u] [i] [s] opening
        if ($s[$i+2] == ']') {
            if (strpos('buis',$s[$i+1]) !== FALSE) {
                $stack[++$sp] = array('i' => $i, 't' => $s[$i+1]);
            } // else 
                // skip unrecognised tag '[x]'
            $i += 3;
            continue;
        }
        // 2 -- check for [code] opening
        if (substr($s,$i,6) == '[code]') {
            $stack[++$sp] = array('i' => $i, 't' => 'code');
            $i += 6;
            continue;
        }
        // 3 -- check for [tag=path] and [tag="path"] (url,img or quote)
        $tag6 = substr($s,$i+1,6);
        $tag4 = substr($tag6,0,4);
        $quote = ($tag6 == 'quote=') ? TRUE : FALSE;
        if (($quote) || ($tag4 == 'url=') || ($tag4 == 'img=')) {
            $p = ($quote) ? 7 : 5; // $p = offset for payload
            if ($s[$i+$p] == '"') {
                if (($j = strpos($s,'"]',$i+$p+1)) === FALSE) {
                    $ci = utf8_strlen(substr($s,0,$i)); // offset in UTF8 characters (not bytes)
                    $aparams = array('{I}' => strval($ci),
                                     '{CONTEXT}' => "'".utf8_substr($s,$ci,24)."'");
                    ++$item['errors'];
                    $item['error_messages'] = t('bbcode_unterminated_str','m_althing',$aparams);
                    return FALSE;
                }
                $param = substr($s,$i+$p+1,$j-$i-$p-1);
                $stack[++$sp] = array('i' => $i, 't' => ($quote) ? $tag6 : $tag4, 'p' => $param);
                $j = $j+2;
            } else {
                if (($j = strpos($s,']',$i+$p+1)) === FALSE) {
                    $ci = utf8_strlen(substr($s,0,$i)); // offset in UTF8 characters (not bytes)
                    $aparams = array('{I}' => strval($ci),
                                     '{CONTEXT}' => "'".utf8_substr($s,$ci,24)."'");
                    ++$item['errors'];
                    $item['error_messages'] = t('bbcode_unterminated_open_tag','m_althing',$aparams);
                    return FALSE;
                }
                $param = substr($s,$i+$p,$j-$i-$p);
                $stack[++$sp] = array('i' => $i, 't' => ($quote) ? $tag6 : $tag4, 'p' => $param);
                $j = $j+1;
            }
            if ((!$quote) && (!bbcode_valid_path($param))) {
                $ci = utf8_strlen(substr($s,0,$i)); // offset in UTF8 characters (not bytes)
                $cj = utf8_strlen(substr($s,0,$j));
                $aparams = array('{I}' => strval($ci),
                                 '{CONTEXT}' => "'".utf8_substr($s,$ci,$cj-$ci)."'");
                ++$item['errors'];
                $item['error_messages'] = t('bbcode_invalid_path','m_althing',$aparams);
                return FALSE;
            }
            if ($tag4 == 'img=') { // [img=path] and [img="path"] don't have an ending [/img]
                --$sp;
            }
            $i = $j;
            continue;
        }
        // 4 -- check for [url]path[/url] or [img]path[/img] ($path checked later)
        if (($tag4 == 'url]') || ($tag4 == 'img]')) {
            $stack[++$sp] = array('i' => $i, 't' => substr($tag4,0,3));
            $i += 5;
            continue;
        }
        // 5 -- check for bare [quote]
        if ($tag6 == 'quote]') {
            $stack[++$sp] = array('i' => $i, 't' => 'quote');
            $i += 7;
            continue;
        }

        // At this point we have seen all valid opening tags and remembered them via stack
        // (except img= which is a special case) We now check for matching closing tags, i.e.
        // IF we find a tag it MUST match the current tag on the stack.

        if ($s[$i+1] == '/') { // Aha! a closing tag starts here
            // 6 -- sanity check: every closing tag must end with a ']' eventually
            if (($j = strpos($s,']',$i+2)) === FALSE) {
                $ci = utf8_strlen(substr($s,0,$i)); // offset in UTF8 characters (not bytes)
                $aparams = array('{I}' => strval($ci),
                                 '{CONTEXT}' => "'".utf8_substr($s,$ci,24)."'");
                ++$item['errors'];
                $item['error_messages'] = t('bbcode_unterminated_close_tag','m_althing',$aparams);
                return FALSE;
            }
            // 7 -- deal with [tag]path[/tag]: check path
            $tag = substr($s,$i+2,$j-$i-2);
            if (((($tag == 'url') || $tag == 'img')) && ($tag == $stack[$sp]['t'])) {
                $path = substr($s,$stack[$sp]['i']+5,$i-$stack[$sp]['i']-5);
                if (!bbcode_valid_path($path)) {
                    $ci = utf8_strlen(substr($s,0,$stack[$sp]['i'])); // offset in UTF8 characters
                    $cj = utf8_strlen(substr($s,0,$j));
                    $aparams = array('{I}' => strval($ci),
                                     '{CONTEXT}' => "'".utf8_substr($s,$ci,$cj-$ci+1)."'");
                    ++$item['errors'];
                    $item['error_messages'] = t('bbcode_invalid_path','m_althing',$aparams);
                    return FALSE;
                }
                --$sp;
                $i += 6;
                continue;
            }
            // 8 -- check out regular (simple) end tags
            if (($tag == $stack[$sp]['t']) || ($tag.'=' ==  $stack[$sp]['t'])) {
                --$sp;
                $i += strlen($tag)+3;
                continue;
            }
            // still here? Oops, tag mismatch
            $ci1 = utf8_strlen(substr($s,0,$stack[$sp]['i'])); // offset in UTF8 characters
            $ci2 = utf8_strlen(substr($s,0,$i));
            $context = min($ci2-$ci1,24);
            $aparams = array('{I1}' => strval($ci1),
                             '{I2}' => strval($ci2),
                             '{CONTEXT1}' => "'".utf8_substr($s,$ci1,$context)."'",
                             '{CONTEXT2}' => "'".utf8_substr($s,$ci2,24)."'");
            ++$item['errors'];
            $item['error_messages'] = t('bbcode_mismatched_tag','m_althing',$aparams);
            return FALSE;
        }
        ++$i; //move pointer past the '[' we detected in the first place
    }

    // At this point the stack should be back at 0. If not, there must be some unfinished
    // business. Tell the user about at least one.

    if ($sp != 0) {
        $ci = utf8_strlen(substr($s,0,$stack[$sp]['i'])); // offset in UTF8 characters
        $aparams = array('{I}' => strval($ci),
                         '{CONTEXT}' => "'".utf8_substr($s,$ci,24)."'");
        ++$item['errors'];
        $item['error_messages'] = t('bbcode_unmatched_tag','m_althing',$aparams);
        return FALSE;
    }
    return TRUE;
} // bbcode_validate()


/** validate a filename for existence and access 
 *
 * this routine checks to see if the current user is allowed to
 * access the file $filename and also that it actually exists.
 * We return FALSE if access is denied or the cleaned up path
 * if all is well.
 *
 * @param string $filename the file to check, relative to CFG->datadir
 * @return bool|string FALSE on access denied/non-existent, clean path otherwise
 */
function bbcode_valid_path($filename) {
    global $CFG,$USER;

    $parts = explode('/',trim(strtr($filename,'\\','/'),'/'));
    $path = '/'.implode('/',$parts);

    // 1 -- always disallow attempts to escape from tree via parent directory tricks
    if (in_array('..',$parts)) {
        logger(sprintf('%s(%s): .. not allowed',__FUNCTION__,$path),WLOG_DEBUG);
        return FALSE;
    }

    // 2 -- check the 1st and 2nd component of the requested file
    $where = array('is_active' => TRUE, 'path' => (isset($parts[1])) ? $parts[1] : '');
    switch ($parts[0]) {
    case 'areas':
        $fields = array('area_id','is_private');
        if (($record = db_select_single_record('areas',$fields,$where)) === FALSE) {
            logger(sprintf("%s(%s): access denied in area",__FUNCTION__,$path),WLOG_DEBUG);
            return FALSE;
        }
        if ((db_bool_is(TRUE,$record['is_private'])) &&
            (!$USER->has_intranet_permissions(ACL_ROLE_INTRANET_ACCESS,intval($record['area_id'])))) {
            logger(sprintf("%s(%s): permission denied in private area",__FUNCTION__,$path),WLOG_DEBUG);
            return FALSE;
        }
        break;

    case 'users':
        if (($record = db_select_single_record('users',array('user_id'),$where)) === FALSE) {
            logger(sprintf("%s(%s): permission denied in user directory",__FUNCTION__,$path),WLOG_DEBUG);
            return FALSE;
        }
        break;

    case 'groups':
        if (($record = db_select_single_record('groups',array('group_id'),$where)) === FALSE) {
            logger(sprintf("%s(%s): permission denied in group directory",__FUNCTION__,$path),WLOG_DEBUG);
            return FALSE;
        }
        break;

    default:
        logger(sprintf("%s(%s): access denied: unknown directory",__FUNCTION__,$path),WLOG_DEBUG);
        return FALSE;
        break;
    }

    // 3 -- check actual file existence
    if (!is_file($CFG->datadir.$path)) {
        logger(sprintf("%s(%s): access denied: file does not exist",__FUNCTION__,$path),WLOG_DEBUG);
        return FALSE;
    }

    // 4 -- OK, good to go
    return $path;
} // bbcode_valid_path()


/** show a table illustrating the possible BBCode
 *
 * this routine is a help function. If JS is enabled,
 * the information is hidden and the user can make it
 * visible by clicking on the title.
 *
 * @param object &$theme collects (html) output
 * @return void output added to $theme
 */
function bbcode_help(&$theme) {
    global $CFG;
    $file_php = $CFG->www_short.'/file.php'.(($CFG->friendly_url) ? '' : '?file=');
    $examples = array(
        array('[b]'.t('bbcode_bold','m_althing').'[/b]',
              '<b>'.t('bbcode_bold','m_althing').'</b>'),
        array('[u]'.t('bbcode_underline','m_althing').'[/u]',
              '<u>'.t('bbcode_underline','m_althing').'</u>'),
        array('[i]'.t('bbcode_italic','m_althing').'[/i]',
              '<i>'.t('bbcode_italic','m_althing').'</i>'),
        array('[s]'.t('bbcode_strike','m_althing').'[/s]',
              '<s>'.t('bbcode_strike','m_althing').'</s>'),
        array('[code]'.t('bbcode_code','m_althing').'[/code]',
              '<pre>'.t('bbcode_code','m_althing').'</pre>'),
        array(sprintf('[url]%s[/url]', t('bbcode_url','m_althing')),
              sprintf('<a target="_blank" href="%s%s">%s</a>',
                      $file_php,
                      t('bbcode_url','m_althing'),
                      t('bbcode_url','m_althing'))),
        array(sprintf('[url=&quot;%s&quot;]%s[/url]',
                      t('bbcode_url','m_althing'),t('bbcode_anchor','m_althing')),
              sprintf('<a target="_blank" href="%s%s">%s</a> (%s)',
                      $file_php,
                      t('bbcode_url','m_althing'),
                      t('bbcode_anchor','m_althing'),
                      t('bbcode_url','m_althing'))),
        array(sprintf('[img=&quot;%s&quot;]',
                      t('bbcode_img','m_althing')),
              sprintf('<img src="%s/graphics/waslogo-140x35.png">',
                      $CFG->progwww_short)),
        array(sprintf('[quote]%s[/quote]', t('bbcode_quoted','m_althing')),
              sprintf('<div class="quote">%s</div>',t('bbcode_quoted','m_althing'))),
        array(sprintf('[quote=&quot;%s&quot;]%s[/quotel]',
                      t('bbcode_quoting','m_althing'),t('bbcode_quoted','m_althing')),
              sprintf('<div class="quote"><div class="quoting">%s:</div>%s</div>',
                      t('bbcode_quoting','m_althing'),t('bbcode_quoted','m_althing')))
        );
    $anchor = t('bbcode_help','m_althing');
    $attr = array(
        'title' => t('bbcode_help_title','m_althing'),
        'id' => 'bbcode_help_header',
        'name' => 'bbcode_help');
    $theme->add_content('<h3>'.html_a('#bbcode_help',NULL,$attr,$anchor).'</h3>');

    $class = 'bbcode_header';
    $theme->add_content('<div id="bbcode_help_container">');
    $attr = array('style' => 'border: 1px solid #000000; border-collapse: collapse;');
    $theme->add_content(html_table($attr));
    $attr = array('style' => 'border: 1px solid #000000; padding: 5px;');
    $theme->add_content('  '.html_table_row(array('class' => $class)));
    $theme->add_content('    '.html_table_head($attr,t('bbcode_bbcode','m_althing')));
    $theme->add_content('    '.html_table_head($attr,t('bbcode_html','m_althing')));
    $theme->add_content('  '.html_table_row_close());
    // 3 -- table rows
    foreach($examples as $example) {
        $class = ($class == 'bbcode_odd') ? 'bbcode_even' : 'bbcode_odd';
        $theme->add_content('  '.html_table_row(array('class' => $class)));
        $theme->add_content('    '.html_table_cell($attr,'<pre>'.$example[0].'</pre>'));
        $theme->add_content('    '.html_table_cell($attr,$example[1]));
        $theme->add_content('  '.html_table_row_close());
    }
    $theme->add_content(html_table_close());
    $theme->add_content('</div>');

    // This suppresses the DIV that holds the BBCode explanation
    // If JavaScript is not enabled, it stays visible, otherwise
    // a click on the title toggels the view
    $js="<script><!--\n".
        "var hdr=document.getElementById('bbcode_help_header');\n".
        "var toggle=0;\n".
        "hdr.onclick=function() {\n".
        "  var div=document.getElementById('bbcode_help_container');\n".
        "  div.style.display=(toggle)?'block':'none';\n".
        "  toggle=1-toggle;\n".
        "}\n".
        "hdr.onclick();\n".
        "--></script>\n";
    $theme->add_content($js);

} // bbcode_help()


?>