<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/althing/althing_search.php - interface to the search-part of this module
 *
 * This file defines the interface with this module for searching content.
 * The interface consists of a (temporary) table and this function:
 *
 * <code>
 * bool althing_search(&$hits,&$results,$qwords, $limit, $offset)
 * </code>
 *
 * The function is called whenever module content is to be searched. At that point
 * the helper table, which is called 'search_nodes', exists and is populated.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_althing
 * @version $Id: althing_search.php,v 1.4 2016/04/05 09:25:26 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }

require_once(dirname(__FILE__).'/althing_common.php'); // common code + constants (bbcode2html())

/** search this module's content in selected nodes for keywords in $qwords 
 *
 * This interface is documented in {@link htmlpage_search()} in {@link htmlpage_search.php}.
 *
 * We have to search the following fields in two tables:
 *
 * a. althings: 1 record/althing: header/introduction/question/conclusion/footer
 *
 * b. althing_posts: 0,1,...N records/althing: subject/author/content
 *
 * Althings always qualify (the 5 fields are always visible)
 * Posts only qualify when althing status is 0 (open) or 1 (frozen) AND the post is published and visible
 *
 * Hits in althings link to the main page (node=$node_id)
 * Hits in the posts link to node=$node_id,post=$post_id
 *
 * The content in posts is converted from BBCode to HTML before starting the search,
 *
 * @param array &$hits receives search results (at most $limit)
 * @param int &$results receives number of hits (negative value on entry means (re-)calculate total) 
 * @param array $qwords holds 1 or more keywords to search for (original, utf8lower, quoted)
 * @param int $limit maximum number of hits to return (could be 0)
 * @param int $offset starting point within this module for returning results
 * @return bool|array FALSE on failure, TRUE on success and $hits, $results updated (maybe)
 */
function althing_search(&$hits, &$results, $qwords, $limit, $offset=0) {
    global $DB;

    //
    // 0A -- do we need to do anything
    //
    if (($results == 0) ||  // cached value from caller says 'no hits'
        (($results > 0) && ($limit <= 0))) { // there are hits but none are requested at this time
        return TRUE;
    }

    //
    // 0B -- yes, initialise and prepare the selection criteria
    //
    $tables = array('althings','althing_posts'); 
    $fields = array(array('a.header','a.introduction','a.question','a.conclusion','a.footer'),
                    array('p.author_full_name','p.subject','p.content'));
    $wheres = array(search_where($qwords,$fields[0]),
                    sprintf('((a.status = %d) OR (a.status = %d)) AND '.
                            '((p.visible) AND (p.published)) AND %s',
                            ALTHING_STATUS_OPENED,ALTHING_STATUS_FROZEN,search_where($qwords,$fields[1])));
    $fields[1][] = 'p.post_id'; // additional post fields to SELECT but not to be used in WHERE-clause
    $fields[1][] = 'p.display_id'; 
    $fields[1][] = 'p.datim'; 
    $counts = array(-1,-1); // cache for # of hits (-1 means: no value yet)
    $context_fields = array(array('header','introduction','question','conclusion','footer'),
                            array('author_full_name','subject','content'));

    //
    // 1 -- count hits in althings and -posts, don't retrieve actual data
    //
    if ($results < 0) {
        $results = 0;
        foreach($tables as $k => $table) {
            if (($count = althing_search_results($table,$wheres[$k])) === FALSE) {
                return FALSE;
            }
            $counts[$k] = $count;
            $results += $count;
        }
    }

    //
    // 2 -- retrieve $limit records at $offset
    //
    if (($limit > 0) && ($offset < $results)) {
         // 2A -- we may have to count table[0] (=althings) hits again...
        if (($count = $counts[0]) < 0) {
            if (($count = althing_search_results($tables[0],$wheres[0])) === FALSE) {
                return FALSE;
            }
        }
        // 2B -- ...and this may have an impact on the limits and offsets to use in the two tables
        $limits = array(min($limit,$count-$offset),($offset<$count) ? $limit+$offset-$count : $limit);
        $offsets = array($offset,($offset < $count) ? 0 : $offset - $count);
        $orders = array('n.node_id', 'n.node_id, p.datim');

        // 2C -- do search the tables if the (recalculated) limit > 0 for this table
        foreach($tables as $k => $table) {
            if (($lim = $limits[$k]) <= 0) {
                continue;
            }
            $off = $offsets[$k];
            if ($k == 0) {
                $sql = sprintf("SELECT n.node_id, n.title AS node_title, %s ".
                               "FROM  %ssearch_nodes s INNER JOIN %snodes n USING (node_id) ".
                               "INNER JOIN %s%s a ON (n.node_id = a.althing_id) ".
                               "WHERE %s ORDER BY %s",implode(',',$fields[$k]),$DB->prefix,
                               $DB->prefix,$DB->prefix,$table,$wheres[$k],$orders[$k]);
            } else {
                $sql = sprintf("SELECT n.node_id, n.title AS node_title, %s ".
                               "FROM  %ssearch_nodes s INNER JOIN %snodes n USING (node_id) ".
                               "INNER JOIN %s%s a ON (n.node_id = a.althing_id) ".
                               "INNER JOIN %s%s p USING (althing_id) ".
                               "WHERE %s ORDER BY %s",implode(',',$fields[$k]),
                               $DB->prefix,$DB->prefix,
                               $DB->prefix,$tables[0],
                               $DB->prefix,$table,$wheres[$k],$orders[$k]);
            }
            if (($DBResult = $DB->query($sql,$lim,$off)) === FALSE) {
                logger(sprintf('%s(): cannot retrieve from %s: %s',__FUNCTION__,$table,db_errormessage()));
                return FALSE;
            }
            // 2D -- step through result set of this query
            while (($record = $DBResult->fetch_row_assoc()) !== FALSE) {
                $parameters = array('node' => $record['node_id']);
                if ($k == 0) {
                    $title = trim($record['header']);
                } else {
                    $parameters['post'] = $record['post_id'];
                    // prevent endless lists of the same header by adding the display_id to the title
                    $title = trim($record['subject']);
                    if (!empty($title)) {
                        $aparams = array('{DISPLAY_ID}' => strval($record['display_id']),
                                         '{POST_ID}'    => strval($record['post_id']));
                        $title .= ' '.t('althing_view_msg_id','m_althing',$aparams);
                    }
                    // suppress the bbcode; we'll lose the html too, lateron in search_context()
                    $record['content'] = bbcode2html($record['content']);
                }
                // 2E -- only selected fields add to context
                $content = '';
                foreach($context_fields[$k] as $f) {
                    $content .= ' '.$record[$f];
                }
                $hits[] = array('title' => (empty($title)) ? $record['node_title'] : $title,
                                'context' => search_context($qwords,$content),
                                'url' => was_node_url(NULL,$parameters));
            }
            $DBResult->close();
        }
    }
    //
    // 3 -- all done
    //
    return TRUE;
} // althing_search()


/** count the number of hits in $table
 *
 * 
 * @param string $table name of table to search
 * @param string $where where clause to select hits in $table
 * @return bool|int FALSE on error, # of hits otherwise (could be 0)
 */
function althing_search_results($table,$where) {
    global $DB;

    if ($table == 'althings') {
        $sql = sprintf("SELECT COUNT(s.node_id) AS results ".
                       "FROM %ssearch_nodes s INNER JOIN %s%s a ON s.node_id = a.althing_id ".
                       "WHERE (%s)",
                       $DB->prefix,$DB->prefix,$table,$where);
    } else {
        $sql = sprintf("SELECT COUNT(s.node_id) AS results ".
                       "FROM %ssearch_nodes s INNER JOIN %salthings a ON s.node_id = a.althing_id ".
                       "INNER JOIN %s%s p USING (althing_id) ".
                       "WHERE (%s)",
                       $DB->prefix,$DB->prefix,$DB->prefix,$table,$where);
    }
    if ((($DBResult = $DB->query($sql,1)) === FALSE) || ($DBResult->num_rows != 1)) {
        logger(sprintf('%s(): cannot retrieve count in %s: %s',__FUNCTION__,$table,db_errormessage()));
        return FALSE;
    }
    $record = $DBResult->fetch_row_assoc();
    $DBResult->close();
    return intval($record['results']);
} // althing_search_results()

?>