<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/althing/althing_view.php - interface to the view-part of the althing module
 *
 * This file defines the interface with the althing-module for viewing content.
 * The interface consists of this function:
 *
 * <code>
 * althing_view(&$output,$area_id,$node_id,$module)
 * </code>
 *
 * This function is called from /index.php when the node to display is connected
 * to this module.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_althing
 * @version $Id: althing_view.php,v 1.23 2016/05/26 13:22:27 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }

define('ALTHING_DIRNAME',dirname(__FILE__));
require_once(ALTHING_DIRNAME.'/althing_common.php'); // common code + constants

define('ALTHING_REFERENCE',sha1(__FILE__.':'.__LINE__));
require_once($CFG->progdir.'/lib/tokenlib.php');


/** display the content of the althing linked to node $node_id
 *
 * @param object &$theme collects the (html) output
 * @param int $area_id identifies the area where $node_id lives
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function althing_view(&$theme,$area_id,$node_id,$module) {
    $theme->add_stylesheet(sprintf('program/modules/%s/%s.css',$module['name'],$module['name']));
    $althing_id = intval($node_id);

    // 0A -- get configuration in core
    $where = array('althing_id' => $althing_id);
    if (($config = db_select_single_record('althings','*',$where)) === FALSE) {
        $theme->add_message(t('error_retrieving_config','m_althing'));
        logger(sprintf("%s(): cannot retrieve althing configuration '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        return FALSE;
    }

    // 0B -- do they want to add (and are they allowed)?
    $status = intval($config['status']);
    if (($posts = check_add_reply_harvest($theme,$althing_id,$status)) === FALSE) {
        return althing_show_overview($theme,$config);
    }
    if (!is_null($posts)) {
        // 0B -- yes, need to show the add post dialog for the 1st time; setup token
        $token_key = '';
        if (($token_id = token_create(ALTHING_REFERENCE,$token_key,20)) === FALSE) { // 20s delay
            $msg = t('error_creating_token','m_althing',array('{NODE}' => strval($althing_id)));
            $theme->add_message($msg);
            $theme->add_content($msg);
            althing_show_overview($theme,$config); // ignore these results...
            return FALSE;                          // ...but tell user about failed token instead
        }
        $dialogdef = althing_view_get_dialogdef($config, $token_key ,$posts);
        // special trick: we want to keep the user from maliciously tampering with the
        // parent post_id's so we save that information in the token container rather than
        // trust the user with a hidden field
        $data = array('parents' => $dialogdef['message']['parents']);
        if (!token_store($token_id, $data)) {
            $theme->add_message(t('error_storing_data','m_althing'));
            logger(sprintf('%s(): token store error in althing %d: %s',
                           __FUNCTION__,$althing_id,db_errormessage()));
            althing_show_overview($theme,$config); // ignore these results...
            return FALSE;                          // ...but tell user about failed token instead
        }
        althing_show_form($theme, $config, $dialogdef);
        return TRUE;
    }

    // If we are still here, there were no new posts to add.
    // We can now check for other buttons that might have been pressed. e.g. 
    // - in the Form dialog: [Preview] or [Cancel]
    // - in the Preview dialog: [Edit] or [Post] or [Cancel]
    // We need to check for a valid token too. If anything fails, fall back
    // to showing the Althing overview.

    //
    // 1 -- do we have a token already?
    //
    $t0 = $t1 = 0;
    $ip_addr = '';
    $data = FALSE;
    $token_id = FALSE;
    if (isset($_POST['token'])) { // lookup valid UTF8 key (or fail with substitute U+FFFD instead)
        $token_key = (utf8_validate($_POST['token'])) ? magic_unquote($_POST['token']) : "\xEF\xBF\xBD";
        $token_id = token_lookup(ALTHING_REFERENCE,$token_key,$t0,$t1,$ip_addr,$data);
    }
    //
    // 2 -- handle cases of expired tokens and Cancel first
    //
    $now = time();
    if (($token_id !== FALSE) && (isset($_POST['button_cancel']))) { // visitor pressed [Cancel]
        $theme->add_message(t('cancelled','admin'));
        token_destroy($token_id);
        $token_id = FALSE;
    }
    if (($token_id !== FALSE) && ($t1 < $now)) { // token expired
	$theme->add_message(t('error_token_expired','m_althing'));
        token_destroy($token_id);
        $token_id = FALSE;
    }
    //
    // 3 -- handle the three remaining buttons from the two dialogs
    //
    if ($token_id !== FALSE) {
        if (isset($_POST['button_preview'])) {
            //
            // 3A -- Preview button
            //
            $dialogdef = althing_view_get_dialogdef($config, $token_key);
            if (!althing_view_dialog_validate($dialogdef)) {
                foreach($dialogdef as $k => $item) {
                    if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                        $theme->add_message($item['error_messages']);
                    }
                }
                althing_show_form($theme, $config, $dialogdef);
            } else {
                $data['dialogdef'] = $dialogdef; // $data['parents'] was already filled in step 0B
                if (!token_store($token_id, $data)) {
                    $theme->add_message(t('error_storing_data','m_althing'));
                    logger(sprintf('%s(): token store error in page %d: %s',
                                   __FUNCTION__,$node_id,db_errormessage()));
                    return FALSE;
                }
                althing_show_preview($theme, $config, $dialogdef, $ip_addr, $data['parents']);
            }
        } elseif (isset($_POST['button_edit'])) {
            //
            // 3B -- Edit button
            //
            if ($data === FALSE) {
                $theme->add_message(t('error_retrieving_data','m_althing'));
                logger(sprintf('%s(): no data after token_lookup()? (page=%d)',__FUNCTION__,$althing_id));
                $data = array('parents' => array(),
                              'dialogdef' => althing_view_get_dialogdef($config, $token_key)
                              );
            }
            althing_show_form($theme, $config, $data['dialogdef']);
        } elseif (isset($_POST['button_done'])) {
            //
            // 3C -- Done button (adds post)
            //
            if ($data === FALSE) {
                $theme->add_message(t('error_retrieving_data','m_althing'));
                logger(sprintf('%s(): no data after token_lookup()? (page=%d)',__FUNCTION__,$althing_id));
                $data = array('parents' => array(),
                              'dialogdef' => althing_view_get_dialogdef($config, $token_key)
                              );
            }
            if ($now < $t0) { // the window of opportunity is still closed; go back to form a la Edit
                $msg = t('error_too_fast','m_althing');
                $theme->add_message($msg);
                $theme->add_popup_top($msg);
                logger(sprintf('%s(): reply too fast (%ds) from %s',__FUNCTION__,$t0-$now,$ip_addr));
                althing_show_form($theme, $config, $data['dialogdef']);
            } else {
                $post = array();
                if (!althing_submit_message($theme,$config, $data, $ip_addr, $now-$t0, $post)) {
                    $theme->add_message(t('error_submitting_message','m_althing'));
                    althing_show_form($theme, $config, $data['dialogdef']);
                } else {
                    token_destroy($token_id);
                    althing_show_thankyou($theme, $config, $ip_addr, $post);
                }
            }
        } else {
            //
            // 3D -- catch all: initiate a new round (shouldn't happen)
            //
            token_destroy($token_id);
            $token_id = FALSE;
        }
    }
    return ($token_id === FALSE) ? althing_show_overview($theme,$config) : TRUE;
} // althing_view()


/** display the althing information and all published posts OR a single post
 *
 * this is the main screen that users/visitors usually see on arrival:
 * the header/intro/question, a list of 0,1 or more published posts,
 * and the conclusion and the footer. There are also buttons to
 * [Post] a message or [Harvest] exiting messages.
 *
 * However, this routine also takes care of requests for a single,
 * individual post. This post is specified via the command line parameter
 * 'post'. If it is specified AND if it identifies a valid (published)
 * message within this Althing that single post is displayed, together
 * with an OK-button. This function is used to direct subscribers and
 * moderators to a specific single message. Note that most links from the
 * single message (parents, children and even self) bring the visitor to
 * the whole althing, ie. they then leave the single post view. (July 2014)
 *
 * @param object &$theme collects (html) output
 * @param array $config holds the althing configration parameters
 * @return bool TRUE on success, FALSE if error retrieving data from DB
 */
function althing_show_overview(&$theme,$config) {
    $retval = TRUE; // assume success
    $althing_id = intval($config['althing_id']);
    $status = intval($config['status']);
    $theme->add_content(html_a('',NULL,array('name' => 'althing_top'),''));

    $sort = get_parameter_int('sort',1); // 0=hide redacted, 1=show redacted
    $show = get_parameter_int('show',1); // 1=sort ascending, -1=sort descending
    $post_id = get_parameter_int('post',NULL); // if specified show only this post

    // 1A -- always start off with the header
    if (!empty($config['header'])) {
        $theme->add_content('<h2 class="althing_header">'.$config['header'].'</h2>');
    }

    // 1B -- get posts and check if we do we need to show but a single post
    if (($status == ALTHING_STATUS_OPENED) || ($status == ALTHING_STATUS_FROZEN)) {
        $posts = array();
        $retval = althing_view_get_posts($theme,$althing_id, $posts,$sort,$show);
        if (!is_null($post_id)) {
            if (isset($posts[$post_id])) {
                // OK. Show single post + [OK] and we're out of here
                $href = was_node_url($theme->node_record); 
                $visible = (db_bool_is(TRUE,$posts[$post_id]['visible'])) ? TRUE : FALSE;
                althing_view_show_post($theme,$posts[$post_id],$visible,TRUE,TRUE,'odd',$href);
                $theme->add_content(html_a('',NULL,array('name' => 'althing_bot'),''));
                $dialogdef = array('button_ok' => dialog_buttondef(BUTTON_OK));
                $href .= '#p'.$post_id; // direct user to position in list via #fragment after pressing OK
                $theme->add_content(dialog_quickform($href,$dialogdef));
                return $retval;
            }
            // Still here? Show error message and fall through to regulare overview
            $theme->add_message(t('error_showing_post','m_althing',array('{POST_ID}'=>strval($post_id))));
            logger(sprintf("%s(): cannot show individual post '%d' in althing '%d'",
                           __FUNCTION__,$post_id,$althing_id));
        }
    }

    // 2 -- continue with introduction and question and the regular overview
    if (!empty($config['introduction'])) {
        $theme->add_content('<div class="althing_introduction">');
        $theme->add_content($config['introduction']);
        $theme->add_content('</div>');
        }
    if (!empty($config['question'])) {
        $theme->add_content('<div class="althing_question">');
        $theme->add_content($config['question']);
        $theme->add_content('</div>');
        }

    // 3 -- maybe show existing posts
    if (($status == ALTHING_STATUS_OPENED) || ($status == ALTHING_STATUS_FROZEN)) { // show existing
        // 4 -- display 1 or 2 buttons or a simple message about Althing being frozen
        $button_post = dialog_buttondef('button_post',t('button_post_value','m_althing'));
        $button_harvest = dialog_buttondef('button_harvest',t('button_harvest_value','m_althing'));
        if ($status == ALTHING_STATUS_FROZEN) {
            $theme->add_content('<div class="althing_buttonbar">');
            $theme->add_content('  <p>'.t('althing_view_is_frozen','m_althing').'</p>');
            $theme->add_content('</div>');
        } else {
            $theme->add_content(html_form(was_node_url($theme->node_record)));
            if (sizeof($posts) > 0) { // there is at least 1 post to reply to
                $theme->add_content('<div class="althing_buttonbar_top">');
                $theme->add_content(dialog_get_widget($button_post));
                $theme->add_content(dialog_get_widget($button_harvest));
                $theme->add_content('<p>'.t('althing_view_post_or_reply','m_althing').'</p>');
            } else {
                $theme->add_content('<div class="althing_buttonbar">');
                $theme->add_content(dialog_get_widget($button_post));
                $theme->add_content('  <p>'.t('althing_view_post_only','m_althing').'</p>');
            }
            $theme->add_content('</div>');
        }

        // 5 -- step through all posts and display the content including checkboxes etc.
        $count_visible = 0;
        $count_hidden = 0;
        if (sizeof($posts) > 0) {
            $frozen = ($status == ALTHING_STATUS_FROZEN) ? TRUE : FALSE;
            $odd_even = "odd";
            $theme->add_content('<div class="althing_posts">'); // container for ALL posts
            foreach($posts as $post_id => $post) {
                $odd_even = ($odd_even == "odd") ? "even" : "odd";
                if (db_bool_is(TRUE,$post['visible'])) {
                    $visible = TRUE;
                    ++$count_visible;
                } else {
                    $visible = FALSE;
                    ++$count_hidden;
                }
                althing_view_show_post($theme,$post,$visible,$frozen, FALSE,$odd_even);
                $theme->add_content(''); // visibility in generated code
            }
            $theme->add_content('</div>'); // end of container for ALL posts
        }
        // 6 -- perhaps we need to show sort/show links
        $show_sort = '';
        if ((sizeof($posts) > 0) || ($show == 0)) { // at least we need to show an option to show redacted
            if ($show) { // currently showing everything so there is at least 1 post currently visible
                if ($count_hidden > 0) { // more precise: there is at least 1 post to hide, do show button
                    $show_anchor = t('hide_redacted','m_althing');
                    $show_params = array('show' => '0', 'sort' => strval($sort));
                } else { // nothing, because there are no redacted posts to hide
                    $show_anchor = '';
                    $show_params = array();
                }
            } else { // currently hiding, so we need to display the show all button
                $show_anchor = t('show_redacted','m_althing');
                $show_params = array('show' => '1', 'sort' => strval($sort));
            }
            if (sizeof($posts) > 1) { // no point in sorting 0 or 1 posts
                if ($sort < 0) {
                    $sort_anchor = t('sort_ascending','m_althing');
                    $sort_params = array('show' => strval($sort), 'sort' => '1');
                } else {
                    $sort_anchor = t('sort_descending','m_althing');
                    $sort_params = array('show' => strval($sort), 'sort' => '-1');
                }
            } else {
                $sort_anchor = '';
                $sort_params = array();
            }
            if (!empty($show_anchor)) {
                $href = was_node_url($theme->node_record,$show_params);
                $show_sort .= html_a($href,NULL,NULL,$show_anchor).' ';
            }
            if (!empty($sort_anchor)) {
                $href = was_node_url($theme->node_record,$sort_params);
                $show_sort .= html_a($href,NULL,NULL,$sort_anchor).' ';
            }
        }

        // 7 -- more or less repeat the button bar but don't repeat lone [Post]
        if ($status == ALTHING_STATUS_FROZEN) {
            $theme->add_content('<div class="althing_buttonbar">');
            $theme->add_content('  <p>'.t('althing_view_is_frozen','m_althing').'</p>');
            if (!empty($show_sort)) {
                $theme->add_content('  <p>'.trim($show_sort).'</p>');
            }
            $theme->add_content('</div>');
        } else {
            if (sizeof($posts) > 0) { // there is at least 1 post to reply to
                $theme->add_content('<div class="althing_buttonbar_bottom">');
                $theme->add_content('  <p>'.t('althing_view_post_or_reply','m_althing').'</p>');
                $theme->add_content(dialog_get_widget($button_post));
                $theme->add_content(dialog_get_widget($button_harvest));
                if (!empty($show_sort)) {
                    $theme->add_content('  <p>'.trim($show_sort).'</p>');
                }
                $theme->add_content('</div>');
            } // else do NOT repeat that lone [Post] button
            $theme->add_content(html_form_close());
        }
    }

    // 8 -- always finish with the conclusion and the footer
    if (!empty($config['conclusion'])) {
        $theme->add_content('<div class="althing_conclusion">');
        $theme->add_content($config['conclusion']);
        $theme->add_content('</div>');
    }
    if (!empty($config['footer'])) {
        $theme->add_content('<div class="althing_footer">');
        $theme->add_content($config['footer']);
        $theme->add_content('</div>');
    }
    $theme->add_content(html_a('',NULL,array('name' => 'althing_bot'),''));
    return $retval;
} // althing_show_overview()


/** output a single post via $theme
 *
 * [ ] $author $datim (#$msgid) Subject: $subject Marbles: $marbles (reply to $msgid)
 * In reply to: $p1 Replies: $c1, $c2, $c3 Harvests: $h1
 *
 * bbcode2html(Content)
 *
 * Note "In reply to: $p1" might also be 'Harvested from: $p2, $p3, $p4' depending on the post.
 *
 * Note: preview mode means that the displayed post does not have links to elsewhere
 * which would let the user leave the current page. This is handy for showing a preview
 * of a message that is not yet posted.
 *
 * Note: the parameter href (default '') is used in the single post display to make sure
 * that the links to the "relations" (parent, children) are absolute, i.e. they link to
 * the full althing page, including headers etc. etc. (the single post page shows only
 * a single post, without the context of the althing).
 *
 * @param object &$theme collects the (html) output
 * @param array &$post a self-contained array with all information about a single post
 * @param bool $visible TRUE is show post, FALSE is show redacted post 
 * @param bool $frozen FALSE indicates if the althing is open for replies, TRUE means althing is frozen
 * @param bool $preview TRUE indicates whether we show this post in preview mode, FALSE=not
 * @param string $odd_even used to link CSS parameters to odd and even posts
 * @param string $href links to this althing page, used to create absolute links to relations
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function althing_view_show_post(&$theme,&$post,$visible=TRUE,$frozen=FALSE,$preview=TRUE,$odd_even='odd',$href='') {
    $fragment = 'p'.$post['post_id'];

    // 0 -- some fields are redacted if the post should be invisible (others are omitted later)
    if ($visible) {
        $attr = array('title' => t((empty($post['author_id'])) ? 'anonymous':'registered','m_althing'));
        $author = html_tag('span',$attr,htmlspecialchars($post['author_full_name']));
        $subject = htmlspecialchars($post['subject']);
        $content = bbcode2html($post['content']);
    } else {
        $author = t('althing_view_redacted_name','m_althing');
        $subject = t('althing_view_redacted_subject','m_althing');
        $content = t('althing_view_redacted_content','m_althing');
    }

    // 1 -- add basic flexibility for display of header fields
    $aparams = array(
        '{AUTHOR}'     => $author,
        '{SUBJECT}'    => $subject,
        '{DISPLAY_ID}' => strval($post['display_id']),
        '{POST_ID}'    => strval($post['post_id']),
        '{DATIM}'      => substr($post['datim'],0,16),
        '{CHECKBOX}'   => '',
        '{MARBLES]'    => '',
        '{MARBLES_BAR}'=> '',
        '{REPLY}'      => '',
        '{DELIMITER}'  => '',
        '{INREPLYTO}'  => '',
        '{HARVESTS}'   => '',
        '{REPLIES}'    => ''
        );

    // 2 -- now construct additional aparams building on earlier substitutions

    // 2A -- checkbox for harvest: only when visible and not frozen
    if (($visible) && (!$frozen)) {
        $checkbox = array(
            'type' => F_CHECKBOX,
            'name' => $fragment,
            'options' => array('1' => ' '), // ' ' is the nearest we get to nothing
            'viewonly' => !$visible,
            'title' => t('althing_view_check_title','m_althing')
            );
        $aparams['{CHECKBOX}'] = dialog_get_widget($checkbox);
    }
    // 2B -- clickable message id that doubles as a named A-tag for future reference
    $anchor = t('althing_view_msg_id','m_althing',$aparams);
    $aparams['{MESSAGE_ID}'] = html_a($href.'#'.$fragment,NULL,array('name' => $fragment),$anchor);

    // 2C -- two variations of datim are available: DATIM and SHORT_DATIM (with full DATIM in title)
    $aparams['{SHORT_DATIM}']=html_tag('span',array('title'=>$post['datim']),short_datim($post['datim']));

    // 2D -- marbles (not for redacted posts)
    if ($visible) {
        $m0 = t('althing_view_marble0','m_althing');
        $m1 = t('althing_view_marble1','m_althing');
        $marbles = intval($post['marbles']);
        $aparams['{MARBLES}'] = strval($marbles);
        $attr = array('class' => 'marbles','title'=>t('althing_view_marbles_bar','m_althing',$aparams));
        $marbles_bar = html_tag('span',$attr,str_repeat($m1,$marbles).str_repeat($m0,ALTHING_MARBLES_LIMIT-$marbles));
        $aparams['{MARBLES_BAR}'] = $marbles_bar;
    }
    // 2E -- clickable reply-link only for visible posts and non-frozen althing
    if (($visible) && (!$frozen)) {
        $params = array('reply' => $post['post_id']);
        $anchor = t('althing_view_msg_reply','m_althing',$aparams);
        $attr = array('title' => t('althing_view_msg_reply_title','m_althing',$aparams));
        $msg_reply = html_a(was_node_url($theme->node_record,$params,'',$preview),NULL,$attr,$anchor);
        $aparams['{REPLY}'] = $msg_reply;
    }
    // 2F -- links to move user to the top/bottom of the page
    $attr = array('width'  => 16,'height' => 16,
        'title'  => t('althing_view_top_title','m_althing'),
        'alt'    => t('althing_view_top_alt','m_althing'));
    $img = html_img(was_url('program/modules/althing/ub.png'),$attr);
    $aparams['{TOP}'] = html_a('#althing_top',NULL,array('title'=>$attr['title']),$img);
    $attr = array('width'  => 16,'height' => 16,
        'title'  => t('althing_view_bottom_title','m_althing'),
        'alt'    => t('althing_view_bottom_alt','m_althing'));
    $img = html_img(was_url('program/modules/althing/db.png'),$attr);
    $aparams['{BOTTOM}'] = html_a('#althing_bot',NULL,array('title'=>$attr['title']),$img);

    // 2G -- $InReplyTo OR $HarvestedFrom $Replies $Harvests
    $relations = array('parents'=>'parent', 'children'=>'child', 'harvesters'=>'harvester');
    foreach($relations as $k => $v) {
        if ((isset($post[$k])) && (($n = sizeof($post[$k])) > 0)) {
            if ($v == 'parent') {
                $list = t(($n > 1)?'althing_view_list_'.$k.'_n':'althing_view_list_'.$k.'_1','m_althing');
            } else {
                $list = t('althing_view_list_'.$k,'m_althing');
            }
            $i = 0;
            foreach($post[$k] as $post_id => $display_id) {
                $comma = (++$i < $n) ? ',' : '';
                $idparams = array('{DISPLAY_ID}' => $display_id,'{POST_ID}' => $post_id);
                $anchor = t('althing_view_list_'.$v,'m_althing',$idparams);
                $list .= ' '.html_a($href.'#p'.$post_id,NULL,NULL,$anchor).$comma;
            }
            // delimiter is only set when there is at least one list
            $aparams['{DELIMITER}'] = t('althing_view_list_delimiter','m_althing');
            if ($v == 'parent') {
                $aparams['{INREPLYTO}'] = $list;
            } elseif ($v == 'child') {
                $aparams['{REPLIES}'] = $list;
            } else {
                $aparams['{HARVESTS}'] = $list;
            }
        }
    }

    // 3 -- wrap it up and put the  whole post in a DIV
    $theme->add_content('<div class="althing_post althing_post_'.$odd_even.'">');
    $theme->add_content('  <div class="althing_post_head althing_post_head_'.$odd_even.'">');
    $theme->add_content('    '.t('althing_view_post_header1','m_althing',$aparams));
    $theme->add_content('    <br>');
    $theme->add_content('    '.t('althing_view_post_header2','m_althing',$aparams));
    $theme->add_content('  </div>'); // post header
    $theme->add_content('  <div class="althing_post_body althing_post_body_'.$odd_even.'">');
    $theme->add_content('    '.$content);
    $theme->add_content('  </div>'); // post body
    $theme->add_content('</div>'); // post
} // althing_view_show_post()


/** get all posts for this althing into an array for easy output
 *
 * this retrieves data from the database in a specific order (if necessary).
 * via $show=0 it is possible to actually suppress the output of the redacted posts
 *
 * @param object &$theme collects (html) output
 * @param int $althing_id aka $node_id
 * @param array &$posts receives all posts (could be empty array)
 * @param int $sort is the sort order of the posts for display: 1=ascending, -1=descending
 * @param int $show indicates actual visibility of redacted posts: 1=show redacted, 0=hide redacted.
 * @return bool TRUE on success + data in &$posts, FALSE otherwise and &$posts == array()
 */
function althing_view_get_posts(&$theme, $althing_id, &$posts,$sort=1,$show=1) {
    $retval = TRUE; // assume success

    // 1 -- get raw data in $posts array
    $table = 'althing_posts';
    $fields = '*';
    $where = array('althing_id' => $althing_id, ($show) ? 'published' : 'visible' => TRUE);
    $order = array(($sort > 0) ? 'post_id' : 'post_id DESC');
    if (($posts = db_select_all_records($table, $fields,$where,$order,'post_id')) === FALSE) {
        $theme->add_message(t('error_retrieving_posts','m_althing'));
        logger(sprintf("%s(): cannot retrieve althing posts '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        $posts = array();
        return FALSE;
    }

    // 2 -- try to fetch the relations between the posts too
    $table = 'althing_relations';
    $fields = array('parent_id','child_id');
    $where = array('althing_id' => $althing_id);
    if (($relations = db_select_all_records($table, $fields,$where)) === FALSE) {
        $theme->add_message(t('error_retrieving_posts','m_althing'));
        logger(sprintf("%s(): cannot retrieve althing post relations '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        $posts = array();
        return FALSE;
    }
  
    // 3 -- Add the relations between the selected posts to array
    // 3A -- 1st round: add all parents to the children
    foreach($relations as $relation) {
        $parent_id = intval($relation['parent_id']);
        $child_id = intval($relation['child_id']);
        if ((isset($posts[$parent_id])) && (isset($posts[$child_id]))) {
            if (!isset($posts[$child_id]['parents'])) {
                $posts[$child_id]['parents'] = array($parent_id=>$posts[$parent_id]['display_id']);
            } else {
                $posts[$child_id]['parents'][$parent_id] = $posts[$parent_id]['display_id'];
            }
        }
    }
    // 3B -- 2nd round: add children OR harvesters to parents
    foreach($relations as $relation) {
        $parent_id = intval($relation['parent_id']);
        $child_id = intval($relation['child_id']);
        if ((isset($posts[$parent_id])) && (isset($posts[$child_id]))) {
            $relation = (sizeof($posts[$child_id]['parents']) > 1) ? 'harvesters' : 'children'; 
            if (!isset($posts[$parent_id][$relation])) {
                $posts[$parent_id][$relation]=array($child_id => $posts[$child_id]['display_id']);
            } else {
                $posts[$parent_id][$relation][$child_id] = $posts[$child_id]['display_id'];
            }
        }
    }
    // At this point all children and harvesters have 1 or more parents in $post['parents']
    // All parents may have $post['harvesters'] or $post['children'] or both. This translates
    // directly to the information in the displayed post header lateron.
    return TRUE;    
} // althing_view_get_posts()


/** construct a dialog definition for the add a post form
 *
 * This defines the add a post form.
 *
 * @param array $config althing configuration record straight from althings table
 * @param string $token_key
 * @return array datadefinition
 */
function althing_view_get_dialogdef($config, $token_key, $posts=NULL) {
    global $USER;
    // 1 -- logged in users are registered users and don't need to enter their name/email
    if ($USER->is_logged_in) {
        $fullname = $USER->full_name;
        $email = $USER->email;
        $registered = TRUE;
    } else {
        $fullname = '';
        $email = '';
        $registered = FALSE;
    }
    // 2A -- prepare some initial values (overruled by $_POST[] values on preview/submit)
    $re = t('re','m_althing');
    $subject = trim($config['header']);
    $parents = array();
    $message = '';

    // 2B -- if we are constructing a reply, we quote the parent post(s) and maybe the subject
    if ((is_array($posts)) && (sizeof($posts) > 0)) {
        $r = strlen($re);
        $i = 0;
        foreach($posts as $post_id => $post) {
            $parents[$post_id] = intval($post['display_id']); // need to remember this for post relations
            $s = trim($post['subject']);  // eat any 'Re:' to prevent meaningless Re: Re: Re: sequences
            if (substr($s,0,$r) == $re) { // substr is utf8-safe here long as $re and $s
                $s = trim(substr($s,$r)); // are valid utf8 (which was enforced when storing $post)
            }
            if (++$i == 1) {
                $subject = $s;
            } elseif ($subject != $s) { // fall back to althing title if different parent subjects
                $subject = trim($config['header']);
            }
            $parms = array('{AUTHOR}' => trim($post['author_full_name']),
                           '{DISPLAY_ID}' => strval($post['display_id']),
                           '{POST_ID}' => strval($post_id),
                           '{DATIM}' => substr($post['datim'],0,16)
                           );
            $quoting = strtr(t('quoting','m_althing',$parms),'"',"'");
            $content = trim($post['content']);
            $message .= sprintf("[quote=\"%s\"]\n%s\n[/quote]\n",$quoting,$content);
        }
    }
    // 2C -- prepend a 'Re:' if the bare subject is not the empty string
    if (!empty($subject)) {
        $subject = ltrim($re.' '.$subject); // length MIGHT be >80, let dialog_validate() handle that later
    }

    // 3 -- actually construct it
    $dialogdef = array(
        'token' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'token',
	    'hidden' => TRUE,
            'value' => $token_key
            ),
        'fullname' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'fullname',
            'minlength' => 1,
            'maxlength' => 255,
            'columns' => 53,
            'label' => t(($registered) ? 'fullname_label' : 'fullname_required_label','m_althing'),
            'title' => t(($registered) ? 'fullname_title' : 'fullname_required_title','m_althing'),
            'value' => $fullname,
            'viewonly' => $registered
            ),
        'email' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'email',
            'minlength' => 3,
            'maxlength' => 255,
            'columns' => 53,
            'label' => t(($registered) ? 'email_label' : 'email_required_label','m_althing'),
            'title' => t(($registered) ? 'email_title' : 'email_required_title','m_althing'),
            'value' => $email,
            'viewonly' => $registered
            ),
        'subject' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'subject',
            'minlength' => 0,
            'maxlength' => 80,
            'columns' => 53,
            'label' => t('subject_label','m_althing'),
            'title' => t('subject_title','m_althing'),
            'value' => $subject,
            ),
        'message' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'message',
            'id' => 'althing_message',
            'minlength' => 1,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 70,
            'rows' => 15,
            'label' => t('message_label','m_althing'),
            'title' => t('message_title','m_althing'),
            'value' => $message,
            'parents' => $parents // trickt to tell the caller about which posts we are respondingn to
            ),
        'button_preview' => dialog_buttondef('button_preview',t('button_preview','m_althing')),
        'button_cancel' => dialog_buttondef(BUTTON_CANCEL)
        );
    return $dialogdef;
} // althing_view_get_dialogdef()


/** check for user action indicating wish to add a post, reply to a post or harvest posts
 *
 * this checks to see if the user has taken action to start adding a new post.
 * This is the case if 
 *  1. the user pressed the [Post] button
 *  2. the user clicks the reply-link in an existing post's header
 *  3. the user selected 0, 1 or more existing posts and pressed the [Harvest] button
 *
 * If we find that the user does want to add/reply/harvest, we collect the
 * 0, 1 or more posts in an array, keyed by post_id. This array is then returned.
 * If the user does NOT want to add/reply/harvest, we return NULL.
 * If the althing is currently closed or frozen (or rather not open), we
 * tell the user about that and the return value is FALSE.
 *
 * Note:
 * In step 3A below we identify the checked posts by looking at the $_POST[] array.
 * This is a shortcut, because usually we first setup a dialogdef and subsequently
 * check all the elements in the $dialogdef for their values. We now sort of trust
 * the data in the POST-array; an ordinary user will only be able to  have just the
 * checkboxes from the overview checked, leading to an element with key 'p' followed
 * by the post_id (in decimal). An attacker might try to fool us by crafting a post
 * with post_id's from another althing, eg. one hidden in a protected area. We mitigate
 * that by explicitly requiring the althing_id in the WHERE-clause we will be constructing.
 * Another possible danger is specifying many, many, many post_id's. We will limit that
 * via the limit of posts that can be harvested in an althing (there must be a limit,
 * somewhere, right?).
 *
 * @param object &$theme collects (html) output
 * @param int $althing_id aka $node_id
 * @param int $status tells us about Open/Frozen/Closed status of the althing
 * @return bool|null|array FALSE on error, NULL if no add/reply/harvest, $posts-array otherwise
 */
function check_add_reply_harvest(&$theme, $althing_id, $status) {
    // 0 -- sentinel to combine cases 2 and 3 in 1 trip to the database
    $reply = FALSE;
    $reply_to = array();

    // 1 -- a simple post quoting nothing at all yields an empty array (or FALSE if althing not open)
    if (isset($_POST['button_post'])) {
        if ($status != ALTHING_STATUS_OPENED) {
            $theme->add_message(t('althing_add_not_possible','m_althing'));
            return FALSE;
        }
        return array();
    }

    // 2 -- a reply to single message via $_GET[]
    $reply_id = get_parameter_int('reply',NULL);
    if (!is_null($reply_id)) {
        if ($status != ALTHING_STATUS_OPENED) {
            $theme->add_message(t('althing_add_not_possible','m_althing'));
            return FALSE;
        }
        $reply_to[$reply_id] = sprintf('(post_id=%d)',$reply_id);
        $reply = TRUE;
    }

    // 3 -- a harvest of 0, 1 or more posts (max ALTHING_HARVEST_LIMIT) via checkboxes in $_POST[]
    if (isset($_POST['button_harvest'])) {
        if ($status != ALTHING_STATUS_OPENED) {
            $theme->add_message(t('althing_add_not_possible','m_althing'));
            return FALSE;
        }
        // 3A -- shortcut: attempt to determine post_id's directly from $_POST[]
        foreach($_POST as $k => $v) {
            if (!utf8_validate($k)) {
                logger(sprintf("%s(): cannot harvest invalid utf8 in althing '%d': showing overview",
                               __FUNCTION__,$althing_id));
                $theme->add_message(t('althing_harvest_not_possible','m_althing'));
                return FALSE;
            }
            if ($k[0] == 'p') { // format is 'X9999999999'
                $reply_id = intval(substr($k,1));
                $reply_to[$reply_id] = sprintf('(post_id=%d)',$reply_id);
            }
        }
        // 3B -- check sanity limit for quoted messages
        if (($n = sizeof($reply_to)) > ALTHING_HARVEST_LIMIT) {
            logger(sprintf("%s(): cannot harvest '%d' posts (max:'%d') in althing '%d': showing overview",
                           __FUNCTION__, $n, ALTHING_HARVEST_LIMIT, $althing_id));
            $theme->add_message(t('althing_harvest_too_many','m_althing',
                                  array('{COUNT}' => strval($n), '{LIMIT}' => ALTHING_HARVEST_LIMIT)));
            return FALSE;
        }
        $reply = TRUE;
    }

    // 4 -- go to the DB and fetch posts (if any are specified)
    if ($reply) { // either a single reply (step 2) or a harvest (step 3)
        if (sizeof($reply_to) <= 0) { // none, equivalent of [Post]; return empty array (not NULL or FALSE)
            return array();
        }
        if (sizeof($reply_to) == 1) {
            list($reply_id,$dummy) = each($reply_to);
            $where = array('althing_id' => $althing_id, 'post_id' => $reply_id, 'visible' => TRUE);
        } else {
            $where = sprintf('((althing_id=%d) AND (visible) AND (%s))',$althing_id,
                             implode(' OR ',$reply_to));
        }
        $table = 'althing_posts';
        $fields = '*';
        $order = 'post_id';
        if (($posts = db_select_all_records($table,$fields,$where,$order,'post_id')) === FALSE) {
            $theme->add_message(t('error_retrieving_posts','m_althing'));
            logger(sprintf("%s(): cannot retrieve althing posts for reply/harvest in althing '%d': %s",
                           __FUNCTION__,$althing_id,db_errormessage()));
            return FALSE;
        }
        return $posts;
    }
    return NULL;
} // check_add_reply_harvest()


/** show the form where the user/visitor can compose a new post
 *
 * this routine displays the form where a registered user or an anonymous visitor
 * can compose a new post. There is a difference between the two groups:
 * - users cannot change their name + email; this is copied from $USER,
 * - visitors are allowed to pick their own nicknames and email address
 * Another difference is that only registered users (with sufficient permissions)
 * have an additional link that leads to the file browser.
 *
 * The file browser allows for inserting a full path into the message area,
 * at the current cursor position. This way the user does not have to type
 * the filename/path when inserting a URL or an IMG.
 *
 * @param object &$theme collects (html) output
 * @param array $config holds the althing configration parameters
 * @param array $dialogdef defines the various fields in the dialog
 * @return void output added to $theme
 */
function althing_show_form(&$theme, $config, $dialogdef) {
    global $CFG,$USER;
    $theme->add_content(html_tag('h2','',t('add_post_header','m_althing')));
    if ($USER->has_job_permissions(JOB_PERMISSION_FILEMANAGER)) {
        $file_php = ($CFG->friendly_url) ? $CFG->www_short.'/file.php' : $CFG->www_short.'/file.php?file=';
        $javascript_code = array(
            '<script type="text/javascript"><!--',
            '  function SetUrl(url) {',
            '    '.sprintf('var k = "%s".length;',$file_php),
            '    var p=url.slice(k);',                                // extract the 'path' part from url
            '    var f=document.getElementById("althing_message");',
            '    var i=f.selectionStart;',
            '    var j=f.selectionEnd;',
            '    var s=f.value;',
            '    f.value=s.slice(0,i)+p;',                            // insert bare URL into textarea
            '    k=f.value.length;',
            '    f.value+=s.slice(j);',
            '    f.setSelectionRange(i,k);',
            '    f.focus();',
            '  }',
            '//--></script>'
            );
        foreach($javascript_code as $line) {
            $theme->add_html_header($line);
        }
        $href = $CFG->www_short.'/admin.php';
        $params = array('job' => 'filebrowser'); // JOB_FILEBROWSER='filebrowser'
        $url = href($href,$params);
        $attr = array('title' => t('filemanager_title','m_althing'),
                      'target' => '_blank',
                      'onclick' => "window.open('$url','','left=100,top=100,height=480,width=640,resizable,scrollbars');return false;");
        $anchor = t('filemanager','m_althing');
        $theme->add_content('<p>'.html_a($href,$params,$attr,$anchor).'</p>');
    }
    $href = was_node_url($theme->node_record);
    $theme->add_content(dialog_quickform($href,$dialogdef));
    bbcode_help($theme);
} // althing_show_form()


/** validate the data entered by the user/visitor

 * Other than the standard validation we check for at least 1 @-sign in
 * the email address. (It is non-trivial to validate an email address,
 * it may not even be possible to do it in a single regular expression.
 * See http://stackoverflow.com/questions/201323 for more information.)
 * Note that we only check this field if it is not viewonly. Reason: the
 * user is not able to edit a viewonly field, so an error there would
 * be a fatal error. We don't want that.
 *
 * Also, we validate the BBCode in the message.
 *
 * @param array &$dialogdef defines the dialog and will hold POSTed values
 * @return bool TRUE if valid, else FALSE + messages added to dialogdef
 */
function althing_view_dialog_validate(&$dialogdef) {
    $retval = TRUE; // assume success
    // 1 -- check the usual suspects (side-effect: fetch POSTed values in dialogdef)
    if (!dialog_validate($dialogdef)) {
        $retval = FALSE;
    }

    // 2 -- special case: there must be at least 1 @-sign in the email field (if it is editable anyway)
    if ((!isset($dialogdef['email']['viewonly'])) || (!$dialogdef['email']['viewonly'])) {
        if (strpos($dialogdef['email']['value'],'@') === FALSE) { // no @-sign in mail address
            $retval = FALSE;
            ++$dialogdef['email']['errors'];
            $params = array('{FIELD}' => str_replace('~','',$dialogdef['email']['label']));
            $dialogdef['email']['error_messages'][] = t('validate_invalid','',$params);
        }
    }

    // 3 -- message must contain valid BBCode
    if (!bbcode_validate($dialogdef['message'])) {
        $retval = FALSE;
    }
    return $retval;
} // althing_view_dialog_validate()


/** show a preview of the post to the user, including references to parents etc.
 *
 * the user can either submit the post or go back to edit the post
 *
 * @param object &$theme collects (html) output
 * @param array $config althing configuration record straight from althings table
 * @param array &$dialogdef defines the dialog and will hold POSTed values
 * @param string $ip_addr is the address the user/visitor is calling from
 * @param array $parents contains parent_id's of this post
 * @return bool FALSE if something went wrong, TRUE on succes + post data stored in $post
 */
function althing_show_preview(&$theme, $config, $dialogdef, $ip_addr, $parents) {
    global $USER;
    $author_id = ($USER->is_logged_in) ? $USER->user_id : NULL;
    $author = ($USER->is_logged_in) ? $USER->full_name : $dialogdef['fullname']['value'];
    $email = ($USER->is_logged_in) ? $USER->email : $dialogdef['email']['value'];
    $now = strftime("%Y-%m-%d %T");
    $post = array(
        'post_id'          => 0,
        'althing_id'       => intval($config['althing_id']),
        'display_id'       => 0,
        'author_id'        => $author_id,
        'author_full_name' => $author,
        'email'            => $email,
        'subject'          => $dialogdef['subject']['value'],
        'datim'            => $now,
        'content'          => $dialogdef['message']['value'],
        'visible'          => SQL_TRUE, // fake visibility...
        'published'        => SQL_TRUE, // ...and publication status too
        'marbles'          => 0,
        'remote_addr'      => $ip_addr,
        'mtime'            => NULL,
        'muser_id'         => NULL
        );
    if (!empty($parents)) {
        $post['parents'] = $parents;
    }
    $forbidden = array(chr(10),chr(13),chr(34),'\\');
    $mailfrom = sprintf('&quot;%s&quot; &lt;%s&gt;',
                        htmlspecialchars(str_replace($forbidden,'',trim($author))),
                        htmlspecialchars(str_replace($forbidden,'',trim($email))));
    $subject = htmlspecialchars(trim($dialogdef['subject']['value']));
    $remote_addr = htmlspecialchars($ip_addr);

    $theme->add_content(html_a('',NULL,array('name' => 'althing_top'),''));
    $theme->add_content(html_tag('h2','',t('preview_header','m_althing')));
    $theme->add_content('<p>');
    $theme->add_content(sprintf('  <strong>%s</strong>: %s<br>',t('from','m_althing'),$mailfrom));
    $theme->add_content(sprintf('  <strong>%s</strong>: %s<br>',t('subject','m_althing'),$subject));
    $theme->add_content(sprintf('  <strong>%s</strong>: %s<br>',t('date','m_althing'),date('r')));
    $theme->add_content(sprintf('  <strong>%s</strong>: %s<br>',t('ip_addr','m_althing'),$remote_addr));
    $theme->add_content('</p>');
    althing_view_show_post($theme,$post,TRUE,TRUE); // visible+frozen: no checkbox+reply link in preview
    $previewdef = array(
        'token' => $dialogdef['token'],
        'button_edit' => dialog_buttondef(BUTTON_EDIT),
        'button_done' => dialog_buttondef(BUTTON_DONE),
        'button_cancel' => dialog_buttondef(BUTTON_CANCEL)
        );
    // try to defeat the cache of the browser and make {LINK_POST} in thankyou screen work
    $href = was_node_url($theme->node_record,array('cache'=>strval(time())));
    $theme->add_content(dialog_quickform($href,$previewdef));
    $theme->add_content(html_a('',NULL,array('name' => 'althing_bot'),''));
} // althing_show_preview()


/** store the new post, add relations and send mail to moderators (and maybe subscribers too)
 *
 * this routine takes care of storing the (validated) data in $data['dialogdef'] and
 * $data['parents'] into the various tables in the database. If all goes well, we
 * also send a mail to the moderators and, if the post was published immediately,
 * the subscribers too. On success, the array $post is filled with the post data,
 * including the newly assigned post_id. It is up to the caller to do something
 * with this returned post data, e.g. display it or whatever.
 *
 * Storing the data takes a few separate steps. First we need to calculate
 * a display_id which is unique within this althing. Once we have that, we can
 * store the actual post, which yields a fresh (unique) post_id. Subsequently we
 * can use that post_id to store the parent-child-relations in the database.
 *
 * If that all works out, we always inform the moderators and, if the post was
 * published, also the subscribers. If the post is not visible, the subscribers
 * will be sent a mail lateron, if and when the post is published by a moderator.
 *
 * @param object &$theme collects (html) output
 * @param array $config althing configuration record straight from althings table
 * @param array $data contains the (validated) dialogdef and also the post's parents if any
 * @param string $ip_addr is the address the user/visitor is calling from
 * @param int $delay the # of seconds since time=t0
 * @param array &$post a self-contained array returning all information about the post
 * @return bool FALSE if something went wrong, TRUE on succes + post data stored in $post
 */
function althing_submit_message(&$theme, $config, $data, $ip_addr, $delay, &$post) {
    global $USER,$CFG;
    $althing_id = intval($config['althing_id']);

    //
    // 1 -- try to calculate a fresh display_id
    //
    $last_display_id = intval($config['last_display_id']);
    $display_id = $last_display_id + 1;
    $table = 'althings';
    $fields = array('last_display_id' => $display_id);
    $where = array('althing_id' => $althing_id, 'last_display_id' => $last_display_id);
    if (($retval = db_update($table,$fields,$where)) !== 1) {
        logger(sprintf("%s(): cannot set last_display_id to %d in '%d'; trying again: %s",
                       __FUNCTION__,$display_id,$althing_id,db_errormessage()));
        if (($record = db_select_single_record($table,'last_display_id')) === FALSE) {
            logger(sprintf("%s(): double fault: cannot calculate display_id in '%d': %s",
                           __FUNCTION__,$althing_id,db_errormessage()));
            return FALSE;
        }
        $last_display_id = intval($record['last_display_id']);
        $display_id = $last_display_id + 1;
        $fields = array('last_display_id' => $display_id);
        $where = array('althing_id' => $althing_id, 'last_display_id' => $last_display_id);
        if (($retval = db_update($table,$fields,$where)) !== 1) {
            logger(sprintf("%s(): cannot set last_display_id to %d in '%d'; failed again: %s",
                           __FUNCTION__,$display_id,$althing_id,db_errormessage()));
            return FALSE;
        }
    }
    //
    // 2 -- insert a new post record
    //
    switch($config['default_visibility']) {
    case ALTHING_DEFAULT_VISIBILITY_REGISTERED:
        $flag = ($USER->is_logged_in) ? TRUE : FALSE;
        $flag_sql = ($USER->is_logged_in) ? SQL_TRUE : SQL_FALSE;
        break;
    case ALTHING_DEFAULT_VISIBILITY_VISIBLE:
        $flag = TRUE;
        $flag_sql = SQL_TRUE;
        break;
    case  ALTHING_DEFAULT_VISIBILITY_HIDDEN:
        $flag = FALSE;
        $flag_sql = SQL_FALSE;
        break;
    }
    $author_id = ($USER->is_logged_in) ? $USER->user_id : NULL;
    $author = ($USER->is_logged_in) ? $USER->full_name : $data['dialogdef']['fullname']['value'];
    $email = ($USER->is_logged_in) ? $USER->email : $data['dialogdef']['email']['value'];
    $now = strftime("%Y-%m-%d %T");
    $post = array(
        'althing_id'       => $althing_id,
        'display_id'       => $display_id,
        'author_id'        => $author_id,
        'author_full_name' => $author,
        'email'            => $email,
        'subject'          => $data['dialogdef']['subject']['value'],
        'datim'            => $now,
        'content'          => $data['dialogdef']['message']['value'],
        'visible'          => $flag,
        'published'        => $flag,
        'marbles'          => 0,
        'remote_addr'      => $ip_addr,
        'mtime'            => NULL,
        'muser_id'         => NULL
        );
    $table = 'althing_posts';
    if (($post_id = db_insert_into_and_get_id($table,$post,'post_id')) === FALSE) {
        logger(sprintf("%s(): cannot insert new post (display_id=%d) into althing '%d': %s",
                       __FUNCTION__,$display_id,$althing_id,db_errormessage()));
        return FALSE; // this leaves a gap in the display_id's. Oh well.
    }
    $post['post_id'] = $post_id; // pretend $post[] is the record straight from the DB...
    $post['visible'] = $flag_sql; // ...so caller can easily create a clean thankyou-screen...
    $post['published'] = $flag_sql;

    //
    // 3 -- update the relations (if any)
    //
    if (sizeof($data['parents']) > 0) {
        $post['parents'] = $data['parents']; // also keep track of family ties (if there are any)
        $table = 'althing_relations';
        foreach($data['parents'] as $parent_id => $parent_display_id) {
            $fields = array(
                'parent_id'  => $parent_id,
                'child_id'   => $post_id,
                'althing_id' => $althing_id
                );
            if (db_insert_into($table,$fields) === FALSE) {
                logger(sprintf("%s(): cannot save relation (p=%d c=%d) in %d but continuing anyway: %s",
                               __FUNCTION__,$parent_id,$post_id,$althing_id,db_errormessage()));
            }
        }
    }

    //
    // 4 -- add an alert for the webmaster about a new post
    //
    $aparams = array(
        '{DISPLAY_ID}' => strval($post['display_id']),
        '{POST_ID}'    => strval($post['post_id']),
        '{SUBJECT}'    => strval($post['subject']),
        '{AREA}'       => strval($theme->area_id),
        '{NODE_FULL_NAME}' => sprintf('%d %s (%s)',$theme->node_id,
                                      $theme->node_record['link_text'],
                                      $theme->node_record['title'])
        );
    $msg = t('alert_new_post','m_althing',$aparams);
    $theme->queue_alert($msg,$author);

    //
    // 4A -- do we tell moderators (and maybe the subscribers) about this post?
    //
    $moderators = althing_get_emails($config['email_moderators']);
    $subscribers = ($flag) ? althing_get_emails($config['email_subscribers']) : array();
    if ((sizeof($moderators) <= 0) && (sizeof($subscribers) <= 0)) { // nothing left to do
        return TRUE;
    }

    //
    // 4B -- yes, we need to send at least 1 mail.
    //
    $private = (db_bool_is(TRUE,$theme->area_record['is_private'])) ? TRUE : FALSE;
    $area_id = intval($theme->area_record['area_id']);
    if (sizeof($moderators) > 0) {
        $moderator = TRUE;
        althing_send_alerts($moderators,$moderator,$private,$flag,$area_id,$post);
    }
    if (sizeof($subscribers) > 0) {
        $moderator = FALSE;
        althing_send_alerts($subscribers,$moderator,$private,$flag,$area_id,$post);
    }

    //
    // 5 -- all done
    //
    return TRUE;
} // althing_submit_message()

/** display a thank-you message and the URL of the newly added post
 *
 *
 * @param object &$theme collects (html) output
 * @param array $config althing configuration record straight from althings table
 * @param string $ip_addr is the address the user/visitor is calling from
 * @param array &$post a self-contained array returning all information about the post
 * @return bool TRUE to indicate success (always)
 */
function althing_show_thankyou(&$theme, $config, $ip_addr, $post) {
    $fragment = 'p'.$post['post_id'];
    $href = was_node_url($theme->node_record,NULL,'',FALSE,TRUE);
    $anchor = $href.'#'.$fragment;
    $params = array(
        '{LINK_POST}' =>  html_a($href,NULL,NULL,$anchor,$fragment),
        '{LIST_URL_OPEN}' => html_a($href),
        '{LIST_URL_CLOSE}' => '</a>'
        );
    $forbidden = array(chr(10),chr(13),chr(34),'\\');
    $mailfrom = sprintf('&quot;%s&quot; &lt;%s&gt;',
                        htmlspecialchars(str_replace($forbidden,'',trim($post['author_full_name']))),
                        htmlspecialchars(str_replace($forbidden,'',trim($post['email']))));
    $subject = htmlspecialchars(trim($post['subject']));
    $remote_addr = htmlspecialchars($ip_addr);
    $display_id = $post['display_id'];
    $theme->add_content(html_a('',NULL,array('name' => 'althing_top'),''));
    $theme->add_content(html_tag('h2','',t('thank_you_header','m_althing')));
    if (db_bool_is(TRUE,$post['published'])) {
        $theme->add_content(html_tag('p','',t('thank_you_published','m_althing',$params)));
    } else {
        $theme->add_content(html_tag('p','',t('thank_you_unpublished','m_althing',$params)));
    }
    $theme->add_content(html_tag('p','',t('thank_you_return','m_althing',$params)));
    $theme->add_content('<p>');
    $theme->add_content(sprintf('  <strong>%s</strong>: %s<br>',t('from','m_althing'),$mailfrom));
    $theme->add_content(sprintf('  <strong>%s</strong>: %s<br>',t('subject','m_althing'),$subject));
    $theme->add_content(sprintf('  <strong>%s</strong>: %s<br>',t('date','m_althing'),date('r')));
    $theme->add_content(sprintf('  <strong>%s</strong>: %s<br>',t('ip_addr','m_althing'),$remote_addr));
    $theme->add_content(sprintf('  <strong>%s</strong>: #%s<br>',t('msg_id','m_althing'),$display_id));
    $theme->add_content('</p>');
    althing_view_show_post($theme,$post,TRUE,TRUE); // visible+frozen: no checkbox+reply link in thankyou
    $theme->add_content(html_a('',NULL,array('name' => 'althing_bot'),''));
    return TRUE;
} // althing_show_thankyou()

?>