<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/confab/confab_admin.php - management interface for confab-module
 *
 * This file defines the administrative interface to this module.
 * The interface consists of the following four functions.
 *
 * <code>
 * confab_disconnect(&$output,$area_id,$node_id,$module)
 * confab_connect(&$output,$area_id,$node_id,$module)
 * confab_show_edit(&$output,$area_id,$node_id,$module,$viewonly,$edit_again,$href,$option)
 * confab_save(&$output,$area_id,$node_id,$module,$viewonly,&$edit_again,$option)
 * </code>
 *
 * These functions are called from pagemanagerlib.php whenever necessary.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_confab
 * @version $Id: confab_admin.php,v 1.16 2016/03/23 09:28:21 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }

require_once(dirname(__FILE__).'/confab_common.php'); // common code + constants


/** disconnect this module from a node
 *
 * this breaks the link between the node $node_id in area $area_id and
 * this module. For now we simply delete the records in all tables relating
 * to 'confabs*' using node_id == confab_id. Note the order in which
 * the tables are emptied (to satisfy the FK constraints).
 * 
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node from which we need to disconnect
 * @param array $module the module record straight from the database
 * @return bool TRUE on success, FALSE otherwise
 */
function confab_disconnect(&$output, $area_id, $node_id, $module) {
    $where = array('confab_id' => intval($node_id));
    $tables = array(
        'confab_participants',
        'confab_messages',
        'confabs');
    $retval = TRUE; // assume success
    foreach($tables as $table) {
        if (db_delete($table,$where) === FALSE) {
            logger(sprintf("%s(): cannot disconnect table '%s' from node '%d': %s",
                           __FUNCTION__,$table,$node_id,db_errormessage()));
            $retval = FALSE;
        }
    }
    return $retval;
} // confab_disconnect()


/** connect this module to a node
 *
 * this makes the link between the node $node_id in area $area_id and this module.
 * In this case we create a single 'confabs' record linked to node_id via confab_id.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which we need to connect
 * @param array $module the module record straight from the database
 * @return bool TRUE on success, FALSE otherwise
 * @uses $USER
 */
function confab_connect(&$output, $area_id, $node_id, $module) {
    global $USER;
    $now = strftime('%Y-%m-%d %T');
    $fields = array(
        'confab_id' => intval($node_id),
        'header' => '',
        'introduction' => '',
        'footer' => '',
        'passcode' => '',
        'max_participants' => 26,
        'max_tries' => 10,
        'refresh' => 6,
        'timeout_idle' => 300,          // 5 minutes after last message from the participant
        'timeout_conversation' => 3600, // 1 hour after last message from any participant
        'status' => CONFAB_STATUS_CLOSED,
        'personal' => CONFAB_PERSONAL_NO,
        'last_conversation_id' => 0,
        'last_participant_id' => 0,
        'ctime' => $now,
        'cuser_id' => $USER->user_id,
        'mtime' => $now,
        'muser_id' => $USER->user_id);
    $retval = db_insert_into('confabs',$fields);
    if ($retval !== 1) {
        logger(sprintf("%s(): cannot connect to node '%d': %s",__FUNCTION__,$node_id,db_errormessage()));
        $retval = FALSE;
    } else {
        $retval = TRUE;
    }
    return $retval;
} // confab_connect()


/** dispatcher for showing various dialog screens
 *
 * there are a few dialogs that are accessed via this routine:
 * - overview: a simple introduction to this module (via $option==NULL)
 * - configuration: all properties of the confab (main confabs record) ($option=configuration)
 * - moderation: show current with option to PM users or close the conversation ($option=moderation)
 * - report: a list of _all_ messages in a conversation in this confab ($option=report)
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @param string $option the selected submenu option or NULL if user used the 'Content' option to get here
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function confab_show_edit(&$output, $area_id, $node_id, $module, $viewonly, $edit_again, $href, $option) {
    if ($option == 'configuration') {
        return confab_show_edit_configuration($output, $node_id, $module, $viewonly,$edit_again, $href);
    } elseif ($option == 'moderation') {
        return confab_show_edit_moderation($output,$area_id,$node_id,$module,$viewonly,$edit_again,$href);
    } elseif ($option == 'report') {
        return confab_show_edit_report($output,$node_id,$module,$edit_again,$href);
    } else { // generic call to Content ($option == NULL)
        confab_show_content($output,$node_id,$module);
    }
    return TRUE;
} // confab_show_edit()


/** show the main content screen (in fact documentation only)
 *
 * this displays the simple overview screen with documentation.
 * It ends with a cancel button that ends the whole edit operation for
 * this node.
 * 
 * @param object &$output collects the html output (if any)
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function confab_show_content(&$output,$node_id,$module) {
    global $WAS_SCRIPT_NAME;
    $params = array(
        'job' => JOB_PAGEMANAGER,
        'task' => TASK_NODE_EDIT_CONTENT,
        'node' => strval($node_id)
        );
    $aparams = array('{NODE}' => strval($node_id));
    if (!empty($module['options'])) {
        $options = explode(',',$module['options']);
        if (!empty($options)) {
            foreach($options as $option) {
                $key = sprintf('{%s}',strtoupper($option));
                $title = t($option.'_title','m_confab',$aparams);
                $params[PARAM_SUBMENU_OPTION] = $option;
                $a_attr = array('title' => $title);
                $anchor = t($option.'_label','m_confab',$aparams);
                $aparams[$key] = html_a($WAS_SCRIPT_NAME,$params,$a_attr,$anchor);
            }
        }
    }
    $output->add_content('<h2>'.t('confab_content_header','m_confab',$aparams).'</h2>');
    $output->add_content(t('confab_content_explanation','m_confab',$aparams));
    $params = array(
        'job'    => JOB_PAGEMANAGER,
        'task'   => TASK_SAVE_CONTENT,
        'node'   => intval($node_id)
        );
    $href = href($WAS_SCRIPT_NAME, $params);
    $dialogdef = array('button_cancel' => dialog_buttondef(BUTTON_CANCEL));
    $output->add_content(dialog_quickform($href,$dialogdef));
    return TRUE;
} // confab_show_content()


/** dispatcher for save routines
 *
 * We use different subroutines for clarity.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which the content is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing and hence saving is not allowed
 * @param bool &$edit_again set to TRUE if we need to edit the content again, FALSE otherwise
 * @param string $option the selected submenu option or NULL if user used the 'Content' option to get here
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function confab_save(&$output, $area_id, $node_id, $module, $viewonly, &$edit_again, $option) {
    if ($option == 'configuration') {
        return confab_save_configuration($output,$node_id,$viewonly,$edit_again);
    } elseif ($option == 'moderation') {
        return confab_save_moderation($output,$viewonly,$edit_again);
    } elseif ($option == 'report') {
        return confab_save_report($output,$viewonly,$edit_again);
    } else { // save NULL or invalid option: indicate [Cancel]
        $edit_again = FALSE;
        return FALSE;
    }
    $edit_again = TRUE;
    return TRUE;
} // confab_save()


// ==================================================================
// =================== CONFAB MAIN CONFIGURATION ===================
// ==================================================================

/** show a simple edit screen for the confab configuration
 *
 *
 * @param object &$output collects the html output (if any)
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function confab_show_edit_configuration(&$output, $node_id, $module, $viewonly, $edit_again, $href) {
    // Cancel and Done from edit configuration should return to content screen
    if ((isset($_POST['button_cancel'])) || (isset($_POST['button_done']))) {
        return confab_show_content($output,$node_id,$module);
    }
    $output->add_content('<h2>'.t('confab_configuration_header','m_confab').'</h2>');
    $output->add_content(t('confab_configuration_explanation','m_confab'));
    $dialogdef = confab_get_dialogdef_configuration($output,$viewonly,$node_id);
    if ($edit_again) {
        confab_config_dialog_validate($dialogdef); // just for side effect of grabbing data from $_POST
    }
    $output->add_content(dialog_quickform($href,$dialogdef));
    return TRUE;
} // confab_show_edit_configuration()


/** validate and save the modified confab configuration linked to node $node_id
 *
 * this validates and saves the data that was submitted by the user.
 *
 * If validation fails, the flag $edit_again is set to TRUE and the return value
 * is FALSE.
 *
 * If the user has cancelled the operation, the flag $edit_again is set to FALSE
 * and the return value is also FALSE.
 *
 * If the modified data is stored successfully, the return value is TRUE.
 * The flag edit_again is set to TRUE if the user saved via the [Save] button,
 * and to FALSE if the user saved via the [Done] button.
 *
 * @param object &$output collects the html output (if any)
 * @param int $node_id the node to which the content is connected
 * @param bool $viewonly if TRUE, editing and hence saving is not allowed
 * @param bool &$edit_again set to TRUE if we need to edit the content again, FALSE otherwise
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 * @uses $USER
 */
function confab_save_configuration(&$output, $node_id, $viewonly, &$edit_again) {
    global $USER;

    // 1 -- bail out if cancelled or viewonly
    if ((isset($_POST['button_cancel'])) || ($viewonly)) {
        $edit_again = TRUE; // trick to let [Cancel] return to the Content screen
        return FALSE;
    }

    // 2 -- redo if invalid data was submitted
    $dialogdef = confab_get_dialogdef_configuration($output,$viewonly,$node_id);
    if (!confab_config_dialog_validate($dialogdef)) {
        // there were errors, show them to the user and do it again
        foreach($dialogdef as $k => $item) {
            if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                $output->add_message($item['error_messages']);
            }
        }
        $edit_again = TRUE;
        return FALSE;
    }

    // 3 -- actually save the new setting
    $now = strftime('%Y-%m-%d %T');
    $table = 'confabs';
    $fields = array(
        'header'               => $dialogdef['header']['value'],
        'introduction'         => $dialogdef['introduction']['value'],
        'footer'               => $dialogdef['footer']['value'],
        'passcode'             => $dialogdef['passcode']['value'],
        'max_participants'     => intval($dialogdef['max_participants']['value']),
        'max_tries'            => intval($dialogdef['max_tries']['value']),
        'refresh'              => intval($dialogdef['refresh']['value']),
        'timeout_idle'         => intval($dialogdef['timeout_idle']['value']),
        'timeout_conversation' => intval($dialogdef['timeout_conversation']['value']),
        'status'               => intval($dialogdef['status']['value']),
        'personal'             => intval($dialogdef['personal']['value']),
        'mtime'                => $now,
        'muser_id'             => $USER->user_id);
    $where = array('confab_id' => intval($node_id));
    if (db_update($table,$fields,$where) === FALSE) {
        logger(sprintf('%s(): error saving config value: %s',__FUNCTION__,db_errormessage()));
        $edit_again = TRUE;
        $retval = FALSE;
    } else {
        $edit_again = TRUE; // we _always_ want to return to an edit screen
        $retval = TRUE;
    }
    return $retval;
} // confab_save_configuration()


// ==================================================================
// ======================= CONFAB MODERATION =======================
// ==================================================================

/** present the user with one of two dialogs for moderation of confab posts at node $node_id
 *
 * the moderation functionality is split into two different dialogs.
 *
 * The Messagelist dialog looks like this:
 *
 * <pre>
 * [Refresh] [Intervene] [Cancel]
 * Message N
 * Message N-1
 * ...
 * Message 1
 * </pre>
 *
 * The Refresh button updates the list with messages.
 * The Intervene button leads to the second dialog (see below).
 * The Cancel button returns to the overarching Content dialog.
 *
 * The Intervention-dialog

 * Show a dialog to the user (the moderator) where she can break into
 * the conversation and send (personal) messages to participants
 * OR remove a participant OR remove _all_ participants from the
 * conversation (or end the conversation).
 *
 * The dialog looks a bit like this:
 *
 * <pre>
 * Action
 * ( ) ~Send a message
 * ( ) ~Remove from conversation
 * Participant
 * {listbox-with-participants}
 * ~Message:
 * [____________] 
 * [~OK] [~Cancel]
 * --------------
 * (list of _all_ messages in reverse order (latest at the top)
 * </pre>
 *
 * The OK button performs the selected action (~Send or ~Remove)
 * with the select Participant (or All participants). If the Message
 * input is not empty, that message is added to the conversation.
 *
 * The possible  scenarios are as follows:
 * - Send the message in 'Message:' to All (using [OK])
 * - Send the message in 'Message:' to the participant selected in {listbox} (using [OK])
 * - Remove the selected participant from the conversation (using [OK])
 * - Effectively end the conversation by removing All participants (using [OK])
 * - [Cancel] takes the moderator back to the Messagelist dialog screen
 *
 * The list of messages is ordered in reverse chronological order.
 * The moderator sees _all_ messages, including personal messages
 * The moderator is NOT visible for the regular participants; the name
 * does not show up in the userlist
 *
 * A moderator can send a personal message to a participant; the participant
 * cannot talk back other than by using a message to all.
 * The moderation only works on the current conversation.
 * Removing all participants from the conversation bumps the conversation too
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 * @todo this routine is too long
 */
function confab_show_edit_moderation(&$output,$area_id,$node_id,$module,$viewonly,$edit_again,$href) {
    global $USER;

    // 1 -- Cancel in Messagelist or NoMessages returns to Content
    if ((isset($_POST['button_cancel'])) &&  (!isset($_POST['message']))) {
        return confab_show_content($output,$node_id,$module);
    }
    // 2 -- get configuration in core, including last_conversation_id
    $confab_id = intval($node_id);
    $where = array('confab_id' => $confab_id);
    if (($config = db_select_single_record('confabs','*',$where)) === FALSE) {
        $output->add_message(t('error_retrieving_config','m_confab'));
        logger(sprintf("%s(): cannot retrieve confab configuration '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        return FALSE;
    }
    // 3 -- perhaps we need to bump the conversation (updates $config in core too)?
    if (confab_conversation_timeout($config) === FALSE) {
        $output->add_message(t('error_retrieving_config','m_confab'));
        return FALSE;
    }
    $conversation_id = intval($config['last_conversation_id']);

    // 4 -- any conversations at all?
    if ($config['last_participant_id'] == 0) {
        // No, leave user at NoMessages dialog with bare [Cancel] and [Refresh] button
        $output->add_content('<h2>'.t('moderation_header','m_confab').'</h2>');
        if ($conversation_id == 0) {
            $output->add_content('<p>'.t('moderation_no_conversations','m_confab').'</p>');
        } else {
            $output->add_content('<p>'.t('moderation_no_messages','m_confab').'</p>');
        }
        $refresh = t('moderation_button_refresh','m_confab');
        $dialogdef = array(
            dialog_csrftoken(),
            array(
                'type' => F_SUBMIT,
                'name' => 'button_refresh',
                'class' => 'button button_refresh',
                'value' => $refresh,
                'title' => t('hotkey_for_button','',array('{HOTKEY}' => accesskey_from_string($refresh)))),
            dialog_buttondef(BUTTON_CANCEL)
            );
        $output->add_content(dialog_quickform($href,$dialogdef));
        return TRUE;
    }
    
    // At this point we have to deal with the following possible buttons:
    // Refresh -> show messagelist dialog
    // Action -> show intervention dialog
    // OK -> validate and perhaps perform action + show message list dialog OR
    //       redo the action dialog if validation failed
    // Cancel (from action dialog) -> show messagelist dialog
    // Otherwise: messagelist dialog

    if ((isset($_POST['button_refresh'])) || (isset($_POST['button_cancel']))) {
        return confab_show_edit_moderation_list($output,$confab_id,$conversation_id,$viewonly,$href);
    }
    $participants = confab_get_participants($config);
    $dialogdef = confab_get_dialogdef_moderation($participants,$viewonly);
    if (isset($_POST['button_intervene'])) {
        $output->add_content('<h2>'.t('moderation_header','m_confab').'</h2>');
        $output->add_content(dialog_quickform($href,$dialogdef));
        confab_show_edit_moderation_messages($output,$confab_id,$conversation_id);
        return TRUE;
    } elseif (!isset($_POST['button_ok'])) {
        return confab_show_edit_moderation_list($output,$confab_id,$conversation_id,$viewonly,$href);
    }

    // 4A -- still here? Take care of OK-action (fetch and validate data from POST)
    if (dialog_validate($dialogdef) === FALSE) {
        // 4B -- oops. error in data, redo...
        foreach($dialogdef as $k => $item) {
            if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                $output->add_message($item['error_messages']);
            }
        }
        $output->add_content('<h2>'.t('moderation_header','m_confab').'</h2>');
        $output->add_content(dialog_quickform($href,$dialogdef));
        confab_show_edit_moderation_messages($output,$confab_id,$conversation_id);
        return TRUE;
    }

    // 5A -- data is valid. we may need to send a message
    $action = $dialogdef['action']['value'];
    $now = strftime('%Y-%m-%d %T');
    if (($participant_id = intval($dialogdef['participant']['value'])) == 0) {
        $personal = FALSE;
        $message_to = NULL;
    } else {
        $personal = TRUE;
        $message_to = $participants[$participant_id]['nickname'];
    }
    // 5B -- send message while everybody is still here
    $table = 'confab_messages';
    $message = trim($dialogdef['message']['value']);
    if (!empty($message)) {
        $fields = array('confab_id' => $confab_id,
                        'conversation_id' => $conversation_id,
                        'message_system' => FALSE,
                        'message_personal' => $personal,
                        'message_datim' => $now,
                        'message_from_id' => 0,
                        'message_from' => $USER->username,
                        'message_to' => $message_to,
                        'message' => $message,
                        'remote_addr' => confab_remote_addr()
                        );
        db_insert_into($table,$fields); // best effort
        $dialogdef['message']['value'] = ''; // dont leave msg lying round (better safe than sorry)
    }
    // 6 -- perhaps dismiss 1 or all users
    if ($action == CONFAB_ACTION_REMOVE) {
        if ($personal) {
            $ex_participants = array($participant_id => $participants[$participant_id]);
            $message_key = 'user_left_forced';
        } else {
            $ex_participants = $participants;
            $message_key = 'user_left_closed';
        }
        foreach($ex_participants as $id => $p) {
            $params = array('{NICKNAME}' => htmlspecialchars($p['nickname']),
                            '{FULL_NAME}' => htmlspecialchars($p['full_name'])
                            );
            $message = t($message_key,'m_confab',$params);
            $fields = array('confab_id' => $confab_id,
                            'conversation_id' => $conversation_id,
                            'message_system' => TRUE,
                            'message_personal' => FALSE,
                            'message_datim' => $now,
                            'message_from_id' => 0,
                            'message_from' => NULL,
                            'message_to' => NULL,
                            'message' => $message,
                            'remote_addr' => confab_remote_addr()
                            );
            db_insert_into($table,$fields); // best effort

            // 6A -- participant will be "marked for deletion"
            $p_fields = array('status' => CONFAB_USER_STATUS_LEAVING);
            $p_where = array('confab_participant_id' => intval($p['confab_participant_id']));
            if(db_update('confab_participants',$p_fields,$p_where) === FALSE) {
                logger(sprintf("%s(): cannot dismiss participant '%s' in confab '%d': %s",
                               __FUNCTION__,$p['nickname'],$confab_id,db_errormessage()));
            }
            $output->add_message($message);
        }
        // 6B -- if all users are gone, we start a new conversation
        if (!$personal) { // bump the conversation id too
            $table = 'confabs';
            $where = array('confab_id' => $confab_id);
            ++$conversation_id;
            $fields = array('last_participant_id' => 0,
                            'last_conversation_id' => $conversation_id);
            if (($retval = db_update($table,$fields,$where)) === FALSE) {
                logger(sprintf("%s(): cannot update confab '%d': %s",
                               __FUNCTION__,$confab_id,db_errormessage()));
            }
        }
    }
    return confab_show_edit_moderation_list($output,$confab_id,$conversation_id,$viewonly,$href);
} // confab_show_edit_moderation()


/** show a button bar and list of messages in the selected conversation
 *
 * @param object &$output collects the html output (if any)
 * @param int $confab_id 
 * @param int $conversation_id
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function confab_show_edit_moderation_list(&$output,$confab_id,$conversation_id,$viewonly,$href) {
    $refresh = t('moderation_button_refresh','m_confab');
    $intervene =  t('moderation_button_intervene','m_confab');
    $dialogdef = array(
        'csrftoken' => dialog_csrftoken(),
        'button_refresh' => array(
            'type' => F_SUBMIT,
            'name' => 'button_refresh',
            'class' => 'button button_refresh',
            'value' => $refresh,
            'title' => t('hotkey_for_button','',array('{HOTKEY}' => accesskey_from_string($refresh)))
            )
        );
    if (!$viewonly) {
        $dialogdef['button_intervene'] = array(
            'type' => F_SUBMIT,
            'name' => 'button_intervene',
            'class' => 'button button_intervene',
            'value' => $intervene,
            'title' => t('hotkey_for_button','',array('{HOTKEY}' => accesskey_from_string($intervene)))
            );
    }
    $dialogdef['button_cancel'] = dialog_buttondef(BUTTON_CANCEL);
    $output->add_content('<h2>'.t('moderation_header','m_confab').'</h2>');
    $output->add_content(dialog_quickform($href,$dialogdef));
    confab_show_edit_moderation_messages($output,$confab_id,$conversation_id);
    return TRUE;
} //  confab_show_edit_moderation_list()


/** create a list of all message in chronological order descending
 *
 * @param object &$output collects the html output (if any)
 * @param int $confab_id 
 * @param int $conversation_id
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function confab_show_edit_moderation_messages(&$output,$confab_id,$conversation_id) {
    // 1 -- prepare to show conversation in reverse order
    $table = 'confab_messages';
    $fields = '*';
    $where = array(
        'confab_id' => $confab_id, 
        'conversation_id' => $conversation_id
        );
    $order = array('confab_message_id DESC');
    if (($messages = db_select_all_records($table, $fields,$where,$order,'confab_message_id')) === FALSE) {
        $output->add_message(t('error_retrieving_conversation','m_confab'));
        logger(sprintf("%s(): cannot retrieve messages in conversation '%d' in confab '%d': %s",
                       __FUNCTION__,$conversation_id,$confab_id,db_errormessage()));
        return FALSE;
    }
    // 2 -- start output (if any)
    $output->add_content('<hr>');
    if (empty($messages)) {
        $output->add_content('<p>'.t('moderation_no_messages','m_confab').'</p>');
        return TRUE;
    }
    // 3 -- step through all available messages
    $oddeven = 'even';
    foreach ($messages as $id => $m) {
        $oddeven = ($oddeven == 'odd') ? 'even' : 'odd';
        $sys = (db_bool_is(TRUE,$m['message_system']));
        $personal = (db_bool_is(TRUE,$m['message_personal']));
        $params = array(
            '{DATIM}' => $m['message_datim'],
            '{TIME}' => substr($m['message_datim'],11,5),
            '{FROM}' => ($sys) ? t('visual_system','m_confab') : htmlspecialchars($m['message_from']),
            '{TO}' => ($personal) ? htmlspecialchars($m['message_to']) : '',
            '{MESSAGE}' => htmlspecialchars($m['message'])
            );
        if ($personal) {
            $output->add_content(sprintf("<div title=\"%s\" class=\"%s\">%s</div>\n",
                                         t('visual_from_to_title','m_confab',$params),
                                         $oddeven, t('visual_from_to','m_confab',$params)));
        } else {
            $output->add_content(sprintf("<div title=\"%s\" class=\"%s\">%s</div>\n",
                                         t('visual_from_title','m_confab',$params),
                                         $oddeven, t('visual_from','m_confab',$params)));
        }
    }
    $output->add_content('<br>');
    return TRUE;
} // confab_show_edit_moderation_messages()


/** a dummy routine that pretends to "save" the moderation selections
 *
 * this is just a dummy routine to tell the caller that we want
 * to return to {@link confab_show_edit_moderation()} which is where the
 * actual work is done. It is not even necessary to check for either
 * Cancel or OK because that too is done in confab_show_edit_report().
 * We do set the edit_again flag.
 *
 * @param object &$output collects the html output (if any)
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @return bool always FALSE to indicate we need to return to confab_show_edit_report(), always
 */
function  confab_save_moderation(&$output,$viewonly,&$edit_again) {
    $edit_again = TRUE; // stay here
    return FALSE;
} // confab_save_moderation()


// ==================================================================
// ========================= CONFAB REPORT =========================
// ==================================================================

/** present the user with a dialog for creating reports on confab conversations at node $node_id
 *
 * This is basically the following dialog followed by the results
 * of the constructed query based on the previous dialog contents:
 *
 * <pre>
 * Con~versation
 * [listbox with conversations]
 *
 * Fields:
 * [ ] ~Message ID
 * [ ] ~Author
 * [ ] ~Remote address
 * [ ] ~Date
 * [ ] ~Text
 * Messages from:
 * ( ) ~System
 * ( ) ~User
 * ( ) ~Both
 * Messages to:
 * ( ) ~Group
 * ( ) ~Individual
 * ( ) Bot~h
 * 
 * [~OK] [~Cancel]
 * </pre>
 *
 * The fields checkboxes determine which data is displayed.
 * The two radio buttons are added to the selection criteria (AND'ed together).
 * If the selections yield no data, message is displayed, otherwise
 * all checked fields of the selected records are displayed.
 *
 * @param object &$output collects the html output (if any)
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function confab_show_edit_report(&$output,$node_id,$module,$edit_again,$href) {

    // 1 -- Cancel from report returns to content screen
    if (isset($_POST['button_cancel'])) {
        return confab_show_content($output,$node_id,$module);
    }

    // 2 -- construct a dialog for selecting report options etc.
    $confab_id = intval($node_id);
    $conversations = confab_get_conversations($confab_id);
    if (empty($conversations)) {
        // tell user there's nothing to see, show minimalistic [Cancel] button as escape route
        $output->add_content('<h2>'.t('report_header','m_confab').'</h2>');
        $output->add_content('<p>'.t('report_no_conversations','m_confab').'</p>');
        $dialogdef = array(dialog_csrftoken(), dialog_buttondef(BUTTON_CANCEL));
        $output->add_content(dialog_quickform($href,$dialogdef));
        return TRUE;
    }

    // 2B -- still here? create the dialog definition & fetch data from POST
    $dialogdef = confab_get_dialogdef_report($conversations);
    if ($edit_again) {
        dialog_validate($dialogdef); // for the side effects: fetch data from POST
        foreach($dialogdef as $k => $item) {
            if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                $output->add_message($item['error_messages']);
            }
        }
    }
    $output->add_content('<h2>'.t('report_header','m_confab').'</h2>');
    $output->add_content(dialog_quickform($href,$dialogdef));

    // 2C -- just the dialog the first time around
    if (!$edit_again) {
        return TRUE;
    }

    // 3 -- actually construct a report for selected conversation either HTML or plain text
    // 3A -- fetch the interesting records from database
    $conversation_id = intval($dialogdef['conversation_id']['value']);
    $table = 'confab_messages';
    $fields = '*';
    $where = array(
        'confab_id' => $confab_id, 
        'conversation_id' => $conversation_id
        );
    if (($show_from = intval($dialogdef['from']['value'])) == 1) { // from system only
        $where['message_system'] = TRUE;
    } elseif ($show_from == 2) { // from user only
        $where['message_system'] = FALSE;
    } // else from both => don't care
    if (($show_to = intval($dialogdef['to']['value'])) == 1) { // public/group only
        $where['message_personal'] = FALSE;
    } elseif ($show_to == 2) { // to private/user only
        $where['message_personal'] = TRUE;
    } // else to both => don't care
    $order = array('confab_message_id');
    if (($records = db_select_all_records($table, $fields,$where,$order,'confab_message_id')) === FALSE) {
        $output->add_message(t('error_retrieving_conversation','m_confab'));
        logger(sprintf("%s(): cannot retrieve messages in confab '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        return FALSE;
    }
    // 3B -- the selection could be empty
    $output->add_content('<hr>');
    if (empty($records)) {
        $output->add_content('<p>'.t('report_no_messages','m_confab').'</p>');
        return TRUE;
    }

    // 4 -- still here? we need a real report in either html or as plain text download
    $report = array();
    $html = (isset($_POST['button_download'])) ? FALSE : TRUE;
    $newline = ($html) ? "<br>" : "\r\n";

    // 4A -- Construct a report header repeating the choices made
    $show_message_id = ($dialogdef['message_id']['value'] == 1) ? TRUE : FALSE;
    $show_ip_addr = ($dialogdef['ip_addr']['value'] == 1) ? TRUE : FALSE;
    $show_date = ($dialogdef['date']['value'] == 1) ? TRUE : FALSE;
    $show_author = ($dialogdef['author']['value'] == 1) ? TRUE : FALSE;
    $show_text = ($dialogdef['text']['value'] == 1) ? TRUE : FALSE;
    $from_system = t('report_system','m_confab');

    // Conversation #nn: first: yyyy-mm-dd hh:mm:ss last: yyyy-mm-dd hh:mm:ss messages: xxx
    $report[] = $dialogdef['conversation_id']['options'][$conversation_id]['title'].$newline;

    // Fields: Field1,Field2,...
    $f = array();
    if ($show_message_id) { $f[] = str_replace('~','',t('report_message_id_option','m_confab')); }
    if ($show_ip_addr) { $f[] = str_replace('~','',t('report_ip_addr_option','m_confab')); }
    if ($show_date) { $f[] = str_replace('~','',t('report_date_option','m_confab')); }
    if ($show_author) { $f[] = str_replace('~','',t('report_author_option','m_confab')); }
    if ($show_text) { $f[] = str_replace('~','',t('report_text_option','m_confab')); }
    $report[] = sprintf('%s %s%s',str_replace('~','',t('report_fields_label','m_confab')),
                                  implode(',',$f),
                                  $newline);
    // Messages from: Choice
    $value = $dialogdef['from']['value'];
    $f = array();
    if (($value == '1') || ($value == '3')) {
        $f[] = str_replace('~','',$dialogdef['from']['options']['1']['option']);
    }
    if (($value == '2') || ($value == '3')) {
        $f[] = str_replace('~','',$dialogdef['from']['options']['2']['option']);
    }
    $report[] = sprintf('%s %s%s',str_replace('~','',t('report_from_label','m_confab')),
                                  implode(',',$f),
                                  $newline);
    // Messages to: Choice
    $value = $dialogdef['to']['value'];
    $f = array();
    if (($value == '1') || ($value == '3')) {
        $f[] = str_replace('~','',$dialogdef['to']['options']['1']['option']);
    }
    if (($value == '2') || ($value == '3')) {
        $f[] = str_replace('~','',$dialogdef['to']['options']['2']['option']);
    }
    $report[] = sprintf('%s %s%s',str_replace('~','',t('report_to_label','m_confab')),
                                  implode(',',$f),
                                  $newline);
    // Separator line
    $report[] = $newline;

    // 4B -- Show all checked fields from records that were selected
    foreach($records as $id => $record) {
        $system = (db_bool_is(TRUE,$record['message_system']));
        $personal = (db_bool_is(TRUE,$record['message_personal']));
        $message_from = ($html) ? htmlspecialchars($record['message_from']) : $record['message_from'];
        $message_to = ($html) ? htmlspecialchars($record['message_to']) : $record['message_to'];
        $message =  ($html) ? htmlspecialchars($record['message']) : $record['message'];
        $ip_addr = ($html) ? htmlspecialchars($record['remote_addr']) : $record['remote_addr'];
        $aparams = array('{MESSAGE_ID}' => $id,
                         '{FROM}' => ($system) ? $from_system : $message_from,
                         '{TO}' => ($personal) ? $message_to : '',
                         '{MESSAGE}' => $message,
                         '{REMOTE_ADDR}' => $ip_addr,
                         '{DATIM}' => $record['message_datim'],
                         '{TIME}' => substr($record['message_datim'],11),
                         '{DATIM_SHORT}' => short_datim($record['message_datim']));
        $f = array();
        if ($show_message_id) {
            $f[] = t('report_id','m_confab',$aparams);
        }
        if ($show_ip_addr) {
            $f[] = t('report_ip_addr','m_confab',$aparams);
        }
        if ($show_date) {
            $f[] = t('report_date','m_confab',$aparams);
        }
        if ($show_author) {
            if ($personal) {
                $f[] = t('report_author_to','m_confab',$aparams);
            } else {
                $f[] = t('report_author','m_confab',$aparams);
            }
        }
        if ($show_text) {
            $f[] = t('report_text','m_confab',$aparams);
        }
        if (!empty($f)) {
            $report[] = implode(' ',$f).$newline;
        }
    }
    if ($html) {
        $report[] = $newline;
        $output->add_content($report);
    } else {
        $output->set_suppress_output(TRUE);
        $filename = sprintf('%s-%d-%d.txt',t('report_filename_download','m_confab'),
                                           $confab_id,
                                           $conversation_id);
        header('Content-Type: text/plain; charset=UTF-8');
        header(sprintf('Content-Disposition: attachment; filename="%s"',$filename));
        echo implode('',$report);
    }
    return TRUE;
} // confab_show_edit_report()


/** a dummy routine that pretends to "save" the report selections
 *
 * this is just a dummy routine to tell the caller that we want
 * to return to {@link confab_show_edit_report()} which is where the
 * actual work is done. It is not even necessary to check for either
 * Cancel or OK because that too is done in confab_show_edit_report().
 * We do set the edit_again flag.
 *
 * @param object &$output collects the html output (if any)
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @return bool always FALSE to indicate we need to return to confab_show_edit_report(), always
 */
function  confab_save_report(&$output,$viewonly,&$edit_again) {
    $edit_again = TRUE; // stay here
    return FALSE;
} // confab_save_report()


// ==================================================================
// =========================== WORKHORSES ===========================
// ==================================================================

/** construct a dialog definition for the main confab configuration
 *
 * @param object &$output collects the html output (if any)
 * @param int $viewonly if TRUE the Save button is not displayed and config values cannot be changed
 * @param int $node_id identifies the current confab
 * @return array dialog definition
 */
function confab_get_dialogdef_configuration(&$output, $viewonly, $node_id) {
    //
    // 1 -- generic configuration parameters
    //
    $table = 'confabs';
    $fields = array('header','introduction','footer','passcode',
                    'max_participants','max_tries','refresh',
                    'timeout_idle','timeout_conversation','status',
                    'personal'
                   );
    $where = array('confab_id' => intval($node_id));
    if (($record = db_select_single_record($table, $fields, $where)) === FALSE) {
        logger(sprintf('%s(): error retrieving configuration: %s',__FUNCTION__,db_errormessage()));
        $output->add_message(t('error_retrieving_config','admin'));
        $record = array(
            'header' => '',
            'introduction' => '',
            'footer' => '',
            'passcode' => '',
            'max_participants' => 26,
            'max_tries' => 10,
            'refresh' => 6,
            'timeout_idle' => 300,          // 5 minutes after last message from the participant
            'timeout_conversation' => 3600, // 1 hour after last message from any participant
            'status' => CONFAB_STATUS_CLOSED,
            'personal' => CONFAB_PERSONAL_NO
            );
    }
    $status_options = array(
        strval(CONFAB_STATUS_OPENED) => array(
            'option' => t('status_opened_label','m_confab'),
            'title' => t('status_opened_title', 'm_confab')
            ),
        strval(CONFAB_STATUS_REGISTERED) => array(
            'option' => t('status_registered_label','m_confab'),
            'title' => t('status_registered_title', 'm_confab')
            ),
        strval(CONFAB_STATUS_CLOSED) => array(
            'option' => t('status_closed_label','m_confab'),
            'title' => t('status_closed_title', 'm_confab')
            )
        );
    $personal_options = array(
        strval(CONFAB_PERSONAL_YES) => array(
            'option' => t('personal_yes_label','m_confab'),
            'title' => t('personal_yes_title', 'm_confab')
            ),
        strval(CONFAB_PERSONAL_REGISTERED) => array(
            'option' => t('personal_registered_label','m_confab'),
            'title' => t('personal_registered_title', 'm_confab')
            ),
        strval(CONFAB_PERSONAL_NO) => array(
            'option' => t('personal_no_label','m_confab'),
            'title' => t('personal_no_title', 'm_confab')
            )
        );
    $dialogdef = array(
        'status' => array(
            'type' => F_RADIO,
            'name' => 'status',
            'value' => strval($record['status']),
            'old_value' => strval($record['status']),
            'options' => $status_options,
            'viewonly' => $viewonly,
            'title' => t('status_title','m_confab'),
            'label' => t('status_label','m_confab')
            ),
        'personal' => array(
            'type' => F_RADIO,
            'name' => 'personal',
            'value' => strval($record['personal']),
            'old_value' => strval($record['personal']),
            'options' => $personal_options,
            'viewonly' => $viewonly,
            'title' => t('personal_title','m_confab'),
            'label' => t('personal_label','m_confab')
            ),
        'passcode' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'passcode',
            'minlength' => 0,
            'maxlength' => 80,
            'columns' => 40,
            'label' => t('passcode_label','m_confab'),
            'title' => t('passcode_title','m_confab'),
            'viewonly' => $viewonly,
            'value' => $record['passcode'],
            'old_value' => $record['passcode']
            ),
        'header' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'header',
            'minlength' => 0,
            'maxlength' => 240,
            'columns' => 40,
            'label' => t('header_label','m_confab'),
            'title' => t('header_title','m_confab'),
            'viewonly' => $viewonly,
            'value' => $record['header'],
            'old_value' => $record['header']
            ),
        'introduction' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'introduction',
            'minlength' => 0,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 50,
            'rows' => 10,
            'label' => t('introduction_label','m_confab'),
            'title' => t('introduction_title','m_confab'),
            'viewonly' => $viewonly,
            'value' => $record['introduction'],
            'old_value' => $record['introduction']
            ),
        'footer' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'footer',
            'minlength' => 0,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 50,
            'rows' => 10,
            'label' => t('footer_label','m_confab'),
            'title' => t('footer_title','m_confab'),
            'viewonly' => $viewonly,
            'value' => $record['footer'],
            'old_value' => $record['footer']
            ),
        'max_participants' => array(
            'type' => F_INTEGER,
            'name' => 'max_participants',
            'columns' => 5,
            'maxlength' => 10,
            'minvalue' => 2, // less is not very useful is it?
            'maxvalue' => CONFAB_MAX_PARTICIPANTS_LIMIT,
            'viewonly' => $viewonly,
            'value' => intval($record['max_participants']),
            'old_value' => intval($record['max_participants']),
            'title' => t('max_participants_title','m_confab'),
            'label' => t('max_participants_label','m_confab')
            ),
        'max_tries' => array(
            'type' => F_INTEGER,
            'name' => 'max_tries',
            'columns' => 5,
            'maxlength' => 10,
            'minvalue' => 3, // arbitrary: at least three tries
            'maxvalue' => CONFAB_MAX_TRIES_LIMIT,
            'viewonly' => $viewonly,
            'value' => intval($record['max_tries']),
            'old_value' => intval($record['max_tries']),
            'title' => t('max_tries_title','m_confab'),
            'label' => t('max_tries_label','m_confab')
            ),
        'refresh' => array(
            'type' => F_INTEGER,
            'name' => 'refresh',
            'columns' => 5,
            'maxlength' => 10,
            'minvalue' => CONFAB_REFRESH_LIMIT,
            'maxvalue' => 600, // arbitrary: 10 minutes
            'viewonly' => $viewonly,
            'value' => intval($record['refresh']),
            'old_value' => intval($record['refresh']),
            'title' => t('refresh_title','m_confab'),
            'label' => t('refresh_label','m_confab')
            ),
        'timeout_idle' => array(
            'type' => F_INTEGER,
            'name' => 'timeout_idle',
            'columns' => 5,
            'maxlength' => 10,
            'minvalue' => 60,  // arbitrary: 1 minute
            'maxvalue' => 3600, // arbitrary: 1 hour
            'viewonly' => $viewonly,
            'value' => intval($record['timeout_idle']),
            'old_value' => intval($record['timeout_idle']),
            'title' => t('timeout_idle_title','m_confab'),
            'label' => t('timeout_idle_label','m_confab')
            ),
        'timeout_conversation' => array(
            'type' => F_INTEGER,
            'name' => 'timeout_conversation',
            'columns' => 5,
            'maxlength' => 10,
            'minvalue' => 600,  // arbitrary: 10 minutes
            'maxvalue' => 36000, // arbitrary: 10 hours
            'viewonly' => $viewonly,
            'value' => intval($record['timeout_conversation']),
            'old_value' => intval($record['timeout_conversation']),
            'title' => t('timeout_conversation_title','m_confab'),
            'label' => t('timeout_conversation_label','m_confab')
            ),
        'csrftoken' => dialog_csrftoken()
        );
    //
    // 2 -- maybe add a save button but always a cancel button
    //
    if (!$viewonly) {
        $dialogdef['button_save'] = dialog_buttondef(BUTTON_SAVE);
        $dialogdef['button_done'] = dialog_buttondef(BUTTON_DONE);
    }
    $dialogdef['button_cancel'] = dialog_buttondef(BUTTON_CANCEL);
    return $dialogdef;
} // confab_get_dialogdef_configuration()


/** validation of configuration dialog + rewriting passcode (no trailing/leading spaces)
 *
 * @param array &$dialogdef the dialog to check
 * @return TRUE if valid, FALSE otherwise (+messages in dialogdef)
 */
function confab_config_dialog_validate(&$dialogdef) {
    $retval = dialog_validate($dialogdef);
    $dialogdef['passcode']['value'] = trim($dialogdef['passcode']['value']);
    return $retval;
} // confab_config_dialog_validate()


/** construct the moderation dialog
 *
 * @param array &$participants contains all participants in this conversation
 * @param bool $viewonly if TRUE, editing and hence saving is not allowed
 * @return array dialog definition
 */
function confab_get_dialogdef_moderation(&$participants, $viewonly) {
    $action_options = array(
        CONFAB_ACTION_MESSAGE => array(
            'option' => t('moderation_action_send_label','m_confab'),
            'title' => t('moderation_action_send_title', 'm_confab')
            ),
        CONFAB_ACTION_REMOVE => array(
            'option' => t('moderation_action_remove_label','m_confab'),
            'title' => t('moderation_action_remove_title', 'm_confab')
            )
        );
    $participant_options = array(0 =>  t('message_to_all','m_confab'));
    foreach ($participants as $id => $p) {
        $participant_options[$id] = sprintf('%s (%s)',
                                            htmlspecialchars($p['nickname']),
                                            htmlspecialchars($p['full_name']));
        }
    $dialogdef = array(
        'action' => array(
            'type' => F_RADIO,
            'name' => 'action',
            'value' => CONFAB_ACTION_MESSAGE,
            'options' => $action_options,
            'viewonly' => $viewonly,
            'label' => t('moderation_action_label','m_confab'),
            'title' => t('moderation_action_title','m_confab')
            ),
        'participant' => array(
            'type' => F_LISTBOX,
            'name' => 'participant',
            'viewonly' => $viewonly,
            'label' => t('moderation_participant_label','m_confab'),
            'title' => t('moderation_participant_title','m_confab'),
            'options' => $participant_options,
            'value' => '0'
            ),
        'message' =>  array(
            'type' => F_ALPHANUMERIC,
            'name' => 'message',
            'viewonly' => $viewonly,
            'label' => t('moderation_message_label','m_confab'),
            'title' => t('moderation_message_title','m_confab'),
            'value' => '',
            'class' => 'wide'
            ),
        'csrftoken' => dialog_csrftoken()
        );
    //
    // 2 -- maybe add a save button but always a cancel button
    //
    if (!$viewonly) {
        $dialogdef['button_ok'] = dialog_buttondef(BUTTON_OK);
    }
    $dialogdef['button_cancel'] = dialog_buttondef(BUTTON_CANCEL);
    return $dialogdef;
} // confab_get_dialogdef_moderation()


/** construct the dialog for the reporting tool
 *
 * @param array &$conversations ordered list of conversations holds data for list box
 * @return array ready to use dialog definition
 */
function confab_get_dialogdef_report(&$conversations) {
    $download = t('report_button_download','m_confab'); // button
    $value = 0;
    $conversation_options = array();
    foreach($conversations as $id => $conversation) {
        $value = max($value,$id);
        $aparam = array(
            '{CONVERSATION_ID}' => $id,
            '{DATIM_FIRST}' => $conversation['d0'],
            '{DATIM_FIRST_SHORT}' => short_datim($conversation['d0']),
            '{DATIM_LAST}' => $conversation['d1'],
            '{DATIM_LAST_SHORT}' => short_datim($conversation['d1']),
            '{MESSAGES}' => $conversation['n']
            );
        $conversation_options[$id] = array(
            'option' => t('report_selection_option_label','m_confab',$aparam),
            'title' => t('report_selection_option_title','m_confab',$aparam)
            );
    }

    $from_options = array(
        '1' => array(
            'option' => t('report_from_system_label','m_confab'),
            'title' => t('report_from_system_title', 'm_confab')
            ),
        '2' => array(
            'option' => t('report_from_user_label','m_confab'),
            'title' => t('report_from_user_title', 'm_confab')
            ),
        '3' => array(
            'option' => t('report_from_both_label','m_confab'),
            'title' => t('report_from_both_title', 'm_confab')
            )
        );
    $to_options = array(
        '1' => array(
            'option' => t('report_to_group_label','m_confab'),
            'title' => t('report_to_group_title', 'm_confab')
            ),
        '2' => array(
            'option' => t('report_to_user_label','m_confab'),
            'title' => t('report_to_user_title', 'm_confab')
            ),
        '3' => array(
            'option' => t('report_to_both_label','m_confab'),
            'title' => t('report_to_both_title', 'm_confab')
            )
        );
    $dialogdef = array(
        'conversation_id' => array(
            'type' => F_LISTBOX,
            'name' => 'conversation_id',
            'options' => $conversation_options,
            'label' => t('report_selection_label','m_confab'),
            'title' => t('report_selection_title','m_confab'),
            'value' => $value,
            ),
        'message_id' => array(
            'type' => F_CHECKBOX,
            'name' => 'message_id',
            'options' => array('1' => array('option' => t('report_message_id_option','m_confab'),
                                            'title' => t('report_message_id_title','m_confab'))),
            'label' => t('report_fields_label','m_confab'), // only in the first checkbox...
            'title' => t('report_fields_title','m_confab'), // ...for layout reasons
            'value' => '0'
            ),
        'ip_addr' => array(
            'type' => F_CHECKBOX,
            'name' => 'ip_addr',
            'options' => array('1' => array('option' => t('report_ip_addr_option','m_confab'),
                                            'title' => t('report_ip_addr_title','m_confab'))),
            'value' => '0'
            ),
        'date' => array(
            'type' => F_CHECKBOX,
            'name' => 'date',
            'options' => array('1' => array('option' => t('report_date_option','m_confab'),
                                            'title' => t('report_date_title','m_confab'))),
            'value' => '0'
            ),
        'author' => array(
            'type' => F_CHECKBOX,
            'name' => 'author',
            'options' => array('1' => array('option' => t('report_author_option','m_confab'),
                                            'title' => t('report_author_title','m_confab'))),
            'value' => '1'
            ),
        'text' => array(
            'type' => F_CHECKBOX,
            'name' => 'text',
            'options' => array('1' => array('option' => t('report_text_option','m_confab'),
                                            'title' => t('report_text_title','m_confab'))),
            'value' => '1'
            ),
        'from' => array(
            'type' => F_RADIO,
            'name' => 'from',
            'value' => '3',
            'options' => $from_options,
            'title' => t('report_from_title','m_confab'),
            'label' => t('report_from_label','m_confab')
            ),
        'to' => array(
            'type' => F_RADIO,
            'name' => 'to',
            'value' => '3',
            'options' => $to_options,
            'title' => t('report_to_title','m_confab'),
            'label' => t('report_to_label','m_confab')
            ),
        'csrftoken' => dialog_csrftoken(),
        'button_ok' => dialog_buttondef(BUTTON_OK),
        'button_cancel' => dialog_buttondef(BUTTON_CANCEL),
        'button_download' => array(
            'type' => F_SUBMIT,
            'name' => 'button_download',
            'class' => 'button button_download',
            'value' => $download,
            'title' => t('hotkey_for_button','',array('{HOTKEY}' => accesskey_from_string($download)))
            )
        );
    return $dialogdef;
} // confab_get_dialogdef_report()


/** retrieve all conversations in reverse chronological order
 *
 * if something goes wrong, we return an empty array
 * but we do write to the logger
 *
 * @param int $confab_id
 * @return array with conversations keyed by conversation_id
 */
function confab_get_conversations($confab_id) {
    global $DB;
    // get conversations data in array keyed by conversation_id
    $sql = sprintf("SELECT conversation_id AS id, ".
                          "MIN(message_datim) AS d0, ".
                          "MAX(message_datim) AS d1, ".
                          "COUNT(confab_message_id) AS n ".
                   "FROM %sconfab_messages ".
                   "WHERE confab_id = %d ".
                   "GROUP BY conversation_id ".
                   "ORDER BY conversation_id DESC",
                   $DB->prefix, $confab_id);
    if (($DBResult = $DB->query($sql)) === FALSE) {
        logger(sprintf("%s(): error retrieving conversations in confab '%d': %s",
                       __FUNCTION__,$confab_id,db_errormessage()));
        $conversations = array();
    } else {
        $conversations = $DBResult->fetch_all_assoc('id');
        $DBResult->close();
    }
    return $conversations;
} // confab_get_conversations()

?>