<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/confab/confab_common.php - code shared between admin and view
 *
 * This file defines various constants and subroutines used from both
 * {@link confab_admin.php} and {@link confab_view.php}.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_confab
 * @version $Id: confab_common.php,v 1.8 2016/03/23 09:28:21 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }

define('CONFAB_STATUS_CLOSED',    0); // visitors can not use this confab at this time
define('CONFAB_STATUS_REGISTERED',1); // only registered users can use the confab
define('CONFAB_STATUS_OPENED',    2); // anyone can use the confab iff they present the correct passcode

define('CONFAB_PERSONAL_NO',        0); // participants can not send personal messages at all
define('CONFAB_PERSONAL_REGISTERED',1); // only registered users can send personal messages
define('CONFAB_PERSONAL_YES',       2); // participants can send a personal message to other participants

define('CONFAB_USER_STATUS_NONE',   0); // participant is logged out
define('CONFAB_USER_STATUS_JOINED', 1); // participant has joined the conversation
define('CONFAB_USER_STATUS_LEAVING',2); // participant is currently leaving the conversation

define('CONFAB_BRAILLE',  0); // Braille interface
define('CONFAB_VISUAL',   1); // Visual interface (standard)
define('CONFAB_VISUAL_BW',2); // Visual interface LowVision (High Contrast black+white)
define('CONFAB_VISUAL_RB',3); // Visual interface Red Grey Blue (Primary colours)
define('CONFAB_VISUAL_BY',4); // Visual interface Mondrian (Black and Yellow)


define('CONFAB_MAX_CONVERSATION_LIMIT',86400);// Absolute MAXIMUM # of seconds after conversation start
define('CONFAB_MAX_PARTICIPANTS_LIMIT',100);  // Absolute MAXIMUM # of participants per conversation
define('CONFAB_MAX_TRIES_LIMIT',25);          // Absolute MAXIMUM # of tries entering correct passcode
define('CONFAB_REFRESH_LIMIT',2);             // Absolute MINIMUM # of seconds between auto-refresh calls
define('CONFAB_ANONYMOUS_POSTFIX','*');       // Postfix distinguishes Anonymous & Registered Participants

define('CONFAB_ACTION_MESSAGE',1); // Send a moderator message to 1 or all participants
define('CONFAB_ACTION_REMOVE', 2); // Remove 1 or all participants from the conversation


// ==================================================================
// ==================== CONFAB UTILITY ROUTINES ====================
// ==================================================================


/** retrieve all participants that joined the current conversation
 *
 * @param array &$config holds the confab configuration
 * @return array participants records from database or array() if no participants
 */
function confab_get_participants(&$config) {
    $confab_id = $config['confab_id'];
    $conversation_id = $config['last_conversation_id'];
    $table = 'confab_participants';
    $fields = '*';
    $where = array(
        'confab_id' => $confab_id,
        'conversation_id' => $conversation_id,
        'status' => CONFAB_USER_STATUS_JOINED);
    $order = array('username','username_ext','participant_id');
    $keyfield = 'participant_id';
    if (($participants = db_select_all_records($table, $fields,$where,$order,$keyfield)) === FALSE) {
        logger(sprintf("%s(): cannot retrieve confab participants in confab '%d': %s",
                       __FUNCTION__,$confab_id,db_errormessage()));
        $participants = array();
    }
    return $participants;
} // confab_get_participants()

/** shorthand for remote IP-address
 *
 * @return string remote IP address
 */
function confab_remote_addr() {
    return (isset($_SERVER['REMOTE_ADDR'])) ? $_SERVER['REMOTE_ADDR'] : ''; 
} // confab_remote_addr()


/** check conversations for timeout and hard limit and maybe create a new participant_id
 *
 * this is a dual-purpose routine. The first task is to 
 * check the validity of the current conversation. If the
 * conversation started longer than CONFAB_MAX_CONVERSATION_LIMIT (24h)
 * ago, we bluntly end the conversation by incrementing the
 * last_conversation_id in the confab configuration.
 * The same effect happens when the latest contribution to the
 * conversation was more than $timeout_conversation (1h) ago.
 * 
 * If the conversation is no longer current, the last_conversation_id
 * is incremented and also the last_participant_id is reset to 0.
 * This information will be recorded in the database and also
 * updated in the $config array (which was provided by reference,
 * and not by value).
 *
 * However... in order to save a trip to the database,
 * we can also provide a new participant_id in the same go.
 * This happens when $bump_participant is TRUE. The caller
 * can be sure that in that case the value $config['last_participant_id']
 * can be used by the caller (provided the routine returns
 * TRUE, indicating success).
 *
 * If somehow something goes wrong, e.g. a database error, the
 * $config record is untouched and the routine returns FALSE.
 *
 * This routine should be called periodically, for instance
 * - when a new participants wants to join a conversation
 * - when a participant contributes to the conversation
 * This keeps the conversation_id in line and makes sure that
 * an ongoing conversation is stopped eventually.
 * 
 * @param array &$config holds the confab configuration record
 * @param bool $bump_participant if TRUE a new participant_id is made available
 * @return bool TRUE on success, FALSE on error
 */
function confab_conversation_timeout(&$config,$bump_participant=FALSE) {
    // 0 -- start with current values
    $confab_id = intval($config['confab_id']);
    $conversation_id = intval($config['last_conversation_id']);
    $participant_id = intval($config['last_participant_id']);
    // 1 -- calculate oldest/youngest message datim in conversation
    $table = 'confab_messages';
    $fields = 'MAX(message_datim) AS latest, MIN(message_datim) AS first';
    $where = array(
        'confab_id' => $confab_id,
        'conversation_id' => $conversation_id
        );
    if (($record = db_select_single_record($table,$fields,$where)) === FALSE) {
        logger(sprintf("%s(): cannot check conversation %d in confab '%d': %s",
                       __FUNCTION__,$conversation_id,$confab_id,db_errormessage()));
        return FALSE;
    }
    $where = array( // already prepare for confabs update; keep old values for step 4 below
        'confab_id' => $confab_id,
        'last_conversation_id' => $conversation_id,
        'last_participant_id' => $participant_id
        );
    // 2 -- is the conversation still within limits?
    $threshold_latest = strftime("%Y-%m-%d %T",time() - intval($config['timeout_conversation']));
    $threshold_first =  strftime("%Y-%m-%d %T",time() - CONFAB_MAX_CONVERSATION_LIMIT);
    if (((!is_null($record['latest'])) && ($record['latest'] < $threshold_latest)) ||
        ((!is_null($record['first'])) && ($record['first'] < $threshold_first))) {
        logger(sprintf("%s(): bump conversation %d in confab '%d' ".
                       "first: %s <-> %s latest: %s <-> %s",
                       __FUNCTION__,$conversation_id,$confab_id,
                       $record['first'],$threshold_first,
                       $record['latest'],$threshold_latest),WLOG_DEBUG);
        // 2A -- this conversation must stop
        $table = 'confab_messages';
        $fields = array(
            'confab_id' => $confab_id,
            'conversation_id' => $conversation_id,
            'message_system' => TRUE,
            'message_personal' => FALSE,
            'message_datim' => $record['latest'],
            'message_from_id' => 0, // system
            'message_from' => NULL, // system
            'message_to' => NULL, // to all
            'message' => t('conversation_will_close','m_confab'),
            'remote_addr' => confab_remote_addr()
            );
        db_insert_into($table,$fields); // best effort
        // 2B -- and a new one can commence
        ++$conversation_id;
        $participant_id = 0;
        $dirty = TRUE;
    } else {
        $dirty = FALSE;
    }
    // 3 -- (side effect) does the caller wants to assign a new participant_id?
    if ($bump_participant) {
        if ($conversation_id == 0) { // special case: this is the 1st participant in the 1st conversation
            $conversation_id = 1;
            $participant_id = 1;
        } else {
            ++$participant_id;
        }
        $dirty = TRUE;
    }
    // 4 -- update existing config record with 'old' $where (mitigate race condition)
    $retval = TRUE; // assume success
    if ($dirty) {
        $table = 'confabs';
        $fields = array('last_conversation_id'=>$conversation_id, 'last_participant_id'=>$participant_id);
        if (($retval = db_update($table,$fields,$where)) === FALSE) {
            logger(sprintf("%s(): cannot update conversation => %d participant => %d in confab '%d': %s",
                           __FUNCTION__,$conversation_id,$participant_id, $confab_id,db_errormessage()));
        } elseif ($retval != 1) {
            logger(sprintf("%s(): failed conversation => %d participant => %d: %d confab-records updated",
                           __FUNCTION__,$conversation_id,$participant_id, $confab_id,$retval));
            $retval = FALSE;
        } else {
            $config['last_conversation_id'] = $conversation_id;
            $config['last_participant_id'] = $participant_id;
        }
    }
    return $retval;
} // confab_conversation_timeout()

?>