<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/confab/confab_view.php - interface to the view-part of the confab module
 *
 * This file defines the interface with the confab-module for viewing content.
 * The interface consists of this function:
 *
 * <code>
 * confab_view(&$output,$area_id,$node_id,$module)
 * </code>
 *
 * This function is called from /index.php when the node to display is connected
 * to this module.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_confab
 * @version $Id: confab_view.php,v 1.24 2016/03/23 09:28:21 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }

define('CONFAB_DIRNAME',dirname(__FILE__));
require_once(CONFAB_DIRNAME.'/confab_common.php'); // common code + constants

define('CONFAB_REFERENCE',sha1(__FILE__.':'.__LINE__));
require_once($CFG->progdir.'/lib/tokenlib.php');


/** display the content of the confab linked to node $node_id
 *
 * This routine is called whenever a visitor visits node $node_id.
 * It depends on the joined-status what is actually sent to the browser.
 * If the participant already joined the conversation, indicated by
 * presenting a valid cookie, the request if further handled in
 * the corresponding protocol dispatcher.
 *
 * If not, this must be another participant attempting to join
 * the conversation. If the Confab is not closed AND the maximum
 * number of simultaneous participants is not yet reached, we
 * enter the procedure to let the visitor present her credentials
 * (passcode) etc. This is done in a separate routine
 * {@link confab_join()}. This keeps this routine short
 * and readable.
 *
 * @param object &$theme collects the (html) output
 * @param int $area_id identifies the area where $node_id lives
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function confab_view(&$theme,$area_id,$node_id,$module) {
    global $CFG;
    $theme->add_stylesheet(sprintf('program/modules/%s/%s.css',$module['name'],$module['name']));
    $confab_id = intval($node_id);

    // 1 -- get configuration in core
    $where = array('confab_id' => $confab_id);
    if (($config = db_select_single_record('confabs','*',$where)) === FALSE) {
        $theme->add_message(t('error_retrieving_config','m_confab'));
        logger(sprintf("%s(): cannot retrieve confab configuration '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        return FALSE;
    }
    $config['cookie_name'] = $CFG->session_name.'CONFAB'; // remember the cookie name for lateron

    // 2 -- someone we know already?
    if (isset($_COOKIE[$config['cookie_name']])) {
        $cookie = get_cookie_string($config['cookie_name'],'');
        $table = 'confab_participants';
        $fields = '*';
        $where = array('confab_participant_key' => $cookie, 'confab_id' => $confab_id);
        if (($participant = db_select_single_record($table,$fields,$where)) !== FALSE) {
            // I know you from before. Continue with either BRAILLE or VISUAL protocol
            if ($participant['interface_id'] == CONFAB_BRAILLE) {
                return confab_braille_dispatcher($theme,$config,$participant);
            } else {
                return confab_visual_dispatcher($theme,$config,$participant);
            }
        }
    }

    // 3 -- is the Confab opened and is there room at all?
    if (intval($config['status']) == CONFAB_STATUS_CLOSED) {
        $reason = 'confab_view_is_closed';
    } elseif ($config['max_participants'] <= confab_view_calc_participants($config)) {
        $reason = 'confab_view_unavailable';
    } else {
        $reason = '';
    }
    if (!empty($reason)) {
        if (!empty($config['header'])) {
            $theme->add_content('<h2 class="confab_header">'.$config['header'].'</h2>');
        }
        $theme->add_content('<div class="confab_introduction">');
        $theme->add_message(t($reason,'m_confab'));
        $theme->add_content(t($reason,'m_confab'));
        $theme->add_content('</div>');
        return TRUE;
    }
    // At this point we are dealing with a still unknown participant / visitor.
    // Depending on the configuration we need to show a login dialog of sorts
    // OR process an incoming form with the information from the login dialog.
    // The former is relatively easy: just create a dialogdef and display it.
    // The latter is more complicated because we must validate the submitted data
    // and maybe setup a cookie, etc. etc. Here we go.

    // 4 -- proceed with the join dialog for this visitor
    return confab_join($theme, $config);
} // confab_view()


// ==================================================================
// ================= JOIN / SAVE / LEAVE PROCEDURES =================
// ==================================================================

/** handle the complete procedure to join the conversation
 *
 * this routine handles joining the conversation. We use the token lib,
 * to force participants to first load an empty join dialog before
 * entering credentials and gaining access.
 *
 * the container $data holds the number of attempts to specifiy the correct
 * passcode. If this exceeds the maximum, we leave the user with an error
 * in an otherwise empty screen. They can try again after that, but they have
 * to start with a fresh token.
 *
 * @param object &$theme collects the (html) output
 * @param array &$config holds the confab configuration including the conversation_id
 * @return bool FALSE on error, TRUE on success + output written via $theme
 */
function confab_join(&$theme, &$config) {
    $confab_id = intval($config['confab_id']);

    // 1 -- prepare for handling of existing token
    $t0 = $t1 = 0;
    $ip_addr = '';
    $data = FALSE;
    $token_id = FALSE;
    if (isset($_POST['token'])) { // lookup valid UTF8 key (or fail with substitute U+FFFD instead)
        $token_key = (utf8_validate($_POST['token'])) ? magic_unquote($_POST['token']) : "\xEF\xBF\xBD";
        $token_id = token_lookup(CONFAB_REFERENCE,$token_key,$t0,$t1,$ip_addr,$data);
    }

    // 2 -- handle case of expired token first
    $now = time();
    if (($token_id !== FALSE) && ($t1 < $now)) { // token expired
	$theme->add_message(t('error_token_expired','m_confab'));
        token_destroy($token_id);
        $token_id = FALSE;
    }

    // 3 -- check all data in dialog now we have a (still valid) valid token
    if ($token_id !== FALSE) {
        $dialogdef = confab_join_get_dialogdef($config, $token_key);
        if (isset($_POST['button_ok'])) {
            if (!confab_join_dialog_validate($dialogdef,$config)) {
                foreach($dialogdef as $k => $item) {
                    if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                        $theme->add_message($item['error_messages']);
                    }
                }
                if (!isset($data['tries'])) {
                    $data['tries'] = 0;
                }
                if (++$data['tries'] > $config['max_tries']) {
                    $msg = t('error_too_many_tries','m_confab');
                    $theme->add_message($msg);
                    $theme->add_popup_top($msg);
                    logger(sprintf('%s(): too many join attempts from %s',__FUNCTION__,$ip_addr));
                    token_destroy($token_id);
                    $token_id = FALSE;
                    return FALSE;
                }
                if (!token_store($token_id, $data)) {
                    $theme->add_message(t('error_storing_data','m_confab'));
                    logger(sprintf('%s(): token store error in page %d: %s',
                                   __FUNCTION__,$node_id,db_errormessage()));
                    return FALSE;
                }
                confab_join_dialog_show($theme, $config, $dialogdef);
            } elseif ($now < $t0) { // the window of opportunity is still closed; go back to form
                $msg = t('error_too_fast','m_confab');
                $theme->add_message($msg);
                $theme->add_popup_top($msg);
                confab_join_dialog_show($theme, $config, $dialogdef);
                logger(sprintf('%s(): reply too fast (%ds) from %s',__FUNCTION__,$t0-$now,$ip_addr));
            } else {
                // success with the login hurdle; clean up token and join the conversation
                token_destroy($token_id);
                $token_id = FALSE;
                return confab_join_conversation($theme, $config, $dialogdef);
            }
        } else {
            // initiate a new round
            token_destroy($token_id);
            $token_id = FALSE;
        }
    }

    // 4 -- Start with a clean slate
    if ($token_id === FALSE) {
        $token_key = '';
        if (($token_id = token_create(CONFAB_REFERENCE,$token_key,1)) === FALSE) { // 1s delay
            $msg = t('error_creating_token','m_confab',array('{NODE}' => strval($confab_id)));
            $theme->add_message($msg);
            $theme->add_content($msg);
            return FALSE;
        }
        $dialogdef = confab_join_get_dialogdef($config, $token_key);
        confab_join_dialog_show($theme, $config, $dialogdef);
    }
    return TRUE;
} // confab_join()


/** output the join dialog including the form and header/introduction/footer
 *
 * this routine can yield a dialog that includes fields for both the user's username
 * and password (as used in a system login). If this is the case (indicated by
 * the presence of the field 'login_password'), we add the paramter 'login=1' to
 * our own URL in order to make a stop at the system login procedure (in
 * {@link main_index()}) before proceeding with our own join dialog (killing
 * two birds with one stone).
 *
 * @param object &$theme collects the (html) output
 * @param array &$config holds the confab configuration including the conversation_id
 * @param array &$dialogdef defines the dialog and will hold POSTed values
 * @return bool FALSE on error, TRUE on success + output written via $theme
 */
function confab_join_dialog_show(&$theme, &$config, &$dialogdef) {
    $parameters = array('c' => strval(time()));
    if (isset($dialogdef['login_password'])) {
        $parameters['login'] = '1';
    }
    $href = was_node_url($theme->node_record,$parameters);
    if (!empty($config['header'])) {
        $theme->add_content('<h2 class="confab_header">'.$config['header'].'</h2>');
    }
    if (!empty($config['introduction'])) {
        $theme->add_content('<div class="confab_introduction">');
        $theme->add_content($config['introduction']);
        $theme->add_content('</div>');
    }

    $theme->add_content(dialog_quickform($href,$dialogdef));

    if (!empty($config['footer'])) {
        $theme->add_content('<div class="confab_footer">');
        $theme->add_content($config['footer']);
        $theme->add_content('</div>');
    }
} // confab_join_dialog_show()


/** validate the data entered by the user/visitor
 *
 * Other than the standard validation we check the passcode
 * and for anonymous visitors we make sure the username is unique
 * within the conversation.
 *
 * Note 1:
 * We distinguish between Anonymous users and Registered users
 * by adding a postfix to the nickname of the anonymous user. This postfix
 * makes it possible to choose a name like an existing name, eg. 'hparkh'
 * without making it impossible for the real 'hparkh' to login using her
 * real username. As a result, there could be two participants, one named
 * 'hparkh' and one named (say) 'hparkh*'. For the name lookup below these
 * names are different.
 *
 * Note 2:
 * There is a possible race condition here if two participants use the same
 * nickname at the same time. However, the database will enforce that only
 * one of the two will 'win' and be entered into the confab_participants table.
 * The other one will get a key violation error and, sadly, has to try again.
 * IMHO an acceptable risk but YMMV.
 *
 * @param array &$dialogdef defines the dialog and will hold POSTed values
 * @param array &$config holds the Confab configuration
 * @return bool TRUE if valid, else FALSE + messages added to dialogdef
 * @todo We may have to enforce usernames cannot end in an asterisk in Account Manager.
 */
function confab_join_dialog_validate(&$dialogdef,&$config) {
    global $USER;
    $retval = TRUE; // assume success

    // 1 -- check the usual suspects (side-effect: fetch POSTed values in dialogdef)
    if (!dialog_validate($dialogdef)) {
        $retval = FALSE;
    }

    // 2 -- the passcode should match
    if ($dialogdef['passcode']['value'] != $config['passcode']) {
        ++$dialogdef['passcode']['errors'];
        $params = array('{FIELD}' => str_replace('~','',$dialogdef['passcode']['label']));
        $dialogdef['passcode']['error_messages'][] = t('validate_invalid','',$params);
        $retval = FALSE;
    }
    
    // 3 -- Extra tests for anonymous participants
    $key = (isset($dialogdef['login_username'])) ? 'login_username' : 'nickname';
    $username = (isset($dialogdef[$key])) ? $dialogdef[$key]['value'] : '';
    if ((!$USER->is_logged_in) || ($USER->username != $username)) {
        // 3A - the username provided must not contain '(', '*' or ')'
        if ((strpos($username,'(') !== FALSE) ||
            (strpos($username,CONFAB_ANONYMOUS_POSTFIX) !== FALSE) ||
            (strpos($username,')') !== FALSE)) {
            ++$dialogdef[$key]['errors'];
            $params = array('{FIELD}' => str_replace('~','',$dialogdef[$key]['label']),
                            '{CHARACTERS}' => '( '.CONFAB_ANONYMOUS_POSTFIX.' )');
            $dialogdef[$key]['error_messages'][] = t('error_characters_not_allowed','m_confab',$params);
            $retval = FALSE;
        }
        // 3B - the nickname must be unique within the conversation
        $nickname = $username.CONFAB_ANONYMOUS_POSTFIX; // nickname = username + postfix
        $table = 'confab_participants';
        $fields = 'COUNT(confab_participant_id) AS n';
        $where = array(
            'confab_id' => intval($config['confab_id']),
            'conversation_id' => intval($config['last_conversation_id']),
            'username' => $nickname);
        if ((($record = db_select_single_record($table,$fields,$where)) === FALSE) ||
            (intval($record['n']) > 0)) {
            ++$dialogdef[$key]['errors'];
            $params = array('{FIELD}' => str_replace('~','',$dialogdef[$key]['label']));
            $dialogdef[$key]['error_messages'][] = t('validate_invalid','',$params);
            $retval = FALSE;
        }
    }
    return $retval;
} // confab_join_dialog_validate()


/** construct a dialog definition for the join screen
 *
 * this defines the join form. There are several flavours, depending on
 * the confab configuration and the fact that the user is (already) logged in or not.
 *
 * The basic join screen looks a bit like this:
 * <pre>
 * Nickname: 
 * __________
 * Passcode:
 * __________
 * Interface:
 * <listbox>
 * [OK]
 * </pre>
 *
 * If a user is currently logged in, we can already enter the username.
 * If above that the confab is of type registered users only, we make the
 * name readonly, since the user is not allowed to join anonymously.
 * (She must present the passcode though).
 *
 * If the user is not logged in, we cannot already pre-load the username.
 * If above that the confab is of type registered users only, we add the
 * username and password fields from the system login so we can login and 
 * join in a single go.
 * The reason to take this approach is that we don't want to bother the user
 * with two screens: one with username/password and another with username
 * passcode. IF the user provides the correct password AND passcode, we're in
 * business. If she misses one of the two, we either 'hang' in the system
 * login, OR in a loop for requesting a passcode from an already logged in user.
 *
 * The difference between the dialogs can be detected from the existence of
 * either the 'username' OR the 'login_username' entry in $dialogdef. 
 * Alternatively, the input field 'login_password' is only added if
 * a username/password authentication is required too.
 *
 * Note:
 * The length of the nickname/username in the Confab tables is set at 64.
 * A system username is limited to a length of 60. This allows for at most
 * 4 characters to distinguish Anonymous and Registered participants in the
 * participants and messages tables. Feels like a kludge. Sorry.
 *
 * @param array $config confab configuration record
 * @param string $token_key
 * @return array datadefinition
 */
function confab_join_get_dialogdef($config,$token_key) {
    global $USER;

    if ($USER->is_logged_in) {
        $username = $USER->username;
        $registered = TRUE;
        $viewonly = ($config['status'] == CONFAB_STATUS_REGISTERED) ? TRUE : FALSE;
        $show_system_login = FALSE; // User is already logged in
    } else {
        $username = '';
        $registered = FALSE;
        $viewonly = FALSE;
        $show_system_login = ($config['status'] == CONFAB_STATUS_REGISTERED) ? TRUE : FALSE;
    }
    $options = array(
        CONFAB_BRAILLE => array(
            'option' => t('interface_braille_option','m_confab'),
            'title' => t('interface_braille_title','m_confab')
            ),
        CONFAB_VISUAL => array(
            'option' => t('interface_visual_option','m_confab'),
            'title' => t('interface_visual_title','m_confab'),
            ),
        CONFAB_VISUAL_BW => array(
            'option' => t('interface_visual_bw_option','m_confab'),
            'title' => t('interface_visual_bw_title','m_confab'),
            ),
        CONFAB_VISUAL_RB => array(
            'option' => t('interface_visual_rb_option','m_confab'),
            'title' => t('interface_visual_rb_title','m_confab'),
            ),
        CONFAB_VISUAL_BY => array(
            'option' => t('interface_visual_by_option','m_confab'),
            'title' => t('interface_visual_by_title','m_confab'),
            )
        );
    $tabindex = 0;
    $dialogdef = array();
    $dialogdef['token'] = array(
        'type' => F_ALPHANUMERIC,
        'name' => 'token',
        'hidden' => TRUE,
        'value' => $token_key
        );
    if ($show_system_login) {
        $dialogdef['login_username'] = array(
            'type' => F_ALPHANUMERIC,
            'name' => 'login_username',
            'minlength' => 1,
            'maxlength' => 60, // see note about length 60 vs length 64 above
            'columns' => 25,
            'label' => t('username','loginlib'),
            'value' => $username,
            'viewonly' => $viewonly,
            'tabindex' => ++$tabindex
            );
        $dialogdef['login_password'] = array(
            'type' => F_PASSWORD,
            'name' => 'login_password',
            'minlength' => 0,
            'maxlength' => 255,
            'columns' => 25,
            'label' => t('password','loginlib'),
            'value' => '',
            'tabindex' => ++$tabindex,
            'autocomplete' => 'off'
            );
    } else {
        $dialogdef['nickname'] = array(
            'type' => F_ALPHANUMERIC,
            'name' => 'nickname',
            'minlength' => 1,
            'maxlength' => 60, // see note about length 60 vs length 64 above
            'columns' => 25,
            'label' => t('join_nickname_label','m_confab'),
            'value' => $username,
            'viewonly' => $viewonly,
            'tabindex' => ++$tabindex
            );
    }
    $dialogdef['passcode'] = array(
        'type' => F_PASSWORD,
        'name' => 'passcode',
        'minlength' => 0,
        'maxlength' => 80,
        'columns' => 25,
        'label' => t('join_passcode_label','m_confab'),
        'value' => '',
        'tabindex' => ++$tabindex,
        'autocomplete' => 'off'
        );
    $dialogdef['interface'] = array(
        'type' => F_LISTBOX,
        'name' => 'interface',
        'value' => CONFAB_VISUAL,
        'options' => $options,
        'label' => t('interface_label','m_confab'),
        'tabindex' => ++$tabindex
        );
    $dialogdef['button_ok'] = dialog_buttondef(BUTTON_OK);
    $dialogdef['button_ok']['tabindex'] = ++$tabindex;
    return $dialogdef;
} // confab_join_get_dialogdef()


/** do housekeeping and setup the opening screen with requested interface
 *
 * this routine is called after the user authenticated successfully.
 * The credentials are in $dialogdef and we are quite confident that the
 * data was valid only moments ago and probably still is.
 *
 * we now need to add the collected data to the confab_participants table.
 * as a side effect we may have to bump the last_conversation_id ie if the
 * current conversation has ended (after a time out), we start a new one.
 * In that case we are the first/only participant so any name clash we may
 * have (due to race condition) is mitigated.
 *
 * @param object &$theme collects the (html) output
 * @param array &$config holds the confab configuration including the conversation_id
 * @param array &$dialogdef defines the dialog and will hold POSTed values
 * @return bool FALSE on error, TRUE on success + output written directly to output ('raw').
 */
function confab_join_conversation(&$theme, &$config, &$dialogdef) {
    global $WAS_SCRIPT_NAME,$USER;

    $confab_id = intval($config['confab_id']);
    $interface_id = intval($dialogdef['interface']['value']);
    $now = strftime("%Y-%m-%d %T");

    // 1 -- which conversation exactly?
    if (confab_conversation_timeout($config,TRUE) === FALSE) {
        logger(sprintf("%s(): cannot join conversation in confab '%d'",__FUNCTION__,$confab_id));
        $theme->add_message(t('error_joining_conversation','m_confab'));
        return FALSE;
    }
    $conversation_id = intval($config['last_conversation_id']);
    $participant_id = intval($config['last_participant_id']);

    // 2 -- who am i?
    $key = (isset($dialogdef['login_username'])) ? 'login_username' : 'nickname';
    $username = (isset($dialogdef[$key])) ? $dialogdef[$key]['value'] : '';
    if ((!$USER->is_logged_in) || ($USER->username != $username)) {
        // 2A -- an anonymous user. Indicate that this anonymous username is taken
        //       by adding an asterisk. The nickname is the same, ie there can only
        //       be a single anonymous user with this username/nickname.
        $full_name = t('anonymous','m_confab');
        $username .= CONFAB_ANONYMOUS_POSTFIX;
        $username_ext = 0;
        $nickname = $username;
    } else {
        // 2B -- a registered user. Because registered users cannot change their
        //       own name (they would become anonymous users) we have to resort
        //       to another trick, ie. we must allow multiple variants of the same
        //       name. We do that via a numeric postfix in parens: ie. the first
        //       instance is 'hparkh' but the second one is 'hparkh(2)'. The value
        //       2 is stored in username_ext in the table..
        $table = 'confab_participants';
        $fields = 'MAX(username_ext) AS m';
        $where = array('confab_id' => $confab_id,
                       'conversation_id' => $conversation_id,
                       'username' => $username
                       );
        if (($record = db_select_single_record($table,$fields,$where)) === FALSE) {
            logger(sprintf("%s(): cannot calculate username_ext in confab '%d': %s",
                           __FUNCTION__,$confab_id,db_errormessage()));
            $theme->add_message(t('error_joining_conversation','m_confab'));
            return FALSE;
        }
        $nickname = $username;
        $username_ext = 1;
        if (!is_null($record['m'])) {
            $username_ext += intval($record['m']);
            $nickname .= sprintf('(%d)',$username_ext);
        }
        $full_name = $USER->full_name;
    }

    // 3A -- announce our presence to the others (if any)
    $table = 'confab_messages';
    $fields = array(
        'confab_id' => $confab_id,
        'conversation_id' => $conversation_id,
        'message_system' => TRUE,
        'message_personal' => FALSE,
        'message_datim' => $now,
        'message_from_id' => $participant_id,
        'message_from' => $nickname,
        'message_to' => NULL,
        'message' => t('join_full','m_confab',array('{FROM}' => $nickname, '{FROM_NAME}' => $full_name)),
        'remote_addr' => confab_remote_addr()
        );
    if (($last_message_id = db_insert_into_and_get_id($table,$fields,'confab_message_id')) === FALSE) {
        logger(sprintf("%s(): cannot add join message for conversation %d in confab '%d': %s",
                       __FUNCTION__,$conversation_id,$confab_id,db_errormessage()));
        $theme->add_message(t('error_joining_conversation','m_confab'));
        return FALSE;
    }

    // 3B -- and make sure their userlist is updated eventually by setting the flag
    $table = 'confab_participants';
    $fields = array('show_list' => TRUE);
    $where = array('confab_id' => $confab_id, 'conversation_id' => $conversation_id);
    db_update($table,$fields,$where); // best effort

    // 4 -- set a cookie if not already done; that is our ticket to the confab
    if (isset($_COOKIE[$config['cookie_name']])) {
        $cookie = get_cookie_string($config['cookie_name'],'');
    } else {
        // calc a unique value (see also {@link tokenlib.php} for a comparable construct)
        $cookie = sprintf('%08x%s', $last_message_id ^ 558781249 , md5(quasi_random_string(32)));
        setcookie($config['cookie_name'],$cookie,0,dirname($WAS_SCRIPT_NAME));
    }

    // 5 -- make an entry in participants table (at last...)
    $table = 'confab_participants';
    $fields = array(
        'confab_id' => $confab_id,
        'conversation_id' => $conversation_id,
        'participant_id' => $participant_id,
        'interlocutor_id' => 0, // default is to talk to all
        'confab_participant_key' => $cookie,
        'username' => $username,
        'username_ext' => $username_ext,
        'nickname' => $nickname,
        'full_name' => $full_name,
        'last_message_id' => $last_message_id,
        'last_seen' => $now,
        'last_posted' => $now,
        'interface_id' => $interface_id,
        'status' => CONFAB_USER_STATUS_JOINED,
        'show_list' => FALSE,
        'is_active' => TRUE
        );
    if (($id = db_insert_into_and_get_id($table,$fields,'confab_participant_id')) === FALSE) {
        logger(sprintf("%s(): cannot add participant %d to conversation %d in confab '%d': %s",
                       __FUNCTION__,$participant_id,$conversation_id,$confab_id,db_errormessage()));
        $theme->add_message(t('error_joining_conversation','m_confab'));
        return FALSE;
    }

    // 6 -- do a little housekeeping: weed out old stuff
    confab_view_delete_stale_participants($config);

    // 7 -- now generate the initial/openingscreen for the selected interface.
    $fields['confab_participant_id'] = $id; // Pretend we just retrieved $fields from database, with $id
    if ($fields['interface_id'] == CONFAB_BRAILLE) {
        --$fields['last_message_id']; // make sure we see our own entrance in Braille interface
        $retval = confab_braille_show_main($theme,$config,$fields);
    } else {
        $fields['last_message_id'] = 0; // make sure we see ALL messages in Visual interface
        $retval = confab_visual_show_main($theme,$config,$fields);
    }
    return $retval;
} // confab_join_conversation()


/** add non-empty message to database
 *
 * Note that there is a chance that we post this message to conversation N
 * (as recorded in our participant-record) whereas the confab has moved on
 * to conversation N+1. Therefore we do hold on to the conversation_id in
 * the participants record rather than the last_conversation_id in the config
 * record. Once this message is processed, the participant will find out that
 * the conversation has ended and that she lands on the login screen...
 *
 * @param object &$theme collects the output
 * @param array &$config holds the confab configuration record
 * @param array &$participant holds the participant record
 * @return bool TRUE on success, FALSE otherwise
 */
function confab_save_message(&$theme,&$config,&$participant) {
    $confab_id = intval($config['confab_id']);
    $conversation_id=intval($participant['conversation_id']);
    $interlocutor_id = 0;
    $message_personal = FALSE;
    $message_to = NULL;    
    $message = '';
    $now = strftime("%Y-%m-%d %T");

    // 1 -- validate message, confirm that it is not empty
    if ((isset($_POST['message'])) && (utf8_validate($_POST['message']))) {
        $message = trim(magic_unquote($_POST['message']));
    }

    // 2 -- maybe check for person-to-person message
    if (!empty($message)) {
        if (($config['personal'] == CONFAB_PERSONAL_YES) ||
            (($config['personal'] == CONFAB_PERSONAL_REGISTERED) && ($participant['username_ext'] > 0))) {
            if ((isset($_POST['message_to'])) && (utf8_validate($_POST['message_to']))) {
                $interlocutor_id = intval(magic_unquote($_POST['message_to']));
            }
            if ($interlocutor_id != 0) {
                // 2A -- fetch the interlocutor record
                $table = 'confab_participants';
                $fields = '*';
                $where = array('confab_id' => $confab_id,
                               'conversation_id' => $conversation_id,
                               'participant_id' => $interlocutor_id);
                // 2B -- and retrieve the nickname of the interlocutor
                if (($record = db_select_single_record($table,$fields,$where)) !== FALSE) {
                    $message_personal = TRUE;
                    $message_to = $record['nickname'];
                }
            }
        }
    }

    // 3 -- add the non-empty message to the database
    if (!empty($message)) {
        $table = 'confab_messages';
        $fields = array('confab_id' => $confab_id,
                        'conversation_id' => $conversation_id,
                        'message_system' => FALSE,
                        'message_personal' => $message_personal,
                        'message_datim' => $now,
                        'message_from_id' => intval($participant['participant_id']),
                        'message_from' => $participant['nickname'],
                        'message_to' => $message_to,
                        'message' => $message,
                        'remote_addr' => confab_remote_addr()
                        );
        db_insert_into($table,$fields);
    }

    // 4 -- update our own record, even when message was empty (trick to "keep alive")
    $table = 'confab_participants';
    $fields = array('interlocutor_id' => $interlocutor_id,
                    'last_posted' => $now,
                    'last_seen' => $now,
                    'is_active' => TRUE
                    );
    $where = array('confab_participant_id' => intval($participant['confab_participant_id']));
    db_update($table,$fields,$where);

    // 5 -- maybe trigger a state change from idle to active
    if (db_bool_is(FALSE,$participant['is_active'])) {
        $table = 'confab_participants';
        $fields = array('show_list' => TRUE);
        $where = array('confab_id' => $confab_id,
                       'conversation_id' => $conversation_id);
        db_update($table,$fields,$where);
        $participant['is_active'] = SQL_TRUE;
        $participant['show_list'] = SQL_TRUE;
    }
    // 6 we're done here
    return TRUE;
} // confab_save_message()


/** end the conversation for $participant (at their own request or forced)
 *
 * this is the opposite of the {@link confab_join_conversation()} 
 * routine: we destroy the cookie and remove the participants record etc.
 * The $reason gives the user feedback as to what just happened: if
 * the system has closed down the conversation due to a time-out
 * or when the moderator has closed the Confab altogether, the $reason
 * explains. If no specific reason is given, we show a default message.
 *
 * @param object &$theme collects the output
 * @param array &$config holds the confab configuration record
 * @param array &$participant holds the participant record
 * @param string $reason indicates why we are leaving (or we were made to)
 * @return bool TRUE on success, FALSE otherwise (it DOES return)
 */
function confab_leave_conversation(&$theme,&$config,&$participant,$reason='') {
    global $WAS_SCRIPT_NAME;

    $confab_id = intval($participant['confab_id']);
    $conversation_id = intval($participant['conversation_id']);
    $now =  strftime("%Y-%m-%d %T");
    $nickname = $participant['nickname'];
    $full_name = $participant['full_name'];

    // 1A -- announce we are leaving to the others (if any)
    $table = 'confab_messages';
    $fields = array(
        'confab_id' => $confab_id,
        'conversation_id' => $conversation_id,
        'message_system' => TRUE,
        'message_personal' => FALSE,
        'message_datim' => $now,
        'message_from_id' => intval($participant['participant_id']),
        'message_from' => $nickname,
        'message_to' => NULL,
        'message' => t('leave_full','m_confab',array('{FROM}' => $nickname, '{FROM_NAME}' => $full_name)),
        'remote_addr' => confab_remote_addr()
        );
    if ((db_insert_into($table,$fields)) === FALSE) {
        logger(sprintf("%s(): cannot add leave message for conversation %d in confab '%d': %s",
                       __FUNCTION__,$conversation_id,$confab_id,db_errormessage()));
        $theme->add_message(t('error_leaving_conversation','m_confab'));
    }
    // 1B -- and flag userlist update
    $table = 'confab_participants';
    $fields = array('show_list' => TRUE);
    $where = array('confab_id' => $confab_id, 'conversation_id' => $conversation_id);
    db_update($table,$fields,$where); // best effort

    // 2 -- remove ourselves from the conversation
    $table = 'confab_participants';
    $where = array('confab_participant_id' => intval($participant['confab_participant_id']));
    if (db_delete($table,$where) === FALSE) {
        logger(sprintf("%s(): cannot remove partipant record from conversation %d in confab '%d': %s",
                       __FUNCTION__,$conversation_id,$confab_id,db_errormessage()));
        $theme->add_message(t('error_leaving_conversation','m_confab'));
    }

    // 3 -- and try to bump the conversation if we were the last one standing
    $table = 'confab_participants';
    $fields = 'COUNT(confab_participant_id) AS n';
    $where = array(
        'confab_id' => $confab_id,
        'conversation_id' => $conversation_id,
        'status' => CONFAB_USER_STATUS_JOINED);
    if ((($record = db_select_single_record($table,$fields,$where)) !== FALSE) && 
        (intval($record['n']) <= 0)) { // this was the last one; bump conversation
        $table = 'confabs';
        $fields = array('last_conversation_id' => $conversation_id + 1, 'last_participant_id' => 0);
        $where = array(
            'confab_id' => $confab_id,
            'last_conversation_id' => $conversation_id,
            'last_participant_id' => intval($config['last_participant_id'])
            );
        if (($retval = db_update($table,$fields,$where)) === FALSE) {
            logger(sprintf("%s(): cannot bump conversation => %d in confab '%d': %s",
                           __FUNCTION__,$conversation_id + 1,$confab_id,db_errormessage()));
        } elseif ($retval != 1) {
            logger(sprintf("%s(): bump conversation in confab %d failed: %d confab-records updated",
                           __FUNCTION__,$confab_id,$retval));
        } else {
            $config['last_conversation_id'] = $conversation_id + 1;
            $config['last_participant_id'] = 0;
        }
    }

    // 4 -- remove a cookie if we are the last confab using it
    $cookie = get_cookie_string($config['cookie_name'],'');
    $table = 'confab_participants';
    $fields = 'COUNT(confab_participant_id) AS n';
    $where = array('confab_participant_key' => $cookie);
    if ((($record = db_select_single_record($table,$fields,$where)) !== FALSE) && 
        (intval($record['n']) <= 0)) { // this was the last one; remove cookie
        setcookie($config['cookie_name'],'',0,dirname($WAS_SCRIPT_NAME));
    }

    // 5 -- and end at the join screen with a fresh token
    $params = array('{NICKNAME}' => htmlspecialchars($participant['nickname']),
                    '{FULL_NAME}' => htmlspecialchars($participant['full_name'])
                    );
    $theme->add_message(t((empty($reason)) ? 'user_left' : $reason,'m_confab',$params));
    return confab_join($theme, $config);
} // confab_leave_conversation()


// ==================================================================
// ======================== BRAILLE PROTOCOL ========================
// ==================================================================

/** dispatcher for braille interface
 *
 * this routine decides what needs to be done (and does it).
 * As a side effect, we check various time-out conditions.
 *
 * 1. we ALWAYS check for timeout of (other) participants, this may
 *    or may not change the status of others from active to idle
 * 2. we ALWAYS check if we are kicked out, e.g. because the
 *    conversation timed out, we were directed to leave via a 
 *    status change (maybe initiated by the moderator) or the
 *    Confab was suddenly closed (by a moderator). In this case
 *    we silently leave the conversation and perform the Quit
 *    action which will return us to the regular $theme.
 * 3. we SOMETIMES check for conversations that time out, i.e.
 *    whenever a contribution is posted (to the current conversation),
 *    we check and maybe increment the conversation_id. This means
 *    that we effectively kick ourselves out the next time we come here.
 *
 * Note that most of the subroutines called handle their own I/O
 * and set the $silent_mode of the $theme to TRUE, effectively
 * suppressing the regular theme. This leaves a blank screen to 'play'
 * with. The notable exception is {@link confab_leave_conversation()}:
 * that routine DOES rely on $theme to build a regular page
 * including navigation, etc.
 *
 * @param object &$theme collects the output
 * @param array &$config holds the confab configuration record
 * @param array &$participant holds the participant record
 * @return bool TRUE on succes, FALSE on error or sometimes never at all
 */
function confab_braille_dispatcher(&$theme,&$config,&$participant) {
    confab_view_participants_timeout($config); // best effort
    if ($participant['status'] == CONFAB_USER_STATUS_LEAVING) {
        $threshold = strftime("%Y-%m-%d %T",time() - intval($config['timeout_conversation']));
        $last_seen = $participant['last_seen'];
        $reason = ($last_seen < $threshold) ? 'user_left_timeout' : 'user_left_forced';
        $retval = confab_leave_conversation($theme,$config,$participant,$reason);
    } elseif ($config['last_conversation_id'] != $participant['conversation_id']) {
        $reason = 'user_left_timeout';
        $retval = confab_leave_conversation($theme,$config,$participant,$reason);
    } elseif (($config['status'] == CONFAB_STATUS_CLOSED) ||
              (($config['status'] == CONFAB_STATUS_REGISTERED) && ($participant['username_ext'] <= 0))) {
        $reason = 'user_left_closed';
        $retval = confab_leave_conversation($theme,$config,$participant,$reason);
    } elseif (isset($_POST['button_ok'])) {
        $retval = confab_braille_show_main($theme,$config,$participant);
    } elseif (isset($_POST['button_talk'])) {
        $retval = confab_braille_show_talk($theme,$config,$participant);
    } elseif (isset($_POST['button_cancel'])) {
        $retval = confab_braille_show_main($theme,$config,$participant);
    } elseif (isset($_POST['button_send'])) {
        confab_conversation_timeout($config); // best effort
        confab_save_message($theme,$config,$participant); // best effort
        $retval = confab_braille_show_main($theme,$config,$participant);
    } elseif (isset($_POST['button_all'])) {
        $retval = confab_braille_show_all($theme,$config,$participant);
    } elseif (isset($_POST['button_users'])) {
        $retval = confab_braille_show_users($theme,$config,$participant);
    } elseif (isset($_POST['button_quit'])) {
        $retval = confab_leave_conversation($theme,$config,$participant);
    } else {
        $retval = confab_braille_show_main($theme,$config,$participant);
    }
    return $retval;
} // confab_braille_dispatcher()


/** construct the main/messages screen for the braille interface
 *
 * this routine checks if there are any new messages for $participant.
 * If so, these are displayed and a basic bare bones button bar is 
 * displayed. Here the user can acknowledge the messages (by pressing
 * [OK]), acknowledge the messages and immediately start typing a
 * new message (by pressing [Talk]) or move to the display of all
 * messages that were exchanged since the conversation starten (by
 * pressing [All]).
 *
 * If there are no new messages a single 'Talk' button (with an asterisk
 * rather than the word 'Talk') is displayed. This page automatically
 * refreshes itself, making sure that the server is regularly polled for
 * any new messages.
 *
 * @param object &$theme collects the output
 * @param array &$config holds the confab configuration record
 * @param array &$participant holds the participant record
 * @return bool TRUE on success, FALSE otherwise
 */
function confab_braille_show_main(&$theme,&$config,&$participant) {

    // 1 -- prepare for main/messages screen
    $http_equiv = confab_braille_page_headers(); // headers are now sent; add http_equiv lateron
    $bsss = $theme->get_bazaar_style_style();
    $fields = array('last_seen' => strftime("%Y-%m-%d %T"));
    $content = '';
    $nickname = $participant['nickname'];

    // 2 -- any messages to show?
    $messages = confab_view_get_messages($config,$nickname,$participant['last_message_id']);
    if (sizeof($messages) > 0) {
        // 2A -- yes, show messages and no refresh
        $head = "<head>\n".$http_equiv."<title>&nbsp;</title>\n".$bsss."</head>\n";
        foreach ($messages as $id => $message) {
            $usertime = $message['message_datim'];
            $sys = (db_bool_is(TRUE,$message['message_system']));
            $personal = (db_bool_is(TRUE,$message['message_personal']));
            $nickfrom = ($sys) ? t('braille_system','m_confab'):htmlspecialchars($message['message_from']);
            $nickto = ($personal) ? htmlspecialchars($message['message_to']):t('braille_all','m_confab'); 
            $params = array('{FROM}' => $nickfrom, '{TO}' => $nickto);
            $fromto = t(($personal) ? 'braille_from_to' : 'braille_from','m_confab',$params);
            $msg = htmlspecialchars($message['message']);
            $content .= sprintf("<p><span title=\"%s\">%s: %s</span></p>\n",$usertime,$fromto,$msg);
        }
        $fields['last_message_id'] = $id; // remember the last message we saw
        $buttons = array( // bare bones button bar
            dialog_buttondef(BUTTON_OK),
            dialog_buttondef('button_talk',t('braille_button_talk','m_confab')),
            dialog_buttondef('button_all',t('braille_button_all','m_confab')));
    } else {
        // 2B -- stay in a loop, refresh and wait for updates
        $head = "<head>\n".$http_equiv.
            sprintf("<meta http-equiv=\"refresh\" content=\"%d\">\n",$config['refresh']).
            "<title>&nbsp;</title>\n".$bsss."</head>\n";

        $hotkey = accesskey_from_string(t('braille_button_talk','m_confab'));

        $buttons = array(array(
            'type' => F_SUBMIT,
            'name' => 'button_talk',
            'class' => 'button button_talk',
            'value' => t('braille_button_talk_asterisk','m_confab'),
            'title' => t('hotkey_for_button','',array('{HOTKEY}' => $hotkey)),
            'accesskey' => $hotkey
            ));
    }

    // 3 -- actually send prepared output directly to browser (note: headers are already sent)
    echo "<html>$head<body>\n".$content;
    $href = was_node_url($theme->node_record,array('c' => time()));
    echo sprintf("<form action=\"%s\" method=\"POST\">\n", $href);
    foreach($buttons as $button) {
        echo dialog_get_widget($button)."\n";
    }
    echo "</form></body>\n<!-- IE hack -->\n<head>\n".$http_equiv."</head>\n<!-- /IE hack -->\n</html>\n";

    // 4 -- update our 'last_seen' and maybe 'last_message_id' fields
    $table = 'confab_participants';
    $where = array('confab_participant_id' => intval($participant['confab_participant_id']));
    if (db_update($table,$fields,$where) === FALSE) {
        logger(sprintf("%s(): cannot update participant '%s' in confab '%d': %s",
                       __FUNCTION__,$nickname,$confab_id,db_errormessage()));
    }
    // 5 -- stop $theme from also sending output
    $theme->silent_mode = TRUE;
    return TRUE;
} // confab_braille_show_main()


/** show a dialog where a user can write a message or can navigate to another confab-screen
 *
 * this routine sends a dialog to the user that looks a bit like this:
 * <pre>
 * Message ________ to <listbox> [Send] [Cancel] [All] [Users] [Quit]
 * </pre>
 *
 * In this dialog the user can enter a message, select a destination from the <listbox>
 * (if enabled) and submit the message via [Send]. Alternatively the user can
 * cancel this screen (and return to the main/messages screen) or navigate to one
 * of the other screens [All] or [Users] or simply end the conversation via [Quit].
 *
 * @param object &$theme collects the output
 * @param array &$config holds the confab configuration record
 * @param array &$participant holds the participant record
 * @return bool TRUE on success, FALSE otherwise
 */
function confab_braille_show_talk(&$theme,&$config,&$participant) {
    // 1 -- prepare for talk screen
    $http_equiv = confab_braille_page_headers(); // headers are now sent; add http_equiv lateron
    $bsss = $theme->get_bazaar_style_style();
    $fields = array('last_seen' => strftime("%Y-%m-%d %T"));
    $head = "<head>\n".$http_equiv."<title>&nbsp;</title>\n".$bsss."</head>\n";
    $nickname = $participant['nickname'];
    echo "<html>$head<body>\n";

    // 2 -- prepare input fields and navigation buttons
    $hotkey = t('braille_message_accesskey','m_confab');
    $message_item =  array(
        'type' => F_ALPHANUMERIC,
        'name' => 'message',
        'label' => t('braille_message_label','m_confab'),
        'title' => t('braille_message_title','m_confab',array('{HOTKEY}' => $hotkey)),
        'accesskey' => $hotkey,
        'value' => ''
        );
    if (($config['personal'] == CONFAB_PERSONAL_YES) ||
        (($config['personal'] == CONFAB_PERSONAL_REGISTERED) && ($participant['username_ext'] > 0))) {
        $participants = confab_get_participants($config);
        $options = array('0' => t('message_to_all','m_confab'));
        foreach ($participants as $id => $p) {
            $options[$id] = sprintf('%s (%s)',htmlspecialchars($p['nickname']),
                                              htmlspecialchars($p['full_name']));
        }
        $hotkey = t('braille_message_to_accesskey','m_confab');
        $personal_item = array(
            'type' => F_LISTBOX,
            'name' => 'message_to',
            'label' => t('braille_message_to_label','m_confab'),
            'title' => t('braille_message_to_title','m_confab',array('{HOTKEY}' => $hotkey)),
            'accesskey' => $hotkey,
            'options' => $options,
            'value' => $participant['interlocutor_id']
            );
    } else {
        $personal_item = array();
    }
    $buttons = array(
        dialog_buttondef('button_send',t('braille_button_send','m_confab')),
        dialog_buttondef(BUTTON_CANCEL),
        dialog_buttondef('button_all',t('braille_button_all','m_confab')),
        dialog_buttondef('button_users',t('braille_button_users','m_confab')),
        dialog_buttondef('button_quit',t('braille_button_quit','m_confab'))
        );
    // 3 -- render the dialog
    $href = was_node_url($theme->node_record,array('c' => time()));
    echo sprintf("<form action=\"%s\" method=\"POST\">\n", $href);
    echo str_replace('~','',$message_item['label']).' '.dialog_get_widget($message_item)."\n";
    if (!empty($personal_item)) {
        echo str_replace('~','',$personal_item['label']).' '.
             implode("\n",dialog_get_widget($personal_item))."\n";
    }
    foreach($buttons as $button) {
        echo dialog_get_widget($button)."\n";
    }
    echo "</form></body>\n<!-- IE hack -->\n<head>\n".$http_equiv."</head>\n<!-- /IE hack -->\n</html>\n";

    // 4 -- update our 'last_seen' field
    $table = 'confab_participants';
    $where = array('confab_participant_id' => intval($participant['confab_participant_id']));
    if (db_update($table,$fields,$where) === FALSE) {
        logger(sprintf("%s(): cannot update participant '%s' in confab '%d': %s",
                       __FUNCTION__,$nickname,$confab_id,db_errormessage()));
    }
    // 5 -- stop $theme from also sending output
    $theme->silent_mode = TRUE;
    return TRUE;
} // confab_braille_show_talk()


/** send all messages to the participant; update last_seen_id
 *
 * this not just shows the new messages, but it shows _all_ message.
 *
 * @param object &$theme collects the output
 * @param array &$config holds the confab configuration record
 * @param array &$participant holds the participant record
 * @return bool TRUE on success, FALSE otherwise
 * @todo this looks a lot like the {@link confab_braille_show_main()},
 *       maybe there is room for improvement.
 */
function confab_braille_show_all(&$theme,&$config,&$participant) {

    // 1 -- prepare for all messages screen
    $http_equiv = confab_braille_page_headers(); // headers are now sent; add http_equiv lateron
    $bsss = $theme->get_bazaar_style_style();
    $fields = array('last_seen' => strftime("%Y-%m-%d %T"));
    $head = "<head>\n".$http_equiv."<title>&nbsp;</title>\n".$bsss."</head>\n";
    $content = '';
    $nickname = $participant['nickname'];
    echo "<html>$head<body>\n";

    // 2 -- any messages to show?
    $messages = confab_view_get_messages($config,$nickname);
    if (sizeof($messages) > 0) {
        // 2A -- yes, show messages and no refresh
        foreach ($messages as $id => $message) {
            $usertime = $message['message_datim'];
            $sys = (db_bool_is(TRUE,$message['message_system']));
            $personal = (db_bool_is(TRUE,$message['message_personal']));
            $nickfrom = ($sys) ? t('braille_system','m_confab'):htmlspecialchars($message['message_from']);
            $nickto = ($personal) ? htmlspecialchars($message['message_to']):t('braille_all','m_confab'); 
            $params = array('{FROM}' => $nickfrom, '{TO}' => $nickto);
            $fromto = t(($personal) ? 'braille_from_to' : 'braille_from','m_confab',$params);
            $msg = htmlspecialchars($message['message']);
            echo sprintf("<p><span title=\"%s\">%s: %s</span></p>\n",$usertime,$fromto,$msg);
        }
        $fields['last_message_id'] = $id; // remember the last message we saw
    }

    // 3 -- add navigation button
    $buttons = array(dialog_buttondef(BUTTON_OK));
    $href = was_node_url($theme->node_record,array('c' => time()));
    echo sprintf("<form action=\"%s\" method=\"POST\">\n", $href);
    foreach($buttons as $button) {
        echo dialog_get_widget($button)."\n";
    }
    echo "</form></body>\n<!-- IE hack -->\n<head>\n".$http_equiv."</head>\n<!-- /IE hack -->\n</html>\n";

    // 4 -- update our 'last_seen' and maybe 'last_message_id' fields
    $table = 'confab_participants';
    $where = array('confab_participant_id' => intval($participant['confab_participant_id']));
    if (db_update($table,$fields,$where) === FALSE) {
        logger(sprintf("%s(): cannot update participant '%s' in confab '%d': %s",
                       __FUNCTION__,$nickname,$confab_id,db_errormessage()));
    }
    // 5 -- stop $theme from also sending output
    $theme->silent_mode = TRUE;
    return TRUE;
} // confab_braille_show_all()


/** send a list of users to the participant
 *
 * @param object &$theme collects the output
 * @param array &$config holds the confab configuration record
 * @param array &$participant holds the participant record
 * @return bool TRUE on success, FALSE otherwise
 */
function confab_braille_show_users(&$theme,&$config,&$participant) {

    // 1 -- prepare for participants screen
    $http_equiv = confab_braille_page_headers(); // headers are now sent; add http_equiv lateron
    $bsss = $theme->get_bazaar_style_style();
    $fields = array(
            'last_seen' => strftime("%Y-%m-%d %T"),
            'show_list' => FALSE);
    $head = "<head>\n".$http_equiv."<title>&nbsp;</title>\n".$bsss."</head>\n";
    $nickname = $participant['nickname'];
    echo "<html>$head<body>\n";

    // 2 -- show list of users
    $participants = confab_get_participants($config);
    if (sizeof($participants) > 0) {
        foreach ($participants as $id => $participant) {
            $params = array('{NICKNAME}' => htmlspecialchars($participant['nickname']),
                            '{FULL_NAME}' => htmlspecialchars($participant['full_name']),
                            );
            echo sprintf("<p><span title=\"%s\">%s</span></p>\n",$participant['last_seen'],
                         (db_bool_is(TRUE,$participant['is_active'])) ?
                         t('braille_participant_active','m_confab',$params) :
                         t('braille_participant_idle','m_confab',$params));
        }
    }

    // 3 -- add navigation button
    $buttons = array(dialog_buttondef(BUTTON_OK));
    $href = was_node_url($theme->node_record,array('c' => time()));
    echo sprintf("<form action=\"%s\" method=\"POST\">\n", $href);
    foreach($buttons as $button) {
        echo dialog_get_widget($button)."\n";
    }
    echo "</form></body>\n<!-- IE hack -->\n<head>\n".$http_equiv."</head>\n<!-- /IE hack -->\n</html>\n";

    // 4 -- update our 'last_seen' field
    $table = 'confab_participants';
    $where = array('confab_participant_id' => intval($participant['confab_participant_id']));
    if (db_update($table,$fields,$where) === FALSE) {
        logger(sprintf("%s(): cannot update participant '%s' in confab '%d': %s",
                       __FUNCTION__,$nickname,$confab_id,db_errormessage()));
    }
    // 5 -- stop $theme from also sending output
    $theme->silent_mode = TRUE;
    return TRUE;
} // confab_braille_show_users()


/** send various headers attempting to defeat caching
 *
 * this is an attempt to force the client not to reuse the
 * data that will be sent shortly. As a bonus, we also construct
 * a string with the corresponding http-equiv statements, to further
 * convince the browser not to cache the page.
 * Reference: http://www.htmlgoodies.com/beyond/reference/article.php/3472881
 * http://stackoverflow.com/questions/49547
 *
 * While we are here, we add a few more useful headers such as Content-Type.
 *
 * @return string headers sent and http-equiv headers in a single string 
 */
function confab_braille_page_headers() {
  $headers = array('Content-Type' => 'text/html; charset=UTF-8',
                   'Cache-Control' => 'no-cache, no-store, must-revalidate',
		   'Pragma' => 'no-cache',
		   'Expires' => '0',
		   'Last-Modified' => gmdate('D, d M Y H:i:s').' GMT',
                   'Content-Type' => 'text/html; charset=UTF-8',
                   'Content-Script-Type' => 'text/javascript',
                   'Content-Style-Type' => 'text/css',
                   'MSSmartTagsPreventParsing' => 'TRUE');
  $http_equiv = '';
  foreach ($headers as $k => $v) {
    $http_equiv .= sprintf("<meta http-equiv=\"%s\" content=\"%s\">\n",$k, $v);
    header($k.': '.$v);
  }
  return $http_equiv;
} // confab_braille_page_headers()


// ==================================================================
// ======================== VISUAL PROTOCOL =========================
// ==================================================================

/** dispatcher for visual interface
 *
 * this routine decides what needs to be done (and does it).
 * As a side effect, we check various time-out conditions.
 *
 * 1. we ALWAYS check for timeout of (other) participants, this may
 *    or may not change the status of others from active to idle
 * 2. we ALWAYS check if we are kicked out, e.g. because the
 *    conversation timed out, we were directed to leave via a 
 *    status change (maybe initiated by the moderator) or the
 *    Confab was suddenly closed (by a moderator). In this case we
 *    a. either silently leave the conversation and perform the Quit
 *       action which will return us to the regular $theme, OR
 *    b. send a Q-instruction to the client, requesting the client
 *       to quit the proper way.
 * 3. we SOMETIMES check for conversations that time out, i.e.
 *    whenever a contribution is posted (to the current conversation),
 *    we check and maybe increment the conversation_id. This means
 *    that we effectively kick ourselves out the next time we come here.
 *
 * Note that some of the subroutines called handle their own I/O
 * and set the $silent_mode of the $theme to TRUE, effectively
 * suppressing the regular theme. This leaves a blank screen or
 * an open connection to 'play' with. The notable exception is
 * {@link confab_leave_conversation()}: that routine DOES rely on
 * $theme to build a regular page including navigation, etc.
 *
 * Note that this routine handles both asynchronous requests (indicated
 * by having 'button_send' or 'button_update' set in $_POST) which
 * take plave behind the scenes or requests that involve a complete
 * HTML-page (e.g. after the user presses [Refresh]).
 *
 * The way to quit the conversation is to kindly ask the client to
 * submit a Quit-request rather than simply showing the quit-screen:
 * the latter would probable confuse the asynchronous handler in
 * the client.
 *
 * @param object &$theme collects the output
 * @param array &$config holds the confab configuration record
 * @param array &$participant holds the participant record
 * @return bool TRUE on succes, FALSE on error or sometimes never at all
 */
function confab_visual_dispatcher(&$theme,&$config,&$participant) {
    confab_view_participants_timeout($config); // best effort
    $threshold = strftime("%Y-%m-%d %T",time() - intval($config['timeout_conversation']));
    $last_seen = $participant['last_seen'];

    if ((isset($_POST['button_send'])) || (isset($_POST['button_update']))) {
        if (isset($_POST['button_send'])) {
            confab_conversation_timeout($config); // best effort
            confab_save_message($theme,$config,$participant); // best effort
        }
        $force_quit = FALSE;
        if ($config['last_conversation_id'] != $participant['conversation_id']) {
            $force_quit = TRUE;
        } elseif (($config['status'] == CONFAB_STATUS_CLOSED) ||
                  (($config['status'] == CONFAB_STATUS_REGISTERED) &&
                   ($participant['username_ext'] <= 0))) {
            $force_quit = TRUE;
        } elseif ($last_seen < $threshold) { // confab_view_participants_timeout() didn't update in core
            $participant['status'] = CONFAB_USER_STATUS_LEAVING;
            $force_quit = TRUE;
        }
        $retval = confab_visual_update($theme,$config,$participant,$force_quit);
    } else {
        if ($participant['status'] == CONFAB_USER_STATUS_LEAVING) {
            $reason = ($last_seen < $threshold) ? 'user_left_timeout' : 'user_left_forced';
            $retval = confab_leave_conversation($theme,$config,$participant,$reason);
        } elseif ($config['last_conversation_id'] != $participant['conversation_id']) {
            $reason = 'user_left_timeout';
            $retval = confab_leave_conversation($theme,$config,$participant,$reason);
        } elseif (($config['status'] == CONFAB_STATUS_CLOSED) ||
                  (($config['status'] == CONFAB_STATUS_REGISTERED) &&
                   ($participant['username_ext'] <= 0))) {
            $reason = 'user_left_closed';
            $retval = confab_leave_conversation($theme,$config,$participant,$reason);
        } elseif (isset($_POST['button_refresh'])) {
            $retval = confab_visual_show_main($theme,$config,$participant);
        } elseif (isset($_POST['button_quit'])) {
            $retval = confab_leave_conversation($theme,$config,$participant);
        } else {
            $retval = confab_visual_show_main($theme,$config,$participant);
        }
    }
    return $retval;
} // confab_visual_dispatcher()


/** construct the main screen for the Visual interfae
 *
 * this sets up the initial page that will be controlled via Javascript
 * lateron (including resizing).
 *
 * We generate the page based on the information that is currently available,
 * i.e. the current list of (all) messages, the current list of (all)
 * participants and an up to date listbox (if applicable).
 *
 * This same routine is called whenever the user presses the Refresh-button,
 * This is a way to refrek both lists if somehow something went wrong underway.
 *
 * @param object &$theme collects the output
 * @param array &$config holds the confab configuration record
 * @param array &$participant holds the participant record
 * @return bool TRUE on succes
 */
function confab_visual_show_main(&$theme,&$config,&$participant) {
    global $WAS_SCRIPT_NAME,$CFG;

    // 0 -- initialiase & sanity check on interface id
    switch($participant['interface_id']) {
    case CONFAB_VISUAL_BW: $css = 'visual_bw.css'; $i = CONFAB_VISUAL_BW; break;
    case CONFAB_VISUAL_RB: $css = 'visual_rb.css'; $i = CONFAB_VISUAL_RB; break;
    case CONFAB_VISUAL_BY: $css = 'visual_by.css'; $i = CONFAB_VISUAL_BY; break;
    default:               $css = 'visual.css';    $i = CONFAB_VISUAL;    break;
    }
    $style_css = $CFG->progwww_short.'/modules/confab/'.$css;
    $bsss = $theme->get_bazaar_style_style();
    $odd = 'odd'.$i;
    $even = 'even'.$i;

    // 1 -- construct a userlist and maybe a listbox for personal messages
    $participants = confab_get_participants($config);

    // 1A -- Userlist
    if (empty($config['header'])) {
        $user_list = '';
    } else {
        $user_list = "<h3>".htmlspecialchars($config['header'])."</h3>\n";
    }
    $oddeven = $odd;
    foreach($participants as $id => $p) {
        $oddeven = ($oddeven == $odd) ? $even : $odd;
        $params = array('{NICKNAME}' => htmlspecialchars($p['nickname']),
                        '{FULL_NAME}' => htmlspecialchars($p['full_name'])
                        );
        if (db_bool_is(TRUE,$p['is_active'])) {
            $user_list .= sprintf("<p class=\"user_%s\">%s</p>\n",
                                  $oddeven,t('visual_userlist','m_confab',$params));
        } else {
            $user_list .= sprintf("<p class=\"user_%s\">%s</p>\n",
                                  $oddeven,t('visual_userlist_idle','m_confab',$params));
        }
    }
    // 1B - Personal listbox
    if (($config['personal'] == CONFAB_PERSONAL_YES) ||
        (($config['personal'] == CONFAB_PERSONAL_REGISTERED) && ($participant['username_ext'] > 0))) {
        $options = array('0' => t('message_to_all','m_confab'));
        foreach ($participants as $id => $p) {
            $options[$id] = sprintf('%s (%s)',htmlspecialchars($p['nickname']),
                                              htmlspecialchars($p['full_name']));
        }
        $hotkey = t('visual_message_to_accesskey','m_confab');
        $personal_item = array(
            'type' => F_LISTBOX,
            'name' => 'message_to',
            'label' => t('visual_message_to_label','m_confab'),
            'title' => t('visual_message_to_title','m_confab',array('{HOTKEY}' => $hotkey)),
            'accesskey' => $hotkey,
            'options' => $options,
            'value' => $participant['interlocutor_id'],
            'id' => 'lstPersonal'.$i,
            'tabindex' => '2'
            );
        $listbox = ltrim(str_replace('~','',$personal_item['label']).' ').
                   implode("\n",dialog_get_widget($personal_item))."\n";
    } else {
        $listbox = '';
    }

    // 1C -- Message input
    $hotkey = t('visual_message_accesskey','m_confab');
    $message_item =  array(
        'type' => F_ALPHANUMERIC,
        'name' => 'message',
        'label' => t('visual_message_label','m_confab'),
        'title' => t('visual_message_title','m_confab',array('{HOTKEY}' => $hotkey)),
        'accesskey' => $hotkey,
        'value' => '',
        'tabindex' => 1,
        'id' => 'txtMessage'.$i
        );
    $message_input = ltrim(str_replace('~','',$message_item['label']).' ').
                     dialog_get_widget($message_item)."\n";

    // 2 -- collect all available messages
    $message_list = '';
    $nickname = $participant['nickname'];
    $messages = confab_view_get_messages($config,$nickname);
    if (sizeof($messages) > 0) {
        $oddeven = $odd;
        foreach ($messages as $id => $m) {
            $oddeven = ($oddeven == $odd) ? $even : $odd;
            $sys = (db_bool_is(TRUE,$m['message_system']));
            $personal = (db_bool_is(TRUE,$m['message_personal']));
            $params = array(
                '{DATIM}' => $m['message_datim'],
                '{TIME}' => substr($m['message_datim'],11,5),
                '{FROM}' => ($sys) ? t('visual_system','m_confab') : htmlspecialchars($m['message_from']),
                '{TO}' => ($personal) ? htmlspecialchars($m['message_to']) : '',
                '{MESSAGE}' => htmlspecialchars($m['message'])
                );
            if ($personal) {
                $message_list .= sprintf("<p title=\"%s\" class=\"personal_%s\">%s</p>\n",
                                         t('visual_from_to_title','m_confab',$params),
                                         $oddeven, t('visual_from_to','m_confab',$params));
            } else {
                $message_list .= sprintf("<p title=\"%s\" class=\"%s\">%s</p>\n",
                                         t('visual_from_title','m_confab',$params),
                                         $oddeven, t('visual_from','m_confab',$params));
            }
        }
    }

    // 3 -- construct user interface
    // 3A -- prepare various buttons
    $sound_v = str_replace('~','',t('visual_button_sound','m_confab'));
    $sound_a = t('visual_button_sound_accesskey','m_confab');
    $sound_t = t('visual_button_sound_title','m_confab');
    $refresh_v = str_replace('~','',t('visual_button_refresh','m_confab'));
    $refresh_a = accesskey_from_string(t('visual_button_refresh','m_confab'));
    $refresh_t = t('visual_button_refresh_title','m_confab');
    $quit_v = str_replace('~','',t('visual_button_quit','m_confab'));
    $quit_a = accesskey_from_string(t('visual_button_quit','m_confab'));
    $quit_t = t('visual_button_quit_title','m_confab');
    $send_v = str_replace('~','',t('visual_button_send','m_confab'));
    $send_a = accesskey_from_string(t('visual_button_send','m_confab'));
    $send_t = t('visual_button_send_title','m_confab');
    // 3B -- and form actions
    $href = was_node_url($theme->node_record,array('c' => time()));
    $xref = was_node_url($theme->node_record);
    // 3C -- and the Javascript
    $progconfab = $CFG->progwww_short.'/modules/confab';
    if (($CFG->debug) || (!file_exists($CFG->progdir.'/modules/confab/visual.min.js'))) {
        $visual_js = htmlspecialchars(str_replace("\n","",$progconfab.'/visual.js'));
    } else {
        $visual_js = htmlspecialchars(str_replace("\n","",$progconfab.'/visual.min.js'));
    }
    $confab_dir = addslashes(htmlspecialchars(str_replace("\n","",$progconfab)));
    // JS variables
    $vars = array('confabURL'   => was_node_url($theme->node_record),
                  'confabDir'   => $confab_dir,
                  'interfaceID' => strval($participant['interface_id']),
                  'refreshTime' => intval($config['refresh']),
                  'confabTitle' => $config['header'],
                  'oddEven' => ($oddeven == $odd) ? 0 : 1,
                  'debug' => ($CFG->debug) ? TRUE : FALSE,
                  );
    // JS translations
    $tras = array(t('visual_from','m_confab'),
                  t('visual_from_title','m_confab'),
                  t('visual_from_to','m_confab'),
                  t('visual_from_to_title','m_confab'),
                  t('visual_userlist','m_confab'),
                  t('visual_userlist_idle','m_confab'),
                  t('visual_js_malformed_message','m_confab'),
                  t('message_to_all','m_confab')
                  );
    $js = '';
    foreach($vars as $k => $v) {
        if (is_bool($v)) {
            $js .= sprintf("  var %s=%s;\n",$k,($v) ? 'true' : 'false');
        } elseif (is_int($v)) {
            $js .= sprintf("  var %s=%d;\n",$k,$v);
        } else {
            $js .= sprintf("  var %s='%s';\n",$k,str_replace("\n","\\n",addslashes($v)));
        }
    }
    $js .= "  var tr=['".str_replace("\n","\\n",addslashes($tras[0]));
    for ($j=1; $j<sizeof($tras); ++$j) {
        $js .= "',\n'".str_replace("\n","\\n",addslashes($tras[$j]));
    }
    $js .= "'];\n";
    $js .= "onLoad(confab_init);\n";

    // 4 -- Now construct the main page with all prepared values and echo immediately
    echo <<<EOT
<html>
<head>
<title></title>
<link rel="stylesheet" type="text/css" href="$style_css">
$bsss</head>
<body>
<div id="divOuter$i">
  <div id="divUpper$i">
    <div id="divMessages$i">
$message_list    </div>
    <div id="divParticipants$i">
$user_list    </div>
    <form action="$href" method="POST" id="frmQuit$i">
      <input type="submit" id="btnSound$i" name="button_sound" value="$sound_v" title="$sound_t" accesskey="$sound_a" class="soundoff$i" tabindex="9">
      <input type="submit" id="btnRefresh$i" name="button_refresh" value="$refresh_v" title="$refresh_t" accesskey="$refresh_a" tabindex="10">
      <input type="submit" id="btnQuit$i" name="button_quit" value="$quit_v" title="$quit_t" accesskey="$quit_a" tabindex="11">
    </form>
  </div>
  <div id="divLower$i">
    <form action="$xref" method="POST" id="frmSend$i">
$message_input$listbox<input type="submit" id="btnSend$i" name="button_send" value="$send_v" title="$send_t" accesskey="$send_a" tabindex="3">
    </form>
  </div>
</div>  
<script type="text/javascript" src="$visual_js"></script>
<script type="text/javascript">
$js</script>
</body>
</html>

EOT;

    // 5 -- stop $theme from also sending output
    $theme->silent_mode = TRUE;
    return TRUE;
} // confab_visual_show_main()



/** send an update to the client in resonse to an XMLHttpRequest
 *
 * Note that we cannot process TAB ('\t') and LF ('\n') in JavaScript because these delimiters
 * are used in our protocol. We have to get rid of those.
 *
 * This routine conditionally sends updates to the client.
 * The following messages are defined:
 *
 * Send a message
 * "M" | datim | nickfrom | nickto | message
 *
 * Send a complete userlist
 * "U" | n | m | nick1 | name1 | id1 | active1
 *             ...
 *             | nickn | namen | idn | activen
 *
 * Nicely ask the client to quit the conversation:
 * "Q"
 *
 * Indicate end of file:
 * "Z"
 *
 * The client is supposed not to process any lines following the "Z" command
 * (which is always sent). That allows us to add a minimal escape
 * to the main page via URL, for clients that have trouble with
 * XMLHttpRequest functionality. Even though it doesn't work the
 * client is able to get to a more readable page...
 *
 * @param object &$theme collects the output
 * @param array &$config holds the confab configuration record
 * @param array &$participant holds the participant record
 * @return bool TRUE on succes
 */
function confab_visual_update(&$theme,&$config,&$participant,$force_quit=FALSE) {
    // 0 -- initialise
    $fields = array();
    $dirty = FALSE;

    // 1 -- construct data for userlist and maybe for a listbox for personal messages too
    $user_data = '';
    if (db_bool_is(TRUE,$participant['show_list'])) {
        $participants = confab_get_participants($config);
        $user_data = sprintf("U\t%d\t%d",sizeof($participants),4);
        // "U" | n | m | nick1 | name1 | attr1 | status1
        //             ...
        //             | nickn | namen | attrn | statusn
        foreach($participants as $id => $p) {
            $nickname = trim(str_replace("\t","",str_replace("\n"," ",$p['nickname'])));
            $full_name = trim(str_replace("\t","",str_replace("\n"," ",$p['full_name'])));
            $user_data .= sprintf("\t%s\t%s\t%d\t%s",$nickname,$full_name,$id,
                                  (db_bool_is(TRUE,$p['is_active'])) ? '1' : '0');
        }
        $user_data .= "\n";
        $fields['show_list'] = FALSE; // only send userlist once
        $dirty = TRUE;
    }

    // 2 -- any (new) messages to show?
    $nickname = $participant['nickname'];
    $messages = confab_view_get_messages($config,$nickname,$participant['last_message_id']);
    $message_data = '';
    if (sizeof($messages) > 0) {
        // 2A -- yes, send messages
        // "M" | datim | nickfrom | nickto | message
        foreach ($messages as $id => $m) {
            $datim = (empty($m['message_datim'])) ? '' : str_replace("\n"," ",$m['message_datim']);
            $nick_from = (empty($m['message_from'])) ? '' : str_replace("\n"," ",$m['message_from']);
            $nick_to = (empty($m['message_to'])) ? '' : str_replace("\n"," ",$m['message_to']);
            $message = (empty($m['message'])) ? '' : str_replace("\n"," ",$m['message']);
            $message_data .= sprintf("M\t%s\t%s\t%s\t%s\n",
                                     trim(str_replace("\t","",$datim)),
                                     trim(str_replace("\t","",$nick_from)),
                                     trim(str_replace("\t","",$nick_to)),
                                     trim(str_replace("\t","",$message)));
        }
        $fields['last_message_id'] = $id; // remember the last message we saw
        $dirty = TRUE;
    }
    // 3 -- are we leaving?
    $quit_data = '';
    if (($participant['status'] != CONFAB_USER_STATUS_JOINED) || ($force_quit)) {
        $quit_data = "Q\n";
        $dirty = TRUE;
    } else {
        $fields['last_seen'] = strftime("%Y-%m-%d %T"); // only update last seen when NOT quitting
    }
    // 4 -- send results to caller
    header('Content-Type: text/plain; charset=utf-8');
    if ($dirty) {
        echo $user_data;
        echo $message_data;
        echo $quit_data;
    }
    echo "Z\n"; // Done

    // 5 -- but add a 'rescue link' too for clients that don't follow
    echo was_node_url($theme->node_record,NULL,'',FALSE,TRUE); // fully qualified URL

    // 6 -- update our 'last_seen' and maybe 'last_message_id' fields
    $table = 'confab_participants';
    $where = array('confab_participant_id' => intval($participant['confab_participant_id']));
    if ((!empty($fields)) && (db_update($table,$fields,$where) === FALSE)) {
        logger(sprintf("%s(): cannot update participant '%s' in confab '%d': %s",
                       __FUNCTION__,$participant['nickname'],$participant['confab_id'],db_errormessage()));
    }

    // 7 -- finally stop $theme from also sending output
    $theme->silent_mode = TRUE;
    return TRUE;
} // confab_visual_update()


// ==================================================================
// =========================== WORKHORSES ===========================
// ==================================================================

/** calculate the number of currently joined participants in the current conversation
 *
 * this calculates the number of participants in the current conversation
 * in the confab. We only count those that have joined the conversation 
 * (status is CONFAB_USER_STATUS_JOINED) and not those that are already
 * marked as leaving the conversation (status is CONFAB_USER_STATUS_LEAVING)
 * or are otherwise not joined (status is CONFAB_USER_STATUS_NONE).
 *
 * In case of error, we return the maximum number of participants. This may be
 * the wrong answer but at least we won't be letting too many participants in.
 *
 * @param array &$config holds the confab configuration including the conversation_id
 * @return int the number of active and idle participants that are not yet leaving
 */
function confab_view_calc_participants(&$config) {
    $confab_id = intval($config['confab_id']);
    $conversation_id = intval($config['last_conversation_id']);

    $table = 'confab_participants';
    $where = array('confab_id' => $confab_id,
                   'conversation_id' => $conversation_id,
                   'status' => CONFAB_USER_STATUS_JOINED);
    $fields = 'COUNT(confab_participant_id) AS n';
    if (($record = db_select_single_record($table,$fields,$where)) === FALSE) {
        $max_participants = intval($config['max_participants']);
        logger(sprintf("%s(): cannot count participants (returning %d) in confab '%d': %s",
                       __FUNCTION__,$max_participants,$confab_id,db_errormessage()));
        return $max_participants; // worst case, at least we won't overflow...
    }
    return intval($record['n']);
} // confab_view_calc_participants()


/** retrieve all messages in the conversation since $last_message_id
 *
 * this routine returns an array with all messages with an id that
 * is greater than $last_message_id, ie. the messages that were added
 * since $last_message_id was recorded. If $last_message_id = 0, _all_
 * messages in the conversation are retrieved.
 *
 * Note that this routine also takes the 'personal messages' into account:
 * if a message is personal, it is only added to the result if $nickname
 * is either the sender or the receiver of the personal message. Others
 * are not allowed to see these personal messages.
 *
 * The result is always an array (could be empty) and it is always
 * keyed by confab_message_id.
 *
 * @param array &$config holds the confab configuration
 * @param string $nickname for who are we retrieving the messages
 * @param int $last_message_id the last message $nickname has already received
 * @return array available messsages (but could be empty)
 */
function confab_view_get_messages(&$config,$nickname, $last_message_id=0) {
    $confab_id = $config['confab_id'];
    $table = 'confab_messages';
    $fields = '*';
    $where = sprintf('(confab_message_id > %d) AND (confab_id = %d) AND (conversation_id = %d) AND '.
                     '((message_personal = %s) OR '.
                     '((message_personal = %s) AND ((message_from = %s) OR (message_to = %s))))',
                     $last_message_id,
                     $confab_id,
                     $config['last_conversation_id'],
                     db_escape_and_quote(FALSE),
                     db_escape_and_quote(TRUE),
                     db_escape_and_quote($nickname),
                     db_escape_and_quote($nickname));
    $order = 'confab_message_id';
    $keyfield = 'confab_message_id';
    if (($messages = db_select_all_records($table, $fields,$where,$order,$keyfield)) === FALSE) {
        logger(sprintf("%s(): cannot retrieve confab messages in confab '%d': %s",
                       __FUNCTION__,$confab_id,db_errormessage()));
        $messages = array();
    }
    return $messages;
} // confab_view_get_messages()


/** check for idle participants and timed out participants
 *
 * we update all participants for which the status is active but that
 * posted a message longer than timeout_idle ago. If there are any
 * participants with a changes status, we flag that to all participants
 * by setting the show_list flag. That means that the next time a
 * participant returns a fresh userlist has to be sent to the browser
 * (if applicable).
 *
 * This routine also flags users that have passed the conversation
 * timeout value, i.e. that have not touched home for a long time.
 * Those participants are more or less kicked out, eventually freeing
 * space for other participants.
 *
 * @param array &$config holds the confab configuration record
 * @return bool|int FALSE on error, # of participants in conversation on success
 */
function confab_view_participants_timeout(&$config) {
    $confab_id = intval($config['confab_id']);
    $conversation_id = intval($config['last_conversation_id']);
    $dirty = FALSE;
    $table = 'confab_participants';

    // 1 -- check for idle participants
    $fields = array('is_active' => FALSE);
    $threshold = strftime("%Y-%m-%d %T",time() - intval($config['timeout_idle']));
    $where = sprintf('(confab_id = %d) AND (conversation_id = %d) AND (is_active) AND (last_posted < %s)',
                     $confab_id,$conversation_id,db_escape_and_quote($threshold));
    if ((($retval = db_update($table,$fields,$where)) !== FALSE) && ($retval > 0)) {
        $dirty = TRUE;
    }
    // 2 -- check for conversation time out of participants
    $fields = array('status' => CONFAB_USER_STATUS_LEAVING);
    $threshold = strftime("%Y-%m-%d %T",time() - intval($config['timeout_conversation']));
    $where = sprintf('(confab_id = %d) AND (conversation_id = %d) AND (last_seen < %s)',
                     $confab_id,$conversation_id,db_escape_and_quote($threshold));
    if ((($retval = db_update($table,$fields,$where)) !== FALSE) && ($retval > 0)) {
        $dirty = TRUE;
    }
    // 3 -- maybe a userlist needs to be refreshed
    if ($dirty) {
        $fields = array('show_list' => TRUE);
        $where = array('confab_id' => $confab_id, 'conversation_id' => $conversation_id);
        $retval = db_update($table,$fields,$where);
    }
    return $retval;
} // confab_view_participants_timeout()


/** check for stale participants and remove corresponding records
 *
 * this routine keeps the database clean; any participants that were
 * last seen more than CONFAB_MAX_CONVERSATION_LIMIT (24h) ago.
 * As long as this routine is called periodically, say every time
 * a participant joins a conversation we are able to weed out all
 * the garbage, eventually.
 *
 * @param array &$config holds the confab configuration record
 * @return bool|int FALSE on error, # of deleted stale participants on success
 */
function confab_view_delete_stale_participants(&$config) {
    $confab_id = intval($config['confab_id']);
    $threshold = strftime("%Y-%m-%d %T",time() - CONFAB_MAX_CONVERSATION_LIMIT); // 24h ago
    $table = 'confab_participants';
    $where = sprintf('(confab_id = %d) AND (last_seen < %s)',$confab_id,db_escape_and_quote($threshold));
    if (($retval = db_delete($table,$where)) === FALSE) {
        logger(sprintf("%s(): cannot delete stale participants from confab '%d': %s",
                           __FUNCTION__,$confab_id,db_errormessage()));
    } else {
        logger(sprintf("%s(): deleted %d stale participants from confab '%d'",
                       __FUNCTION__,$retval,$confab_id),WLOG_DEBUG);
    }
    return $retval;
} // confab_view_delete_stale_participants()

?>