/*
 * This file is part of Website@School, a Content Management System especially designed for schools.
 * Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation supplemented with the Additional Terms, as set
 * forth in the License Agreement for Website@School (see /program/license.html).
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
 * for more details.
 *
 * You should have received a copy of the License Agreement for Website@School
 * along with this program. If not, see http://websiteatschool.eu/license.html
 */

/** /program/modules/confab/visual.js - javascript for the Confab Visual interface
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_crew
 * @version $Id: visual.js,v 1.5 2016/03/23 09:28:21 peter Exp $
 */

/* Parameters set in the initial screen are
 *
 * var confabURL;	// the place to post our XMLHttpRequest
 * var interfaceID;	// distinguish between various Visual variants
 * var refreshTime;	// the time to wait between calls (in seconds)
 * var confabTitle;	// the title of the Confab (could be empty)
 * var oddEven;		// keeps track of the # of messages modulo 2
 * var debug;		// Enable logger for development purposes, set to false for production
 */
var divOuter;		// Overarching container div
var divUpper;		// Upper part of the screen
var divLower;		// Lower part of the screen
var divMessages;	// Collects all incoming messages
var divParticipants;	// Shows the current set of participants (reloaded from userList[])
var btnSound;		// Toggle a sound on an incoming message
var btnRefresh;		// Refresh the complete page
var btnQuit;		// Leave the Confab
var frmQuit;		// Form contains Quit, Refresh buttons
var txtMessage;		// Inputfield for contributions
var lstPersonal;	// (optional) Listbox for sending personal messages
var btnSend;		// Submit message
var frmSend;		// Form contains Message, Listbox and Send button

var userList=[];	// 0-based array with user information (for both lstPersonal and divParticipants)
var beep;		// an audio object used to play a sound whenever a confab message is received
var beeps=false;	// toggle to enable/disable beeps (we start with false and toggle once in init)

var classOddEven=[];	// Helper for constructing odd/even classes

//        ===INITIALISATION===
 
/** initialise the confab
 */
function confab_init() {
  divOuter = document.getElementById('divOuter'+interfaceID);
  divUpper = document.getElementById('divUpper'+interfaceID);
  divLower = document.getElementById('divLower'+interfaceID);
  divMessages = document.getElementById('divMessages'+interfaceID);
  divParticipants = document.getElementById('divParticipants'+interfaceID);
  btnSound = document.getElementById('btnSound'+interfaceID);
  btnRefresh = document.getElementById('btnRefresh'+interfaceID);
  btnQuit = document.getElementById('btnQuit'+interfaceID);
  frmQuit = document.getElementById('frmQuit'+interfaceID);
  txtMessage = document.getElementById('txtMessage'+interfaceID);
  lstPersonal = document.getElementById('lstPersonal'+interfaceID);
  btnSend = document.getElementById('btnSend'+interfaceID);
  frmSend = document.getElementById('frmSend'+interfaceID);

  classOddEven = ['even'+interfaceID,'odd'+interfaceID];
  document.title = confabTitle;
  btnSend.onclick=doSend;
  btnSound.onclick=doSoundToggle;

  // setup the beep functionality
  beep=new Audio();
  if (beep.canPlayType('audio/ogg')) beep.src=confabDir+'/beep.ogg';
  else if (beep.canPlayType('audio/x-wav')) beep.src=confabDir+'/beep.wav';
  else if (beep.canPlayType('audio/mpeg')) beep.src=confabDir+'/beep.mp3';
  beep.load();
  doSoundToggle(); // initially switch sound on

  // more housekeeping: maximum screen size, focus on input field, etc.
  divMessages.scrollTop=divMessages.scrollHeight;
  txtMessage.focus();
  logger(hhmm()+'init');
  window.onresize=resizeVisual;
  resizeVisual();
  doUpdate();
} // confab_init()

/** add function f to list of functions to execute on load */
function onLoad(f) {
  if (window.addEventListener) {
    window.addEventListener('load', f, false);
  } else if (window.attachEvent) {
    window.attachEvent('load', f);
  }
} // onLoad()

//        ===Handlers===

/** toggle the sound on/off */
function doSoundToggle() {
  beeps=!(beeps);
  btnSound.className=((beeps)?'soundon':'soundoff')+interfaceID;
  txtMessage.focus();
  return false;
} // doSoundToggle()

/** send a message to all participants or maybe a single interlocutor
 *
 * Eventually this yields an M-reply from the server with this message.
 * Note that we reset the input field after sending this message.
 */
function doSend() {
  var s=txtMessage.value.trim();
  if (s.length>0) {
      var p = (lstPersonal) ? lstPersonal.value : "0";
      var data={"message":s,"message_to":p,"button_send":btnSend.value};
      var request = new XMLHttpRequest();
      request.open("POST",confabURL);
      request.onreadystatechange = function() {
          if ((request.readyState === 4) && (request.status === 200)) {
              receiveData(request.responseText);
          }
      }
      request.setRequestHeader("Content-Type","application/x-www-form-urlencoded");
      request.send(encodeFormData(data));
  }
  txtMessage.value='';
  txtMessage.focus();
  return false;
} // doSend()

/** send a request to the server to see if there is news. */
function doUpdate() {
    var data={"button_update":1};
    var request = new XMLHttpRequest();
    request.open("POST",confabURL);
    request.onreadystatechange = function() {
        if ((request.readyState === 4) && (request.status === 200)) {
            receiveData(request.responseText);
            setTimeout(doUpdate,1000*refreshTime);
        }
    }
    request.setRequestHeader("Content-Type","application/x-www-form-urlencoded");
    request.send(encodeFormData(data));
} // doUpdate()

/** callback routine that handles incoming response */
function receiveData(data) {
    var lines=data.split('\n');
    var n=lines.length;
    var dirty=false;
    for (var l=0; l<n; ++l) {
        switch (lines[l][0]) {
        case 'U': handleUserList(lines[l]); dirty=true; break;
        case 'M': handleMessage(lines[l]); dirty=true; break;
        case 'Q': handleQuit(); break;
        case 'Z': if ((dirty) && (beeps)) { beep.play(); } return; break;
        default: break; // cannot parse, ignore error
        }
    }
    // fall through if no 'Z' seen
    if ((dirty) && (beeps)) { beep.play(); }
} // receiveData()

//        ===Helpers===

/** process an incoming message
 *
 * format: "M" | datim | nickfrom | nickto | message
 */
function handleMessage(data) {
    var a=data.split('\t');
    if (a.length < 5) {
        var r={'{DATA}':data};
        logger(hhmm()+str(6,r)); // ERROR: malformed message {DATA}
        return;
    }
    var r={'{DATIM}':a[1],
           '{TIME}':a[1].substring(11,16),
           '{FROM}':htmlSpecialChars(a[2]),
           '{TO}':htmlSpecialChars(a[3]),
           '{MESSAGE}':htmlSpecialChars(a[4])};
    var personal = (a[3].length <= 0) ? false : true;
    var p=document.createElement('P');
    p.title = (personal) ? str(3,r) : str(1,r);
    p.className=((personal) ? 'personal_' : '') + classOddEven[oddEven];
    p.innerHTML=(personal) ? str(2,r) : str(0,r);
    divMessages.appendChild(p);
    divMessages.scrollTop=divMessages.scrollHeight;
    oddEven = 1 - oddEven;
} // handleMessage()

/** send a form to the server only containing the parameter 'button_quit'
 *
 * Append a hidden field to the form because the name/value of the real
 * submit button is not sent to server. Server depends on the parameter
 * being set (or not) hence the value is unimportant (don't care).
 */
function handleQuit() {
    var h=document.createElement('input');
    h.type = 'hidden';
    h.name = 'button_quit';
    h.value = '1';
    frmQuit.appendChild(h);
    frmQuit.submit();
} // handleQuit()

/** process an incoming userlist
 *
 * Format: "U" | n | m 
 *             | nick1 | full_name1 | id1 | active1
 *             | nick2 | full_name2 | id2 | active2
 *             ...
 *             | nickn | full_namen | idn | activen
 *
 * where n identifies the number of users and m the
 * number of properties of a user (here: 4).
 *
 * Part of handling the userlist is to simply overwrite the current
 * userList and replacing it with this new version. The new version
 * of the list is rendered in the pane with participant names via
 * redrawUserList().
 */
function handleUserList(data) {
  var a=data.split('\t');
  if (a.length < 3) {
    var r={'{DATA}':data};
    logger(hhmm()+str(6,r)); // ERROR: malformed message {DATA}
    return;
  }
  var n=parseInt(a[1],10);
  var m=parseInt(a[2],10);
  if (a.length != n * m + 3) {
    var r={'{DATA}':data};
    logger(hhmm()+str(6,r)); // ERROR: malformed message {DATA}
    return;
  }
  userList=[];
  var k=3;
  for (var i=0; i<n; ++i) {
    userList[i]=[];
    for (var j=0; j<m; ++j) {
      userList[i][j]=a[k++];
    }
  }
  redrawUserList();
  if (lstPersonal) {
      redoPersonal();
  }
} // handleUserList()

/** refresh the content of the Participants div
 *
 * - show confab name in h3
 * - show each user in a p using even/odd class
 */
function redrawUserList() {
  var p; // generic helper
  var s; // span element
  var c; // class helper

  // 0--clean up existing pane
  while (divParticipants.childNodes.length > 0) {
    divParticipants.removeChild(divParticipants.firstChild);
  }

  // 1--Confab name (if any)
  if (confabTitle.length > 0) {
      p=document.createElement('h3');
      p.style.wordwrap='break-word';
      p.innerHTML=confabTitle;
      divParticipants.appendChild(p);
  }

  // 2--List of all members
  for (var i=0; i<userList.length; ++i) {
      r = {'{FULL_NAME}':htmlSpecialChars(userList[i][1]),
           '{NICKNAME}':htmlSpecialChars(userList[i][0])};
      var t = (userList[i][3] == '1') ? 4 : 5; // index in tr[]
      p=document.createElement('P');
      p.className='user_' + classOddEven[i % 2];
      p.innerHTML=str(t,r);
      divParticipants.appendChild(p);
  }
} // redrawUserList()

/** re-create the options in the 'personal' listbox
 *
 * strategy:
 * 1. which option was currently selected (default)?
 * 2. lose all existing options
 * 3. construct a new list based on userList[]
 * 4. done!
 */
function redoPersonal() {
    // 1 -- which one is selected?
    var selected = '0';
    for (var i=0; i<lstPersonal.options.length; ++i) {
        if (lstPersonal.options[i].selected) {
            selected = lstPersonal.options[i].value;
        }
    }

    // 2 -- get rid of all options
    while (lstPersonal.options.length > 0) {
        lstPersonal.removeChild(lstPersonal.firstChild);
    }
    // 3 -- and reconstruct the complete list w/ correct selected option
    p=document.createElement('option');
    p.innerHTML = htmlSpecialChars(tr[7]);
    p.value = '0';
    p.selected = (selected == '0');
    lstPersonal.appendChild(p);
    for (var i=0; i<userList.length; ++i) {
        p=document.createElement('option');
        p.innerHTML = htmlSpecialChars(userList[i][0] + ' (' + userList[i][1] + ')');
        p.value = userList[i][2];
        p.selected = (selected == userList[i][2]);
        lstPersonal.appendChild(p);
    }
} // redoPersonal()

/** resize the whole interface to use available space as good as possible
 *
 * Note:
 * Code for determining the viewport inspired by an article by Andy Langton
 * http://andylangton.co.uk/articles/javascript/get-viewport-size-javascript/
 * and subsequent refinements by 'elstcb'.
 */
function resizeVisual() {
  // determine viewport in (vx,vy) (see note above)
  var d=document;
  var e=d.documentElement;
  var b=d.getElementsByTagName('body')[0];
  var vx=window.innerWidth||e.clientWidth||b.clientWidth;
  var vy=window.innerHeight||e.clientHeight||b.clientHeight;
  // var s = '(vx,vy) = ('+vx+','+vy+')'; logger(s);
  // enforce minimum dimensions of 400x300
  vx=(vx<400)?400:vx;
  vy=(vy<300)?300:vy;
  // 45px to allow for a scrollbar
  w=vx-45;
  divOuter.style.width=w+'px';
  divUpper.style.width=w+'px';
  txtMessage.style.width=w+'px';
  w-=225;
  divMessages.style.width=w+'px';
  // 45px to allow for a scrollbar
  h=vy-45;
  divOuter.style.height=h+'px';
  // adjust for height of txtMessage and btnSend; add 22px for good measure
  h-=22+boxHeight(txtMessage)+boxHeight(btnSend);
  divUpper.style.height=h+'px';
  divMessages.style.height=h+'px';
  // adjust for the height of two buttons underneath divParticipants; add 4px for good measure
  h-=boxHeight(btnSound)+boxHeight(btnQuit)+4;
  divParticipants.style.height=h+'px';
  divMessages.scrollTop=divMessages.scrollHeight;
} // resizeVisual()

//        ===Utilities===

/** construct a translated string optionally with parameters
 *
 * this returns the i'th string in the global list of translations in tr[].
 * If r is specified it is supposed to be an associative array with 1 or
 * more properties.
 * Example:
 *
 * tr[17]='{USER} has left the building';
 * r={'{USER}':'Elvis'};
 * s=str(17,r); // s='Elvis has left the building'
 *
 * @param int i identifies the message
 * @param object r is an optional associative array with parameters
 * @return string translated string with optional parameters
 * @uses tr[]
 */
function str(i,r) {
  var s=tr[i]||'?['+i+']?';
  if (r) { // if passed we use key-value pairs to substitute inside string s
    var re;
    for (k in r) {
      re=new RegExp(k,'g');
      s=s.replace(re,r[k]);
    }
  }
  return s;
} // str()

/** escape some HTML-characters
 *
 * utility routine escapes HTML-characters much the
 * same as the PHP-function with a similar name.
 * Characters '&' '<' '>' and '"' are converted to
 * their corresponding entities.
 *
 * @param string s
 * @return string escaped string s
 */
function htmlSpecialChars(s) { 
  return s.replace(/&/g,'&amp;'). 
           replace(/</g,'&lt;').
           replace(/>/g,'&gt;').
           replace(/"/g,'&quot;');
  return s.replace(/"/g,'&quot;'); // FixMe: stupid line to match the " in the previous
} // htmlSpecialChars()

/** write an error message to the list of messages (debug only)
 *
 * @param string s
 * @param string a attribute (color) code or #rgb triplet (default '@')
 * @return void msg written to list, list pruned
 */
function logger(s) {
    if (!debug) return;
    var p=document.createElement('p');
    p.className='error'+interfaceID;
    p.appendChild(document.createTextNode(s));
    divMessages.appendChild(p);
    divMessages.scrollTop=divMessages.scrollHeight;
} // logger()

/** utility routine to show the current hour and minute as 'hh:mm ' including trailing space */
function hhmm() {
    var d=new Date;
    var h=d.getHours();
    var m=d.getMinutes();
    return ((h<10)?'0'+h:h)+':'+((m<10)?'0'+m:m)+' ';
} // hhmm()

/** utility routine to prepare data for transmission to server */
function encodeFormData(data) {
    if (!data) return "";
    var pairs = [];
    for (var name in data) {
        if (!data.hasOwnProperty(name)) continue;
        if (typeof data[name] === "function") continue;
        var value = data[name].toString();
        name = encodeURIComponent(name).replace("%20","+");
        value = encodeURIComponent(value).replace("%20","+");
        pairs.push(name + "=" + value);
    }
    return pairs.join('&');
} // encodeFormData()

/** utility to calculate the height of element e */
function boxHeight(e) {
  var box=e.getBoundingClientRect();
  var y=box.height || (box.right - box.left);
  return y;
} // boxHeight()

/* eof visual.js */
