<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/htmlpage/htmlpage_admin.php - management interface for htmlpage-module
 *
 * This file defines the administrative interface to this module.
 * The interface consists of the following four functions.
 *
 * <code>
 * htmlpage_disconnect(&$output,$area_id,$node_id,$module)
 * htmlpage_connect(&$output,$area_id,$node_id,$module)
 * htmlpage_show_edit(&$output,$area_id,$node_id,$module,$viewonly,$edit_again,$href,$option)
 * htmlpage_save(&$output,$area_id,$node_id,$module,$viewonly,&$edit_again,$option)
 * </code>
 *
 * These functions are called from pagemanagerlib.php whenever necessary.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_htmlpage
 * @version $Id: htmlpage_admin.php,v 1.8 2016/03/23 09:28:22 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }


/** disconnect this module from a node
 *
 * this breaks the link between the node $node_id in area $area_id and this module.
 * For now we simply delete the record with page data. In a future version
 * we might want to retain the page data, 'for future reference' (but: what do
 * we do with it? Oh well).
 * 
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node from which we need to disconnect
 * @param array $module the module record straight from the database
 * @return bool TRUE on success, FALSE otherwise
 */
function htmlpage_disconnect(&$output,$area_id,$node_id,$module) {
    $where = array('node_id' => intval($node_id));
    $retval = (db_delete('htmlpages',$where) === FALSE) ? FALSE : TRUE;
    return $retval;
} // htmlpage_disconnect()


/** connect this module to a node
 *
 * this makes the link between the node $node_id in area $area_id and this module.
 * In this case we simply link a data container to node $node_id in a 1-to-1 relation.
 * Note that we might decide lateron to keep versions of pages around, e.g. by inserting
 * new records every save rather than updating the existing.
 * 
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which we need to connect
 * @param array $module the module record straight from the database
 * @return bool TRUE on success, FALSE otherwise
 */
function htmlpage_connect(&$output,$area_id,$node_id,$module) {
    global $USER;
    $now = strftime('%Y-%m-%d %T');
    $fields = array(
        'node_id' => intval($node_id),
        'version' => 1,
        'page_data' => '',
        'ctime' => $now,
        'cuser_id' => $USER->user_id,
        'mtime' => $now,
        'muser_id' => $USER->user_id);
    if (db_insert_into('htmlpages',$fields) === FALSE) {
        logger(sprintf("%s(): cannot create record for page '%d': %s",
                       __FUNCTION__,$node_id.db_errormessage()));
        $retval = FALSE;
    } else {
        $retval = TRUE;
    }
    return $retval;
} // htmlpage_connect()


/** present the user with a dialog to modify the content that is connected to node $node_id
 *
 * this prepares a dialog for the user filled with existing data (if any), possibly allowing
 * the user to modify the content. If the flag $viewonly is TRUE, this routine should only
 * display the content rather than let the user edit it. If the flag $edit_again is TRUE,
 * the routine should use the data available in the $_POST array, otherwise it should read
 * the data from the database (or wherever the data comes from). The parameter $href is the
 * place where the form should be POST'ed.
 *
 * The dialog should be added to the $output object. Useful routines are:
 * <code>
 * $output->add_content($content): add $content to the content area
 * $output->add_message($message): add $message to the message area (feedback to the user)
 * $output->add_popup_bottom($message): popup $message in browser after loading the page (uses javascript)
 * $output->add_popup_top($message): popup $message in browser before loading the page (uses javascript)
 * </code>
 * 
 * Note: the field version is basically unused, allthough the version number gets to
 * be incremented on save. We keep the original functionality, including picking the
 * latest version of the page_data (via $order = version DESC).
 *
 * The parameter $option is not used in this module.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @param string $href the action property of the HTML-form, the placa where data will be POST'ed
 * @param string $option indicates which submenu-option was selected (NULL for none)
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 * @todo get rid of 'version' field and related cruft (or do something useful with it). 2014-05-05/PF
 */
function htmlpage_show_edit(&$output,$area_id,$node_id,$module,$viewonly,$edit_again,$href, $option) {
    $dialogdef = htmlpage_get_dialogdef($viewonly);
    if ($edit_again) { // retrieve and validate the POSTed values
        dialog_validate($dialogdef); // no need to show messages; did that already in htmlpage_save()
    } else {
        $where = array('node_id' => intval($node_id));
        $order = array('version DESC');
        $retval = db_select_single_record('htmlpages','*',$where,$order);
        if ($retval === FALSE) {
            $output->add_message(t('error_retrieving_data','admin'));
            $dialogdef['htmlpage_content']['value'] = '??error??';
        } else {
            $dialogdef['htmlpage_content']['value'] = $retval['page_data'];
        }
    }
    $output->add_content(dialog_quickform($href,$dialogdef));
    return TRUE;
} // htmlpage_show_edit()


/** save the modified content data of this module linked to node $node_id
 *
 * this validates and saves the data that was submitted by the user.

 * If validation fails, or storing the data doesn't work, the flag $edit_again
 * is set to TRUE and the return value is FALSE.
 *
 * If the user has cancelled the operation, the flag $edit_again is set to FALSE
 * and the return value is also FALSE.
 *
 * If the modified data is stored successfully, the return value is TRUE.
 * The value of $edit_again is determined by the key used to save: [Save]
 * brings the user back to edit ($edit_again = TRUE) where [Done] leaves
 * editing alltogether ($edit_again = FALSE).
 *
 * Here is a summary of return values.
 *
 * <pre>
 * $retval | $edit_again | Action
 * ------- | ----------- | ------
 * FALSE   | FALSE       | [Cancel] Stop editing, unlock & return to tree view
 * FALSE   | TRUE        | [Save] or [Done] failed: redo the dialog and let user correct mistakes
 * TRUE    | FALSE       | [Done] pressed, everything saved succesfully, we're done
 * TRUE    | TRUE        | [Save] pressed, everything saved succesfully but continue editing
 * </pre>
 *
 * The parameter $option is not used in this module.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which the content is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing and hence saving is not allowed
 * @param bool &$edit_again set to TRUE if we need to edit the content again, FALSE otherwise
 * @param string $option indicates which submenu-option was selected (NULL for none)
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function htmlpage_save(&$output,$area_id,$node_id,$module,$viewonly,&$edit_again,$option) {
    global $USER;

    // 1 -- bail out if cancelled or viewonly
    if ((isset($_POST['button_cancel'])) || ($viewonly)) {
        $edit_again = FALSE;
        return FALSE;
    }
   
    // 2 -- redo if invalid data was submitted
    $dialogdef = htmlpage_get_dialogdef($viewonly);
    if (!dialog_validate($dialogdef)) {
        // there were errors, show them to the user and do it again
        foreach($dialogdef as $k => $item) {
            if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                $output->add_message($item['error_messages']);
            }
        }
        $edit_again = TRUE;
        return FALSE;
    }
 
    // 3 -- actually save the data (a 2-step process for historic reasons)
    // 3A -- pick up the latest version + unique htmlpage id.
    $table = 'htmlpages';
    $fields = array('htmlpage_id','version');
    $where = array('node_id' => intval($node_id));
    $order = array('version DESC');
    if (($record = db_select_single_record($table,$fields,$where,$order)) === FALSE) {
        logger(sprintf("%s(): cannot retrieve existing record for node '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        $output->add_message(t('error_saving_node','admin'));
        $edit_again = TRUE;
        return FALSE;
    }
    // 3B -- bump version and write back (again, this way due to historic reasons).
    $now = strftime('%Y-%m-%d %T');
    $version = intval($record['version']);
    $where = array('htmlpage_id' => intval($record['htmlpage_id']));
    $fields = array('version' => $version + 1,
                    'page_data' => $dialogdef['htmlpage_content']['value'],
                    'mtime' => $now,
                    'muser_id' => $USER->user_id);
    $retval = db_update('htmlpages',$fields,$where);
    if ($retval === FALSE) {
        logger(sprintf('%s(): error saving htmlpage \'%d\': %s',
                       __FUNCTION__,$node_id,db_errormessage()));
        $output->add_message(t('error_saving_node','admin'));
        $edit_again = TRUE;
        return FALSE;
    }
    $edit_again =  (isset($_POST['button_done'])) ? FALSE : TRUE;
    return TRUE;
} // htmlpage_save()


/** construct the dialog definition for editing a plain HTML page
 *
 * @param bool $viewonly if TRUE make dialog readonly (and suppress [Save] / [Done])
 * @return array dialog definition
 * @todo should we add a title here?
 */
function htmlpage_get_dialogdef($viewonly) {
    $dialogdef = array(
        'htmlpage_content' => array(
            'type' => F_RICHTEXT,
            'name' => 'htmlpage_content',
            'value' => '',
            'rows' => 20,
            'columns' => 80,
            'maxlength' => 655360, // arbitrary limit, 640kB is enough for anyone (famous last words)
            'viewonly' => $viewonly,
            ),
        'csrftoken' => dialog_csrftoken()
        );
    if (!$viewonly) {
        $dialogdef['button_save'] = dialog_buttondef(BUTTON_SAVE);
        $dialogdef['button_done'] = dialog_buttondef(BUTTON_DONE);
    }
    $dialogdef['button_cancel'] = dialog_buttondef(BUTTON_CANCEL);
    return $dialogdef;
} // htmlpage_get_dialogdef()

?>