<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/htmlpage/htmlpage_search.php - interface to the search-part of this module
 *
 * This file defines the interface with this module for searching content.
 * The interface consists of a (temporary) table and this function:
 *
 * <code>
 * bool htmlpage_search(&$hits,&$results,$qwords, $limit, $offset)
 * </code>
 *
 * The function is called whenever module content is to be searched. At that point
 * the helper table, which is called 'search_nodes', exists and is populated.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_htmlpage
 * @version $Id: htmlpage_search.php,v 1.8 2016/04/04 06:59:17 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }


/** search all htmlpages linked to selected nodes for keywords in $qwords
 *
 *
 * <b>TASKS</b>
 *
 * This function needs to perform these tasks.
 *
 *  - If ($results < 0) on entry, the function must calculate and set
 *    $results to the number of results in this module's data (could be 0).
 *
 *  - If ($limit > 0) on entry, the function must return at most $limit
 *    search hits in the $hits array.
 *
 *  - If ($results > 0) on entry the function is allowed to use that
 *    number as a hint of the total # of hits in this module's data.
 *
 *  - If ($results == 0) on entry the function can decide to return
 *    immediately because (according to the caller) there are no hits
 *    anyway.
 *
 *
 * <b>KEYWORDS TO SEARCH FOR</b>
 *
 * On entry the array $qwords contains the sanitised user-supplied
 * keywords, as follows:
 *
 * <code>
 * $qword[$i] = array($original, $utf8lower, $quoted);
 *
 * $original: exactly what the user entered into the search box.
 * $utf8lower: lowercase version of $original
 * $quoted: escaped/quoted $original ready for LIKE '%$quoted%'
 * </code>
 *
 *
 * <b>HELPER TABLE</b>
 *
 * On entry the table 'search_nodes' [*] exists and is populated
 * with node_id's that are qualified to be searched. The structure
 * of this table is as follows:
 *
 * <code>
 * CREATE TABLE $search_nodes (node_id int(11) NOT NULL DEFAULT '0' PRIMARY KEY (node_id));
 * </code>
 *
 * [*] Note that as with every other table the name of the table is
 * prefixed with the prefix found in $DB->prefix or $CFG->prefix.
 *
 * The search results of this module can be limited to the qualifying
 * nodes by using an inner join, e.g.
 *
 * <code>
 * SELECT m.data FROM $search_nodes s INNER JOIN $module m USING (node_id) WHERE $where;
 * </code>
 *
 *
 * <b>RETURNING SEARCH RESULTS</b>
 *
 * If there are hits, these are collected in $hits as follows:
 *
 * <code>
 * $hits[] = array('title' => $title,
 *                 'context' => array($snippet1, $snippet2, ...),
 *                 'url' => $url);
 * </code>
 *
 * Eventually this yields HTML-code like this:
 *
 * <code>
 * <div>
 *   <div><a href="$url">$title</a></div>
 *   <div>$snippet1 $snippet2 ...</div>
 *   <div><a href="$url">$url</a></div>
 * </div>
 * </code>
 *
 * Tip: function {@link search_context()} can be used to extract
 * context snippets from hits.
 *
 * If search is not relevant to the module, the NOP can be this:
 *
 * <code>
 * $results = 0;
 * return TRUE;
 * </code>
 *
 * See also {@link search_nodes()}.
 *
 * @param array &$hits receives search results (at most $limit)
 * @param int &$results receives number of hits (negative value on entry means (re-)calculate total) 
 * @param array $qwords holds 1 or more keywords to search for (original, utf8lower, quoted)
 * @param int $limit maximum number of hits to return (could be 0)
 * @param int $offset starting point within this module for returning results
 * @return bool|array FALSE on failure, TRUE on success and $hits, $results updated (maybe)
 * @uses search_context()
 * @uses search_where()
 */
function htmlpage_search(&$hits, &$results, $qwords, $limit, $offset=0) {
    global $DB;

    // 0A -- do we need to do anything
    if (($results == 0) ||  // cached value from caller says 'no hits'
        (($results > 0) && ($limit <= 0))) { // there are hits but none are requested at this time
        return TRUE;
    }

    // 0B -- prepare the selection criteria (same for both step 1 and 2)
    $where = search_where($qwords,array('page_data'));

    // 1 -- calculate the # of hits, no retrieving of data
    if ($results < 0) {
        $sql = sprintf("SELECT COUNT(n.node_id) AS results ".
                       "FROM %ssearch_nodes s INNER JOIN %snodes n USING (node_id) ".
                       "INNER JOIN %shtmlpages m USING (node_id) ".
                       "WHERE %s",$DB->prefix,$DB->prefix,$DB->prefix,$where);
        if ((($DBResult = $DB->query($sql,1)) === FALSE) || ($DBResult->num_rows != 1)) {
            logger(sprintf('%s(): cannot retrieve record count: %s',__FUNCTION__,db_errormessage()));
            $results = 0;
            return FALSE;
        }
        $record = $DBResult->fetch_row_assoc();
        $results = intval($record['results']);
        $DBResult->close();
    }

    // 2 -- retrieve $limit records at $offset
    if (($limit > 0) && ($offset < $results)) {
        $sql = sprintf("SELECT n.node_id, n.title, n.link_text, m.page_data ".
                       "FROM  %ssearch_nodes s INNER JOIN %snodes n USING (node_id) ".
                       "INNER JOIN %shtmlpages m USING (node_id) ".
                       "WHERE %s",$DB->prefix,$DB->prefix,$DB->prefix,$where);
        if (($DBResult = $DB->query($sql,$limit,$offset)) === FALSE) {
            logger(sprintf('%s(): cannot retrieve selected records: %s',__FUNCTION__,db_errormessage()));
            return FALSE;
        }
        while (($record = $DBResult->fetch_row_assoc()) !== FALSE) {
            $url = was_node_url(NULL,array('node' => $record['node_id']));
            $hits[] = array('title' => $record['title'],
                            'context' => search_context($qwords,$record['page_data']),
                            'url' => $url);
        }
        $DBResult->close();
    }

    // 3 -- all done
    return TRUE;
} // htmlpage_search()

?>