<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/mailpage/mailpage_admin.php - management interface for mailpage-module
 *
 * This file defines the administrative interface to this module.
 * The interface consists of the following four functions.
 *
 * <code>
 * mailpage_disconnect(&$output,$area_id,$node_id,$module)
 * mailpage_connect(&$output,$area_id,$node_id,$module)
 * mailpage_show_edit(&$output,$area_id,$node_id,$module,$viewonly,$edit_again,$href,$option)
 * mailpage_save(&$output,$area_id,$node_id,$module,$viewonly,&$edit_again,$option)
 * </code>
 *
 * These functions are called from pagemanagerlib.php whenever necessary.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_mailpage
 * @version $Id: mailpage_admin.php,v 1.11 2016/03/23 09:28:23 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }


/** disconnect this module from a node
 *
 * this breaks the link between the node $node_id in area $area_id and
 * this module. For now we simply delete the records in 'mailpages' and
 * 'mailpages_addresses' linked to node node_id. Note that the order
 * is not important (in this case) because there are no FK constraints
 * between the two tables. However, there are FK constraints with the
 * nodes table. The db_delete() here is a simple best effort.
 * 
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node from which we need to disconnect
 * @param array $module the module record straight from the database
 * @return bool TRUE on success, FALSE otherwise
 */
function mailpage_disconnect(&$output, $area_id, $node_id, $module) {
    $where = array('node_id' => intval($node_id));
    $retval = db_delete('mailpages_addresses',$where);
    if (db_delete('mailpages',$where) === FALSE) {
        $retval = FALSE;
    }
    return ($retval === FALSE) ? FALSE : TRUE;
} // mailpage_disconnect()


/** connect this module to a node
 *
 * this makes the link between the node $node_id in area $area_id and this module.
 * In this case we create a single 'mailpages' record linked to node_id.
 * Any addresses can be added later as desired and records in 'mailpages_addresses'
 * will be created as necessary.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which we need to connect
 * @param array $module the module record straight from the database
 * @return bool TRUE on success, FALSE otherwise
 */
function mailpage_connect(&$output, $area_id, $node_id, $module) {
    global $USER;
    $now = strftime('%Y-%m-%d %T');
    $fields = array(
        'node_id'      => intval($node_id),
        'header'       => '',
        'introduction' => '',
        'ctime'        => $now,
        'cuser_id'     => $USER->user_id,
        'mtime'        => $now,
        'muser_id'     => $USER->user_id);
    $retval = db_insert_into('mailpages',$fields);
    if ($retval !== 1) {
        logger(sprintf("%s(): cannot connect to node '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        $retval = FALSE;
    } else {
        $retval = TRUE;
    }
    return $retval;
} // mailpage_connect()


/** dispatcher for presenting the user with one of the mailpage dialogs
 *
 * this routine looks at $option to see which dialog is requested and
 * subsequently displays that dialog.
 *
 * There are three possibilities:
 *
 * - $option == NULL: an overview/help screen basically pointing the user to the submenu items
 * - $option == 'configuration': main configuration via table 'mailpages'
 * - $option == 'destinations': destination addresses config via 'mailpages_addresses'
 *
 * We use different subroutines for clarity.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @param string $option the selected submenu option or NULL if user used the 'Content' option to get here
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function mailpage_show_edit(&$output,$area_id,$node_id,$module,$viewonly,$edit_again,$href,$option) {
    // Always make the user aware of the missing addresses
    $records = mailpage_get_addresses($node_id);
    if (sizeof($records) <= 0) {
        $output->add_message(t('warning_no_addresses','m_mailpage'));
    }
    if ($option == 'configuration') {
        return mailpage_show_edit_configuration($output,$area_id,$node_id,$module,
                                                $viewonly,$edit_again,$href);
    } elseif ($option == 'destinations') {
        return mailpage_show_edit_destinations($output,$area_id,$node_id,$module,
                                                $viewonly,$edit_again,$href);
    } else { // generic call to Content ($option == NULL)
        mailpage_show_content($output,$node_id,$module);
    }
    return TRUE;
} // mailpage_show_edit()


/** show the main content screen (in fact documentation only)
 *
 * this displays the simple overview screen with documentation.
 * It ends with a cancel button that ends the whole edit operation for
 * this node.
 * 
 * @param object &$output collects the html output (if any)
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function mailpage_show_content(&$output,$node_id,$module) {
    global $WAS_SCRIPT_NAME;
    $params = array(
        'job' => JOB_PAGEMANAGER,
        'task' => TASK_NODE_EDIT_CONTENT,
        'node' => strval($node_id)
        );
    $aparams = array('{NODE}' => strval($node_id));
    if (!empty($module['options'])) {
        $options = explode(',',$module['options']);
        if (!empty($options)) {
            foreach($options as $option) {
                $key = sprintf('{%s}',strtoupper($option));
                $title = t($option.'_title','m_mailpage',$aparams);
                $params[PARAM_SUBMENU_OPTION] = $option;
                $a_attr = array('title' => $title);
                $anchor = t($option.'_label','m_mailpage',$aparams);
                $aparams[$key] = html_a($WAS_SCRIPT_NAME,$params,$a_attr,$anchor);
            }
        }
    }
    $output->add_content('<h2>'.t('mailpage_content_header','m_mailpage',$aparams).'</h2>');
    $output->add_content(t('mailpage_content_explanation','m_mailpage',$aparams));
    $params = array(
        'job'    => JOB_PAGEMANAGER,
        'task'   => TASK_SAVE_CONTENT,
        'node'   => intval($node_id)
        );
    $href = href($WAS_SCRIPT_NAME, $params);
    $dialogdef = array('button_cancel' => dialog_buttondef(BUTTON_CANCEL));
    $output->add_content(dialog_quickform($href,$dialogdef));
    return TRUE;
} // mailpage_show_content()


/** dispatcher for save routines
 *
 * We use different subroutines for clarity.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which the content is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing and hence saving is not allowed
 * @param bool &$edit_again set to TRUE if we need to edit the content again, FALSE otherwise
 * @param string $option the selected submenu option or NULL if user used the 'Content' option to get here
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function mailpage_save(&$output, $area_id, $node_id, $module, $viewonly, &$edit_again, $option) {
    if ($option == 'configuration') {
        return mailpage_save_configuration($output,$area_id,$node_id,$module,$viewonly,$edit_again);
    } elseif ($option == 'destinations') {
        return mailpage_save_destinations($output,$area_id,$node_id,$module,$viewonly,$edit_again);
    } else { // save NULL or invalid option: indicate [Cancel]
        $edit_again = FALSE;
        return FALSE;
    }
    $edit_again = TRUE;
    return TRUE;
} // mailpage_save()


// ==================================================================
// =================== MAILPAGE MAIN CONFIGURATION ==================
// ==================================================================


/** present the user with the main configuration dialog for mailpage at node $node_id
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function mailpage_show_edit_configuration(&$output,$area_id,$node_id,$module,$viewonly,$edit_again,$href) {
    global $WAS_SCRIPT_NAME;
    // Cancel and Done from edit configuration should return to content screen
    if ((isset($_POST['button_cancel'])) || (isset($_POST['button_done']))) {
        return mailpage_show_content($output,$node_id,$module);
    }
    $dialogdef = mailpage_get_dialogdef_config($output,$viewonly,$node_id);
    if ($edit_again) { // retrieve and validate POSTed values
        dialog_validate($dialogdef); // don't show messages; already did that in save() routine
    }
    $output->add_content('<h2>'.t('mailpage_configuration_header','m_mailpage').'</h2>');
    $output->add_content(t('mailpage_configuration_explanation','m_mailpage'));
    $output->add_content(dialog_quickform($href,$dialogdef));
    return TRUE;
} // mailpage_show_edit_configuration()


/** save the modified configuration data of this mailpage linked to node $node_id
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which the content is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing and hence saving is not allowed
 * @param bool &$edit_again set to TRUE if we need to edit the content again, FALSE otherwise
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function mailpage_save_configuration(&$output, $area_id, $node_id, $module, $viewonly, &$edit_again) {
    global $USER;

    // 1 -- bail out if cancelled or viewonly
    if ((isset($_POST['button_cancel'])) || ($viewonly)) {
        $edit_again = TRUE; // but do return to Content
        return FALSE;
    }

    // 2 -- redo if invalid data was submitted
    $dialogdef = mailpage_get_dialogdef_config($output,$viewonly,$node_id);
    if (!dialog_validate($dialogdef)) {
        // there were errors, show them to the user and do it again
        foreach($dialogdef as $k => $item) {
            if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                $output->add_message($item['error_messages']);
            }
        }
        $edit_again = TRUE;
        return FALSE;
    }

    // 3 -- actually save the settings
    $retval = TRUE; // assume success
    $now = strftime('%Y-%m-%d %T');
    $table = 'mailpages';
    $fields = array('header'       => trim($dialogdef['header']['value']),
                    'introduction' => trim($dialogdef['introduction']['value']),
                    'message'      => trim($dialogdef['message']['value']),
                    'mtime'        => $now,
                    'muser_id'     => $USER->user_id);
    $where = array('node_id' => $node_id);
    if (db_update($table,$fields,$where) === FALSE) {
        logger(sprintf("%s(): error saving config values '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        $output->add_message(t('error_saving_data','m_mailpage'));
        $edit_again = TRUE;
        $retval = FALSE;
    } else {
        $edit_again = TRUE; // we _always_ want to return to an edit screen
        $retval = TRUE;
    }
    return $retval;
} // mailpage_save_configuration()


// ==================================================================
// ================ MAILPAGE DESTINATION ADDRESSES  =================
// ==================================================================


// Hack to implement the transition from [Save] in the Add address dialog to the Edit address dialog
$mailpage_new_address_id = NULL;

/** present the user with a dialog for mailpage destination addresses at node $node_id
 *
 * this is a complex routine that is responsible for constructing and showing
 * the dialogs associated with destination addresses.
 *
 *
 *The following dialogs exist.
 *<pre>
 * #1. Mailpage Destinations Overview
 *
 * Add a new destination address
 * [D] Principal (10)
 * [D] Webmaster (20)
 * [Cancel]
 *
 * #2. Add new destination address/Edit existing destination address
 *
 * Name:        __________
 * E-Mail:      __________
 * Description: __________
 * Thankyou:    __________
 * Sort order:  ____
 * [Save] [Done] [Cancel]
 *
 * #3. Delete address confirmation
 *
 * [Delete] [Cancel]
 *</pre>
 *
 * There are some complex interactions between these screens and the corresponding 
 * save routine  {@link mailpage_save_destinations()}. The effects we try to achieve
 * are as follows.
 * The link 'Add a new destination address' leads to dialog #2 with address_id = 0
 * The links 'Principal' and 'Webmaster' lead also to dialog #2 with valid, existing address_id != 0
 * The delete button '[D]' leads to dialog #3 with valid, existing delete_id != 0
 *
 * The [Cancel] button in dialog #1 ends the editing of this node; ie. returns to Page Manager.
 * The other [Cancel] buttons bring the user back to dialog #1. In effect this is a kind of
 * zoom function: you can zoom in on an address and go back to the list without leaving this
 * node, etc. etc.
 *
 * The [Done] and [Delete] buttons perform an action (saving or deleting a destination)
 * and also return to dialog #1.
 *
 * The [Save] button always has the meaning "save the data and continue editing". For the
 * Edit dialog #2 this works perfect. For the Add dialog #2 there is a slight problem,
 * because the Add dialog received address_id = 0 and saving the new address changes
 * that: the address_id should now be the last_id as inserted in the database instead of 0.
 * Unfortuinately there is no clean way to convey this fact via the module interface with
 * Page Manager. In order to provide a smooth user experience I had to use an ugly hack
 * in the form of a global variable called $mailpage_new_address_id. This variable is
 * usually set to NULL, but after successful insertion into the DB of the new address,
 * the new address_id is stored here (from the save() routine). We pick it up here
 * and use that value instead of the value 0 and present the user with an Edit dialog #2
 * for the newly added address when she presses the [Save] button. This magic works
 * because we know that after Page Manager calls the save() routine, which sets the
 * $edit_again flag, Page Manager will call show_view() again. Ugly, but it works. Sorry.
 *
 * Note that we also take $viewonly into account: if this flag is set, there is no
 * 'Add a destination' link because $viewonly does not allow anything to be changed in the
 * node configuration, including adding (and editing and deleting) destinations.
 *
 * Note: the parameter that holds the address_id is called $_GET['address'] in case of
 * adding or editing an address. If we are deleting we use $_GET['delete'] in order
 * to distinguish these cases. I use the variable name $delete_id to keep things separated.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function mailpage_show_edit_destinations(&$output,$area_id,$node_id,$module,$viewonly,$edit_again,$href) {
    global $WAS_SCRIPT_NAME;
    global $mailpage_new_address_id;

    $records = mailpage_get_addresses($node_id); // side effect: all sort_orders are renumbered 10,20,...
    $sort_order = 10 * (1 + sizeof($records)); // ...so this is next available sort_order for new records

    $address_id = get_parameter_int('address',NULL);
    $delete_id = get_parameter_int('delete',NULL);

    // 1 -- handle cancellation from dialogs #2 and #3; return to dialog #1
    if ((isset($_POST['button_cancel'])) && ((!is_null($address_id)) || (!is_null($delete_id)))) {
        mailpage_show_edit_destinations_overview($output,$node_id,$viewonly,$href,$records);
        return TRUE;
    }
    if (isset($_POST['button_cancel'])) { // other cancels lead to Content
        return mailpage_show_content($output,$node_id,$module);
    }

    // 2 -- handle the case for a fresh add #2, a redo add #2, a fresh edit #2 or a redo edit #2
    if (!is_null($address_id)) {
        // $mailpage_new_address_id is set to the newly inserted address_id in save()
        // This is our cue to start editing that address rather than continue with
        // the address in $_GET['address'] that was presented to us (before the save()).
        // We do this only once: only in the case that a new one was added in this run.
        if (($address_id == 0) && (!is_null($mailpage_new_address_id))) {
            $address_id = $mailpage_new_address_id;
            $mailpage_new_address_id = NULL;
        }
        if (($address_id == 0) || (isset($records[$address_id]))) {
            $dialogdef=mailpage_get_dialogdef_address($output,$viewonly,$node_id,$address_id,$sort_order);
            if ($edit_again) {
                if ((mailpage_dialog_validate_address($dialogdef)) && (isset($_POST['button_done']))) {
                    mailpage_show_edit_destinations_overview($output,$node_id,$viewonly,$href,$records);
                    return TRUE;
                }
            }
            if ($address_id == 0) {
                $output->add_content('<h2>'.t('mailpage_add_address_header','m_mailpage').'</h2>');
                $output->add_content(t('mailpage_add_address_explanation','m_mailpage'));
            } else {
                $output->add_content('<h2>'.t('mailpage_edit_address_header','m_mailpage').'</h2>');
                $output->add_content(t('mailpage_edit_address_explanation','m_mailpage'));
            }
            // Alas, we need to construct our own href because of adding $address_id
            $href = href($WAS_SCRIPT_NAME, array(
                'job'=>JOB_PAGEMANAGER,
                'task'=>TASK_SAVE_CONTENT,
                'node'=>$node_id,
                'option' => 'destinations',
                'address' => strval($address_id)));
            $output->add_content(dialog_quickform($href,$dialogdef));
            return TRUE;
        }
        mailpage_show_edit_destinations_overview($output,$node_id,$viewonly,$href,$records);
        return TRUE;
    } // !is_null($address_id)

    // 3 -- handle the case for deletion of an address
    $address_id = $delete_id;
    if ((!is_null($address_id)) && (isset($records[$address_id]))) {
        $dialogdef = mailpage_get_dialogdef_delete($viewonly);
        if ($edit_again) {
            dialog_validate($dialogdef);
        }
        $params = array();
        $params = array('{NAME}' => $records[$address_id]['name'],
                         '{EMAIL}' => $records[$address_id]['email'],
                         '{SORT_ORDER}' => $records[$address_id]['sort_order'],
                         '{ADDRESS_ID}' => strval($address_id));

        $output->add_content('<h2>'.t('mailpage_delete_address_header','m_mailpage').'</h2>');
        $output->add_content(t('mailpage_delete_address_explanation','m_mailpage'));
        $output->add_content('<ul>');
        $output->add_content('<li class="level0">'.t('address_delete_confirm','m_mailpage',$params));
        $output->add_content('</ul>');
        $output->add_content(t('address_delete_are_you_sure','m_mailpage'));

        // Alas, we need to construct our own href because of adding $address_id
        $href = href($WAS_SCRIPT_NAME, array(
            'job'=>JOB_PAGEMANAGER,
            'task'=>TASK_SAVE_CONTENT,
            'node'=>$node_id,
            'option' => 'destinations',
            'delete' => strval($address_id)));
        $output->add_content(dialog_quickform($href,$dialogdef));
        return TRUE;
    }
    mailpage_show_edit_destinations_overview($output,$node_id,$viewonly,$href,$records);
    return TRUE;
} // mailpage_show_edit_destinations()


/** show a screen with a links to add/edit/delete destination addresses
 *
 * @param object &$output collects the html output (if any)
 * @param int $node_id the node to which this module is connected
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @param array $records the currently defined destination adresses
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function mailpage_show_edit_destinations_overview(&$output,$node_id,$viewonly,$href,$records) {
    global $WAS_SCRIPT_NAME;
    // Show a list of existing addresses if any
    $params = array('job' => JOB_PAGEMANAGER,
                    'task' => TASK_NODE_EDIT_CONTENT,
                    'node' => $node_id,
                    'option' => 'destinations',
                    'address' => '0');
    $items = (($viewonly) ? 0 : 1) + sizeof($records);
    $output->add_content('<h2>'.t('mailpage_destinations_header','m_mailpage').'</h2>');
    $output->add_content(t('mailpage_destinations_explanation','m_mailpage'));
    if ($items > 0) {
        $output->add_content('<ul>');
        if (!$viewonly) { // cannot add records in view-only mode
            $output->add_content('  <li class="level0">');
            if (!$output->text_only) {
                $output->add_content('    '.$output->skin->get_icon('blank'));
            }
            $attributes = array('title' => t('add_new_address_title','m_mailpage'));
            $anchor = t('add_new_address_label','m_mailpage');
            $output->add_content('    '.html_a($WAS_SCRIPT_NAME,$params,$attributes,$anchor).'<br>');
        }
        foreach($records as $record) {
            $address_id = intval($record['mailpage_address_id']);
            $aparams = array('{NAME}' => $record['name'],
                             '{EMAIL}' => $record['email'],
                             '{SORT_ORDER}' => $record['sort_order'],
                             '{ADDRESS_ID}' => strval($address_id));
            $output->add_content('  <li class="level0">');
            // 1 -- delete link via icon (text or graphics)
            $output->add_content('    '.
                                 mailpage_get_icon_delete($output,$node_id,$viewonly,$address_id,$aparams));
            $params['address'] = strval($address_id);
            // 2 -- edit link
            $anchor = t('edit_address_label','m_mailpage',$aparams);
            $attributes = array('title' => t('edit_address_title','m_mailpage',$aparams));
            $output->add_content('    '.html_a($WAS_SCRIPT_NAME,$params,$attributes,$anchor).'<br>');
        }
        $output->add_content('</ul>');
    }
    // Finish with a lone [Cancel] that ends the whole Content edit operation
    $dialogdef = array('button_cancel' => dialog_buttondef(BUTTON_CANCEL));
    $output->add_content(dialog_quickform($href,$dialogdef));
    return TRUE;
} // mailpage_show_edit_destinations_overview()


/** process the POSTed destination address data
 *
 * this is another complex routine to fit saving of three different dialogs into one.
 * See also the comprehensive documentation in function {@link mailpage_show_edit_destinations()}.
 *
 * Basically this routine handles all database operations like inserting new addresses, editing
 * or deleting existing ones. We arrive here because the user has either pressed [Save] or
 * [Done], indicating that the data should be stored, or [Cancel], indicating the wish to end
 * the adding/editing/deleting operation (dialogs #2 and #3 as we called them in 
 * function {@link mailpage_show_edit_destinations()}).
 *
 * However, there is one other [Cancel] button: the one in the destinations addresses overview
 * (dialog #1). This button ends the editing of the node alltogether and returns to the Page
 * Manager. This specific situation is indicated by returning FALSE after setting $edit_again
 * to FALSE. All other cases always return FALSE with $edit_again set to TRUE.
 *
 * This means that we do have to check for the validity of the dialog data in show_edit()
 * again, because we abuse the returned value to stay in the overview screen (dialog #1).
 * This is ugly, and I am sorry about that.
 *
 * Here too we distinguish between add/edit and delete via address_id and delete_id respectively
 * (linked to $_GET['address'] and $_GET['delete']). We do check for invalid address_id's:
 * if we don't recognise the address_id/delete_id we simply do nothing (but send a message
 * to the log nevertheless).
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @return bool TRUE on success, FALSE otherwise (but see documentation above)
 */
function mailpage_save_destinations(&$output,$area_id,$node_id,$module,$viewonly,&$edit_again) {
    global $mailpage_new_address_id;

    // 0 -- setup relevant data
    $records = mailpage_get_addresses($node_id); // side effect: all sort_orders are renumbered 10,20,...
    $sort_order = 10 * (1 + sizeof($records)); // ...so this is next available sort_order for new records
    $address_id = get_parameter_int('address',NULL);
    $delete_id = get_parameter_int('delete',NULL);

    // 1 -- bail out if cancelled or viewonly
    if ((isset($_POST['button_cancel'])) || ($viewonly)) {
        if ((is_null($address_id)) && (is_null($delete_id))) { // [Cancel] button on overview screen...
            $edit_again = TRUE; // stay here so we can show Content again
        } else {
            $output->add_message(t('cancelled','admin')); // feedback about this 'partial' Cancel...
            $edit_again = TRUE; // We want to see the addresses overview again, do not exit to Content
        }
        return FALSE;
    }

    // 2 -- do we need to save a new/edited address?
    if (!is_null($address_id)) {
        $dialogdef = mailpage_get_dialogdef_address($output,$viewonly,$node_id, $address_id, $sort_order);
        if (!mailpage_dialog_validate_address($dialogdef)) { // show errors to the user + do it again
            foreach($dialogdef as $k => $item) {
                if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                    $output->add_message($item['error_messages']);
                }
            }
            $edit_again = TRUE;
            return FALSE;
        }
        // actually save the settings
        $table = 'mailpages_addresses';
        $fields = array(
            'node_id'     => $node_id,
            'sort_order'  => intval($dialogdef['sort_order']['value']),
            'name'        => trim($dialogdef['name']['value']),
            'email'       => trim($dialogdef['email']['value']),
            'description' => trim($dialogdef['description']['value']),
            'thankyou'    => trim($dialogdef['thankyou']['value'])
            );
        $params = array('{NAME}' => $fields['name'],
                        '{EMAIL}' => $fields['email'],
                        '{SORT_ORDER}' => $fields['sort_order'],
                        '{ADDRESS_ID}' => strval($address_id));
        if ($address_id <= 0) { // new record needs to be saved.
            $keyfield = 'mailpage_address_id';
            if (($id = db_insert_into_and_get_id($table,$fields,$keyfield)) === FALSE) {
                logger(sprintf('%s(): error adding address: %s',__FUNCTION__,db_errormessage()));
                $output->add_message(t('error_saving_data','m_mailpage'));
            } else {
                $mailpage_new_address_id = $id; // hack: indicate a new address was added via global
                $params['{ADDRESS_ID}'] = strval($id);
                $output->add_message(t('success_saving_address','m_mailpage',$params));
            }
        } elseif (isset($records[$address_id])) { // OK, that is an existing record
            $where = array('mailpage_address_id' => $address_id);
            if (db_update($table,$fields,$where) === FALSE) {
                logger(sprintf('%s(): error updating address: %s',__FUNCTION__,db_errormessage()));
                $output->add_message(t('error_saving_data','m_mailpage'));
            } else {
                $output->add_message(t('success_saving_address','m_mailpage',$params));
            }
        } else { // non-existing record? Hmmm...
            $output->add_message(t('error_saving_data','m_mailpage'));
            logger(sprintf("%s(): cannot save address '%d' in node '%d'",
                           __FUNCTION__,$address_id,$node_id));
        }
        $edit_again = TRUE;
        return FALSE;
    }
    // 3 -- do we need to delete an address?
    $address_id = $delete_id;
    if (!is_null($address_id)) {
        $dialogdef = mailpage_get_dialogdef_delete($viewonly);
        if (!dialog_validate($dialogdef)) { // show errors to the user + do it again
            foreach($dialogdef as $k => $item) {
                if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                    $output->add_message($item['error_messages']);
                }
            }
            $edit_again = TRUE;
            return FALSE;
        }
        if (isset($records[$address_id])) { // OK, that is an existing record
            $params = array('{NAME}' => $records[$address_id]['name'],
                            '{EMAIL}' => $records[$address_id]['email'],
                            '{SORT_ORDER}' => $records[$address_id]['sort_order'],
                            '{ADDRESS_ID}' => strval($address_id));
            $table = 'mailpages_addresses';
            $where = array('mailpage_address_id' => $address_id);
            if (db_delete($table,$where) === FALSE) {
                logger(sprintf('%s(): error deleting address: %s',__FUNCTION__,db_errormessage()));
                $output->add_message(t('error_deleting_data','m_mailpage'));
            } else {
                $output->add_message(t('success_deleting_address','m_mailpage',$params));
            }
        } else { // non-existing record? Hmmm...
            $output->add_message(t('error_deleting_data','m_mailpage'));
            logger(sprintf("%s(): cannot delete address '%d' in node '%d'",
                           __FUNCTION__,$address_id,$node_id));
        }
        $edit_again = TRUE;
        return FALSE;
    }
    logger(sprintf("%s(): weird: no address to save and no address to delete. Hmmm...",__FUNCTION__));
    $edit_again = TRUE; // Make sure we return to show_edit; handle [Done] case there instead of here
    return FALSE; // Dirty trick to return to the destinations dialog
} // mailpage_save_destinations()


// ==================================================================
// =========================== WORKHORSES ===========================
// ==================================================================


/** validate the data entered by the user
 *
 * Other than the standard validation we check for at least 1 @-sign in the email field
 * As a side effect, the dialogdef is filled with the POST'ed values.
 *
 * @param object &$dialogdef defines the dialog and will hold POSTed values
 * @return bool TRUE if valid, else FALSE + messages added to $dialogdef
 */
function mailpage_dialog_validate_address(&$dialogdef) {
    $retval = TRUE; // assume success
    // 1 -- check the usual suspects (side-effect: fetch POSTed values in $dialogdef)
    if (!dialog_validate($dialogdef)) {
        $retval = FALSE;
    }
    // 2 -- special case: there must be at least 1 @-sign in the email field
    if (strpos($dialogdef['email']['value'],'@') === FALSE) { // no @-sign in mail address
        $retval = FALSE;
        ++$dialogdef['email']['errors'];
        $params = array('{FIELD}' => str_replace('~','',$dialogdef['email']['label']));
        $dialogdef['email']['error_messages'][] = t('validate_invalid','',$params);
    }
    return $retval;
} // mailpage_dialog_validate_address()


/** construct a dialog definition for the main mailpage configuration
 *
 * this constructs the dialogdef and also fills it with data from the DB (if any)
 *
 * @param object &$output collects the html output (if any)
 * @param int $viewonly if TRUE the Save button is not displayed and config values cannot be changed
 * @param int $node_id identifies the current mailpage
 * @return array dialog definition
 */
function mailpage_get_dialogdef_config(&$output, $viewonly, $node_id) {
    //
    // 1 -- generic configuration parameters
    //
    $table = 'mailpages';
    $fields = array('header','introduction', 'message');
    $where = array('node_id' => intval($node_id));
    if (($record = db_select_single_record($table, $fields, $where)) === FALSE) {
        logger(sprintf("%s(): error retrieving configuration '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        $output->add_message(t('error_retrieving_data','admin'));
        $header = '';
        $introduction = '';
        $message = '';
    } else {
        $header = $record['header'];
        $introduction = $record['introduction'];
        $message = $record['message'];
    }
    $dialogdef = array(
        'header' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'header',
            'minlength' => 0,
            'maxlength' => 240,
            'columns' => 40,
            'label' => t('header_label','m_mailpage'),
            'title' => t('header_title','m_mailpage'),
            'viewonly' => $viewonly,
            'value' => $header,
            'old_value' => $header,
            ),
        'introduction' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'introduction',
            'minlength' => 0,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 50,
            'rows' => 10,
            'label' => t('introduction_label','m_mailpage'),
            'title' => t('introduction_title','m_mailpage'),
            'viewonly' => $viewonly,
            'value' => $introduction,
            'old_value' => $introduction
            ),
        'message' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'message',
            'minlength' => 0,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 50,
            'rows' => 10,
            'label' => t('default_message_label','m_mailpage'),
            'title' => t('default_message_title','m_mailpage'),
            'viewonly' => $viewonly,
            'value' => $message,
            'old_value' => $message
            ),
        'csrftoken' => dialog_csrftoken()
        );
    //
    // 2 -- maybe add a save and a done button but always a cancel button
    //
    if (!$viewonly) {
        $dialogdef['button_save'] = dialog_buttondef(BUTTON_SAVE);
        $dialogdef['button_done'] = dialog_buttondef(BUTTON_DONE);
    }
    $dialogdef['button_cancel'] = dialog_buttondef(BUTTON_CANCEL);
    return $dialogdef;
} // mailpage_get_dialogdef_config()


/** construct a dialog definition for a mailpage destination address (add/edit)
 *
 * this constructs a dialog definition which handles adding and editing destination addresses.
 * In the case of adding a new address we start with empty fields (mostly), otherwise
 * we attempt to load the existing data from the DB.

 * @param object &$output collects the html output (if any)
 * @param int $viewonly if TRUE the Save button is not displayed and config values cannot be changed
 * @param int $node_id identifies the current mailpage
 * @param int $address_id identifies the address, 0 for a new one
 * @param int $sort_order is the next available sort order for new records
 * @return array dialog definition
 */
function mailpage_get_dialogdef_address(&$output, $viewonly, $node_id, $address_id, $sort_order=10) {
    //
    // 1 -- maybe get existing data from table
    //
    $values = array(
        'name'        => '',
        'sort_order'  => $sort_order,
        'email'       => '',
        'description' => '',
        'thankyou'    => '',
        'message'     => ''
        );
    if ($address_id > 0) {
        $table = 'mailpages_addresses';
        $keyfield = 'mailpage_address_id';
        $fields = '*';
        $where = array('node_id' => intval($node_id), $keyfield => intval($address_id));
        if (($record = db_select_single_record($table, $fields, $where)) === FALSE) {
            logger(sprintf('%s(): error retrieving configuration: %s',__FUNCTION__,db_errormessage()));
            $output->add_message(t('error_retrieving_data','admin'));
        } else {
            $values = $record;
        }
    }
    //
    // 2 -- define the dialog
    //
    $dialogdef = array(
        'name' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'name',
            'minlength' => 1,
            'maxlength' => 80,
            'columns' => 40,
            'label' => t('address_name_label','m_mailpage'),
            'title' => t('address_name_title','m_mailpage'),
            'viewonly' => $viewonly,
            'value' => $values['name'],
            'old_value' => $values['name'],
            ),
        'email' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'email',
            'minlength' => 1,
            'maxlength' => 255,
            'columns' => 40,
            'label' => t('address_email_label','m_mailpage'),
            'title' => t('address_email_title','m_mailpage'),
            'viewonly' => $viewonly,
            'value' => $values['email'],
            'old_value' => $values['email']
            ),
        'description' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'description',
            'minlength' => 0,
            'maxlength' => 240,
            'columns' => 40,
            'label' => t('address_description_label','m_mailpage'),
            'title' => t('address_description_title','m_mailpage'),
            'viewonly' => $viewonly,
            'value' => $values['description'],
            'old_value' => $values['description']
            ),
        'thankyou' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'thankyou',
            'minlength' => 0,
            'maxlength' => 240,
            'columns' => 40,
            'label' => t('address_thankyou_label','m_mailpage'),
            'title' => t('address_thankyou_title','m_mailpage'),
            'viewonly' => $viewonly,
            'value' => $values['thankyou'],
            'old_value' => $values['thankyou']
            ),
        'sort_order' => array(
            'type' => F_INTEGER,
            'name' => 'sort_order',
            'minlength' => 0,
            'maxlength' => 10,
            'columns' => 7,
            'minvalue' => 0,
            'maxvalue' => $sort_order,
            'label' => t('address_sort_order_label','m_mailpage'),
            'title' => t('address_sort_order_title','m_mailpage'),
            'viewonly' => $viewonly,
            'value' => $values['sort_order'],
            'old_value' => $values['sort_order']
            ),
        'csrftoken' => dialog_csrftoken()
        );
    //
    // 3 -- add buttons: save (mabye), cancel and delete (maybe)
    //
    if (!$viewonly) {
        $dialogdef['button_save'] = dialog_buttondef(BUTTON_SAVE);
        $dialogdef['button_done'] = dialog_buttondef(BUTTON_DONE);
    }
    $dialogdef['button_cancel'] = dialog_buttondef(BUTTON_CANCEL);
    return $dialogdef;
} // mailpage_get_dialogdef_address()


/** construct a dialog definition for deletion of a destination address
 *
 * this constructs a dialog definition which handles deleting of a destination address.
 * There is not much to do: the address_id to delete is conveyed via $_GET[].

 * @param int $viewonly if TRUE the Save button is not displayed and config values cannot be changed
 * @return array dialog definition
 */
function mailpage_get_dialogdef_delete($viewonly) {
    $dialogdef = array(
        'csrftoken' => dialog_csrftoken()
        );
    if (!$viewonly) {
        $dialogdef['button_delete'] = dialog_buttondef(BUTTON_DELETE);
    }
    $dialogdef['button_cancel'] = dialog_buttondef(BUTTON_CANCEL);
    return $dialogdef;
} // mailpage_get_dialogdef_delete()


/** retrieve current list of addresses in an array (could be empty)
 *
 * this retrieves the addresses associated with $node_id from the
 * database. As an important side effect all entries' sort order values
 * are renumbered. This means that we always have a neat set of sort
 * order numbers 10, 20, ...
 *
 * On error we return an empty array as a sort of best effort.
 * The error is logged though.
 *
 * @param int $node_id indicates page
 * @return array records from the database with sort order renumbered OR an empty array
 */
function mailpage_get_addresses($node_id) {
    $table = 'mailpages_addresses';
    $keyfield = 'mailpage_address_id';
    $fields = '*';
    $where = array('node_id' => intval($node_id));
    $order = array('sort_order',$keyfield);
    if (($records = db_select_all_records($table,$fields,$where,$order,$keyfield)) === FALSE) {
        logger(sprintf('%s(): error retrieving addresses: %s',__FUNCTION__,db_errormessage()));
        return array();
    }
    // side-effect: renumber the sort order of the addresses if not a nice list of 10, 20, ...
    $sort_order = 0;
    foreach($records as $k => $record) {
        $sort_order += 10;
        if ($record['sort_order'] == $sort_order) {
            continue;
        }
        $records[$k]['sort_order'] = $sort_order;
        $where = array($keyfield => intval($record[$keyfield]));
        $fields = array('sort_order' => $sort_order);
        if (db_update($table,$fields,$where) === FALSE) {
            logger(sprintf('%s(): error renumbering addresses: %s',__FUNCTION__,db_errormessage()));
        }
    }
    return $records;
} // mailpage_get_addresses()




/** construct a delete button for deletion of a  destination address
 *
 * this constructs a clickable icon/button that leads to the delete confirmation
 * dialog. We use the same code as does the Page Manager (picking up the icon from
 * the current skin et al). This makes that the icon automagically adapts to the
 * skin, eg. the textonly skin.

 * @param object &$output collects the html output (if any)
 * @param int $node_id identifies the current mailpage
 * @param int $viewonly if TRUE the Save button is not displayed and config values cannot be changed
 * @param int $address_id identifies the address, 0 for a new one
 * @param int $aparms contains values to substitute in translations, e.g. {ADDRESS_ID}
 * @return array dialog definition
 */
function mailpage_get_icon_delete(&$output,$node_id,$viewonly,$address_id,$aparams) {
    global $WAS_SCRIPT_NAME;
    // 1 -- prepare an icon (text or graphics)
    $title = t(($viewonly) ? 'icon_delete_access_denied' : 'icon_delete','m_mailpage',$aparams);
    $alt = t('icon_delete_alt','m_mailpage');
    $text = t('icon_delete_text','m_mailpage');
    $anchor = $output->skin->get_icon('delete',$title,$alt,$text);

    // 2 -- construct the A tag using the keyword 'delete' instead of 'address' to address the address
    $a_params = array('job' => JOB_PAGEMANAGER,
                      'task' => TASK_NODE_EDIT_CONTENT,
                      'node' => $node_id,
                      'option' => 'destinations',
                      'delete' => strval($address_id));
    $a_attr = array('title' => $title);
    if ($viewonly) {
        $a_attr['class'] = 'dimmed';
    }
    return html_a($WAS_SCRIPT_NAME,$a_params,$a_attr,$anchor);
} // mailpage_get_icon_delete()

?>