<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/mailpage/mailpage_install.php - installer of the mailpage module
 *
 * This file contains the mailpage module installer.
 * The interface consists of these functions:
 *
 * <code>
 * mailpage_install(&$messages,$module_id)
 * mailpage_upgrade(&$messages,$module_id)
 * mailpage_uninstall(&$messages,$module_id)
 * mailpage_demodata(&$messages,$module_id,&$config,$manifest)
 * </code>
 *
 * These functions can be called from the main installer and/or admin.php.
 *
 * Note
 * You cannot be sure about the environment from which these routines are called.
 * If the caller is the Install Wizard, you do not have all subroutines available.
 * However, it IS possible to manipulate the database via the db_*() routines and/or
 * the global $DB object. Therefore you have to keep the install routine extremly
 * simple. You also have no option to interact with the user; the install has to
 * be a silent install; you can only indicate success (TRUE) or failure (FALSE)
 * and maybe an error message in $messages[] but that's it. Good luck.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_mailpage
 * @version $Id: mailpage_install.php,v 1.6 2016/03/23 09:28:23 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }


/** install the module
 *
 * this routine installs the module. For this module
 * there is nothing to install, so we simply return success.
 *
 * Note that the record for this module is already created in the
 * modules table; the pkey is $module_id.
 *
 * @param array &$messages collects the (error) messages
 * @param int $module_id the key for this module in the modules table
 * @return bool TRUE on success + output via $messages, FALSE otherwise
 */
function mailpage_install(&$messages,$module_id) {
    return TRUE;
} // mailpage_install()


/** upgrade the module
 *
 * this routine performs an upgrade to the installed module.
 * This updates the list of submenu options in the module record.
 * The submenu-feature was added 2014-05-05 and appears in release
 * 0.90.6.
 *
 * @param array &$messages collects the (error) messages
 * @param int $module_id the key for this module in the modules table
 * @return bool TRUE on success + output via $messages, FALSE otherwise
 */
function mailpage_upgrade(&$messages,$module_id) {
    $version = 2014050600;
    $table = 'modules';
    $fields = array('options' => 'configuration,destinations', 'version' => $version);
    $where = sprintf('(module_id = %d) AND (version < %d)', $module_id, $version);
    if (($retval = db_update($table,$fields,$where)) === FALSE) {
        $msg = sprintf("%s(): cannot set field 'options' to %s: %s",
                       __FUNCTION__, $fields['options'], db_errormessage());
        $retval = FALSE;
    } elseif ($retval != 1) {
        $msg = sprintf("%s(): field 'options' not updated, version already greater than/equal to %d",
                       __FUNCTION__, $fields['version']);
        $retval = TRUE;
    } else {
        $msg = sprintf("%s(): success setting field 'options' to '%s'",__FUNCTION__,$fields['options']);
        $retval = TRUE;
    }
    logger($msg);
    $messages[] = htmlspecialchars($msg);
    return $retval;
} // mailpage_upgrade()


/** uninstall the module
 *
 * this is a hook for future extensions of Website@School.
 * For now we simply return success.
 * 
 * @param array &$messages collects the (error) messages
 * @param int $module_id the key for this module in the modules table
 * @return bool TRUE on success + output via $messages, FALSE otherwise
 */
function mailpage_uninstall(&$messages,$module_id) {
    return TRUE;
} // mailpage_uninstall()


/** add demonstration data to the system
 *
 * this routine is a no-op because there is no mailpage demodata
 * and if it is there, it is already created in the main
 * demodata-routine in /program/install/demodata.php.
 *
 * This routine is retained here as an example and also because a routine
 * by this name should exist (even if it does nothing).
 *
 * Note
 * If the module is installed via the Install Wizard, this routine is
 * called. However, if a module is installed as an additional module
 * after installation, the {$mode}_demodata() routine is never called.
 * This is because the only time you know that demodata is installed is
 * when the Install Wizard runs. If we're called from admin.php, the
 * webmaster may have already deleted existing (core) demodata so you
 * never can be sure what to expect. To put it another way: it is hard
 * to re-construct $config when we're NOT the Instal Wizard.
 *
 * @param array &$messages collects the (error) messages
 * @param int $module_id the key for this module in the modules table
 * @param array &$config pertinent data for the new website + demodata foundation
 * @param array $manifest a copy from the manifest for this module
 * @return bool TRUE on success + output via $messages, FALSE otherwise
 */
function mailpage_demodata(&$messages,$module_id,&$config,$manifest) {
    return TRUE;
} // mailpage_demodata()

?>