<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/mypage/mypage_view.php - interface to the view-part of the mypage module
 *
 * This file defines the interface with the mypage-module for viewing content.
 * The interface consists of this function:
 *
 * <code>
 * mypage_view(&$output,$area_id,$node_id,$module)
 * </code>
 *
 * This function is called from /index.php when the node to display is connected
 * to this module.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_mypage
 * @version $Id: mypage_view.php,v 1.4 2016/06/21 06:16:58 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }


/** display the content mypage linked to node $node_id
 *
 * there are 4 different views:
 * one if the user is not yet logged in:
 *<pre>
 * 1. a login dialogue (see {@link mypage_view_login()})
 *</pre>
 * and three when the user is logged in:
 *<pre>
 * 2. an overview (see {@link mypage_view_home()}),
 * 3. a change profile dialogue (see {@link mypage_view_profile()}), and
 * 4. a change password dialogue (see {@link mypage_view_password()}).
 *</pre>
 *
 * The idea is to present the user with a list of handy links,
 * first links to managing the user's profile, then (if the
 * user has privileges) links to admin.php and finally a list
 * of areas that are accessible for the user (now she is logged in).
 *
 * @param object &$theme collects the (html) output
 * @param int $area_id identifies the area where $node_id lives
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function mypage_view(&$theme,$area_id,$node_id,$module) {
    global $USER;
    if ($USER->is_logged_in) {
        if (!is_null($value = get_parameter_int('profile'))) {
            return mypage_view_profile($theme,$area_id,$node_id);
        } elseif (!is_null($value = get_parameter_int('password'))) {
            return mypage_view_password($theme,$area_id,$node_id);
        } else {
            return mypage_view_home($theme,$area_id,$node_id);
        }
    } else {
        return mypage_view_login($theme,$area_id,$node_id);
    }
} // mypage_view()


/** show a login dialog 
 *
 * build on loginlib: show login dialog. Actual processing
 * is done in main_index() but eventually we end up at the
 * node node_id.
 *
 * @param object &$theme collects the (html) output
 * @param int $node_id the node to which this module is connected
 * @param int $area_id identifies the area where $node_id lives
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function mypage_view_login(&$theme,$area_id,$node_id) {
    global $CFG;
    include_once($CFG->progdir.'/lib/loginlib.php');
    $dialogdef = login_dialogdef(LOGIN_PROCEDURE_NORMAL);
    $a_params = array('login' => '1');
    $href = was_node_url($theme->node_record,$a_params);
    $theme->add_content(html_tag('h2',NULL,t('login','loginlib')));
    $theme->add_content(t('please_enter_username_password','loginlib'));
    $theme->add_content(dialog_quickform($href,$dialogdef));
    return TRUE;
} // mypage_view_login()


/** show links to profile, admin.php and available areas
 *
 * this personal page contains links to places the user is allowed to go.
 *
 * @param object &$theme collects the (html) output
 * @param int $area_id identifies the area where $node_id lives
 * @param int $node_id the node to which this module is connected
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function mypage_view_home(&$theme,$area_id,$node_id) {
    global $CFG,$USER,$WAS_SCRIPT_NAME;
    $preview = $theme->preview_mode;

    // 1 -- Profile, password & logout
    $theme->add_content('<h3>'.t('my_profile','m_mypage').'</h3>');
    $theme->add_content('<ul>');
    $href = was_node_url($theme->node_record,array('profile' => '1'),'',$preview);
    $theme->add_content('  <li>'.html_a($href,NULL,NULL,t('my_profile_edit','m_mypage')));
    $href = was_node_url($theme->node_record,array('password' => '1'),'',$preview);
    $theme->add_content('  <li>'.html_a($href,NULL,NULL,t('my_password_edit','m_mypage')));
    $a_params = array('{USERNAME}' => htmlspecialchars($USER->username),
                      '{FULL_NAME}' => htmlspecialchars($USER->full_name));
    $href = was_node_url($theme->node_record,array('logout' => '1'),'',$preview);
    $theme->add_content('  <li>'.html_a($href,NULL,NULL,t('logout_username','m_mypage',$a_params)));
    $theme->add_content('</ul>');

    // 2 -- Website administration
    if ($USER->is_admin()) {
        $href = ($preview) ? '#' : $CFG->www_short.'/admin.php';
        $theme->add_content('<h3>'.t('website_administration','m_mypage').'</h3>');
        $theme->add_content('<ul>');
        $params = array('job' => 'start');
        $anchor = t('description_startcenter','admin'); // Website@School Startcenter
        $theme->add_content('  <li>'.html_a($href,$params,NULL,$anchor));
        if ($USER->has_job_permissions(JOB_PERMISSION_PAGEMANAGER)) {
            $params = array('job' => 'pagemanager');
            $anchor = t('description_pagemanager','admin'); // Page Manager
            $theme->add_content('  <li>'.html_a($href,$params,NULL,$anchor));
        }
        if ($USER->has_job_permissions(JOB_PERMISSION_FILEMANAGER)) {
            $params = array('job' => 'filemanager');
            $anchor = t('description_filemanager','admin'); // File Manager
            $theme->add_content('  <li>'.html_a($href,$params,NULL,$anchor));
        }
        $href = $CFG->progwww_short.'/manual.php';
        $params = array('language' => $USER->language_key);
        $attributes = array('target' => '_blank');
        $anchor = t('description_help','admin'); // Help
        $theme->add_content('  <li>'.html_a($href,$params,$attributes,$anchor));
        $theme->add_content('</ul>');
    }

    // 3 -- Available areas
    if (($areas = get_area_records()) !== FALSE) {
        $theme->add_content('<h3>'.t('available_areas','m_mypage').'</h3>');
        $theme->add_content('<ul>');
        foreach($areas as $id => $area) {
            $public = db_bool_is(FALSE,$area['is_private']);
            if ((db_bool_is(TRUE,$area['is_active'])) &&
                (($public) || ($USER->has_intranet_permissions(ACL_ROLE_INTRANET_ACCESS,$id)))) {
                     $a_params = array('{AREA_FULL_NAME}' => $area['title'], 
                                       '{AREA}' => $id,'{SORT_ORDER}' => $area['sort_order']);
                     if ($public) {
                         $anchor = t('public_area_name','m_mypage',$a_params);
                         $attributes = array('title' => t('public_area_title','m_mypage',$a_params));
                     } else {
                         $anchor = t('private_area_name','m_mypage',$a_params);
                         $attributes = array('title' => t('private_area_title','m_mypage',$a_params));
                     }
                     $params = array('area' => $id);
                     $href = was_node_url(NULL,$params,$area['title'],$preview);
                     $theme->add_content('  <li>'.html_a($href,NULL,$attributes,$anchor));
                 }
        }
        $theme->add_content('</ul>');
    }
    return TRUE; // indicate success
} // mypage_view_home()


/** show/process profile dialogue
 *
 * display a dialogue where the user can change a few properties of
 * her own account. Notably missing: password (see see {@link mypage_view_password()}).
 * Note that currently the username and the also the directory are
 * not editable. The information is displayed, however.
 *
 * Processing is also done via this routine. The dialog ends
 * with three submit buttons: Save, Done and Cancel. Save and Done
 * store the new values in the user record in the database AND in
 * the global $USER object too. Save then shows the dialogue again,
 * while Done returns to the overview. Cancel always returns to the
 * overview without changing anything.
 *
 * @param object &$theme collects the (html) output
 * @param int $area_id identifies the area where $node_id lives
 * @param int $node_id the node to which this module is connected
 * @return bool TRUE on success, FALSE otherwise
 */
function mypage_view_profile(&$theme,$area_id, $node_id) {
    global $USER,$CFG;
    include_once($CFG->progdir.'/lib/loginlib.php');

    // 1 -- cancel?
    if (isset($_POST['button_cancel'])) {
        $theme->add_message(t('cancelled','admin'));
        return mypage_view_home($theme,$area_id,$node_id);
    }
    // 2 -- save?
    $dialogdef = mypage_profile_get_dialogdef();
    if ((isset($_POST['button_save'])) || (isset($_POST['button_done']))) {
        $user_id = intval($USER->user_id);
        if (mypage_profile_dialog_validate($dialogdef,$user_id)) {
            if (mypage_profile_save($theme,$dialogdef,$user_id)) {
                if (isset($_POST['button_done'])) {
                    return mypage_view_home($theme,$area_id,$node_id);
                }
            }
        } else {
            // prepare for redo w/errors
            foreach($dialogdef as $k => $item) {
                if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                    $theme->add_message($item['error_messages']);
                }
            }
        }
    }

    // 3 -- show dialogue (maybe again w/errors)
    $a_params = array('{USERNAME}' => htmlspecialchars($USER->username),
                      '{FULL_NAME}' => htmlspecialchars($USER->full_name));
    $href = was_node_url($theme->node_record,array('profile' => 1));
    $theme->add_content(html_tag('h2',NULL,t('profile_header','m_mypage',$a_params)));
    $theme->add_content(t('profile_explanation','m_mypage',$a_params));
    $theme->add_content(dialog_quickform($href,$dialogdef));
    $href = was_node_url($theme->node_record,array('profile' => '1'),'',$theme->preview_mode);
    $theme->breadcrumb_addendum[] = html_a($href,NULL,NULL,t('my_profile_edit','m_mypage'));
    return TRUE; // indicate success
} // mypage_view_profile()


/** show/process password dialogue
 *
 * this dialogue requires both the old and the new password.
 * the username is implied (because the user is logged in)
 * The extra authentication is necessary to prevent a passer
 * by to change a password on an unguarded screen.
 *
 * The logic applied is re-used from see {@link loginlib.php},
 * including the blacklist mechanism; it is therefore not
 * possible to brute force a new password.
 *
 * We have only 1 submit button for save: we always return
 * to the overview screen after changing the password.
 *
 * Note that the new password has to be entered twice. This
 * excercises the muscle memory for better remembering the
 * password.
 *
 * @param object &$theme collects the (html) output
 * @param int $area_id identifies the area where $node_id lives
 * @param int $node_id the node to which this module is connected
 * @return bool TRUE on success, FALSE otherwise
 * @uses loginlib.php
 * @uses login_change_password()
 */
function mypage_view_password(&$theme,$area_id,$node_id) {
    global $USER, $CFG;
    include_once($CFG->progdir.'/lib/loginlib.php');

    // 1 -- cancel?
    if (isset($_POST['button_cancel'])) {
        $theme->add_message(t('cancelled','admin'));
        return mypage_view_home($theme,$area_id,$node_id);
    }
    // 2 -- save?
    $dialogdef = mypage_password_get_dialogdef();
    if (isset($_POST['button_ok'])) {
        if (mypage_password_dialog_validate($dialogdef)) {
            if (login_change_password(intval($USER->user_id),$dialogdef['password1']['value'])) {
                $theme->add_message(t('password_changed','loginlib'));
            } else {
                $theme->add_message(t('password_error','m_mypage'));
            }
            // return to home even when save fails
            return mypage_view_home($theme,$area_id,$node_id);
        } else {
            // redo the dialog w/errors
            foreach($dialogdef as $k => $item) {
                if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                    $theme->add_message($item['error_messages']);
                }
            }
        }
    }
    // 3 -- show (again)
    $a_params = array('{USERNAME}' => htmlspecialchars($USER->username),
                      '{FULL_NAME}' => htmlspecialchars($USER->full_name));
    $href = was_node_url($theme->node_record,array('password' => 1));
    $theme->add_content(html_tag('h2',NULL,t('password_header','m_mypage',$a_params)));
    $theme->add_content(t('password_explanation','m_mypage',$a_params));
    $theme->add_content(dialog_quickform($href,$dialogdef));
    $href = was_node_url($theme->node_record,array('password' => '1'),'',$theme->preview_mode);
    $theme->breadcrumb_addendum[] = html_a($href,NULL,NULL,t('my_password_edit','m_mypage'));
    return TRUE; // indicate success
} // mypage_view_password()


/* ======================================================================
 * WORKERS
 *======================================================================
 */

/** construct the edit dialog for the profile of the current user 
 *
 * user must reauthenticate to let changes take effect hence the password field
 *
 * @return bool|array FALSE on error or array with dialog definition and data from $USER
 * @uses $LANGUAGE
 * @uses $USER
 * @todo should we remove username and datadir altogether?
 */
function mypage_profile_get_dialogdef() {
    global $LANGUAGE,$USER;

    // add the language key to the language name for easier selection
    $options_languages = $LANGUAGE->get_active_language_names();
    foreach($options_languages as $k => $v) {
        $options_languages[$k] = $v.' ('.$k.')';
    }

    // construct dialog definition including current values from $USER
    $dialogdef = array(
        'full_name' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'full_name',
            'minlength' => 1,
            'maxlength' => 255,
            'columns' => 50,
            'label' => t('usermanager_edit_user_fullname_label','admin'),
            'title' => t('usermanager_edit_user_fullname_title','admin'),
            'value' => $USER->full_name,
            'old_value' => $USER->full_name
            ),
        'email' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'email',
            'minlength' => 0,
            'maxlength' => 255,
            'columns' => 50,
            'label' => t('usermanager_edit_user_email_label','admin'),
            'title' => t('usermanager_edit_user_email_title','admin'),
            'value' => $USER->email,
            'old_value' => $USER->email
            ),
        'redirect' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'redirect',
            'minlength' => 0,
            'maxlength' => 255,
            'columns' => 50,
            'label' => t('usermanager_edit_user_redirect_label','admin'),
            'title' => t('usermanager_edit_user_redirect_title','admin'),
            'value' => $USER->redirect,
            'old_value' => $USER->redirect
            ),
        'language_key' => array(
            'type' => F_LISTBOX,
            'name' => 'language_key',
            'options' => $options_languages,
            'label' => t('usermanager_edit_user_language_label','admin'),
            'title' => t('usermanager_edit_user_language_title','admin'),
            'value' => $USER->language_key,
            'old_value' => $USER->language_key
            ),
        'editor' => array(
            'type' => F_LISTBOX,
            'name' => 'editor',
            'options' => get_editor_names(),
            'label' => t('usermanager_edit_user_editor_label','admin'),
            'title' => t('usermanager_edit_user_editor_title','admin'),
            'value' => $USER->editor,
            'old_value' => $USER->editor
            ),
        'skin' => array(
            'type' => F_LISTBOX,
            'name' => 'skin',
            'options' => get_skin_names(),
            'label' => t('usermanager_edit_user_skin_label','admin'),
            'title' => t('usermanager_edit_user_skin_title','admin'),
            'value' => $USER->skin,
            'old_value' => $USER->skin
            ),
        'password' => array( // used for transaction authentication
            'type' => F_PASSWORD,
            'name' => 'password',
            'minlength' => 1,
            'maxlength' => 255,
            'columns' => 30,
            'label' => t('password_label','m_mypage'),
            'title' => t('password_title','m_mypage'),
            'value' => '',
            ),
        'csrftoken' => dialog_csrftoken(),
        'button_save' => dialog_buttondef(BUTTON_SAVE),
        'button_done' => dialog_buttondef(BUTTON_DONE),
        'button_cancel' => dialog_buttondef(BUTTON_CANCEL)
        );
    return $dialogdef;
} // mypage_profile_get_dialogdef()


/** validate entries in user profile dialogue
 *
 * exta validation:  email address should contain at least one '@' (very crude)
 * Also the user MUST reauthorise to let the changes have effect
 *
 * @param object &$dialogdef defines the dialog and will hold POSTed values
 * @param int $user_id user to check
 * @return bool TRUE if valid, else FALSE + messages added to dialogdef
 */
function mypage_profile_dialog_validate(&$dialogdef,$user_id) {
    $retval = TRUE;

    // 1 -- first the generic tests
    if (!dialog_validate($dialogdef)) {
        $retval = FALSE;
    }

    // 2 -- see if this change is authorised with the user's password
    if (mypage_authorisation($dialogdef) === FALSE) {
        return FALSE;
    }

    // 3 -- additional check: email
    $email = $dialogdef['email']['value'] = trim(utf8_strtolower($dialogdef['email']['value']));
    if ((strlen($email) > 0) && (strpos($email,'@') === FALSE)) { // no @-sign in mail address
        $retval = FALSE;
        ++$dialogdef['email']['errors'];
        $params = array('{FIELD}' => str_replace('~','',$dialogdef['email']['label']));
        $dialogdef['email']['error_messages'][] = t('validate_invalid','',$params);
    }
    return $retval;
} // mypage_profile_dialog_validate()


/** save modified user profile in database and keep a copy in core
 *
 * save the modified data to database and also in core (in $USER) and
 * even in $_SESSION for a few selected parameters.
 * Note that the username is always updated, even if the field is
 * read/only in the dialogue. The source for that data is $USER so there should ne
 * no harm.
 *
 * @param object &$theme collects the (html) output
 * @param object &$dialogdef defines the dialog and holds POSTed values to save
 * @param int $user_id user to check
 * @return bool TRUE on success, FALSE on error
 */
function mypage_profile_save(&$theme,&$dialogdef,$user_id) {
    global $USER;
    $retval = TRUE;
    $fields = array(
        'full_name'    => $dialogdef['full_name']['value'],
        'email'        => $dialogdef['email']['value'],
        'redirect'     => $dialogdef['redirect']['value'],
        'language_key' => $dialogdef['language_key']['value'],
        'editor'       => $dialogdef['editor']['value'],
        'skin'         => $dialogdef['skin']['value']
        );
    $params = array('{USERNAME}' => htmlspecialchars($USER->username),
                    '{FULL_NAME}' => htmlspecialchars($dialogdef['full_name']['value']));
    $retval = db_update('users',$fields,array('user_id' => $user_id));
    if ($retval === FALSE) {
        $theme->add_message(t('usermanager_save_user_failure','admin',$params));
        logger(sprintf("%s(): error saving data user '%d': %s",
                       __FUNCTION__,$user_id,db_errormessage()));
    } else {
        $retval = TRUE;
        $theme->add_message(t('usermanager_save_user_success','admin',$params));
        logger(sprintf("%s(): success saving changes to '%d' in 'users'",
                       __FUNCTION__,$user_id),WLOG_DEBUG);
        // Update data already in core
        foreach($fields as $k => $v) {
            $USER->$k = $v;
        }
        $_SESSION['redirect'] = $fields['redirect'];
        $_SESSION['skin'] = $fields['skin'];
    }
    return $retval;
} // mypage_profile_save()


/** construct the edit dialog for the password of the current user 
 *
 * heavily leans on see {@link loginlib.php}.
 *
 * @return bool|array FALSE on error or array with dialog definition and existing data from database
 */
function mypage_password_get_dialogdef() {
    $params = array('{MIN_LENGTH}' => MINIMUM_PASSWORD_LENGTH,
                    '{MIN_LOWER}' => MINIMUM_PASSWORD_LOWERCASE,
                    '{MIN_UPPER}' => MINIMUM_PASSWORD_UPPERCASE,
                    '{MIN_DIGIT}' => MINIMUM_PASSWORD_DIGITS);
    $dialogdef = array(
        'password' => array( // used for transaction authentication
            'type' => F_PASSWORD,
            'name' => 'password',
            'minlength' => 1,
            'maxlength' => 255,
            'columns' => 30,
            'label' => t('old_password_label','m_mypage'),
            'title' => t('old_password_title','m_mypage'),
            'value' => '',
            ),
        'password1' => array(
            'type' => F_PASSWORD,
            'name' => 'password1',
            'minlength' => 0,
            'maxlength' => 255,
            'columns' => 30,
            'label' => t('new_password1_label','m_mypage'),
            'title' => t('usermanager_edit_user_password1_title','admin',$params),
            'value' => '',
            ),
        'password2' => array(
            'type' => F_PASSWORD,
            'name' => 'password2',
            'minlength' => 0,
            'maxlength' => 255,
            'columns' => 30,
            'label' => t('new_password2_label','m_mypage'),
            'title' => t('usermanager_edit_user_password2_title','admin',$params),
            'value' => '',
            ),
        'csrftoken' => dialog_csrftoken(),
        'button_ok' => dialog_buttondef(BUTTON_OK),
        'button_cancel' => dialog_buttondef(BUTTON_CANCEL)
        );
    return $dialogdef;
} // mypage_password_get_dialogdef()


/** validate entries in user password dialogue
 *
 * @param object &$dialogdef defines the dialog and will hold POSTed values
 * @return bool TRUE if valid, else FALSE + messages added to dialogdef
 * @uses mypage_authorisation()
 * @uses acceptable_new_password()
 */
function mypage_password_dialog_validate(&$dialogdef) {
    $retval = TRUE;

    // 1 -- first the generic tests + get data from POST
    if (!dialog_validate($dialogdef)) {
        $retval = FALSE;
    }
    // 2 -- see if this change is authorised with the user's password
    if (mypage_authorisation($dialogdef) === FALSE) {
        $dialogdef['password1']['value'] = '';
        $dialogdef['password2']['value'] = '';
        return FALSE;
    }

    // 3 -- additional check: new passwords
    $password1 = $dialogdef['password1']['value'];
    $password2 = $dialogdef['password2']['value'];
    if ($password1 != $password2) {
        $params = array('{FIELD1}' => str_replace('~','',$dialogdef['password1']['label']),
                        '{FIELD2}' => str_replace('~','',$dialogdef['password2']['label']));
        ++$dialogdef['password1']['errors'];
        ++$dialogdef['password2']['errors'];
        $dialogdef['password1']['error_messages'][] = t('validate_different_passwords','',$params);
        $dialogdef['password1']['value'] = '';
        $dialogdef['password2']['value'] = '';
        $retval = FALSE;
    } elseif (!acceptable_new_password($password1,$password2)) {
        $params = array('{MIN_LENGTH}' => MINIMUM_PASSWORD_LENGTH,
                        '{MIN_LOWER}' => MINIMUM_PASSWORD_LOWERCASE,
                        '{MIN_UPPER}' => MINIMUM_PASSWORD_UPPERCASE,
                        '{MIN_DIGIT}' => MINIMUM_PASSWORD_DIGITS,
                        '{FIELD}'=> str_replace('~','',$dialogdef['password1']['label']));
        ++$dialogdef['password1']['errors'];
        $dialogdef['password1']['error_messages'][] = t('validate_bad_password','',$params);
        $params['{FIELD}'] = str_replace('~','',$dialogdef['password2']['label']);
        ++$dialogdef['password2']['errors'];
        $dialogdef['password1']['value'] = '';
        $dialogdef['password2']['value'] = '';
        $retval = FALSE;
    }
    return $retval;
} // mypage_password_dialog_validate()


/** check the authorisation for a submitted dialogue
 *
 * check if the remote caller is not blacklisted and see
 * if $USER->username and $dialogdef['password']['value']
 * match.
 * 
 * @param object &$dialogdef defines the dialog and will hold POSTed values
 * @return bool TRUE if authorised, else FALSE + messages added to dialogdef
 * @uses login_is_blacklisted()
 * @uses authenticate_user()
 * @uses login_failure_delay()
 * @uses login_failure_blacklist_address()
 * @uses acceptable_new_password()
 * @uses login_failure_reset()
 */
function mypage_authorisation(&$dialogdef) {
    global $CFG,$USER;

    // 1 -- no access for blacklisted addresses
    $remote_addr = $_SERVER['REMOTE_ADDR'];
    if (login_is_blacklisted($remote_addr)) {
        ++$dialogdef['password']['errors'];
        $dialogdef['password']['error_messages'][] = t('access_denied','loginlib');
        return FALSE; // Fatal, return immediately
    }

    // 2 -- user must authenticate again
    $username = $USER->username;
    $password = $dialogdef['password']['value'];
    $dialogdef['password']['value'] = ''; // make sure user retypes password in next round
    if (($user = authenticate_user(BY_PASSWORD,$username,$password)) === FALSE) {
        // Invalid credentials; pretend we're busy (slow user down), increment failure count...
        $blacklist_seconds = 60 * intval($CFG->login_blacklist_interval);
        ++$dialogdef['password']['errors'];
        $failure_count = login_failure_increment($remote_addr,LOGIN_PROCEDURE_NORMAL,$username);
        login_failure_delay($remote_addr);
        if ($failure_count < intval($CFG->login_max_failures)) {
            $dialogdef['password']['error_messages'][] = t('invalid_credentials_please_retry','loginlib');
        } elseif ($failure_count > intval($CFG->login_max_failures)) {
            login_failure_blacklist_address($remote_addr,$blacklist_seconds,$username);
            $dialogdef['password']['error_messages'][]=t('contact_webmaster_for_new_password','loginlib');
        } else {
            $dialogdef['password']['error_messages'][] = t('access_denied','loginlib');
        }
        return FALSE; // Fatal, return immediately
    }

    // 3 -- password was correct, forget previous failures (if any)
    login_failure_reset($remote_addr);
    return TRUE;
} // mypage_authorisation()

?>