<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/newsletter/newsletter_common.php - code shared between admin and view
 *
 * This file defines various constants and subroutines used from both
 * {@link newsletter_admin.php} and {@link newsletter_view.php}.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_newsletter
 * @version $Id: newsletter_common.php,v 1.18 2016/03/23 09:28:24 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }

define('NEWSLETTER_UNPUBLISHED',  0); // this issue is unpublished
define('NEWSLETTER_FULLMAIL_ONLY',1); // only send full mail, no archive visible on site
define('NEWSLETTER_ALERT_SITE',   2); // send mail alert, lure subscribers to newsletter on site
define('NEWSLETTER_FULLMAIL_SITE',3); // send full mail and also make archive visible on site

define('NEWSLETTER_CONTRIBUTION_POLICY_NO',   0); // No contributions to the next newsletter accepted
define('NEWSLETTER_CONTRIBUTION_POLICY_USERS',1); // Only registered users can contribute
define('NEWSLETTER_CONTRIBUTION_POLICY_YES',  2); // Anyone can contribute an article to the newsletter

define('NEWSLETTER_SUBSCRIPTION_POLICY_NONE',0); // No self-service subscriptions possible
define('NEWSLETTER_SUBSCRIPTION_POLICY_ONE', 1); // One-pass subscription policy; only email validation
define('NEWSLETTER_SUBSCRIPTION_POLICY_TWO', 2); // Two-pass subscription policy; applications are approved

// Note: do NOT change actual values of definitions below; we use the numerical value in subscriber upload
define('NEWSLETTER_SUBSCRIPTION_SKIP',     0); // Skip this address in upload (same value as 'new')
define('NEWSLETTER_SUBSCRIPTION_NEW',      0); // New subscription; waiting for email validation
define('NEWSLETTER_SUBSCRIPTION_CONFIRMED',1); // Email was validated; waiting for admin approval
define('NEWSLETTER_SUBSCRIPTION_APPROVED', 2); // Subscription is approved; subscriber receives newsletters
define('NEWSLETTER_SUBSCRIPTION_BLACKLIST',3); // Address is on the blacklist; no newsletters at all


define('NEWSLETTER_CHORE_ADD',     'add');     // Add an article to the current issue
define('NEWSLETTER_CHORE_DELETE',  'delete');  // Delete an article from the list
define('NEWSLETTER_CHORE_DESELECT','deselect');// Postpone article to a later issue
define('NEWSLETTER_CHORE_SELECT',  'select');  // Select article for the current issue
define('NEWSLETTER_CHORE_UP',      'up');      // Move article up in the list
define('NEWSLETTER_CHORE_DOWN',    'down');    // Move articel down in the list
define('NEWSLETTER_CHORE_EDIT',    'edit');    // Edit existing article
define('NEWSLETTER_CHORE_PREVIEW', 'preview'); // Show a preview of the newsletter being composed
define('NEWSLETTER_CHORE_TESTMAIL','testmail');// Send an example of the newsletter to newsletter reply=to
define('NEWSLETTER_CHORE_LIST',    'list');    // Show a list of records
define('NEWSLETTER_CHORE_DOWNLOAD','download');// Download a CSV list of subscribers 
define('NEWSLETTER_CHORE_UPLOAD',  'upload'  );// Upload a CSV list of subscribers 

define('NEWSLETTER_CONFIRM_ADD', 'A'); // Action code for adding a subscriber ('subscribe')
define('NEWSLETTER_CONFIRM_DEL', 'D'); // Action code for deleting a subscriber ('unsubscribe')
define('NEWSLETTER_CONFIRM_LEN', 12);  // Length of quasi-random part in subscriber confirm


// ==================================================================
// =================== NEWSLETTER UTILITY ROUTINES ==================
// ==================================================================


/** create a neat array with email-addresses from a (possibly messy) text
 *
 * this routine converts a (messy) list of email addresses to an array with
 * a single address per array element. If there are no addresses, an empty
 * array is returned. The addresses are supposed to be comma-delimited and/or
 * newline and/or space delimited. Note that we expect only plain email
 * addresses, no readable names etc. (KISS)
 *
 * @param string $buffer holds the messy list of email addresses
 * @return array with one email address per element (or an empty array)
 */
function newsletter_get_emails($buffer) {
    $s = trim(preg_replace("/[\s,]+/","\n",$buffer));
    $a = (empty($s)) ? array() : explode("\n",$s);
    return $a;
} // newsletter_get_emails()

/** retrieve the newsletter configuration
 *
 * @param int $node_id the node to which the content is connected
 * @return bool|array FALSE on error, otherwise the configuration from table
 */
function newsletter_config($node_id) {
    $table = 'newsletters';
    $fields = '*';
    $where = array('newsletter_id' => intval($node_id));
    if (($config = db_select_single_record($table, $fields, $where)) === FALSE) {
        logger(sprintf('%s(): error retrieving configuration for newsletter %d: %s',
                       __FUNCTION__,$node_id,db_errormessage()));
    }
    return $config;
} // newsletter_config()


/** validate the data entered by the (un)subscriber
 *
 * Apart from the standard validation we check for at least 1 @-sign in
 * the email address. (It is non-trivial to validate an email address,
 * it may not even be possible to do it in a single regular expression.
 * See http://stackoverflow.com/questions/201323 for more information.)
 *
 * As an added bonus the email address is trimmed and standardised in
 * lowercase.
 *
 * Note that this routine is used both in {@link newsletter_view.php} 
 * where visitors can subscribe/unsubscribe themselves and
 * {@link newsletter_admin.php} where the administrator is responsible. 
 *
 * @param object &$dialogdef defines the dialog and will hold POSTed values
 * @return bool TRUE if valid, else FALSE + messages added to dialogdef
 */
function newsletter_dialog_validate(&$dialogdef) {
    $retval = dialog_validate($dialogdef);
    if (isset($dialogdef['email'])) {
        if ((!isset($dialogdef['email']['viewonly'])) || (!($dialogdef['email']['viewonly']))) {
            // massage email: no trailing/leading spaces and always lowercase
            $email = $dialogdef['email']['value'] = trim(utf8_strtolower($dialogdef['email']['value']));
            if ((strlen($email) > 0) && (strpos($email,'@') === FALSE)) { // no @-sign in mail address
                $retval = FALSE;
                if (!isset($dialogdef['email']['errors'])) {
                    $dialogdef['email']['errors'] = 0;
                }
                if (!isset($dialogdef['email']['error_messages'])) {
                    $dialogdef['email']['error_messages'] = array();
                }
                ++$dialogdef['email']['errors'];
                $params = array('{FIELD}' => str_replace('~','',$dialogdef['email']['label']));
                $dialogdef['email']['error_messages'][] = t('validate_invalid','',$params);
            }
        }
    }
    return $retval;
} // newsletter_dialog_validate()


/** remove obsolete subscriber data from the database
 *
 * this removes superfluous subscribers from the database;
 * eventually there should only be a single instance of an email address
 * per newsletter. However, during the process of subscribing
 * or re-subscribing it may be possible for different records
 * with the same email address (in the same newsletter) to
 * co-exist, be it with different values for status.
 *
 * This routine is supposedly called after a particular record
 * reaches the status of either NEWSLETTER_SUBSCRIPTION_APPROVED (2)
 * or NEWSLETTER_SUBSCRIPTION_BLACKLIST (3). In those cases all
 * other records with the same email address should simply be
 * removed.
 *
 * On difficulty may be that there can still be records in the
 * queue that point to 'old' subscribers that are about to be
 * deleted; that would break the foreign key. So, the strategy
 * here is to
 *  - change the current records in the queue so that they are
 *    linked to the new record (if the status is 'APPROVED')
 *    OR
 *  - delete the current records in the queue if the status
 *    is 'BLACKLIST'.
 *  - delete all other records with the same email address.
 *
 * @param int $newsletter_id the newsletter of interest
 * @param int $subscriber_id the subscriber record we want to keep
 * @param string $email the corresponding email address 
 * @param int $status the new status of the subscriber
 * @return bool TRUE on success, FALSE otherwise
 */
function newsletter_consolidate_subscribers($newsletter_id,$subscriber_id,$email,$status) {
    if (($status != NEWSLETTER_SUBSCRIPTION_APPROVED) && ($status != NEWSLETTER_SUBSCRIPTION_BLACKLIST)) {
        return TRUE; // nothing to do
    }

    // 1 -- are there any other records with this email address at all?
    $table = 'newsletter_subscribers';
    $fields = array('subscriber_id','full_name','status');
    $where = sprintf('((newsletter_id = %d) AND (subscriber_id <> %d) AND (email = %s))',
                     $newsletter_id,$subscriber_id,db_escape_and_quote($email));
    if (($records = db_select_all_records($table,$fields,$where)) === FALSE) {
        logger(sprintf("%s(): cannot identify other records with '%s' in newsletter %d: %s",
                       __FUNCTION__,$email,$newsletter_id,db_errormessage()));
        return FALSE;
    }
    if (sizeof($records) <= 0) {
        logger(sprintf("%s(): no other records with '%s' in newsletter %d; nothing to do",
                       __FUNCTION__,$email,$newsletter_id),WLOG_DEBUG);
        return TRUE;
    }

    // 2 -- construct a where-clause to manipulate them all at once
    $ids = array();
    $details = array();
    foreach($records as $record) {
        $ids[] = sprintf('(subscriber_id = %d)',$record['subscriber_id']);
        $details[] = sprintf('%s (id=%d,status=%d)',
                             $record['full_name'],$record['subscriber_id'],$record['status']);
    }

    // 3 -- change or get rid of pending entries in the queue
    $retval = TRUE;
    $table = 'newsletter_queue';
    $where = implode(' OR ',$ids).sprintf(' OR (subscriber_id = %d)',$subscriber_id);
    if ($status == NEWSLETTER_SUBSCRIPTION_BLACKLIST) {
        if (($count = db_delete($table,$where)) === FALSE) {
            logger(sprintf("%s(): error deleting pending records for '%s' in queue %d: %s",
                           __FUNCTION__,$email,$newsletter_id,db_errormessage()));
            $retval = TRUE;
        } else {
            logger(sprintf("%s(): success deleting %d pending records for '%s' in queue %d",
                           __FUNCTION__,$count,$email,$newsletter_id),WLOG_DEBUG);
        }
    } else {
        $fields = array('subscriber_id' => $subscriber_id);
        if (($count = db_update($table,$fields,$where)) === FALSE) {
            logger(sprintf("%s(): error consolidating pending records for '%s' in queue %d: %s",
                           __FUNCTION__,$email,$newsletter_id,db_errormessage()));
            $retval = TRUE;
        } else {
            logger(sprintf("%s(): success consolidating %d pending records for '%s' in queue %d",
                           __FUNCTION__,$count,$email,$newsletter_id),WLOG_DEBUG);
        }
    }
    // 4 -- actually delete the other subscriber records
    $table = 'newsletter_subscribers';
    $where = implode(' OR ',$ids);
    if (($count = db_delete($table,$where)) === FALSE) {
        logger(sprintf("%s(): error deleting other subscribers at '%s' in newsletter %d: %s",
                       __FUNCTION__,$email,$newsletter_id,db_errormessage()));
            $retval = TRUE;
    } else {
        logger(sprintf("%s(): success deleting %d other subscribers at '%s' in newsletter %d: %s",
                       __FUNCTION__,$count,$email,$newsletter_id,implode(', ',$details)),WLOG_DEBUG);
    }
    // 5 -- done
    return $retval;
} // newsletter_consolidate_subscribers()


/** attempt to send out as many queued newsletters as possible
 *
 * this constructs a list of all queued mails ordered by
 * - issue_id
 * - retry time (oldest go first; fifo)
 * - queue_id
 *
 * This attempts to send out as many queued newsletters as possible.
 * Depending on the value of $node_id we are looking at a specific
 * newsletter ($node_id == $newsletter_id) or any newsletter 
 *($node_id = NULL). Either way we retrieve a list of at most
 * $limit records from the queue. We select only records that
 * have a retry-time in the past, Within that selection, we order
 * the records by issue_id and subsequently the retry time.
 * This means that we can re-use the issue data if there are
 * more queued requests for the same issue. In the same way
 * we also cache the newsletter configuration; read once and
 * hopefully use many.
 *
 * Note that we start each queue record with incrementing the
 * error count before we do anything else; this means that if
 * somehow the first record gets 'stuck' we can continue with
 * the queue on the next run without hitting the same error
 * over and over again.
 *
 * This is a best effort; if there's an error we simply bail..
 *
 * Note that this routine is not interactive (no output to
 * a browser), so it can easily be called from a cronjob.

 * @param int $node_id indicates which newsletter to handle
 * @param int $limit maximum number of mails to send this time
 * @return bool|int FALSE on error, number of sent messages otherwise
 */
function newsletter_queue_run($node_id=NULL,$limit=50) {
    global $DB,$CFG;
    if (($limit = intval($limit)) <= 0) { // nothing to do, at all
        return 0;
    }

    if (($max_execution_time = intval(ini_get('max_execution_time'))) <= 0) {
        $max_execution_time = 900; // 0=infinite, go for the interval value of 15 minutes
    }
    $start = time();
    $finish = $start + $max_execution_time / 2; // limit ourselves to half the available time
    $now = strftime('%Y-%m-%d %T');
    if (is_null($node_id)) {
        $sql = sprintf('SELECT q.*, s.newsletter_id, s.email, s.full_name '.
                       'FROM %snewsletter_queue q '.
                       'INNER JOIN %snewsletter_subscribers s USING (subscriber_id) '.
                       'WHERE q.retry <= %s '.
                       'ORDER BY q.issue_id, q.retry, q.queue_id',
                       $DB->prefix, $DB->prefix,
                       db_escape_and_quote($now));
        $node = 'any';
    } else {
        $sql = sprintf('SELECT q.*, s.newsletter_id, s.email, s.full_name '.
                       'FROM %snewsletter_queue q '.
                       'INNER JOIN %snewsletter_subscribers s USING (subscriber_id) '.
                       'WHERE s.newsletter_id = %d AND q.retry <= %s '.
                       'ORDER BY q.issue_id, q.retry, q.queue_id',
                       $DB->prefix, $DB->prefix,
                       $node_id,
                       db_escape_and_quote($now));
        $node = strval($node_id);
    }
    $offset = 0;
    if (($DBResult = $DB->query($sql,$limit,$offset)) === FALSE) {
        logger(sprintf('%s(): error  retrieving queue (node = %s): %s',
                       __FUNCTION__,$node,db_errormessage()));
        return FALSE;
    }
    // 2 -- anything available in the queue at all?
    $counter = 0;
    $records = $DBResult->fetch_all_assoc('queue_id');
    $DBResult->close();
    if (sizeof($records) <= 0) {
        logger(sprintf('%s(): nothing to do at this time (node = %s)',__FUNCTION__,$node));
        return $counter;
    }

    // 3 -- prepare for (re)constructing mail
    include_once($CFG->progdir.'/lib/email.class.php');
    $mailer = new Email;
    $issue = array('issue_id' => 0); // sentinel for cache hit/miss
    $config = array('newsletter_id' => 0); // sentinel for cache hit/miss

    // 4 -- step through the at most $limit items in queue
    foreach($records as $queue_id => $record) {
        $newsletter_id = intval($record['newsletter_id']);
        $issue_id = intval($record['issue_id']);
        $queue_id = intval($queue_id);

        // 4A - prepare for disaster; anticipate to do it again next run
        $errors = $record['errors'] + 1;
        $delta = 1800 << $errors; // every error doubles the delay: 1h, 2h, ..., 1024h (~6 weeks)
        $retry = strftime('%Y-%m-%d %T',time() + $delta);
        if (db_update('newsletter_queue',
                      array('errors' => $errors,'retry' => $retry),
                      array('queue_id' => $queue_id)) === FALSE) {
            logger(sprintf('%s(): cannot prepare queue %d in newsletter %d for disaster; bailing out: %s',
                           __FUNCTION__,$queue_id,$newsletter_id,db_errormessage()));
            return FALSE;
        }

        // 4B -- now, do we have the correct newsletter configuration in our cache?
        if ($config['newsletter_id'] != $newsletter_id) {
            $table = 'newsletters';
            $fields = array('newsletter_address','title');
            $where = array('newsletter_id' => $newsletter_id);
            if (($config = db_select_single_record($table, $fields, $where)) === FALSE) {
                logger(sprintf('%s(): error retrieving configuration for newsletter %d: %s',
                               __FUNCTION__,$newsletter_id,db_errormessage()));
                return FALSE;
            }
            $mailer->set_mailreplyto($config['newsletter_address'],$config['title']);
            $mailer->set_mailfrom($config['newsletter_address'],$config['title']);
            logger(sprintf('%s(): cache-miss: (re-)reading configuration newsletter %d (node=%s)',
                           __FUNCTION__,$newsletter_id,$node),WLOG_DEBUG);
            $config['newsletter_id'] = $newsletter_id;
        }

        // 4C -- how about the correct issue in our cache?
        if ($issue['issue_id'] != $issue_id) { // cache miss
            $table = 'newsletter_issues';
            $fields = array('mail_subject','mail_body','mail_headers');
            $where = array('newsletter_id' => $newsletter_id, 'issue_id' => $issue_id);
            if (($issue = db_select_single_record($table, $fields, $where)) === FALSE) {
                logger(sprintf('%s(): error retrieving issue %d from newsletter %d; bailing out: %s',
                               __FUNCTION__,$issue_id,$newsletter_id,db_errormessage()));
                return FALSE;
            }
            $lines = explode("\n",trim($issue['mail_headers']));
            $headers = array();
            foreach($lines as $line) {
                list($k,$v) = explode(':',$line,2);
                $headers[$k] = trim($v);
            }
            unset($lines);
            $message_id = $headers['Message-ID'];
            $mailer->set_subject($issue['mail_subject']);
            logger(sprintf('%s(): cache-miss: (re-)reading issue %d in newsletter %d (node=%s)',
                           __FUNCTION__,$issue_id,$newsletter_id,$node),WLOG_DEBUG);
            $issue['issue_id'] = $issue_id;
        }

        // 5 -- attempt to send message
        $mailer->set_mailto($record['email'],$record['full_name']);
        if ($mailer->send_body($issue['mail_body'],$headers)) {
            // 5A - success! remove record from queue
            db_delete('newsletter_queue',array('queue_id' => $queue_id));
            logger(sprintf('%s(): success sending issue %d %s to %s (%s) in newsletter %d',
                           __FUNCTION__,$issue_id,$message_id,$record['email'],$record['full_name'],
                           $newsletter_id));
            ++$counter;
        } elseif ($errors >= 10) {
            //   5B - the umptieth failure, get rid of this one
            db_delete('newsletter_queue',array('queue_id' => $queue_id));
            ++$counter; // even though we didn't send the actual newsletter on this final attempt
            logger(sprintf('%s(): too many attempts sending issue %d %s to %s (%s) in newsletter %d; fail',
                           __FUNCTION__,$issue_id,$message_id,$record['email'],$record['full_name'],
                           $newsletter_id));
        } else {
            // 5C - failed, do it again later (retry was already set in 4A)
            logger(sprintf('%s(): failed sending issue %d %s to %s (%s) in newsletter %d; will try again',
                           __FUNCTION__,$issue_id,$message_id,$record['email'],$record['full_name'],
                           $newsletter_id));
        }
        // 6 -- don't overstay our welcome
        if ($finish < time()) {
            logger(sprintf('%s(): time has run out (%d s) after processing %d queue entries (node=%s)',
                           __FUNCTION__,$max_execution_time / 2,$counter,$node));
            return $counter;
        }
    }
    logger(sprintf('%s(): processed %d queue entries  in %d s (node=%s); bailing out',
                   __FUNCTION__,$counter,time() - $start,$node));
    return $counter;
} // newsletter_queue_run()

?>