<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/newsletter/newsletter_search.php - interface to the search-part of this module
 *
 * This file defines the interface with this module for searching content.
 * The interface consists of a (temporary) table and this function:
 *
 * <code>
 * bool newsletter_search(&$hits,&$results,$qwords, $limit, $offset)
 * </code>
 *
 * The function is called whenever module content is to be searched. At that point
 * the helper table, which is called 'search_nodes', exists and is populated.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_newsletter
 * @version $Id: newsletter_search.php,v 1.5 2016/04/05 06:34:25 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }


/** search this module's content in selected nodes for keywords in $qwords 
 *
 * This interface is documented in {@link htmlpage_search()} in {@link htmlpage_search.php}.
 *
 * We have to search the following fields in two tables:
 *
 * a. newsletters: 1 record/newsletter: title/subtitle/header/introduction/footer
 *
 * b. newsletter_issues: 0,1,...N records/newsletter: html_title/html_body
 *
 * Newsletters and -issues qualify only when the channel/status is 2 (website) or 3 (both)
 * Hits in newsletters link to the main page (node=$node_id)
 * Hits in the newsletter_issues link to node=$node_id,volume=$volume,number=$number
 *
 * @param array &$hits receives search results (at most $limit)
 * @param int &$results receives number of hits (negative value on entry means (re-)calculate total) 
 * @param array $qwords holds 1 or more keywords to search for (original, utf8lower, quoted)
 * @param int $limit maximum number of hits to return (could be 0)
 * @param int $offset starting point within this module for returning results
 * @return bool|array FALSE on failure, TRUE on success and $hits, $results updated (maybe)
 */
function newsletter_search(&$hits, &$results, $qwords, $limit, $offset=0) {
    global $DB;

    //
    // 0A -- do we need to do anything
    //
    if (($results == 0) ||  // cached value from caller says 'no hits'
        (($results > 0) && ($limit <= 0))) { // there are hits but none are requested at this time
        return TRUE;
    }

    //
    // 0B -- yes, initialise and prepare the selection criteria
    //
    $tables = array('newsletters','newsletter_issues'); 
    $fields = array(array('m.title','m.subtitle','m.header','m.introduction','m.footer'),
                    array('m.html_title','m.html_body'));
    $wheres = array(sprintf('((m.newsletter_channel = 2) OR (m.newsletter_channel = 2)) AND %s',
                            search_where($qwords,$fields[0])),
                    sprintf('((m.status = 2) OR (m.status = 3)) AND %s',search_where($qwords,$fields[1])));
    $fields[1][] = 'm.volume'; // additional issue fields to SELECT but not to be used in WHERE-clause
    $fields[1][] = 'm.number';
    $counts = array(-1,-1); // cache for # of hits (-1 means: no value yet)
    $context_fields = array(array('title','subtitle','header','introduction','footer'),
                            array('html_title','html_body'));

    //
    // 1 -- count hits in newsletters and -issues, don't retrieve actual data
    //
    if ($results < 0) {
        $results = 0;
        foreach($tables as $k => $table) {
            if (($count = newsletter_search_results($table,$wheres[$k])) === FALSE) {
                return FALSE;
            }
            $counts[$k] = $count;
            $results += $count;
        }
    }

    //
    // 2 -- retrieve $limit records at $offset
    //
    if (($limit > 0) && ($offset < $results)) {
         // 2A -- we may have to count table[0] (=newsletters) hits again...
        if (($count = $counts[0]) < 0) {
            if (($count = newsletter_search_results($tables[0],$wheres[0])) === FALSE) {
                return FALSE;
            }
        }
        // 2B -- ...and this may have an impact on the limits and offsets to use in the two tables
        $limits = array(min($limit,$count-$offset),($offset<$count) ? $limit+$offset-$count : $limit);
        $offsets = array($offset,($offset < $count) ? 0 : $offset - $count);
        $orders = array('n.node_id', 'n.node_id, m.datim DESC');

        // 2C -- do search the tables if the (recalculated) limit > 0 for this table
        foreach($tables as $k => $table) {
            if (($lim = $limits[$k]) <= 0) {
                continue;
            }
            $off = $offsets[$k];
            $sql = sprintf("SELECT n.node_id, n.title AS node_title, %s ".
                           "FROM  %ssearch_nodes s INNER JOIN %snodes n USING (node_id) ".
                           "INNER JOIN %s%s m ON (n.node_id = m.newsletter_id) ".
                           "WHERE %s ORDER BY %s",implode(',',$fields[$k]),$DB->prefix,
                           $DB->prefix,$DB->prefix,$table,$wheres[$k],$orders[$k]);
            if (($DBResult = $DB->query($sql,$lim,$off)) === FALSE) {
                logger(sprintf('%s(): cannot retrieve from %s: %s',__FUNCTION__,$table,db_errormessage()));
                return FALSE;
            }
            // 2D -- step through result set of this query
            while (($record = $DBResult->fetch_row_assoc()) !== FALSE) {
                $parameters = array('node' => $record['node_id']);
                if ($k == 0) {
                    $title = trim($record['title']);
                } else {
                    $parameters['volume'] = $record['volume'];
                    $parameters['number'] = $record['number'];
                    $title = trim($record['html_title']);
                }
                // 2E -- only selected fields add to context
                $content = '';
                foreach($context_fields[$k] as $f) {
                    $content .= ' '.$record[$f];
                }
                $hits[] = array('title' => (empty($title)) ? $record['node_title'] : $title,
                                'context' => search_context($qwords,$content),
                                'url' => was_node_url(NULL,$parameters));
            }
            $DBResult->close();
        }
    }
    //
    // 3 -- all done
    //
    return TRUE;
} // newsletter_search()


/** count the number of hits in $table
 *
 * 
 * @param string $table name of table to search
 * @param string $where where clause to select hits in $table
 * @return bool|int FALSE on error, # of hits otherwise (could be 0)
 */
function newsletter_search_results($table,$where) {
    global $DB;
    $sql = sprintf("SELECT COUNT(s.node_id) AS results ".
                   "FROM %ssearch_nodes s INNER JOIN %s%s m ON (s.node_id = m.newsletter_id) ".
                   "WHERE (%s)",
                   $DB->prefix,$DB->prefix,$table,$where);
    if ((($DBResult = $DB->query($sql,1)) === FALSE) || ($DBResult->num_rows != 1)) {
        logger(sprintf('%s(): cannot retrieve count in %s: %s',__FUNCTION__,$table,db_errormessage()));
        return FALSE;
    }
    $record = $DBResult->fetch_row_assoc();
    $DBResult->close();
    return intval($record['results']);
} // newsletter_search_results()

?>