<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/newsletter/newsletter_view.php - interface to the view-part of the newsletter module
 *
 * This file defines the interface with the newsletter-module for viewing content.
 * The interface consists of this function:
 *
 * <code>
 * newsletter_view(&$output,$area_id,$node_id,$module)
 * </code>
 *
 * This function is called from /index.php when the node to display is connected
 * to this module.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_newsletter
 * @version $Id: newsletter_view.php,v 1.13 2016/05/26 13:22:27 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }

define('NEWSLETTER_DIRNAME',dirname(__FILE__));
require_once(NEWSLETTER_DIRNAME.'/newsletter_common.php'); // common code + constants

define('NEWSLETTER_REFERENCE',sha1(__FILE__.':'.__LINE__));
require_once($CFG->progdir.'/lib/tokenlib.php');


/** display the content of the newsletter linked to node $node_id
 *
 * this routine is basically a dispatcher. First we decide which
 * policies are in place for this visitor (who might be a user):
 *
 *  - flag_a: the newsletter archive is available
 *  - flag_c: the visitor is allowed to contribute
 *  - flag_s: the visitor is allowd to subscribe/unsubscribe
 *
 * After that we decide which routine to perform based on the
 * parameters that are specified. The following combinations
 * are recognised:
 *
 *  - archive: volume=v and number=n; shows 0, 1 or more old newsletters
 *  - contribution: contribution=c: handle a visitor contribution
 *  - subscribe: subscribe=s or unsubscribe=u: handle subscriptions
 *
 * If none of those are applicable, the default page is displayed,
 * possibly with the latest newsletter (if policy allows it).
 *
 * The actual work is done in subroutines. The values of these
 * flags are added to the config record which is passed to the
 * subroutines.
 *
 * @param object &$theme collects the (html) output
 * @param int $area_id identifies the area where $node_id lives
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function newsletter_view(&$theme,$area_id,$node_id,$module) {
    global $USER;
    $node_id = intval($node_id);

    // 1 -- we always need the configuration
    if (($config = newsletter_config($node_id)) === FALSE) {
        $theme->add_message(t('error_retrieving_config','m_newsletter'));
        return FALSE;
    }

    // 2 -- policies
    $config['flag_a'] = (($config['newsletter_channel'] == NEWSLETTER_ALERT_SITE) ||
                         ($config['newsletter_channel'] == NEWSLETTER_FULLMAIL_SITE));

    $config['flag_c'] = (($config['contribution_policy']==NEWSLETTER_CONTRIBUTION_POLICY_YES) ||
                         (($config['contribution_policy']==NEWSLETTER_CONTRIBUTION_POLICY_USERS) &&
                          ($USER->is_logged_in)));

    $config['flag_s'] = (($config['subscription_policy'] == NEWSLETTER_SUBSCRIPTION_POLICY_ONE) ||
                         ($config['subscription_policy'] == NEWSLETTER_SUBSCRIPTION_POLICY_TWO));

    // 3 -- dispatcher
    if ($config['flag_a']) {
        if (!is_null($issue_id = get_parameter_int('issue'))) {
            return newsletter_view_issue($theme,$config,$node_id,$issue_id);
        } elseif (!is_null($volume = get_parameter_int('volume'))) {
            $number = get_parameter_int('number',0);
            return newsletter_view_archive($theme,$config,$node_id,$volume,$number);
        }
    }
    if ($config['flag_c']) {
        if (!is_null($value = get_parameter_int('contribution'))) {
            return newsletter_view_contribution($theme,$config,$node_id,$value);
        }
    }
    if ($config['flag_s']) {
        if (!is_null($value = get_parameter_int('subscribe'))) {
            return newsletter_view_subscribe($theme,$config,$node_id);
        } elseif (!is_null($value = get_parameter_int('unsubscribe'))) {
            return newsletter_view_unsubscribe($theme,$config,$node_id);
        } elseif (!is_null($code = get_parameter_string('confirm'))) {
            return newsletter_view_confirm($theme,$config,$node_id,$code);
        }
    }
    return newsletter_view_home($theme,$config,$node_id);
} // newsletter_view()


/** display the default page, possibly with the latest newsletter embedded
 *
 * @param object &$theme collects the (html) output
 * @param array $config newsletter configuration + a few policy-flags
 * @param int $node_id the node to which this module is connected
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function newsletter_view_home(&$theme,$config,$node_id) {
    if (!empty($config['header'])) {
        $theme->add_content(html_tag('h2',array('id' => 'newsletter_view_header'),$config['header']));
    }
    if (!empty($config['introduction'])) {
        $theme->add_content(html_tag('div',array('id' => 'newsletter_view_introduction'),$config['introduction']));
    }
    $issue_id = NULL; // sentinel for 'print-version' in navigation bar
    if ($config['flag_a']) {
        $table = 'newsletter_issues';
        $fields = array('issue_id','html_title','html_body');
        $where = sprintf('newsletter_id = %d AND ((status = %d) OR (status = %d))',
                         $node_id,NEWSLETTER_ALERT_SITE,NEWSLETTER_FULLMAIL_SITE);
        $order = array('volume DESC','number DESC','issue_id');
        if (($record = db_select_single_record($table,$fields,$where,$order)) === FALSE) {
            $theme->add_content(html_tag('p','',t('view_no_newsletters_yet','m_newsletter')));
        } else {
            $issue_id = intval($record['issue_id']);
            $theme->add_content(html_tag('div', array('id' => 'newsletter_view_body')));
            $theme->add_content($record['html_body']);
            $theme->add_content(html_tag_close('div'));
        }

    }
    if (!empty($config['footer'])) {
        $theme->add_content(html_tag('div',array('id' => 'newsletter_view_footer'),$config['footer']));
    }
    newsletter_view_add_navbar($theme,$config,$issue_id);
    return TRUE;
} // newsletter_view_home()


/** add a navigation bar to the output
 *
 * if the corresponding policy allows this creates links to the archive, the contribution
 * dialog and the subscribe/unsubscribe logic. We also (optionally) show a link to display
 * the specified issue 'stand-alone' (in a new window).
 *
 * @param object &$theme collects the (html) output
 * @param array $config newsletter configuration + a few policy-flags
 * @param int $issue_id identify issue to show 'stand-alone' or NULL for none
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function newsletter_view_add_navbar(&$theme,$config,$issue_id=NULL) {
    if ((is_null($issue_id)) && !($config['flag_a'])  && !($config['flag_c'])  && !($config['flag_s'])) {
        return; // nothing to do
    }
    $theme->add_content(html_tag('div',array('id' => 'newsletter_view_navigation')));
    if ($config['flag_a']) {
        $href = was_node_url($theme->node_record,array('volume' => '0'));
        $attributes = array('title' => t('view_archive_title','m_newsletter'));
        $theme->add_content(html_a($href,NULL,$attributes,t('view_archive','m_newsletter')));
    }
    if ($config['flag_c']) {
        $href = was_node_url($theme->node_record,array('contribution' => '1'));
        $attributes = array('title' => t('view_contribution_title','m_newsletter'));
        $theme->add_content(html_a($href,NULL,$attributes,t('view_contribution','m_newsletter')));
    }
    if ($config['flag_s']) {
        $href = was_node_url($theme->node_record,array('subscribe' => '1'));
        $attributes = array('title' => t('view_subscribe_title','m_newsletter'));
        $theme->add_content(html_a($href,NULL,$attributes,t('view_subscribe','m_newsletter')));
        $href = was_node_url($theme->node_record,array('unsubscribe' => '1'));
        $attributes = array('title' => t('view_unsubscribe_title','m_newsletter'));
        $theme->add_content(html_a($href,NULL,$attributes,t('view_unsubscribe','m_newsletter')));
        $href = was_node_url($theme->node_record,array('confirm' => '1'));
        $attributes = array('title' => t('view_confirm_title','m_newsletter'));
        $theme->add_content(html_a($href,NULL,$attributes,t('view_confirm','m_newsletter')));
    }
    if (!is_null($issue_id)) {
        $href = was_node_url($theme->node_record,array('issue' => $issue_id));
        $attributes = array('title' => t('view_print_title','m_newsletter'),'target' => '_blank');
        $theme->add_content(html_a($href,NULL,$attributes,t('view_print','m_newsletter')));
    }
    $theme->add_content(html_tag_close('div'));
    return;
} // newsletter_view_add_navbar()


/** show a 'stand-alone' version of the newsletter
 *
 * this is almost the same routine as {@link newsletter_publish_show_preview()}
 * and {@link newsletter_compose_show_preview()}. The difference is that we
 * look for a specific issue in the archives (could be the latest issue).
 * However, the issue must be in the archive as indicated by the status.
 * If this issue was sent by mail only, we exit with fatal error 099.
 * We assume we have our own window (arrived here via target="_blank").
 *
 * @param object &$theme collects the (html) output
 * @param array $config newsletter configuration + a few policy-flags
 * @param int $node_id the node to which this module is connected
 * @param int $issue_id explicit issue to show 'stand-alone'
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function newsletter_view_issue(&$theme,$config,$node_id,$issue_id) {
    $table = 'newsletter_issues';
    $fields = array('issue_id','html_style','html_title','html_body');
    $where = sprintf('newsletter_id = %d AND issue_id = %d AND ((status = %d) OR (status = %d))',
                     $node_id,$issue_id,NEWSLETTER_ALERT_SITE,NEWSLETTER_FULLMAIL_SITE);
    if (($record = db_select_single_record($table,$fields,$where)) === FALSE) {
        error_exit('099'); // shouldn't happen
    }
    echo "<html>\n".
         "<head><title>".$record['html_title']."</title>\n";
    if (!empty($record['html_style'])) {
        echo "<style><!--\n".$record['html_style']."\n--></style>\n";
    }
    echo "</head>\n".
         "<body>\n";
    echo $record['html_body'];
    echo "</body>\n</html>\n";
    $theme->silent_mode = TRUE;
    return TRUE;
} // newsletter_view_issue()


/** display two levels of newsletter issues: volumes+number or numbers+toc
 *
 * if $volume == 0 we show all available volumes with clickable links to
 * the numbers (including a print link) as nested lists. If $volume == V
 * we show all numbers in volume V (+print link) AND the clickable ToC
 * of every number. We keep everything in at most two nested lists to 
 * prevent cluttering.
 *
 * @param object &$theme collects the (html) output
 * @param array $config newsletter configuration + a few policy-flags
 * @param int $node_id the node to which this module is connected
 * @param int $volume is the volume number > 0 or a wildcard (<= 0)
 * @param int $number is the issue number > 0 or a wildcard (<= 0)
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function newsletter_view_archive(&$theme,$config,$node_id,$volume,$number) {
    if (!empty($config['header'])) {
        $theme->add_content(html_tag('h2',array('id' => 'newsletter_view_header'),$config['header']));
    }
    if (!empty($config['introduction'])) {
        $theme->add_content(html_tag('div',array('id' => 'newsletter_view_introduction'),$config['introduction']));
    }
    $issue_id = NULL; // sentinel for 'print-version' in navigation bar

    // 1 or 2 breadcrumbs
    $params = array('{VOLUME}' => $volume, '{NUMBER}' => $number);
    $href = was_node_url($theme->node_record,array('volume' => '0'));
    $attributes = array('title' => t('breadcrumb_archive_title','m_newsletter'));
    $anchor = t('breadcrumb_archive','m_newsletter');
    $theme->breadcrumb_addendum[] = html_a($href,NULL,$attributes,$anchor);
    if ($volume > 0) {
        $href = was_node_url($theme->node_record,array('volume' => $volume));
        $attributes = array('title' => t('breadcrumb_volume_title','m_newsletter',$params));
        $anchor = t('breadcrumb_volume','m_newsletter',$params);
        $theme->breadcrumb_addendum[]=html_a($href,NULL,$attributes,$anchor);
    }

    // 1 -- show a specific volume+number in embedded form
    if (($volume > 0) && ($number > 0)) {
        $table = 'newsletter_issues';
        $fields = array('issue_id','html_title','html_body');
        $where = sprintf('newsletter_id = %d AND '.
                         'volume = %d AND number = %d AND '.
                         '((status = %d) OR (status = %d))',
                         $node_id,$volume,$number,
                         NEWSLETTER_ALERT_SITE,NEWSLETTER_FULLMAIL_SITE);
        if (($record = db_select_single_record($table,$fields,$where)) === FALSE) {
            $theme->add_content(t('view_newsletter_not_found','m_newsletter',$params));
        } else {
            $issue_id = intval($record['issue_id']);
            $theme->add_content(html_tag('div', array('id' => 'newsletter_view_body')));
            $theme->add_content($record['html_body']);
            $theme->add_content(html_tag_close('div'));
            // final breadcrumb
            $href = was_node_url($theme->node_record,array('volume' => $volume,'number' => $number));
            $attributes = array('title' => t('breadcrumb_number_title','m_newsletter',$params));
            $anchor = t('breadcrumb_number','m_newsletter',$params);
            $theme->breadcrumb_addendum[] = html_a($href,NULL,$attributes,$anchor);
        }
    } else {
        // 2 -- show links to issues in the archive
        $table = 'newsletter_issues';
        $fields = array('issue_id','volume','number','datim','toc');
        if ($volume > 0) {
            $where = sprintf('newsletter_id = %d AND volume = %d AND ((status = %d) OR (status = %d))',
                             $node_id,$volume,NEWSLETTER_ALERT_SITE,NEWSLETTER_FULLMAIL_SITE);
        } else {
            $where = sprintf('newsletter_id = %d AND ((status = %d) OR (status = %d))',
                             $node_id,NEWSLETTER_ALERT_SITE,NEWSLETTER_FULLMAIL_SITE);
        }
        $order = array('volume','number','datim','issue_id');
        if (($issues = db_select_all_records($table,$fields,$where,$order,'issue_id')) === FALSE) {
            $theme->add_content(t('view_no_newsletters_found','m_newsletter'));
        } else {
            $last_volume = 0;
            $theme->add_content(html_tag('ul'));
            foreach($issues as $id => $issue) {
                // maybe start a new volume + first number
                if ($last_volume != $issue['volume']) {
                    if ($last_volume != 0) {
                        $theme->add_content('    '.html_tag_close('ul'));
                        $theme->add_content('  '.html_tag_close('li'));
                    }
                    $last_volume = $issue['volume'];
                    $href = was_node_url($theme->node_record,array('volume' => $last_volume));
                    $params = array('{VOLUME}' => $last_volume);
                    $theme->add_content('  '.html_tag('li').
                                        html_a($href,NULL,NULL,t('view_volume','m_newsletter',$params)));
                    $theme->add_content('    '.html_tag('ul'));
                }
                // details of number follow
                $href = was_node_url($theme->node_record,array('volume' => $last_volume,
                                                               'number' => $issue['number']));
                $params['{NUMBER}'] = $issue['number'];
                $params['{DATIM}'] = $issue['datim'];
                $params['{DATE}'] = substr($issue['datim'],0,10);
                $params['{TIME}'] = substr($issue['datim'],11,5);
                $attributes = array('title' => t('view_print_title','m_newsletter'),'target' => '_blank');
                $anchor = html_a($href,NULL,NULL,t('view_number','m_newsletter',$params)).' '.
                          html_a(was_node_url($theme->node_record,array('issue' => $issue['issue_id'])),
                                 NULL,$attributes,t('view_print','m_newsletter'));
                if ($volume == 0) {
                    $theme->add_content('      '.html_tag('li',NULL,$anchor));
                } else {
                    $theme->add_content('      '.html_tag('li').$anchor);
                    $toc = explode("\n",trim($issue['toc']));
                    $theme->add_content('        '.html_tag('ul'));
                    foreach($toc as $k => $v) {
                        $fragment = sprintf('h%d',++$k);
                        $theme->add_content('          '.html_tag('li',NULL,
                                            html_a($href,NULL,NULL,$v,$fragment)));
                    }
                    $theme->add_content('        '.html_tag_close('ul'));
                    $theme->add_content('      '.html_tag_close('li'));
                }
            }
            if ($last_volume != 0) {
                $theme->add_content('    '.html_tag_close('ul'));
                $theme->add_content('  '.html_tag_close('li'));
            }
            $theme->add_content(html_tag_close('ul'));

        }
    }
    if (!empty($config['footer'])) {
        $theme->add_content(html_tag('div',array('id' => 'newsletter_view_footer'),$config['footer']));
    }
    newsletter_view_add_navbar($theme,$config,$issue_id);
    return TRUE;
} // newsletter_view_archive()


/** handle unsubscriptions (phase 1)
 *
 * this routine shows the dialog where the subscriber can enter
 * her email address to unsubscribe OR processes the same dialog.
 * In the latter case, if the oncecode was sent successfully, the
 * phase 2 dialg (confirm) is displayed.
 *
 * The email that is sent contains the oncecode which can be
 * entered in that dialog. The email also contains a convenient
 * clickable link.

 * @param object &$theme collects the (html) output
 * @param array $config newsletter configuration + a few policy-flags
 * @param int $node_id the node to which this module is connected
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function newsletter_view_unsubscribe(&$theme,$config,$node_id) {

    // 1 -- do they want to cancel?
    if (isset($_POST['button_cancel'])) {
        $theme->add_message(t('cancelled','admin'));
        return newsletter_view_home($theme,$config,$node_id);
    }

    // 2 -- do they want to submit their unsubscription?
    $dialogdef = newsletter_view_unsubscribe_get_dialogdef();
    if (isset($_POST['button_ok'])) {
        if (!newsletter_dialog_validate($dialogdef)) {
            foreach($dialogdef as $k => $item) {
                if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                    $theme->add_message($item['error_messages']);
                }
            }
        } else { // dialog is valid; go process, send a unique unsubscribe code to $email
            $email = $dialogdef['email']['value'];
            if (newsletter_view_unsubscribe_send_code($theme,$config,$email) === FALSE) {
                $theme->add_message(t('error_contact_admin','m_newsletter'));
                return newsletter_view_home($theme,$config,$node_id);
            } else {
                // already show the dialog where the code can be entered
                $params = array('{EMAIL}' => htmlspecialchars($email));
                $theme->add_message(t('confirmation_code_was_sent','m_newsletter',$params));
                $a_params = array('confirm' => '1');
                $href = was_node_url($theme->node_record,$a_params);
                $dialogdef = newsletter_view_confirm_get_dialogdef();
                $theme->add_content(html_tag('h2',NULL,t('confirm_header','m_newsletter')));
                $theme->add_content(t('confirm_explanation','m_newsletter'));
                $theme->add_content(dialog_quickform($href,$dialogdef));
                newsletter_view_add_navbar($theme,$config);
                return TRUE;
            }
        }
    }

    // 3 -- just show the dialog
    $a_params = array('unsubscribe' => '1');
    $href = was_node_url($theme->node_record,$a_params);
    $theme->add_content(html_tag('h2',NULL,t('unsubscribe_header','m_newsletter')));
    $theme->add_content(t('unsubscribe_explanation','m_newsletter'));
    $theme->add_content(dialog_quickform($href,$dialogdef));
    newsletter_view_add_navbar($theme,$config);
    return TRUE;
} // newsletter_view_unsubscribe()


/** construct a dialog definition for unsubscribing
 *
 * this constructs a dialog definition which presents the user with a 1-field 
 * dialog where she can enter her email-address (to unsubscribe).
 *
 * @return array dialog definition
 */
function newsletter_view_unsubscribe_get_dialogdef() {
    $dialogdef = array(
        'email' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'email',
            'minlength' => 3,
            'maxlength' => 255,
            'columns' => 53,
            'label' => t('view_subscriber_email_label','m_newsletter'),
            'title' => t('view_subscriber_email_title','m_newsletter'),
            'value' => ''),
        'button_ok' => dialog_buttondef(BUTTON_OK),
        'button_cancel' => dialog_buttondef(BUTTON_CANCEL)
        );
    return $dialogdef;
} // newsletter_view_unsubscribe_get_dialogdef()


/** send a quasi-random unsubscribe confirmation code to $email
 *
 * @param object &$theme collects the (html) output
 * @param array $config newsletter configuration + a few policy-flags
 * @param string $email destination email address
 * @return bool|string FALSE on error, TRUE on success
 */
function newsletter_view_unsubscribe_send_code(&$theme,$config,$email) {
    global $CFG;

    // 1 -- get subscriber information, bail out on error
    $newsletter_id = intval($config['newsletter_id']);
    $table = 'newsletter_subscribers';
    $fields = '*';
    $where = sprintf('((newsletter_id = %d) AND (email = %s) AND (status != %d))',
                     $newsletter_id,
                     db_escape_and_quote($email),
                     NEWSLETTER_SUBSCRIPTION_BLACKLIST);
    $order = array('subscriber_id');
    if (($records = db_select_all_records($table,$fields,$where,$order)) === FALSE) {
        logger(sprintf('%s(): error retrieving subscriber info \'%s\' in newsletter %d: %s',
                       __FUNCTION__,$email,$newsletter_id,db_errormessage()));
        return FALSE;
    } elseif (sizeof($records) < 1) {
        logger(sprintf('%s(): no subscribers with email \'%s\' found in newsletter %d',
                       __FUNCTION__,$email,$newsletter_id));
        return FALSE;
    }

    // 2 -- set a oncecode in the subscriber records and assume it will go wrong (errors += 1)
    $subscriber_id = intval($records[0]['subscriber_id']); // use the 1st for a uniquemaker
    $code = sprintf('%s%s%d',
                    quasi_random_string(NEWSLETTER_CONFIRM_LEN),
                    NEWSLETTER_CONFIRM_DEL,
                    $subscriber_id);
    $minutes = intval($CFG->login_bypass_interval);
    $datim = strftime("%Y-%m-%d %T",time() + $minutes * 66); // 60 seconds plus 10% for good measure
    // $where can stay exactly the same

    // 2A -- did it go wrong already?
    if ($records[0]['oncecode_errors'] > $CFG->login_max_failures) {
        $fields = array('oncecode' => NULL,
                        'oncecode_expiry' => NULL,
                        'oncecode_errors' => 0,
                        'status' => NEWSLETTER_SUBSCRIPTION_BLACKLIST);
        db_update($table,$fields,$where);
        logger(sprintf('%s(): too many errors for \'%s\' in newsletter %d; subscriber %d blacklisted',
                       __FUNCTION__,$email,$newsletter_id,$subscriber_id));
        newsletter_consolidate_subscribers($newsletter_id,$subscriber_id,$email,$fields['status']);
        return FALSE;
    }

    // 2B -- nope, carry on...
    $fields = array('oncecode' => $code,
                    'oncecode_expiry' => $datim,
                    'oncecode_errors' => intval($records[0]['oncecode_errors']+1));
    if (db_update($table,$fields,$where) === FALSE) {
        logger(sprintf('%s(): cannot set oncecode for email \'%s\' in newsletter %d: %s',
                       __FUNCTION__,$email,$newsletter_id,db_errormessage()));
        return FALSE;
    }

    // 3 -- construct mail to unsubscriber
    $a_params = array('confirm' => '1');
    $manual_url = was_node_url($theme->node_record,$a_params,'',FALSE,TRUE);
    $a_params = array('confirm' => $code);
    $auto_url = was_node_url($theme->node_record,$a_params,'',FALSE,TRUE);
    $params = array('{AUTO_URL}' => $auto_url,
                    '{MANUAL_URL}' => $manual_url,
                    '{CODE}' => $code,
                    '{INTERVAL}' => $minutes,
                    '{REMOTE_ADDR}' => $_SERVER['REMOTE_ADDR'],
                    '{TITLE}' => $config['title'],
                    '{NEWSLETTER}' => $newsletter_id,
                    '{EMAIL}' => $email,
                    '{FULL_NAME}' => $records[0]['full_name']);

    include_once($CFG->progdir.'/lib/email.class.php');
    $mailer = new Email;
    $mailer->set_mailreplyto($config['newsletter_address'],$config['title']);
    $mailer->set_mailfrom($config['newsletter_address'],$config['title']);
    $mailer->set_mailto($records[0]['email'],$records[0]['full_name']);
    $mailer->set_message(t('unsubscribe_mail_body','m_newsletter',$params));
    $mailer->set_subject(replace_crlf(t('unsubscribe_mail_subject','m_newsletter',$params),' '));

    if ($mailer->send()) {
        logger(sprintf('%s(): success sending oncecode to \'%s\'  (subscriber %d) in newsletter %d',
                       __FUNCTION__,$email,$subscriber_id,$newsletter_id));
        return TRUE;
    } else {
        logger(sprintf('%s(): failure sending oncecode to \'%s\'  (subscriber %d) in newsletter %d',
                       __FUNCTION__,$email,$subscriber_id,$newsletter_id));
        return FALSE;
    }
} // newsletter_view_unsubscribe_send_code()


/** handle the confirmation code for subscribing/unsubscribing
 *
 * @param object &$theme collects the (html) output
 * @param array $config newsletter configuration + a few policy-flags
 * @param int $node_id the node to which this module is connected
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function newsletter_view_confirm(&$theme,$config,$node_id,$code=1) {

    // 1 -- do they want to cancel?
    if (isset($_POST['button_cancel'])) {
        $theme->add_message(t('cancelled','admin'));
        return newsletter_view_home($theme,$config,$node_id);
    }
    
    // 2 -- do they want to confirm their (un)subscription?
    $dialogdef = newsletter_view_confirm_get_dialogdef();
    if (isset($_POST['button_ok'])) {
        if (!newsletter_dialog_validate($dialogdef)) {
            foreach($dialogdef as $k => $item) {
                if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                    $theme->add_message($item['error_messages']);
                }
            }
        } else { // dialog is valid; go process, send a unique unsubscribe code to $email
            $code = $dialogdef['code']['value'];
        }
    }
    if (strlen($code) > 13) { // either from GET or via POST/dialogdef
        if (($retval = newsletter_view_confirm_process_code($theme,$config,$code)) === FALSE) {
            $theme->add_message(t('error_contact_admin','m_newsletter'));
        } else {
            $theme->add_message(t('success_confirm_'.$retval,'m_newsletter'));
        }
        return newsletter_view_home($theme,$config,$node_id);
    }

    // 3 -- just show the dialog
    $a_params = array('confirm' => '1');
    $href = was_node_url($theme->node_record,$a_params);
    $theme->add_content(html_tag('h2',NULL,t('confirm_header','m_newsletter')));
    $theme->add_content(t('confirm_explanation','m_newsletter'));
    $theme->add_content(dialog_quickform($href,$dialogdef));
    newsletter_view_add_navbar($theme,$config);
    return TRUE;
} // newsletter_view_confirm()


/** construct a dialog definition for confirm code
 *
 * this constructs a dialog definition which presents the user with a 1-field 
 * dialog where she can enter the oncecode that was emailed to her earlier.
 *
 * @return array dialog definition
 */
function newsletter_view_confirm_get_dialogdef() {
    $dialogdef = array(
      'code' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'code',
            'minlength' => 13,
            'maxlength' => 255,
            'columns' => 53,
            'label' => t('view_confirm_code_label','m_newsletter'),
            'title' => t('view_confirm_code_title','m_newsletter'),
            'value' => ''
            ),
      'button_ok' => dialog_buttondef(BUTTON_OK),
      'button_cancel' => dialog_buttondef(BUTTON_CANCEL)
      );
    return $dialogdef;
} //  newsletter_view_confirm_get_dialogdef()


/** process a code confirming a subscribe or unsubscribe action
 *
 * we arrive here with a code that has the action to perform
 * embedded. It consists of a quasi-random string, the action to
 * perform ('d' means unsubscribe, 'a' means subscribe)
 * and finally a uniquemaker based on a subscriber_id.

 * @param object &$theme collects the (html) output
 * @param array $config newsletter configuration + a few policy-flags
 * @param string $code the oncecode entered by the user
 * @return bool FALSE on error, non-0 identification of message on success
 */
function newsletter_view_confirm_process_code($theme,$config,$code) {
    global $CFG;

    // 0 -- basic sanity code should look like ABCDEF123456?123 where ? is the action code
    if (strlen($code) <= NEWSLETTER_CONFIRM_LEN) {
        logger(sprintf('%s(): code \'%s\' is too short; bailing out',__FUNCTION__,$code));
        return FALSE;
    }
    $now = strftime('%Y-%m-%d %T');
    $newsletter_id = intval($config['newsletter_id']);
    $action = $code[NEWSLETTER_CONFIRM_LEN];
    if (($action != NEWSLETTER_CONFIRM_DEL) && ($action != NEWSLETTER_CONFIRM_ADD)) {
        logger(sprintf('%s(): action is neither \'%s\' nor \'%s\': \'%s\' in newsletter %d; bailing out',
                       __FUNCTION__,NEWSLETTER_CONFIRM_ADD,NEWSLETTER_CONFIRM_DEL,$code,$newsletter_id));
    }

    // 1 -- first locate record(s) to manipulate in $records[]
    $table = 'newsletter_subscribers';
    $fields = '*';
    $where = sprintf('(newsletter_id = %d) AND '.
                     '(oncecode = %s) AND (%s < oncecode_expiry) AND (status != %d)',
                     $newsletter_id,
                     db_escape_and_quote($code),
                     db_escape_and_quote($now),
                     NEWSLETTER_SUBSCRIPTION_BLACKLIST);    
    
    if (($records = db_select_all_records($table,$fields,$where)) === FALSE) {
        logger(sprintf('%s(): cannot select subscribers in newsletter %d: %s',
                       __FUNCTION__,$newsletter_id,db_errormessage()));
        return FALSE;
    } elseif (sizeof($records) < 1) {
        logger(sprintf('%s(): cannot find subscribers code \'%s\' to unsubscribe in newsletter %d',
                       __FUNCTION__,$code,$newsletter_id));
        return FALSE;
    }

    // 2 -- now carry out the requested action
    if ($action == NEWSLETTER_CONFIRM_ADD) { //  subscribe
        if ($config['subscription_policy'] == NEWSLETTER_SUBSCRIPTION_POLICY_ONE) {
            // self-approve
            $retval = 1;
            $status = NEWSLETTER_SUBSCRIPTION_APPROVED;
        } else {
            // moderator must approve
            $retval = 2;
            $status = NEWSLETTER_SUBSCRIPTION_CONFIRMED;
        }
        foreach($records as $record) {
            $subscriber_id = intval($record['subscriber_id']);
            $where = array('subscriber_id' => $subscriber_id);
            $fields = array('status' => $status,
                            'oncecode' => NULL,
                            'oncecode_expiry' => NULL,
                            'oncecode_errors' => 0);
            if (db_update($table,$fields,$where) === FALSE) {
                logger(sprintf('%s(): cannot update status of subscriber %d in newsletter %d: %s',
                               __FUNCTION__,$subscriber_id,$newsletter_id,db_errormessage()));
                $retval = FALSE;
            } else {
                logger(sprintf('%s(): status of subscriber %d <%s> (%s) set to %d  in newsletter %d',
                               __FUNCTION__,$subscriber_id,$record['email'],$record['full_name'],
                               $status,$newsletter_id));
            }
        }
        // get rid of superfluous records with the same email (best effort)
        if ($status == NEWSLETTER_SUBSCRIPTION_APPROVED) {
            foreach($records as $record) {
                $subscriber_id = intval($record['subscriber_id']);
                $email = $record['email'];
                newsletter_consolidate_subscribers($newsletter_id,$subscriber_id,$email,$status);
            }
        }
    } else { // unsubscribe
        $retval = 3;
        foreach($records as $record) {
            $subscriber_id = intval($record['subscriber_id']);
            $where = array('subscriber_id' => $subscriber_id);
            if (db_delete('newsletter_queue',$where) === FALSE) {
                logger(sprintf('%s(): cannot delete subscriber %d from queue in newsletter %d: %s',
                               __FUNCTION__,$subscriber_id,$newsletter_id,db_errormessage()));
                $retval = FALSE;
            }
            $where['newsletter_id'] = $newsletter_id;
            if (db_delete($table,$where) === FALSE) {
                logger(sprintf('%s(): cannot delete subscriber %d from newsletter %d: %s',
                               __FUNCTION__,$subscriber_id,$newsletter_id,db_errormessage()));
                $retval = FALSE;
            } else {
                logger(sprintf('%s(): subscriber %d <%s> (%s) removed from newsletter %d',
                               __FUNCTION__,$subscriber_id,$record['email'],$record['full_name'],
                               $newsletter_id));
            }
        }
    }
    if ($retval === FALSE) {
        return FALSE;
    }
    // 3 -- tell administrator about the events; use the first subscriber in $records[] as pars pro toto
    $params = array(
        '{SUBSCRIBER}' => $records[0]['subscriber_id'],
        '{EMAIL}' => $records[0]['email'],
        '{FULL_NAME}' => $records[0]['full_name'],
        '{REMARKS}' => $records[0]['remarks'],
        '{NEWSLETTER}' => $newsletter_id,
        '{TITLE}' => $config['title']);
    include_once($CFG->progdir.'/lib/email.class.php');
    $mailer = new Email;
    $mailer->set_mailreplyto($config['newsletter_address'],$config['title']);
    $mailer->set_mailfrom($config['newsletter_address'],$config['title']);
    $mailer->set_message(t('admin_mail_body_'.$retval,'m_newsletter',$params));
    $mailer->set_subject(t('admin_mail_subject_'.$retval,'m_newsletter',$params));;
    $admins = explode("\n",trim($config['email_administrators']));
    foreach($admins as $admin) {
        $mailer->set_mailto($admin,t('newsletter_administrator','m_newsletter'));
        $mailer->send(); // best effort; no error checking
    }
    return $retval;
} // newsletter_view_confirm_process_code()


/** show and process a subscription request (phase 1)
 *
 * @param object &$theme collects the (html) output
 * @param array $config newsletter configuration + a few policy-flags
 * @param int $node_id the node to which this module is connected
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function newsletter_view_subscribe(&$theme,$config,$node_id) {

    // 1 -- do they want to cancel?
    if (isset($_POST['button_cancel'])) {
        $theme->add_message(t('cancelled','admin'));
        return newsletter_view_home($theme,$config,$node_id);
    }

    // 2 -- do they want to submit their subscription?
    $show_remarks = ($config['subscription_policy'] == NEWSLETTER_SUBSCRIPTION_POLICY_TWO) ? TRUE : FALSE;
    $dialogdef = newsletter_view_subscribe_get_dialogdef($show_remarks);
    if (isset($_POST['button_ok'])) {
        if (!newsletter_dialog_validate($dialogdef)) {
            foreach($dialogdef as $k => $item) {
                if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                    $theme->add_message($item['error_messages']);
                }
            }
        } else { // dialog is valid; go add a new record and send a unique subscribe confirmation to $email
            $email = $dialogdef['email']['value'];
            $full_name = $dialogdef['full_name']['value'];
            $remarks = ($show_remarks) ? $dialogdef['remarks']['value'] : '';
            if (newsletter_view_subscribe_send_code($theme,$config,$email,$full_name,$remarks) === FALSE) {
                $theme->add_message(t('error_contact_admin','m_newsletter'));
                return newsletter_view_home($theme,$config,$node_id);
            } else {
                // already show the dialog where the code can be entered
                $params = array('{EMAIL}' => htmlspecialchars($email));
                $theme->add_message(t('confirmation_code_was_sent','m_newsletter',$params));
                $a_params = array('confirm' => '1');
                $href = was_node_url($theme->node_record,$a_params);
                $dialogdef = newsletter_view_confirm_get_dialogdef();
                $theme->add_content(html_tag('h2',NULL,t('confirm_header','m_newsletter')));
                $theme->add_content(t('confirm_explanation','m_newsletter'));
                $theme->add_content(dialog_quickform($href,$dialogdef));
                newsletter_view_add_navbar($theme,$config);
                return TRUE;
            }
        }
    }

    // 3 -- just show the dialog
    $a_params = array('subscribe' => '1');
    $href = was_node_url($theme->node_record,$a_params);
    $theme->add_content(html_tag('h2',NULL,t('subscribe_header','m_newsletter')));
    $theme->add_content($config['subscription_introduction']);
    $theme->add_content(dialog_quickform($href,$dialogdef));
    newsletter_view_add_navbar($theme,$config);
    return TRUE;
} // newsletter_view_subscribe()


/** construct a dialog definition for subscribing
 *
 * this constructs a dialog definition which presents the user
 * with a 2 or 3 field dialog where she can enter her email-address,
 * full name and perhaps additional remarks to subscribe.
 *
 * @param bool $show_remarks if TRUE we show the remarks field too
 * @return array dialog definition
 */
function newsletter_view_subscribe_get_dialogdef($show_remarks) {
    $dialogdef = array(
        'email' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'email',
            'minlength' => 3,
            'maxlength' => 255,
            'columns' => 53,
            'label' => t('view_subscriber_email_label','m_newsletter'),
            'title' => t('view_subscriber_email_title','m_newsletter'),
            'value' => ''),
        'full_name' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'full_name',
            'minlength' => 0,
            'maxlength' => 255,
            'columns' => 53,
            'label' => t('view_subscriber_full_name_label','m_newsletter'),
            'title' => t('view_subscriber_full_name_title','m_newsletter'),
            'value' => ''
            )
        );
    if ($show_remarks) {
        $dialogdef['remarks'] = array(
            'type' => F_ALPHANUMERIC,
            'name' => 'remarks',
            'minlength' => 0,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 70,
            'rows' => 8,
            'label' => t('view_subscriber_remarks_label','m_newsletter'),
            'title' => t('view_subscriber_remarks_title','m_newsletter'),
            'value' => ''
            );
    }
    $dialogdef['button_ok'] = dialog_buttondef(BUTTON_OK);
    $dialogdef['button_cancel'] = dialog_buttondef(BUTTON_CANCEL);
    return $dialogdef;
} // newsletter_view_subscribe_get_dialogdef()


/** send a quasi-random subscribe confirmation code to $email
 *
 * @param object &$theme collects the (html) output
 * @param array $config newsletter configuration + a few policy-flags
 * @param string $email destination email address
 * @return bool|string FALSE on error, TRUE on success
 */
function newsletter_view_subscribe_send_code(&$theme,$config,$email,$full_name,$remarks) {
    global $CFG;

    // 1 -- is this address on the blacklist?
    $newsletter_id = intval($config['newsletter_id']);
    $table = 'newsletter_subscribers';
    $fields = 'subscriber_id';
    $where = array('newsletter_id' => $newsletter_id,
                   'email' => $email,
                   'status' => NEWSLETTER_SUBSCRIPTION_BLACKLIST);
    if (($records = db_select_all_records($table,$fields,$where)) === FALSE) {
        logger(sprintf('%s(): error retrieving subscriber info \'%s\' in newsletter %d: %s',
                       __FUNCTION__,$email,$newsletter_id,db_errormessage()));
        return FALSE;
    } elseif (sizeof($records) > 0) {
        logger(sprintf('%s(): subscriber email \'%s\' already on blacklist in newsletter %d',
                       __FUNCTION__,$email,$newsletter_id));
        return FALSE;
    }

    // 2 -- no blacklist, go add a new record w/ oncecode
    // 2A - add new record and get id
    $fields = array('newsletter_id' => $newsletter_id,
                    'email' => $email,
                    'full_name' => $full_name,
                    'remarks' => $remarks,
                    'status' => NEWSLETTER_SUBSCRIPTION_NEW);
    if (($subscriber_id = db_insert_into_and_get_id($table,$fields,'subscriber_id')) === FALSE) {
        logger(sprintf('%s(): cannot insert new subscriber <%s> (%s): %s',
                       __FUNCTION__,$email,$full_name,db_errormessage()));
        return FALSE;
    }
    // 2B - construct oncecode with id + update record
    $code = sprintf('%s%s%d',
                    quasi_random_string(NEWSLETTER_CONFIRM_LEN),
                    NEWSLETTER_CONFIRM_ADD,
                    $subscriber_id);
    $minutes = intval($CFG->login_bypass_interval);
    $datim = strftime("%Y-%m-%d %T",time() + $minutes * 66); // 60 seconds plus 10% for good measure
    $fields = array('oncecode' => $code,
                    'oncecode_expiry' => $datim,
                    'oncecode_errors' => 0);
    $where = array('subscriber_id' => $subscriber_id);
    if (db_update($table,$fields,$where) === FALSE) {
        logger(sprintf('%s(): cannot set oncecode for subscriber %d <%s> in newsletter %d: %s',
                       __FUNCTION__,$subscriber_id,$email,$newsletter_id,db_errormessage()));
        return FALSE;
    }

    // 3 -- construct mail to subscriber
    $a_params = array('confirm' => '1');
    $manual_url = was_node_url($theme->node_record,$a_params,'',FALSE,TRUE);
    $a_params = array('confirm' => $code);
    $auto_url = was_node_url($theme->node_record,$a_params,'',FALSE,TRUE);
    $params = array('{AUTO_URL}' => $auto_url,
                    '{MANUAL_URL}' => $manual_url,
                    '{CODE}' => $code,
                    '{INTERVAL}' => $minutes,
                    '{REMOTE_ADDR}' => $_SERVER['REMOTE_ADDR'],
                    '{TITLE}' => $config['title'],
                    '{NEWSLETTER}' => $newsletter_id,
                    '{EMAIL}' => $email,
                    '{FULL_NAME}' => $full_name);

    include_once($CFG->progdir.'/lib/email.class.php');
    $mailer = new Email;
    $mailer->set_mailreplyto($config['newsletter_address'],$config['title']);
    $mailer->set_mailfrom($config['newsletter_address'],$config['title']);
    $mailer->set_mailto($email,$full_name);
    $mailer->set_message(t('subscribe_mail_body','m_newsletter',$params));
    $mailer->set_subject(replace_crlf(t('subscribe_mail_subject','m_newsletter',$params),' '));

    if ($mailer->send()) {
        logger(sprintf('%s(): success sending oncecode to \'%s\'  (subscriber %d) in newsletter %d',
                       __FUNCTION__,$email,$subscriber_id,$newsletter_id));
        return TRUE;
    } else {
        logger(sprintf('%s(): failure sending oncecode to \'%s\'  (subscriber %d) in newsletter %d',
                       __FUNCTION__,$email,$subscriber_id,$newsletter_id));
        return FALSE;
    }
} // newsletter_view_subscribe_send_code()


/** handle user contribution of articles
 *
 * this eventually yields a user contribution in the form of a text-only
 * article added to the list of UNselected articles in the database.
 * we use the logic from tokenlib as we did in {@link mailpage_view.php}.
 * Data entry is a two-step process: one screen with editable text and
 * a preview-button, another screen with uneditable text and a submit-button
 * This makes it impossible to simply keep submitting a story; the first
 * screen provides a token, and the second screen only works after a minimum
 * delay.

 * @param object &$theme collects the (html) output
 * @param array $config newsletter configuration + a few policy-flags
 * @param int $node_id the node to which this module is connected
 * @param int $value indicates the phase in the contribution process
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 */
function newsletter_view_contribution(&$theme,$config,$node_id,$value) {

    // 1 -- do we have a token already?
    $t0 = $t1 = 0;
    $ip_addr = '';
    $data = FALSE;
    $token_id = FALSE;
    if (isset($_POST['token'])) { // lookup valid UTF8 key (or fail with substitute U+FFFD instead)
        $token_key = (utf8_validate($_POST['token'])) ? magic_unquote($_POST['token']) : "\xEF\xBF\xBD";
        $token_id = token_lookup(NEWSLETTER_REFERENCE,$token_key,$t0,$t1,$ip_addr,$data);
    }
    // 2 -- handle cases of expired tokens and Cancel first
    $now = time();
    if (($token_id !== FALSE) && (isset($_POST['button_cancel']))) { // visitor pressed [Cancel]
        $theme->add_message(t('cancelled','admin'));
        token_destroy($token_id);
        $token_id = FALSE;
    }
    if (($token_id !== FALSE) && ($t1 < $now)) { // token expired
	$theme->add_message(t('error_token_expired','m_newsletter'));
        token_destroy($token_id);
        $token_id = FALSE;
    }

    // 3 -- handle the three remaining buttons from the two dialogs
    if ($token_id !== FALSE) {
        if (isset($_POST['button_preview'])) {
            // 3A -- Preview button
            $dialogdef = newsletter_view_contribution_get_dialogdef($config, $token_key);
            if (!newsletter_dialog_validate($dialogdef)) {
                foreach($dialogdef as $k => $item) {
                    if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                        $theme->add_message($item['error_messages']);
                    }
                }
                newsletter_view_contribution_form($theme, $config, $dialogdef);
            } else {
                if (!token_store($token_id, $dialogdef)) {
                    $theme->add_message(t('error_storing_data','m_newsletter'));
                    logger(sprintf('%s(): token store error in page %d: %s',
                                   __FUNCTION__,$node_id,db_errormessage()));
                    return FALSE;
                }
                newsletter_view_contribution_preview($theme, $config, $dialogdef, $ip_addr);
            }
        } elseif (isset($_POST['button_edit'])) {
            // 3B -- Edit button
            if ($data === FALSE) {
                $theme->add_message(t('error_retrieving_data','m_newsletter'));
                logger(sprintf('%s(): no data after token_lookup()? (page=%d)',__FUNCTION__,$node_id));
                $data = newsletter_view_contribution_get_dialogdef($config, $token_key);
            }
            newsletter_view_contribution_form($theme, $config, $data);
        } elseif (isset($_POST['button_send'])) {
            // 3C -- Send button
            if ($data === FALSE) {
                $theme->add_message(t('error_retrieving_data','m_newsletter'));
                logger(sprintf('%s(): no data after token_lookup()? (page=%d)',__FUNCTION__,$node_id));
                $data = newsletter_view_contribution_get_dialogdef($config, $token_key);
            }
            if ($now < $t0) { // the window of opportunity is still closed; go back to form a la Edit
                $msg = t('error_too_fast','m_newsletter');
                $theme->add_message($msg);
                $theme->add_popup_top($msg);
                newsletter_view_contribution_form($theme, $config, $data);
                logger(sprintf('%s(): reply too fast (%ds) from %s',__FUNCTION__,$t0-$now,$ip_addr));
            } elseif (!newsletter_view_contribution_send($theme, $config, $data, $ip_addr, $now-$t0)) {
                $theme->add_message(t('error_sending_message','m_newsletter'));
                newsletter_view_contribution_form($theme, $config, $data);
            } else {
                token_destroy($token_id);
                newsletter_view_contribution_thankyou($theme, $config, $data, $ip_addr);
            }
        } else {
            // 3D -- catch all: initiate a new round (shouldn't happen)
            token_destroy($token_id);
            $token_id = FALSE;
        }
    }
    // 4 -- Start with a clean slate
    if ($token_id === FALSE) {
        $token_key = '';
        if (($token_id = token_create(NEWSLETTER_REFERENCE,$token_key,20)) === FALSE) { // 20s delay
            $msg = t('error_creating_token','m_newsletter',array('{NODE}' => strval($node_id)));
            $theme->add_message($msg);
            $theme->add_content($msg);
            return FALSE;
        }
        $dialogdef = newsletter_view_contribution_get_dialogdef($config, $token_key);
        newsletter_view_contribution_form($theme, $config, $dialogdef);
    }
    return TRUE;
} // newsletter_view_contribution()


/** construct a dialog definition for the visitor's contribution form
 *
 * @param array $config newsletter configuration
 * @param string $token_key
 * @return array datadefinition
 */
function newsletter_view_contribution_get_dialogdef($config, $token_key) {
    global $USER;
    $registered = ($config['contribution_policy'] == NEWSLETTER_CONTRIBUTION_POLICY_USERS) ? TRUE : FALSE;
    $dialogdef = array(
        'token' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'token',
	    'hidden' => TRUE,
            'value' => $token_key
            ),
        'full_name' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'full_name',
            'minlength' => 1,
            'maxlength' => 255,
            'columns' => 53,
            'label' => ($registered) ? t('contrib_full_name_label','m_newsletter') :
                                       t('contrib_full_name_required_label','m_newsletter'),
            'title' => ($registered) ? t('contrib_full_name_title','m_newsletter') :
                                       t('contrib_full_name_required_title','m_newsletter'),
            'value' => ($USER->is_logged_in) ? $USER->full_name : '',
            'viewonly' => $registered
            ),
        'email' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'email',
            'minlength' => 3,
            'maxlength' => 255,
            'columns' => 53,
            'label' => ($registered) ? t('contrib_email_label','m_newsletter') :
                                       t('contrib_email_required_label','m_newsletter'),
            'title' => ($registered) ? t('contrib_email_title','m_newsletter') :
                                       t('contrib_email_required_title','m_newsletter'),
            'value' =>  ($USER->is_logged_in) ? $USER->email : '',
            'viewonly' =>  ($registered)
            ),
        'title' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'title',
            'minlength' => 1,
            'maxlength' => 255,
            'columns' => 53,
            'label' => t('contrib_title_label','m_newsletter'),
            'title' => t('contrib_title_title','m_newsletter'),
            'value' => '',
            ),
        'content' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'content',
            'minlength' => 1,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 70,
            'rows' => 10,
            'label' => t('contrib_content_label','m_newsletter'),
            'title' => t('contrib_content_title','m_newsletter'),
            'value' => ''
            ),
        'remarks' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'remarks',
            'minlength' => 0,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 70,
            'rows' => 8,
            'label' => t('contrib_remarks_label','m_newsletter'),
            'title' => t('contrib_remarks_title','m_newsletter'),
            'value' => ''
            ),
        'button_preview' => dialog_buttondef('button_preview',t('button_preview','m_newsletter')),
        'button_cancel' => dialog_buttondef(BUTTON_CANCEL)
        );
    return $dialogdef;
} // newsletter_view_contribution_get_dialogdef()


/** display the contribution form
 *
 * this displays the contribution form.
 *
 * @param object &$theme collects the (html) output
 * @param array mnewsletter configuration data
 * @param array $dialogdef array that defines the input fields
 * @return void output writted to $theme
 */
function newsletter_view_contribution_form(&$theme, $config, $dialogdef) {
    $href = was_node_url($theme->node_record,array('contribution' => '1'));
    $theme->add_content(html_tag('h2','',t('contrib_header','m_newsletter')));
    $theme->add_content($config['contribution_introduction']);
    $theme->add_content(dialog_quickform($href,$dialogdef));
    newsletter_view_add_navbar($theme,$config);
    return TRUE;
} // newsletter_view_contribution_form()


/** show a preview of the message to the visitor
 *
 * this shows a preview of the article to the visitor.
 * Nothing is editable, it is view-only. The only option
 * is to either press the Send-button to actually submit
 * the messate OR to press the Edit button to go back
 * to the editable form.
 *
 * Sending a message is a two-step procedure by design.
 *
 * @param object &$theme collects the (html) output
 * @param array newsletter configuration data
 * @param array $dialogdef array that defines the input fields
 * @param string $ip_addr the originating IP-address
 * @return void output writted to $theme
 */
function newsletter_view_contribution_preview(&$theme, $config, $dialogdef, $ip_addr) {

    // 1 -- prepare the information to show
    $forbidden = array(chr(10),chr(13),chr(34),'\\');
    $author = sprintf('&quot;%s&quot; &lt;%s&gt;',
                      htmlspecialchars(str_replace($forbidden,'',trim($dialogdef['full_name']['value']))),
                      htmlspecialchars(str_replace($forbidden,'',trim($dialogdef['email']['value']))));
    $remarks = nl2br(htmlspecialchars(strip_tags(trim($dialogdef['remarks']['value']))));
    $title = htmlspecialchars(strip_tags(trim($dialogdef['title']['value'])));
    $content = nl2br(htmlspecialchars(strip_tags(trim($dialogdef['content']['value']))));
    $remote_addr = htmlspecialchars($ip_addr);

    // 2 -- actually output the preview
    $theme->add_content(html_tag('h2','',t('preview_header','m_newsletter')));
    $theme->add_content(html_tag('div',array('class'=>'newsletter_article_outer')));
    $theme->add_content(html_tag('h3',array('class'=>'newsletter_article_header'),$title));
    $theme->add_content(html_tag('div',array('class'=>'newsletter_article_author'),$author));
    $theme->add_content(html_tag('div',array('class'=>'newsletter_article_inner'),$content));
    $theme->add_content(html_tag('div',array('class'=>'newsletter_article_remarks'),$remarks));
    $theme->add_content(sprintf('<strong>%s</strong>: %s<br>',
                                t('contrib_ip_addr','m_newsletter'),$remote_addr));
    $theme->add_content(html_tag_close('div'));

    // 3 -- finish with navigation buttons for the visitor
    $previewdef = array(
        'token' => $dialogdef['token'],
        'button_edit' => dialog_buttondef(BUTTON_EDIT),
        'button_send' => dialog_buttondef('button_send',t('button_send','m_newsletter')),
        'button_cancel' => dialog_buttondef(BUTTON_CANCEL)
        );
    $href = was_node_url($theme->node_record,array('contribution' => '1'));
    $theme->add_content(dialog_quickform($href,$previewdef));
    // maybe we don't want to distract the user at this point:
    // newsletter_view_add_navbar($theme,$config);
} // newsletter_view_contribution_preview()


/** add the contribution to the list of unselected articles and mail the admins
 *
 * In order to get a feeling for the time a visitor needs, we also
 * record the delay (in seconds) next to the visitor's IP address.
 *
 * @param object &$theme collects the (html) output and provides queue_alert()
 * @param array $config mailpage configuration data in a (nested) array
 * @param array $dialogdef array that defines the data fields including values
 * @param string $ip_addr the originating IP-address
 * @param int $delay the # of seconds since time=t0
 * @return bool FALSE on error, TRUE on success + message sent
 * @todo extra validation of set_mailreplyto and set_subject?
 * @todo more available parameters in subject_line?
 * @todo make body of mail configuratble?
 */
function newsletter_view_contribution_send(&$theme,$config, $dialogdef, $ip_addr, $delay) {
    global $CFG,$USER;
    $newsletter_id = intval($config['newsletter_id']);

    // 1 -- store in database as UNselected article
    $table = 'newsletter_articles';

    // 1A -- calculate next highest sort order
    $fields = 'MAX(sort_order) AS sort_order';
    $where = array('newsletter_id' => $newsletter_id, 'selected' => FALSE);
    if (($record = db_select_single_record($table,$fields,$where)) === FALSE) {
        $sort_order = 10;
    } else {
        $sort_order = 10 + intval($record['sort_order']);
    }
    $now =       strftime('%Y-%m-%d %T');
    $title =     htmlspecialchars(strip_tags(trim($dialogdef['title']['value'])));
    $content =   htmlspecialchars(strip_tags(trim($dialogdef['content']['value'])));
    $full_name = htmlspecialchars(strip_tags(trim($dialogdef['full_name']['value'])));
    $remarks =   htmlspecialchars(strip_tags(trim($dialogdef['remarks']['value'])));
    $email =     htmlspecialchars(strip_tags(trim($dialogdef['email']['value'])));

    // 1B -- save the new article
    $retval = TRUE;
    $fields = array(
        'newsletter_id' => $newsletter_id,
        'title' => $title,
        'content' => $content,
        'remarks' => $remarks,
        'author_id' => ($USER->is_logged_in) ? $USER->user_id : NULL,
        'author_full_name' => $full_name,
        'email' => $email,
        'datim' => $now,
        'remote_addr' => $ip_addr,
        'selected' => FALSE,
        'sort_order' => $sort_order,
        'mtime' => $now,
        'muser_id' => ($USER->is_logged_in) ? $USER->user_id : NULL
        );

    if (($article_id = db_insert_into_and_get_id($table,$fields,'article_id')) === FALSE) {
        logger(sprintf('%s(): cannot insert newly contributed article %s from <%s> (%s): %s',
                       __FUNCTION__,$title,$email,$full_name,db_errormessage()));
        $retval = FALSE;
        $article_id = 0; // make it visible in mail to admin
    }
    // 2 -- alerts
    $params = array(
        '{TITLE}' => $config['title'],
        '{NEWSLETTER}' => $newsletter_id,
        '{ARTICLE}' => $article_id,
        '{ARTICLE_TITLE}' => $title,
        '{AUTHOR}' => $full_name,
        '{AUTHOR_EMAIL}' => $email,
        '{CONTENT}' => $content,
        '{REMARKS}' => $remarks,
        '{DELAY}' => $delay,
        '{REMOTE_ADDR}' => $ip_addr,
        '{AREA}' => strval($theme->area_id),
        '{NODE_FULL_NAME}' => sprintf('%d %s (%s)',$theme->node_id,
                                      $theme->node_record['link_text'],
                                      $theme->node_record['title'])
        );
    // 2A -- system alert
    $msg = t('alert_new_article','m_newsletter',$params);
    $theme->queue_alert($msg,$full_name);
    // 2B -- tell newsletter administrators about this new article via mail
    include_once($CFG->progdir.'/lib/email.class.php');
    $mailer = new Email;
    $mailer->set_mailreplyto($config['newsletter_address'],$config['title']);
    $mailer->set_mailfrom($config['newsletter_address'],$config['title']);
    $mailer->set_message(t('admin_mail_body_contrib','m_newsletter',$params));
    $mailer->set_subject(t('admin_mail_subject_contrib','m_newsletter',$params));;
    $admins = explode("\n",trim($config['email_administrators']));
    foreach($admins as $admin) {
        $mailer->set_mailto($admin,t('newsletter_administrator','m_newsletter'));
        $mailer->send(); // best effort; no error checking
    }
    return $retval;
} // newsletter_view_contribution_send()


/** thank the visitor for the article and show a text copy too
 *
 * Almost the same as {@see newsletter_view_contribution_preview()}.
 *
 * @param object &$theme collects the (html) output
 * @param array mailpage configuration data in a (nested) array
 * @param array $dialogdef array that defines the input fields
 * @param string $ip_addr the originating IP-address
 * @return void output writted to $theme
 * @todo should we have an OK button at all???
 */
function newsletter_view_contribution_thankyou(&$theme, $config, $dialogdef, $ip_addr) {

    // 1 -- prepare the information to show
    $forbidden = array(chr(10),chr(13),chr(34),'\\');
    $author = sprintf('&quot;%s&quot; &lt;%s&gt;',
                      htmlspecialchars(str_replace($forbidden,'',trim($dialogdef['full_name']['value']))),
                      htmlspecialchars(str_replace($forbidden,'',trim($dialogdef['email']['value']))));
    $remarks = nl2br(htmlspecialchars(strip_tags(trim($dialogdef['remarks']['value']))));
    $title = htmlspecialchars(strip_tags(trim($dialogdef['title']['value'])));
    $content = nl2br(htmlspecialchars(strip_tags(trim($dialogdef['content']['value']))));
    $remote_addr = htmlspecialchars($ip_addr);

    // 2 -- actually output the preview
    $theme->add_content(html_tag('h2','',t('thankyou_header','m_newsletter')));
    $theme->add_content(html_tag('p','',t('contrib_here_is_a_copy','m_newsletter')));
    $theme->add_content(html_tag('div',array('class'=>'newsletter_article_outer')));
    $theme->add_content(html_tag('h3',array('class'=>'newsletter_article_header'),$title));
    $theme->add_content(html_tag('div',array('class'=>'newsletter_article_author'),$author));
    $theme->add_content(html_tag('div',array('class'=>'newsletter_article_inner'),$content));
    $theme->add_content(html_tag('div',array('class'=>'newsletter_article_remarks'),$remarks));
    $theme->add_content(sprintf('<strong>%s</strong>: %s<br>',
                                t('contrib_ip_addr','m_newsletter'),$remote_addr));
    $theme->add_content(html_tag_close('div'));

    // 3 -- navigation so the user can move away from here
    newsletter_view_add_navbar($theme,$config);
    return TRUE;
} // newsletter_view_contribution_thankyou()


?>