<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/search/search_admin.php - management interface for search-module
 *
 * This file defines the administrative interface to this module.
 * The interface consists of the following four functions.
 *
 * <code>
 * search_disconnect(&$output,$area_id,$node_id,$module)
 * search_connect(&$output,$area_id,$node_id,$module)
 * search_show_edit(&$output,$area_id,$node_id,$module,$viewonly,$edit_again,$href,$option)
 * search_save(&$output,$area_id,$node_id,$module,$viewonly,&$edit_again,$option)
 * </code>
 *
 * These functions are called from pagemanagerlib.php whenever necessary.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_search
 * @version $Id: search_admin.php,v 1.1 2016/04/01 18:31:50 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }


/** disconnect this module from a node
 *
 * this breaks the link between the node $node_id in area $area_id and
 * this module. For now we simply delete the record in 'search'.
 * Note that the db_delete() here is a simple best effort: if it fails
 * we do nothing but return FALSE;
 * 
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node from which we need to disconnect
 * @param array $module the module record straight from the database
 * @return bool TRUE on success, FALSE otherwise
 */
function search_disconnect(&$output, $area_id, $node_id, $module) {
    $where = array('node_id' => intval($node_id));
    $retval = db_delete('search',$where);
    return ($retval === FALSE) ? FALSE : TRUE;
} // search_disconnect()


/** connect this module to a node
 *
 * this makes the link between the node $node_id in area $area_id and this module.
 * In this case we create a single 'search' record linked to node_id.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which we need to connect
 * @param array $module the module record straight from the database
 * @return bool TRUE on success, FALSE otherwise
 */
function search_connect(&$output, $area_id, $node_id, $module) {
    global $USER;
    $now = strftime('%Y-%m-%d %T');
    $fields = array(
        'node_id'      => intval($node_id),
        'header'       => '',
        'introduction' => '',
        'scope'        => 0, /* 0=only this area, 1=all public areas, 2=all areas including intranets */
        'results'      => 20, /* maximum # of results per page */
        'ctime'        => $now,
        'cuser_id'     => $USER->user_id,
        'mtime'        => $now,
        'muser_id'     => $USER->user_id);
    $retval = db_insert_into('search',$fields);
    if ($retval !== 1) {
        logger(sprintf("%s(): cannot connect to node '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        $retval = FALSE;
    } else {
        $retval = TRUE;
    }
    return $retval;
} // search_connect()


/** present the user with a dialog to modify the search that is connected to node $node_id
 *
 * this prepares a dialog for the user filled with existing data (if any), possibly allowing
 * the user to modify the content. If the flag $viewonly is TRUE, this routine should only
 * display the content rather than let the user edit it. If the flag $edit_again is TRUE,
 * the routine should use the data available in the $_POST array, otherwise it should read
 * the data from the database (or wherever the data comes from). The parameter $href is the
 * place where the form should be POST'ed.
 *
 * The dialog should be added to the $output object. Useful routines are:
 * <code>
 * $output->add_content($content): add $content to the content area
 * $output->add_message($message): add $message to the message area (feedback to the user)
 * $output->add_popup_bottom($message): popup $message in browser after loading the page (uses javascript)
 * $output->add_popup_top($message): popup $message in browser before loading the page (uses javascript)
 * </code>
 *
 * The parameter $option is not used in this module.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing is not allowed (but simply showing the content is allowed)
 * @param bool $edit_again if TRUE start with data from $_POST, else use data from database
 * @param string $href the action property of the HTML-form, the place where data will be POST'ed
 * @param string $option indicates which submenu-option was selected (NULL for none)
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function search_show_edit(&$output,$area_id,$node_id,$module,$viewonly,$edit_again,$href,$option) {
    $dialogdef = search_get_dialogdef($viewonly);
    if ($edit_again) {
        // retrieve and validate the POSTed values
        dialog_validate($dialogdef); // no need to show messages; did that already in search_save() below
    } else {
        // make a fresh start with data from the database
        $where = array('node_id' => intval($node_id));
        if (($record = db_select_single_record('search','*',$where)) === FALSE) {
            logger(sprintf("%s(): error retrieving search configuration '%d': %s",
                           __FUNCTION__,$node_id,db_errormessage()));
        } else {
            foreach($dialogdef as $name => $item) {
                switch($name) {
                    case 'header':
                    case 'introduction':
                    case 'scope':
                    case 'results':
                        $dialogdef[$name]['value'] = strval($record[$name]);
                        break;
                }
            }
        }
    }
    $output->add_content('<h2>'.t('search_configuration_header','m_search').'</h2>');
    $output->add_content(t('search_configuration_explanation','m_search'));
    $output->add_content(dialog_quickform($href,$dialogdef));
    return TRUE;
} // search_show_edit()


/** save the modified content data of this module linked to node $node_id
 *
 * this validates and saves the data that was submitted by the user.
 * If validation fails, or storing the data doesn't work, the flag $edit_again
 * is set to TRUE and the return value is FALSE.
 *
 * If the user has cancelled the operation, the flag $edit_again is set to FALSE
 * and the return value is also FALSE.
 *
 * If the modified data is stored successfully, the return value is TRUE and
 * the value of $edit_again is based on the button used to save ([Save] or [Done]).
 *
 * Here is a summary of return values.
 *
 * <pre>
 * $retval | $edit_again | Action
 * ------- | ----------- | ------
 * FALSE   | FALSE       | [Cancel] Stop editing, unlock & return to tree view
 * FALSE   | TRUE        | [Save] or [Done] failed: redo the dialog and let user correct mistakes
 * TRUE    | FALSE       | [Done] pressed, everything saved succesfully, we're done
 * TRUE    | TRUE        | [Save] pressed, everything saved succesfully but continue editing
 * </pre>
 *
 * The parameter $option is not used in this module.
 *
 * @param object &$output collects the html output (if any)
 * @param int $area_id the area in which $node_id resides
 * @param int $node_id the node to which the content is connected
 * @param array $module the module record straight from the database
 * @param bool $viewonly if TRUE, editing and hence saving is not allowed
 * @param bool &$edit_again set to TRUE if we need to edit the content again, FALSE otherwise
 * @param string $option indicates which submenu-option was selected (NULL for none)
 * @return bool TRUE on success + output stored via $output, FALSE otherwise
 */
function search_save(&$output,$area_id,$node_id,$module,$viewonly,&$edit_again) {
    global $USER;

    // 1 -- bail out if cancelled or viewonly
    if ((isset($_POST['button_cancel'])) || ($viewonly)) {
        $edit_again = FALSE;
        return FALSE;
    }

    // 2 -- redo if invalid data was submitted
    $dialogdef = search_get_dialogdef($viewonly);
    if (!dialog_validate($dialogdef)) {
        // there were errors, show them to the user and do it again
        foreach($dialogdef as $k => $item) {
            if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                $output->add_message($item['error_messages']);
            }
        }
        $edit_again = TRUE;
        return FALSE;
    }

    // 3 -- actually save the new setting
    $now = strftime('%Y-%m-%d %T');
    $table = 'search';
    $fields = array(
        'header'       => $dialogdef['header']['value'],
        'introduction' => $dialogdef['introduction']['value'],
        'scope'        => intval($dialogdef['scope']['value']),
        'results'      => intval($dialogdef['results']['value']),
        'mtime'        => $now,
        'muser_id'     => $USER->user_id);
    $where = array('node_id' => intval($node_id));
    if (db_update($table,$fields,$where) === FALSE) {
        logger(sprintf("%s(): error saving config value '%d': %s",
                       __FUNCTION__,$node_id,db_errormessage()));
        $edit_again = TRUE;
        return FALSE;
    } else {
        $edit_again =  (isset($_POST['button_done'])) ? FALSE : TRUE;
        return TRUE;
    }
} // search_save()


/** construct a dialog definition for the search 'scope' and other parameters
 *
 * @param int $viewonly if TRUE the Save button is not displayed and values cannot be changed by the user
 * @return array dialog definition
 */
function search_get_dialogdef($viewonly) {
    $options = array(
        '0' => array('option' => t('scope_small_label', 'm_search'),
                     'title'  => t('scope_small_title', 'm_search')),
        '1' => array('option' => t('scope_medium_label','m_search'),
                     'title'  => t('scope_medium_title','m_search')),
        '2' => array('option' => t('scope_large_label', 'm_search'),
                     'title'  => t('scope_large_title', 'm_search')));
    $dialogdef = array(
        'header' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'header',
            'minlength' => 0,
            'maxlength' => 240,
            'columns' => 30,
            'label' => t('header_label','m_search'),
            'title' => t('header_title','m_search'),
            'value' => '',
            ),
        'introduction' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'introduction',
            'minlength' => 0,
            'maxlength' => 32768, // arbitrary; 32 kB
            'columns' => 50,
            'rows' => 10,
            'label' => t('introduction_label','m_search'),
            'title' => t('introduction_title','m_search'),
            'value' => '',
            ),
        'scope' => array(
            'type' => F_RADIO,
            'name' => 'scope',
            'value' => 0,
            'options' => $options,
            'viewonly' => $viewonly,
            'title' => t('scope_title','m_search'),
            'label' => t('scope_label','m_search')
            ),
        'results' => array(
            'type' => F_INTEGER,
            'name' => 'results',
            'columns' => 10,
            'maxlength' => 10,
            'minvalue' => 1,
            'maxvalue' => 100, // arbitrary but seems sane limit 
            'viewonly' => $viewonly,
            'title' => t('results_title','m_search'),
            'label' => t('results_label','m_search')
            ),
        'csrftoken' => dialog_csrftoken()
        );
    if (!$viewonly) {
        $dialogdef['button_save'] = dialog_buttondef(BUTTON_SAVE);
        $dialogdef['button_done'] = dialog_buttondef(BUTTON_DONE);
    }
    $dialogdef['button_cancel'] = dialog_buttondef(BUTTON_CANCEL);
    return $dialogdef;
} // search_get_dialogdef()

?>