<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

/** /program/modules/search/search_view.php - interface to the view-part of the search module
 *
 * This file defines the interface with the search-module for viewing content.
 * The interface consists of this function:
 *
 * <code>
 * search_view(&$output,$area_id,$node_id,$module)
 * </code>
 *
 * This function is called from /index.php when the node to display is connected
 * to this module.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_search
 * @version $Id: search_view.php,v 1.6 2016/04/07 12:06:27 peter Exp $
 */
if (!defined('WASENTRY')) { die('no entry'); }


/** display the content of the search linked to node $node_id
 *
 * @param object &$theme collects the (html) output
 * @param int $area_id identifies the area where $node_id lives
 * @param int $node_id the node to which this module is connected
 * @param array $module the module record straight from the database
 * @return bool TRUE on success + output via $theme, FALSE otherwise
 * @todo should we use the token routines here too?
 */
function search_view(&$theme,$area_id,$node_id,$module) {
    $theme->add_stylesheet(sprintf('program/modules/%s/%s.css',$module['name'],$module['name']));

    // 0 -- basic sanity checks
    if (($config = search_view_get_config($node_id,$area_id)) === FALSE) {
        $theme->add_message(t('error_retrieving_config','m_search'));
        return FALSE;
    }

    // 1 -- need to perform a search?
    $quest = FALSE; // assume no search at first
    $dialogdef = search_view_get_dialogdef();
    if (isset($_POST['button_ok'])) {
        // 1A -- need to perform a fresh quest?
        if (!search_view_dialog_validate($dialogdef)) { // invalid q => do not perform search
            foreach($dialogdef as $k => $item) {
                if ((isset($item['errors'])) && ($item['errors'] > 0)) {
                    $theme->add_message($item['error_messages']);
                }
            }
        } else {
            $q = $dialogdef['q']['value'];
            $r = $config['results'];
            $o = 0; 
            $t = '';
            $quest = TRUE; // initiate new search, starts at offset 0
        }
    } else {
        // 1B -- continue an earlier quest?
        $q = trim(get_parameter_string('q',''));
        if (strlen($q) > 0) {
            $r = max(1,min(100,get_parameter_int('r',intval($config['results']))));
            $o = max(0,min(65536,get_parameter_int('o',0))); // arbitrary: limits 0,...,64k
            $t = get_parameter_string('t','');
            $quest = TRUE;
            $dialogdef['q']['value'] = $q;
        } // else
            // invalid q => do not perform search
    }
    // 1C -- always show the question in dialog for easy changing by the user
    search_show_form($theme, $config, $dialogdef);

    // 2 -- start the quest
    if ($quest) {
        // what are we looking for anyway?
        $qwords = search_get_qwords($q);
        if (sizeof($qwords) <= 0) {
            $theme->add_content('<div class="search_more">'.t('no_results','m_search').'</div>');
        } else {
            search($theme,$config,$qwords,$q,$r,$o,$t);
        }
    }
    return TRUE;
} // search_view()


/** retrieve all configuration data for this search
 *
 * this retrieves all configuration data for this search.
 * Here is a quick reminder of the structure in $config.
 * <pre>
 * 'node_id'
 * 'header'
 * 'introduction'
 * 'scope'
 * 'results'
 * </pre>
 *
 * As a bonus we also add the current area $area_id to the resulting array.
 *
 * @param int $node_id identifies the page
 * @param int $area_id identifies the current area
 * @return bool|array configuration data in an array or FALSE on error
 */
function search_view_get_config($node_id,$area_id) {
    // 1 -- generic configuration
    $table = 'search';
    $fields = array('node_id','header','introduction', 'scope', 'results');
    $where = array('node_id' => intval($node_id));
    if (($config = db_select_single_record($table, $fields, $where)) === FALSE) {
        logger(sprintf('%s(): error retrieving configuration: %s',__FUNCTION__,db_errormessage()));
        return FALSE;
    }
    // 2 -- add a bonus
    $config['area_id'] = intval($area_id);
    return $config;
} // search_view_get_config()


/** construct a dialog definition for the visitor's search form
 *
 * this defines the searchform, basically a single text input
 * and a [Search] button.
 *
 * @return array datadefinition
 */
function search_view_get_dialogdef() {
    $dialogdef = array(
        'q' => array(
            'type' => F_ALPHANUMERIC,
            'name' => 'q',
            'minlength' => 1,
            'maxlength' => 255,
            'columns' => 44,
            'label' => t('query_label','m_search'),
            'title' => t('query_title','m_search'),
            'value' => '',
            ),
        'button_ok' => dialog_buttondef(BUTTON_OK)
    );
    return $dialogdef;
} // search_view_get_dialogdef()


/** validate the data entered by the visitor
 *
 * @param object &$dialogdef defines the dialog and will hold POSTed values
 * @return bool TRUE if valid, else FALSE + messages added to dialogdef
 */
function search_view_dialog_validate(&$dialogdef) {
    $retval = TRUE; // assume success
    // 1 -- check the usual suspects (side-effect: fetch POSTed values in dialogdef)
    if (!dialog_validate($dialogdef)) {
        $retval = FALSE;
    }
    // 2 -- the query must not be empty (after stripping tags and trim()'ing)
    $q = trim(strip_tags($dialogdef['q']['value']));
    $dialogdef['q']['value'] = $q;
    if ($retval) {
        $minlength = intval($dialogdef['q']['minlength']);
        if (strlen($q) < $minlength) {
            $retval = FALSE;
            ++$dialogdef['q']['errors'];
            $dialogdef['q']['error_messages'][] = t('validate_too_short','',
                array('{FIELD}' => str_replace('~','',$dialogdef['q']['label']),
                      '{MIN}'=>strval($minlength)));
        }
    }
    return $retval;
} // search_view_dialog_validate()


/** display the search form
 *
 *
 * @param object &$theme collects the (html) output
 * @param array $config search configuration data
 * @param array $dialogdef array that defines the input fields
 * @return void output written to $theme
 */
function search_show_form(&$theme, $config, $dialogdef) {
    //
    // 1 -- maybe output a header and an introduction
    //
    $header = trim($config['header']);
    if (!empty($header)) {
        $theme->add_content(html_tag('h2','',$header));
    }
    $introduction = trim($config['introduction']);
    if (!empty($introduction)) {
        $theme->add_content($introduction);
    }
    $href = was_node_url($theme->node_record);
    $theme->add_content('<div class="search_form">');
    $theme->add_content(dialog_quickform($href,$dialogdef));
    $theme->add_content('</div>');
    return;
} // search_show_form()


/** perform an actual search
 *
 * At this point we have collected the following parameters:
 *
 *  - string $q: the question, not empty
 *  - int $r: # of results per page
 *  - int $o: offset relative to the very first result
 *  - string $t: a comma-delimited string hinting at hits per module OR empty
 *
 * What needs to be done is to query the relevant search function from
 * one or more modules and present them to the visitor via $theme.
 *
 * Note: array $qwords contains 1 or more arrays with the original keyword
 * (in $qwords[i][0]), the lowercase version (in $qwords[i][1]) and also
 * the quoted version ready for 'LIKE' in a where-clause (in $qwords[i][2]).
 * This may come in handy when preparing the search context lateron.
 *
 * The original query is also available, in $q. This is used to construct
 * Previous and Next links below the search results overview.
 *
 * The user requests $r results. However, we try to find $r+1 results.
 * If that succeeds, we know that after the current $r results there will be
 * at least 1 more result (on the next page). Also this further fills the
 * cache with hints (in parameter $t) for the next iteration.
 * 
 * @param object &$theme collects the (html) output
 * @param array $config search configuration data
 * @param array $qwords list of 1 or more keywords to search for (original AND lowercase each)
 * @param string $q original unprocessed query
 * @param int $r results per page
 * @param int $o offset
 * @param string $t comma separated list with hints about # of hits in various modules
 * @return output written to $theme
 */
function search(&$theme,$config,$qwords,$q,$r,$o,$t) {
    // 1A -- prepare a collection of nodes to search in well known table 'search_nodes'...
    if (($nodes = search_construct_nodeslist($config)) === FALSE) {
        $theme->add_message(t('error_retrieving_nodes','m_search'));
        return;
    }
    // 1B -- ...or bail out if there are no nodes to search
    if ($nodes <= 0) {
        $theme->add_content('<div class="search_more">'.t('no_results','m_search').'</div>');
        return;
    }

    // 2 -- list of modules to search, bail out on error
    if (($places = search_get_places($t)) === FALSE) {
        $theme->add_message(t('error_retrieving_places','m_search'));
        return;
    }

    // 3A -- actually step through the modules and retrieve at most $r+1 results...
    $hits = array();
    $count = 0;
    $results = 0;
    $totals = array();
    foreach($places as $k => $place) {
        $count += max(0,$results);
        $results = $places[$k]['results']; // -1 means unknown, >= 0 is # of hits we counted before
        if (($results != 0) && (sizeof($hits) <= $r)){
            // there is an unknown (results  == -1) or a known non-0 # of results AND we are still 
            // a few hits short of $r; ask the relevant module for answers.
            $off = max(0,$o - $count); // start at 0 if we already had hits in the previous module
            $lim = max(0,$r-sizeof($hits)+1); // ask enough results to fill up this page, plus 1 extra
            switch ($k) {
            case 'search_areas': search_areas($hits,$results,$qwords, $lim, $off, $config); break;
            case 'search_nodes': search_nodes($hits,$results,$qwords, $lim, $off); break;
            default:             search_module($hits,$results,$qwords, $lim, $off, $place); break;
            }
            $places[$k]['results'] = $results;
        } // else 
            // skip loading & calling this module again because $results is 0 anyway (or so we believe)
        $totals[$k] = ($results < 0) ? '' : strval($results);
    }
    $count += max(0,$results); // total number of hits we found sofar (could be the grand total)

    // 3B -- ...or bail out after all if there are no results
    if ((sizeof($hits) <= 0) && ($o <= 0)) {
        $theme->add_content('<div class="search_more">'.t('no_results','m_search').'</div>');
        return;
    }

    // 4 -- report results, perhaps starting with a navigation bar
    $params = array('node'=>$config['node_id'],'q'=>$q,'t'=>implode(',',$totals));
    $navbar = search_navbar($count,$r,$o,$params);
    $theme->add_content($navbar);
    
    // 4A -- list the first $r hits we have collected (this leaves 1 spare if there are more hits)
    $index = $o;
    $oddeven = "search_even";
    foreach($hits as $hit) {
        if (++$index > $o+$r) {
            break;
        }
        $oddeven = ($oddeven == "search_odd") ? "search_even" : "search_odd";
        $url = $hit['url'];
        $theme->add_content(sprintf('<div class="search_result %s">',$oddeven));
        $theme->add_content(sprintf('  <h4 class="search_result_title">%d. <a href="%s">%s</a></h4>',
                                    $index,$url,$hit['title']));
        if (sizeof($hit['context']) > 0) {
            $theme->add_content('  <div class="search_result_context">');
            foreach($hit['context'] as $context) {
                $theme->add_content('    '.$context);
            }
            $theme->add_content('  </div>');
        }
        $theme->add_content(sprintf('  <div class="search_result_url"><a href="%s">%s</a></div>',
                                    $url,htmlspecialchars($url)));
        $theme->add_content("</div>\n");
    }

    // 4B -- this may have been the last screen of results, show 'no more' message
    if (sizeof($hits) <= $r) {
        $theme->add_content(sprintf('<div class="search_more">%s</div>',t('no_more_results','m_search')));
    }

    // 4C -- finish with a navigation bar
    $theme->add_content($navbar);
    return;
} // search()


/** construct a navigation bar with 'previous' 1 2 3 4 ... 'next'
 *
 * construct a DIV with 'previous' 1 2 3 4 ... 'next' links to navigate the search results
 * If there is less than 1 page worth of results no navigation bar is necessary (an
 * empty array is returned). If we're on the first page, the 'previous' link is text
 * (not a link), if we are on the last page the 'next' link is text (not a link).
 *
 * The page numbers 1 2 3 4 ... form a sliding window within all pages; width
 * is determined by the global pagination setting (default 7).
 *
 * @param int $count total number of hits counted sofar (could be grand total)
 * @param int $r results per page
 * @param int $o offset
 * @param array $params basic parameters for the URLs to generate
 * @return array ready to use HTML lines
 */
function search_navbar($count,$r,$o,$params) {
    global $CFG;
    $navbar = array();
    if (($o < $r) && ($count <= $o + $r)) { // no navbar for a single page
        return $navbar;
    }
    // calculate pagination
    $pages_max = $CFG->pagination_width;
    $pages = intval(($count+$r-1)/$r);
    $page_cur = intval(($o+$r-1)/$r);
    if ($pages <= $pages_max) {
        $page_start = 0;
        $page_end =  $pages;
    } else {
        $page_start = min(max(0,$page_cur - intval($pages_max / 2)),$pages - $pages_max);
        $page_end = $page_start + $pages_max;
    }
    // construct navbar in a separate div
    $navbar[] = '<div class="search_navigation">';
    // Previous
    $attr = array('class'=>'search_previous');
    $anchor = t('previous','m_search');
    if ($o >= $r) {
        $params['o'] = strval(min($pages*$r,$o)-$r);
        $navbar[] = html_a($CFG->www_short.'/index.php',$params,$attr,$anchor);
    } else {
        $navbar[] = html_tag('span',$attr,$anchor);
    }
    // 1 2 3 4 ...
    for ($page=$page_start; $page<$page_end; ++$page) {
        $params['o'] = strval($page * $r);
        $anchor = sprintf('%d',$page+1);
        $attr = array('class'=>($page == $page_cur) ? 'search_page current' : 'search_page');
        $navbar[] = html_a($CFG->www_short.'/index.php',$params,$attr,$anchor);
    }
    // Next
    $attr = array('class'=>'search_next');
    $anchor = t('next','m_search');
    if ($o + $r < $count) {
        $params['o'] = strval($o+$r);
        $navbar[] = html_a($CFG->www_short.'/index.php',$params,$attr,$anchor);
    } else {
        $navbar[] = html_tag('span',$attr,$anchor);
    }
    $navbar[] = '</div>';
    return $navbar;
} // search_navbar()


/** kickstarter for calling modules' search function (after loading the source files)
 * 
 * The interface with the modules' search functions is fully documented
 * in {@link htmlpage_search()} in {@link htmlpage_search.php}.
 * 
 * @param array &$hits receives search results (at most $limit)
 * @param int &$results receives number of hits (negative value on entry means (re-)calculate total) 
 * @param array $qwords holds 1 or more keywords to search for (original, utf8lower, quoted)
 * @param int $limit maximum number of hits to return (could be 0)
 * @param int $offset starting point within this module for returning results
 * @param array $module holds module information (name, search_script, module_id, etc.)
 * @return bool|array FALSE on failure, TRUE on success and $hits, $results updated (maybe)
 */
function search_module(&$hits,&$results,$qwords,$limit,$offset,$module) {
    global $CFG;

    // 0 -- anything to do at all?
    if (($results == 0) || (($results > 0) && (($limit <= 0) || ($offset >= $results)))) {
        return TRUE;
    }

    // 1A -- load relevant source file
    $module_search_script = $CFG->progdir.'/modules/'.$module['name'].'/'.$module['search_script'];
    if (!file_exists($module_search_script)) {
        logger(__FUNCTION__."(): weird: file '$module_search_script' does not exist. Huh?");
        return FALSE;
    }
    require_once($module_search_script);

    // 1B -- check the search function's existence
    $module_search = $module['name'].'_search';
    if (!function_exists($module_search)) {
        logger(__FUNCTION__."(): weird: function '$module_search' does not exist. Huh?");
        return FALSE;
    }

    // 2 -- actually go search this module
    return $module_search($hits,$results,$qwords,$limit,$offset);
} // search_module()


/** construct an ordered list of places to search perhaps with hints about # of hits
 *
 * the list of places to search starts with areas and nodes: these are special cases
 * and these places are searched first. After that we give priority to modules htmlpage,
 * newsletter and althing. After that the order is chosen so that the 'dynamic' pages
 * (periodic newsletter, active weblog) are searched first. All current modules (April
 * 2016) are added to the list in a particular order. Additional modules will be added
 * automatically at the end of the list. This could differ between W@S installations,
 * but the order is the same between searches in the same installation of W@S.
 * We do set results to 0 for all modules; when a module of that name exists the
 * results will be set to -1 indicating 'unknonwn'. This way a non-existing module
 * will not yield errors.
 *
 * The parameter $hints contains the search results from a previous search. This
 * is basically a comma delimited list of integers where the first integer is the # of hits
 * for the areas search, the second one for the nodes search, the third one for the
 * search in htmlpage, etc. This saves trips to the database if the user moves forward in
 * the results list. We do have to trust the user submitted data though. OTOH: if the
 * numbers in $hints are too high, the search will simply fail and no harm is done.
 *
 * Note: 1-on-1 means: one module record is linked to a node, 1-on-N means that
 * more than 1 results could emanate from the same node (eg. different issues of the
 * newsletter, or different posts in the althing). NOP indicates that that module
 * always performs No OPeration: always 0 results and never any hits.
 *
 * @param string $hints comma delimited list of hits per module
 * @return bool|array FALSE on error, array with places to search otherwise
 */
function search_get_places($hints) {
    // 1 -- bail out if we cannot retrieve a list of modules
    if (($modules = get_module_records()) === FALSE) {
        logger(sprintf('%s(): cannot retrieve list of modules',__FUNCTION__));
        return FALSE;
    }
    // 2 -- start list of places with 'interactive' + 'dynamic' modules (most likely to have hits) first
    $places = array('search_areas' => array('name' => 'search_areas','results' => -1), // area titles
                    'search_nodes' => array('name' => 'search_nodes','results' => -1), // node titles
                    'htmlpage' => array('results' => 0),   // 1-on-1 likely to have hits
                    'newsletter' => array('results' => 0), // 1-on-N with dynamic content
                    'althing' => array('results' => 0),    // 1-on-N with dynamic content
                    'crew' => array('results' => 0),       // 1-on-1 with user content
                    'confab' => array('results' => 0),     // 1-on-1
                    'snapshots' => array('results' => 0),  // 1-on-1
                    'aggregator' => array('results' => 0), // 1-on-1
                    'mailpage' => array('results' => 0),   // 1-on-1
                    'search' => array('results' => 0),     // 1-on-1
                    'sitemap' => array('results' => 0),    // 1-on-1
                    'redirect' => array('results' => 0),   // NOP
                    'mypage' => array('results' => 0),     // NOP
                    );
    // 3 -- complete the list with all the other modules (if any), keeping initial order intact
    foreach ($modules as $module) {
        $name = $module['name'];
        $places[$name] = array('name' => $name,
                               'search_script' => $module['search_script'],
                               'function' => $name.'_search',
                               'results' => -1);
    }
    // 4 -- perhaps incorporate the available hints
    if (trim($hints) != '') {
        $totals = explode(',',$hints);
        $t = 0;
        foreach($places as $k => $place) {
            if ((!isset($totals[$t])) || (!is_numeric($totals[$t])) || (($v = intval($totals[$t])) < 0)) {
                break; // bail out on first 'unknown' hint, ie. non-existing, empty or <0 value
            }
            $places[$k]['results'] = $v;
            ++$t;
        }
    }
    return $places;
} // search_get_places()


/** construct a helper table with a list of node_id's that are to be searched
 *
 * prepare a temporary table 'search_nodes' with all the nodes that are to be
 * searched. Nodes that are NOT to be searched are those under embaro, expired
 * or those in areas where we are not allowed to look.
 *
 * This is quite a complex routine, mainly because MySQL won't let me use
 * the same temporary table twice in the same SQL statement. The work-around
 * is to have more than on temporary table to build a list of searchable nodes.
 *
 * The strategy is as follows.
 *
 *
 * 1. Determine the areas  selected to be searched based on $config
 *
 * The search module is configured to either
 * <pre>
 *  - search the current (single) area (config['scope'] = 0)
 *  - search all public areas (config['scope'] = 1)
 *  - search all areas where the current user has access (config['scope'] = 2)
 * </pre>
 *
 * Note that if the current user is not logged in, the latter two yield the same result.
 * The area selection is kept in $where which looks like
 * <pre>
 *  - ((a.is_active) AND (a.area_id=AA))
 *  - ((a.is_active) AND (a.is_private=FALSE))
 *  - ((a.is_active) AND ((a.is_private=FALSE) OR (a.area_id=A) OR (area_id=B) OR ...))
 * </pre>
 *
 *
 * 2. Select all embargoed/expired nodes in the selected areas (into new temp table '$accu')
 *
 * This gives a list of otherwise valid nodes that happen to be embargoed/expired
 * and therefore are not to be searched. If this list is empty we can proceed with
 * step 6 below. Note that the temp table '$accu' is always created, even if it is empty.
 *
 *
 * 3. Select all the childeren from the nodes selected in the previous step into new temp table '$work'
 *
 * The effect of embargo/expiry effects all descendants of the embargoed/expired nodes.
 * We have to work our way through all offspring of the embargoed/expired nodes to find all the
 * nodes that are not to be searched. If the list of this generation of descendants is empty
 * we can proceed with step 6 below.
 *
 *
 * 4. Add the generation in temp table 2 ('$work') to the family in temp table 1 ('$accu')
 * 
 * This is the workaround with two temp tables. The new generation in $work is added to the
 * list of ancestors in $accu and $work is cleared for the next round.
 *
 *
 * 5. Select the next generation in $work
 *
 * If there is a next generation it is stored in the existing table $work and we continue
 * with step 4 (again). If there are no new generations at this time, we fall through to
 * step 6.
 * 
 * This loop 4+5 eventually leaves all embargoed/expired nodes and all their descendants
 * in temp table 1 '$accu'. Table temp table '$work' is no longer required at that point.
 *
 *
 * 6. Invert the selection of nodes currently in temp table 1 '$accu' taking areas into account
 *
 * This selects all the node_id's in the selected areas (see step 1) that are NOT in temp table 1
 * '$accu' into temp table 3 ('$nodes').
 *
 *
 * 7. Return the number of records in temp_table3 '$nodes' as the # of searchable nodes.
 *
 * @param array $config search module configuration data
 * @return FALSE on error, otherwis # of nodes + temporary table 'search_nodes' created
 */
function search_construct_nodeslist($config) {
    global $USER, $DB;
    $prefix = $DB->prefix; // shorthand
    $accu  = $prefix.'search_temp1'; // temporary helper #1 is the accumulator
    $work  = $prefix.'search_temp2'; // temporary helper #2 is the worker
    $nodes = $prefix.'search_nodes'; // this is the 'well-known table' used in all module searches
    $now = strftime('%Y-%m-%d %T');  // calculate embargo/expiry against this time
    $area_id = $config['area_id'];   // current area (needed when scope == 0)

    // 1 -- which area(s) are we searching?
    switch (intval($config['scope'])) {
    case 0: $where = sprintf('a.area_id = %d', $area_id); break;
    case 1: $where = sprintf('a.is_private = %s',db_escape_and_quote(FALSE)); break;
    case 2:
        if (($areas = get_area_records()) === FALSE) {
            logger(sprintf('%s(): cannot retrieve list of areas',__FUNCTION__));
            return FALSE;
        }
        $where = sprintf('a.is_private = %s',db_escape_and_quote(FALSE));
        if ($USER->is_logged_in) {
            $first = TRUE;
            foreach($areas as $id => $area) {
                if ((db_bool_is(FALSE,$area['is_private'])) ||
                    (db_bool_is(FALSE,$area['is_active'])) ||
                    (!$USER->has_intranet_permissions(ACL_ROLE_INTRANET_ACCESS,$id))) {
                    continue;
                }
                if ($first) {
                    $where = '('.$where.')';
                    $first = FALSE;
                }
                $where .= sprintf(' OR (a.area_id = %d)',$id);
            }
        }
        break;
    default:
        logger(sprintf('%(): unknown scope %d; bailing out',__FUNCTION__,$config['scope']));return FALSE;
    }
    $where = '(a.is_active) AND ('.$where.')';

    // 2 -- CREATE new helper $accu with 'real' embargoed/expired nodes 
    $generation = 0;
    $sql = sprintf('CREATE TEMPORARY TABLE %s ('.
                   'generation integer DEFAULT 1, '.
                   'PRIMARY KEY (node_id)) '.
                   'SELECT n.node_id,n.parent_id '.
                   'FROM %snodes n INNER JOIN %sareas a USING(area_id) '.
                   'WHERE ((n.expiry < \'%s\' ) OR (\'%s\' < n.embargo)) AND (%s)',
                   $accu,$prefix,$prefix,$now,$now,$where);
    if (($num_rows = $DB->exec($sql)) === FALSE) {
        logger(sprintf('%s(): cannot get embargoed/expired nodes in accu: %s',
                       __FUNCTION__,db_errormessage()));
        return FALSE;
    }

    // 3 -- CREATE new helper $work with next generation embargoed/expired nodes
    if ($num_rows > 0) {
        $sql = sprintf('CREATE TEMPORARY TABLE %s ('.
                       'generation integer DEFAULT 2, '.
                       'PRIMARY KEY (node_id)) '.
                       'SELECT n.node_id, n.parent_id '.
                       'FROM %snodes n INNER JOIN %s p ON n.parent_id = p.node_id '.
                       'WHERE ((p.generation = %d) AND (n.node_id <> n.parent_id))',
                       $work,$prefix,$accu,++$generation);
        if (($num_rows = $DB->exec($sql)) === FALSE) {
            logger(sprintf('%s(): cannot get embargoed/expired nodes in work: %s',
                           __FUNCTION__,db_errormessage()));
            return FALSE;
        }
    }
    while (($num_rows > 0) && ($generation < MAXIMUM_ITERATIONS)) {
        // 4 -- copy results from existing $work to existing $accu (workaround database limitation)...
        $sql = sprintf('INSERT INTO %s SELECT * FROM %s',$accu,$work);
        if ($DB->exec($sql) === FALSE) {
            logger(sprintf('%s(): cannot copy embargoed/expired childeren from work to accu: %s',
                           __FUNCTION__,db_errormessage()));
            return FALSE;
        }
        // ...and cleanup helper table $work
        $sql = sprintf('DELETE FROM %s',$work);
        if ($DB->exec($sql) === FALSE) {
            logger(sprintf('%s(): cannot delete from work: %s',__FUNCTION__,db_errormessage()));
            return FALSE;
        }

        // 5 -- fill existing helper $work with next generation
        $sql = sprintf('INSERT INTO %s '.
                       'SELECT p.generation+1 AS generation, n.node_id,n.parent_id '.
                       'FROM %s p INNER JOIN %snodes n ON n.parent_id = p.node_id '.
                       'WHERE p.generation = %d',
                       $work,$accu,$prefix,++$generation);
        if (($num_rows = $DB->exec($sql)) === FALSE) {
            logger(sprintf('%s(): cannot get embargoed/expired descendants in work: %s',
                           __FUNCTION__,db_errormessage()));
            return FALSE;
        }
    }
    // 6 -- invert the selection w/areas too
    $sql = sprintf('CREATE TEMPORARY TABLE %s ('.
                   'PRIMARY KEY (node_id)) '.
                   'SELECT n.node_id '.
                   'FROM %sareas a INNER JOIN %snodes n USING (area_id) '.
                   'LEFT JOIN %s e USING (node_id) '.
                   'WHERE (e.node_id IS NULL) AND (%s)',
                   $nodes,$prefix,$prefix,$accu,$where);
    if (($num_rows = $DB->exec($sql)) === FALSE) {
        logger(sprintf('%s(): cannot get final set of nodes into search_nodes: %s',
                       __FUNCTION__,db_errormessage()));
        return FALSE;
    }

    // 7 -- tell caller about # of nodes to search
    return $num_rows;
} // search_construct_nodelist()


/** construct an array with keywords to look for
 *
 * construct an array with 0, 1 or more keywords to look for in the database
 * we prepare different versions: the original casing, the lowercase variant
 * and a version prepared for the LIKE-operator.
 * The second comes in handy when we are highlighting the context in
 * the results list. The third one is handy when performing database queries.
 *
 * Note: we KISS and do not take quoted strings into account (for now).
 * Note that earlier we already stripped HTML-tags from $q.
 * (see {@link search_view_dialog_validate()}).
 *
 * @param string $q holds a list of words to search for
 * @return array
 * @todo better parse of keywords taking quotes into account too.
 */
function search_get_qwords($q) {
    global $DB;
    $words = explode(' ',$q);
    $qwords = array();
    foreach($words as $word) {
        if (($word = trim($word)) != '') {
            $qwords[] = array($word,
                              utf8_strtolower($word),
                              str_replace(array('%','_'),array('\\%','\\_'),$DB->escape($word)));
        }
    }
    return $qwords;
} // search_get_qwords()


/** search through area titles
 *
 * @param array &$hits collects the requested search results if any
 * @param int &$results collects the # of hits in this module
 * @param array $qwords holds 1 or more keywords to search for (original, utf8lower, quoted)
 * @param int $limit maximum number of hits to return
 * @param int $offset starting point within this module for returning results
 * @param array $config search configuration data (contains scope and area_id)
 * @return FALSE on error, TRUE on succes + results and hits updated (maybe)
 */
function search_areas(&$hits,&$results,$qwords,$limit,$offset,$config) {
    global $USER;
    // 1 -- use cached results (no $hits[] requested)
    if (($results == 0) || (($results > 0) && ($limit <= 0))) {
        return TRUE;
    }

    if (($areas = get_area_records()) === FALSE) {
        logger(sprintf('%s(): cannot retrieve list of areas',__FUNCTION__));
        return FALSE;
    }
    if (($scope = intval($config['scope'])) == 0) {
        $area_id = $config['area_id'];
        $areas = array($area_id => $areas[$area_id]);
    }
    $counter = 0; 
    foreach($areas as $id => $area) {
        if (db_bool_is(FALSE,$area['is_active'])) {
            continue;
        }
        if ((db_bool_is(FALSE,$area['is_private'])) OR
            (($scope == 2) && ($USER->has_intranet_permissions(ACL_ROLE_INTRANET_ACCESS,$id)))) {
            $title = $area['title'];
            $lower = utf8_strtolower($title);
            $hit = TRUE; // assume success
            foreach($qwords as $words) {
                if (strpos($lower,$words[1]) === FALSE) {
                    $hit = FALSE;
                    break;
                }
            }
            if ($hit) {
                if (++$counter <= $offset) {
                    continue;
                }
                if ($counter <= $limit + $offset) {
                    $hits[] = array('title' => search_highlight($qwords,$title),
                                    'context' => search_context($qwords,$lower),
                                    'url' => was_node_url(NULL,array('area' => $id)));
                } elseif ($results > 0) {
                    break;
                }
            }
        }
    }
    if ($results < 0) {
        $results = $counter; 
    }
    return TRUE;
} // search_areas()


/** search through node titles and link_texts
 *
 * there are a few possibilities
 *<pre>
 * 1 ($results  < 0) && ($limit <= 0): calc $results, return
 * 2 ($results  < 0) && ($limit >  0): calc $results, retrieve $hits, return
 * 3 ($results == 0) && ($limit <= 0): return
 * 4 ($results == 0) && ($limit >  0): return
 * 5 ($results  > 0) && ($limit <= 0): return
 * 6 ($results  > 0) && ($limit >  0): retrieve selected $hits, return
 *</pre>
 *
 * Cases 3, 4 and 5 are easy: we simply trust the caller's cached $results and
 * we do not have to return any $hits.
 *
 * Case 1 can be done via counting the number of matches, but not retrieving them,
 * e.g. sql == SELECT count(node_id) AS n FROM $table WHERE field LIKE '%expression%';
 *
 * Case 2 is a little more difficult because we need to retrieve the records AND
 * calculate the total number of hits. The latter is the # of records in the result
 * set and $limit could be (substantially) less than that.
 *
 * Case 6 is a matter of selecting a $limit hits, starting at offset. No
 * calculations necessary and thus relatively easy.
 *
 * In order to keep this code simple we combine cases 1 and 2 (partly)
 * and cases 6 and 2. In other words:
 * if necessary we calculate the # of records (case 1, case 2)
 * if necessary we snatch $limit records at $offset (case 2, case 6)
 * This is two trips to the database in case 2, but only on the initial search.
 * (the subsquent runs use the cached # of hits via the $hints)
 *
 * @param array &$hits collects the requested search results if any
 * @param int &$results collects the # of hits in this module
 * @param array $qwords holds 1 or more keywords to search for (original, utf8lower, quoted)
 * @param int $limit maximum number of hits to return
 * @param int $offset starting point within this module for returning results
 * @param array $config search configuration data (contains scope and area_id)
 * @return FALSE on error, TRUE on succes + results and hits updated (maybe)
 */
function search_nodes(&$hits,&$results,$qwords,$limit,$offset) {
    global $DB;

    // handle cases 3, 4 and 5
    if (($results == 0) || (($results > 0) && ($limit <= 0))) {
        return TRUE;
    }

    // 0 -- the where clause is the same for all cases
    $where = search_where($qwords,array('title','link_text'));

    // 1 -- handle case 1 and 2a: calculate the # of hits, no retrieving
    if ($results < 0) {
        $sql = sprintf("SELECT COUNT(n.node_id) AS results ".
                       "FROM %ssearch_nodes s ".
                       "INNER JOIN %snodes n USING (node_id) ".
                       "WHERE %s",$DB->prefix,$DB->prefix,$where);
        if ((($DBResult = $DB->query($sql,1)) === FALSE) || ($DBResult->num_rows != 1)) {
            logger(sprintf('%s(): cannot retrieve record count: %s',__FUNCTION__,db_errormessage()));
            $results = 0;
            return FALSE;
        }
        $record = $DBResult->fetch_row_assoc();
        $results = intval($record['results']);
    }

    // 2 -- handle case 6 and 2b - retrieve $limit records at $offset
    if ($limit > 0) {
        $sql = sprintf("SELECT n.node_id, n.title, n.link_text ".
                       "FROM  %ssearch_nodes s INNER JOIN %snodes n USING (node_id) ".
                       "WHERE %s",$DB->prefix,$DB->prefix,$where);
        if (($DBResult = $DB->query($sql,$limit,$offset)) === FALSE) {
            logger(sprintf('%s(): cannot retrieve selected records: %s',__FUNCTION__,db_errormessage()));
            return FALSE;
        }
        while (($record = $DBResult->fetch_row_assoc()) !== FALSE) {
            $hits[] = array('title' => search_highlight($qwords,$record['title']),
                            'context' => search_context($qwords,utf8_strtolower(
                                                        $record['title'].' ('.$record['link_text'].')')),
                            'url' => was_node_url(NULL,array('node' => $record['node_id'])));
        }
    }
    $DBResult->close();
    return TRUE;
} // search_nodes()


/** construct a where-clause to look for keywords in 1 or more fields
 *
 * Possible outcomes for fields f,f1,f2 and qwords q,q1,q2:
 *
 *  - (f LIKE %q%)
 *  - ((f LIKE %q1%) AND (f LIKE %q2%))
 *  - ((f1 LIKE %q%) OR (f2 LIKE %q%))
 *  - (((f1 LIKE %q1%) AND (f1 LIKE %q2%)) OR ((f2 LIKE %q1%) AND (f2 LIKE %q2%)))
 *
 * @param array $qwords holds 1 or more keywords to search for (original, utf8lower, quoted)
 * @param array $fields holds 1 or more fieldnames
 * @return string ready to use where-clause (including quoting etc.)
 */
function search_where($qwords,$fields) {
    $f = '';
    foreach($fields as $field) {
        $f .= ($f=='') ? '' : ' OR ';
        $q = '';
        foreach($qwords as $words) {
            $q .= sprintf("%s(%s LIKE '%%%s%%')",($q=='')?'':' AND ',$field,$words[2]);
        }
        $f .= (sizeof($qwords)>1) ? '('.$q.')' : $q;
    }
    $where = (sizeof($fields)>1) ? '('.$f.')' : $f;
    return  $where;
} // search_where()


/** return a ready-to-use context with highlights
 *
 * search $content for keywords in $qwords and return highlighted snippets (if any).
 *
 * Strategy is as follows:
 *
 * 1. prepare the $content: make lowercase, strip HTML, etc.
 *    if the string is now shorter than the $window we're done:
 *    return the highlighted string.
 *
 * 2. iterate through $words and the string, rember the offset and length
 *    of any matches in array hits[]. If there are no hits we're done:
 *    return an empty array.
 *
 * 3. step through $hits and attempt to combine $hits within a $window
 *    into a single run of hits in $runs[] (storing starting and ending
 *    offsets).
 *
 * 4. step through $runs and extract the specified substrings from the string,
 *    adding about half a window context before and about half a window context
 *    after the run. Store the highlighted strings in an array and return that
 *    array.
 *
 * @param array $qwords holds 1 or more keywords to search for (original, utf8lower, quoted)
 * @param string $content original content to search
 * @param int $window arbitrary base for size of context snippets
 * @param int $maxruns arbitrary limit for number of runs to return
 * @param int $maxhits arbitrary limit for # of hits/qword
 * @return array with 0 or more highlighted lowercase context snippets
 */
function search_context($qwords, $content, $window=100, $maxruns=5, $maxhits=10) {
    $contexts = array(); // collects 0, 1 or more context snippets (maximum $maxruns)

    //
    // 1 -- prepare the string for actual processing:
    //      maybe insert a few more spaces at 'natural' places, assuming $content is HTML-code
    //      strip all HTML-tags and make the UTF8 string lowercase 
    $search =  array('/<br.*?>/i', '/<p.*?>/i', '/<div.*?>/i', '/&nbsp;/i');
    $replace = array(' <br>',' <p>',' <div>',' ');
    $lower = utf8_strtolower(strip_tags(preg_replace($search,$replace,strtr($content,"\r\n\t\f","    "))));
    $length = strlen($lower); // counting bytes, not UTF8, for now (see also step 3 below)
    if ($length <= $window) {
        return array(search_highlight($qwords,$lower));
    }

    //
    // 2 -- collect at most $maxhits hits per keyword as $offset => $length pairs
    //      if two keywords start at the same offset (e.g. 'foo' and 'foobar')
    //      we record the longest $length as not to break/overlap the longer keyword
    //      note: we use the UTF8 lowercase variant of the keywords for matching
    $hits = array();
    foreach($qwords as $words) {
        $qword = $words[1];
        $i = $maxhits;
        $off = 0;
        $qlen = strlen($qword);
        while (($i-- > 0) && (($off = strpos($lower,$qword,$off)) !== FALSE)) {
            $hits[$off] = (isset($hits[$off])) ? max($hits[$off],$qlen) : $qlen;
            ++$off;
        }
    }
    if (sizeof($hits) <= 0) {
        return array();
    }

    //
    // 3 -- try to combine hits into 'runs' of hits
    //      if the distance between two (or more) hits is less than
    //      the window we remember the starting offset of the first
    //      hit within the run and the offset of the first character
    //      after the last hit within the run.
    ksort($hits);         // order by offsets ascending
    $left = -1 - $window; // headstart; the test always fails 1st time, good, starts the first run.
    $run = array();
    foreach($hits as $k => $v) {
        if ($left < $k - $window) {
            $left = $k;
        }
        $run[$left] = $k + $v + 1;
    }

    //
    // 4 -- extract the various runs from $lower including a valid context before and after
    //      no more than $maxruns snippets are returned
    $i = $maxruns;
    foreach($run as $k => $v) {
        if ($i-- <= 0) {
            break;
        }
        //
        // 4A - find a suitable starting point
        //      add about half a window of context, before the first hit in this run
        //      if we're too close to the start of the string, we start at the beginning
        //      of the string. If not, we fill $before with ellipses, indicating that
        //      there was more before the snippet.
        $left = $k - $window / 2;
        if ($left <= 0) {
            $before = '';
            $left = 0;
        } else {
            $before = '...';
            if ((($left = strpos($lower,' ',$left)) !== FALSE) && ($left < $k)) {
                ++$left; // eat the space
            } else {
                while ((ord($lower[$left]) & 0xC0) == 0x80) { // move to next multibyte UTF8-char boundary
                    ++$left;
                }
                // at this point $left points to the first valid byte of a multibyte character
                // or to a plain ascii character. We can safely use $left with substr()
            }
        }
        //
        // 4B - find a suitable ending point
        //      here too we attempt to add about half a window of context (if available)
        //      we also try to break the string at a space character if there is one
        $right = $v + $window / 2;
        if ($right >= $length) {
            $right = $length;
            $snippet = substr($lower,$left);
            $after = '';
        } else {
            while ((ord($lower[$right]) & 0xC0) == 0x80) { // move to prev multibyte UTF8-char boundary
                --$right;
            }
            // at this point $right points to the first valid byte of a multibyte character
            // or to a plain ascii character. We can safely use $right with substr()
            $snippet = substr($lower,$left,$right-$left);
            if ((($right = strrpos($snippet,' ')) !== FALSE) && ($right > strlen($snippet)-$window/2)) {
                $snippet = substr($snippet,0,$right);
            }
            $after = '...';
        }
        $contexts[] = $before.trim(search_highlight($qwords,$snippet)).$after;
    }

    //
    // 5 -- all done
    return $contexts;
} // search_context()


/** highlight words in a lowercase search context snippet
 *
 * highlight keywords in $snippet by sandwiching them between '<b>' and '</b>'
 * we first create a list of (unique) words ordered by the lenght and subsequently use
 * that list for matching. This list is based on the utf8lower variant of the keyword.
 *
 * @param array $qwords holds 1 or more keywords to search for (original, utf8lower, quoted)
 * @param string $snippet lowercase version of a context snippet
 * @return string with highlights inserted
 * @todo what to do with overlapping qwords, eg. "foo" and "foobar"? result depends on
 *       search/replace execution order: either "<b>foo</b>bar" or "<b><b>foo</b>bar</b>
 * @todo is there a chance that our static cache $q fails if somehow a different $qwords is used
 *       on subsequent calls? Mmmmm...
 */
function search_highlight($qwords,$snippet) {
    static $q = NULL;
    if (is_null($q)) {
        $q = array();
        foreach ($qwords as $words) {
            $q[$words[1]] = strlen($words[1]);
        }
        arsort($q);
    }
    foreach ($q as $k => $v) {
        $snippet = str_replace($k,'<b>'.$k.'</b>',$snippet);
    }
    return $snippet;
} // search_highlight()


/** perform a simple linear search in a module table
 *
 * link tables search_nodes s and nodes n to $table m, using $node as the link field.
 * retrieves hits in $hits, using either "m.$ttl" or "n.title" as title.
 * keys in $fields are plain fieldnames, values are prefixed with 'm.' preventing $DB ambiguities
 *
 * @param array &$hits receives search results (at most $limit)
 * @param int &$results receives number of hits (negative value on entry means (re-)calculate total) 
 * @param array $qwords holds 1 or more keywords to search for (original, utf8lower, quoted)
 * @param int $limit maximum number of hits to return (could be 0)
 * @param int $offset starting point within this module for returning results
 * @param string $table module table linked 1 : 1 to nodes 
 * @param array $fields data to retrieve, key="$fieldname", value="m.$fieldname"
 * @param string $where full where-clause without the word WHERE
 * @param string $order sort order
 * @param string $node name of field to link to nodes.node_id 1:1
 * @param string $ttl name of field to use as search result title
 * @return bool|array FALSE on failure, TRUE on success and $hits, $results updated (maybe)
 */
function search_simple(&$hits,&$results,$qwords,$limit,$offset,$table,$fields,$where,$order,$node,$ttl) {
    global $DB;

    // 1 -- calculate the # of hits, no retrieving of data
    if ($results < 0) {
        $results = 0;
        $sql = sprintf("SELECT COUNT(n.node_id) AS results ".
                       "FROM %ssearch_nodes s INNER JOIN %snodes n USING (node_id) ".
                       "INNER JOIN %s%s m ON (n.node_id = m.%s) ".
                       "WHERE %s",$DB->prefix,$DB->prefix,$DB->prefix,$table,$node,$where);
        if ((($DBResult = $DB->query($sql,1)) === FALSE) || ($DBResult->num_rows != 1)) {
            logger(sprintf('%s(): cannot count %s: %s',__FUNCTION__,$table,db_errormessage()));
            return FALSE;
        }
        $record = $DBResult->fetch_row_assoc();
        $DBResult->close();
        $results = intval($record['results']);
    }

    // 2 -- retrieve $limit records at $offset
    if (($limit > 0) && ($offset < $results)) {
        $sql = sprintf("SELECT n.node_id, n.title, %s ".
                       "FROM  %ssearch_nodes s INNER JOIN %snodes n USING (node_id) ".
                       "INNER JOIN %s%s m ON (n.node_id = m.%s) WHERE %s ORDER BY %s",
                       implode(',', $fields),$DB->prefix,$DB->prefix,
                       $DB->prefix,$table,$node,$where,$order);
        if (($DBResult = $DB->query($sql,$limit,$offset)) === FALSE) {
            logger(sprintf('%s(): cannot retrieve from %s: %s',__FUNCTION__,$table,db_errormessage()));
            return FALSE;
        }
        while (($record = $DBResult->fetch_row_assoc()) !== FALSE) {
            $title = trim($record[$ttl]);
            $content = '';
            foreach($fields as $k => $v) {
                $content .= ' '.$record[$k];
            }
            $url = was_node_url(NULL,array('node' => $record['node_id']));
            $hits[] = array('title' => (empty($title)) ? $record['node_title'] : $title,
                            'context' => search_context($qwords,$content),
                            'url' => $url);
        }
        $DBResult->close();
    }

    // 3 -- all done
    return TRUE;
} // search_simple()


?>