/* 
 * This file is part of Website@School, a Content Management System especially designed for schools.
 * Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation supplemented with the Additional Terms, as set
 * forth in the License Agreement for Website@School (see /program/license.html).
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
 * for more details.
 *
 * You should have received a copy of the License Agreement for Website@School
 * along with this program. If not, see http://websiteatschool.eu/license.html
 */

/** /program/modules/snapshots/slideshow.js - simple snapshots slideshow function
 *
 * This file contains the Javascript-code that implements a simple
 * slideshow function for the snapshots module. It should be included
 * at the top of the HTML-page (in the head-section). Initial configuration
 * is done via a few global variables, as follows.
 *
 * - var img[]
 *
 *   The initial list of snapshots is created dynamically from
 *   {@link snapshots_view.php} in the form of a 2D array.The n'th image
 *   (starting at 0) in that array is defined as follows:
 *
 *   img[n][0] = width of the image (in pixels)
 *   img[n][1] = height of the image (in pixels)
 *   img[n][2] = the url of the image file (src-attribute of the img tag)
 *   img[n][3] = the number of seconds to display this image
 *   img[n][4] = title to add to the display (document title)
 *
 *   The array with images is periodically updated using XMLHttpRequest.
 *
 * - var msg[]
 *
 *   Initially this array contains (English) messages. This array is
 *   overruled lateron from {@link snapshots_view.php} with the localised
 *   (translated) messages.
 *
 * - var url
 *
 *   This variable is used in conjunction with the XMLHttpRequest that
 *   periodically updates the list of images.
 *
 * - var timeout
 *
 *   This variable holds the number of seconds between periodical updates
 *   of the list of images.
 *
 * @author Peter Fokker <peter@berestijn.nl>
 * @copyright Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker
 * @license http://websiteatschool.eu/license.html GNU AGPLv3+Additional Terms
 * @package wasmod_snapshots
 * @version $Id: slideshow.js,v 1.5 2016/04/08 12:06:30 peter Exp $
 */
var img=new Array(); // contains image definitions (see above for syntax)
var msg=['loading...','no images']; // error/warning messages
var url; // location for retrieving updates
var timeout = 300;  // # of seconds between refreshes (default: 5m=300s)


/** create new window to show images and start the loop
 *
 * @param int n indicates the image to start with (0=first)
 * @return void
 */
function show_start(n) {
  if (img.length <= 0) {
    alert(msg[1]);
    return;
  }
  features = 'width='+screen.width+',height='+screen.height+
             ',location=0,menubar=0,scrollbars=0,status=0,titlebar=0,toolbar=0';
  hwnd = window.open('','',features);
  var html='<!DOCTYPE html>\n'+
           '<html>\n<head>\n<meta name="viewport" content="width=device-width, initial-scale=1.0">\n' +
           '<title></title>\n<style>\nbody { margin: 0; padding: 0; }\n'+
           'img { border: 0; max-width: 100%; height: auto; }\n</style>\n'+
           '</head>\n<body text="#FFFF00" bgcolor="#000000">\n'+
           '<div style="text-align:center;cursor:none;"><img id="image" width="1" height="1"></div>\n'+
           '</body>\n</html>';
  hwnd.document.open();
  hwnd.document.writeln(html);
  hwnd.document.close();
  hwnd.slides=makeSlideShow(hwnd,img,msg,url,timeout);
  hwnd.slides.show(n);
} // show_start()


/** construct self-contained slideshow
 *
 * This constructs an automatic slideshow that also
 * responds to these keyboard events:
 *
 * [space] => refresh list if images
 * [left]  => show first image
 * [up]    => show next image
 * [down]  => show previous image
 * [right] => stop the slideshow
 *
 * There is also a timer that initiates a refresh (just
 * like the spacebar) but that only refreshes the list of
 * image to display. It does not actively bumps to the
 * next image in the list, unlike the spacebar.
 * 
 * @param handle hwnd the window where the show takes place
 * @param array img initial collection of images
 * @param array msg (translated) messages
 * @param string url used for refreshing img[] via XMLHttpRequest
 * @param int timeout # of seconds before refreshing img[] again
 * @return object with a few useful functions
 */
function makeSlideShow(hwnd,img,msg,url,timeout) {
    var next=0; // next image to show on screen
    var cur=0; // current image on screen
    var timerNext; // keep track of timeout for showing next image
    var timerFresh;  // keep track of timeout for refresh
    var waitasec=0; // counter for tracking loading of images in queue
    var refreshTime=timeout; // # of seconds between refreshes (default: 5m)
    // 32=space,37=left,38=up,39=right,40=down
    var keymap={32:snapFresh, 37:snapFirst, 38:snapNext, 39:snapStop, 40:snapPrev};

    /** show the next image from the list and also start (pre)loading another one
     *
     * Note:
     * Code for determining the viewport inspired by an article by Andy Langton
     * http://andylangton.co.uk/articles/javascript/get-viewport-size-javascript/
     * and subsequent refinements by 'elstcb'.
     *
     * @return void
     */
    function snapNext() {

        // 0 - initialise - are we still here?
        if (!hwnd) return;
        hwnd.clearTimeout(timerNext);
        var d=hwnd.document;
        if (!d) return;

        // 1 - can we show something?
        var i=d.getElementById('image'); // visible image
        var q=d.getElementById('image'+next.toString()); // next image in queue

        // 1A - if no (hidden) image by that Id; create + append
        if (!q) {
            q=d.createElement('IMG');
            q.id='image'+next.toString();
            q.style.display='none';
            i.parentNode.appendChild(q);
        }

        // 1B - make sure we use the latest src of img[next]
        if (q.src!=img[next][2]) {
            q.src=img[next][2];
        }

        // 1C - wait at most 30 seconds for loading to complete
        if (!q.complete) {
            if (++waitasec < 30) {
                d.title=img[next][4]+' '+msg[0]+' ('+waitasec.toString()+')';
                timerNext=hwnd.setTimeout(snapNext,1000);
                return;
            }
            waitasec=0;
        }

        // 2A - determine viewport in (vx,vy) (see note above)
        var e=d.documentElement;
        var b=d.getElementsByTagName('body')[0];
        var vx=hwnd.innerWidth||e.clientWidth||b.clientWidth;
        var vy=hwnd.innerHeight||e.clientHeight||b.clientHeight;

        // 2B - scale the image (px,py) to maximum that fits within viewport
        var px=img[next][0];
        var py=img[next][1];
        var r=Math.min(px*vy,py*vx);
        i.width=Math.ceil(r/py);
        i.height=Math.ceil(r/px);

        // 2C - actually replace visible image
        i.src=img[next][2];
        i.alt=img[next][4];
        d.title=img[next][4];

        // 3 - bump to next (forward) but remember current
        cur=next++;
        next%=img.length;

        // 4 - attempt to preload the next image (maybe create new img along the way)
        q=d.getElementById('image'+next.toString());
        if (!q) {
            q=d.createElement('IMG');
            q.id='image'+next.toString();
            q.style.display='none';
            i.parentNode.appendChild(q);
        }
        if (q.src!=img[next][2]) {
            q.src=img[next][2];
        }
        // 5 - see ya later...
        timerNext=hwnd.setTimeout(snapNext,1000*img[cur][3]);
    }

    /** show image i */
    function snapShow(i) {
        next=i%img.length;
        snapNext();
    }

    /** show previous image */
    function snapPrev() {
        snapShow(cur+img.length-1);
    }

    /** show first image */
    function snapFirst() {
        snapShow(0);
    }
    /** request update from timer */
    function snapFreshTimer() {
        snapFreshWorker(0);
    }

    /** request update from keyboard */
    function snapFresh() {
        snapFreshWorker(1);
    }

    /** prepare a request for update of the images (including IE-support)
     *
     * request a fresh list of images from the server. This can be initiated
     * from the timer (flag=0) or from user interaction (flag=1). In case
     * of user interaction we want something to happen visually. If the
     * timer expired we silently update the list which will eventually be
     * showed.
     *
     * @param int flag 0=just update the list, 1=update list + show next too 
     */
    function snapFreshWorker(flag) {
        hwnd.clearTimeout(timerFresh);
        if (hwnd.XMLHttpRequest) {
            var r=new hwnd.XMLHttpRequest();
        } else {
            var v=['MSXML2.XmlHttp.6.0','MSXML2.XmlHttp.5.0','MSXML2.XmlHttp.4.0',
                   'MSXML2.XmlHttp.3.0','MSXML2.XmlHttp.2.0','Microsoft.XmlHttp'];
            for(var i=0; i<v.length; ++i) {
                try { var r=new ActiveXObject(v[i]); break; }
                catch(e) { }
            }
        }
        r.open("GET",url);
        r.onreadystatechange=function() {
            if ((r.readyState===4) && (r.status===200)) {
                refreshHandler(r.responseText,flag);
            }
        };
        r.send(null);
        // update again in refreshTime seconds...
        timerFresh=hwnd.setTimeout(snapFreshTimer,1000*refreshTime);
    }

    /** process incoming refresh and maybe update main images array
     *
     * data format: width | height | src | showtime | title
     *
     * If the list shrunk, we possibly adjust next, if the list
     * is longer there is no need. If the list would become empty,
     * we don't update the list but stick to the old one in place.
     * This old list will keep on running until at least 1 image
     * surfaces in a fresh list.
     *
     * @param array data new tab-delimited list of images
     * @param int flag 0=just update the list, 1=update list + show next too 
     */
    function refreshHandler(data,flag) {
        var lines=data.split('\n');
        var n=lines.length;
        var i=new Array;
        for (var l=0; l<n; ++l) {
            var a=lines[l].split('\t');
            if (a.length != 5) continue; // ignore malformed lines
            a[0]=parseInt(a[0],10);
            a[1]=parseInt(a[1],10);
            a[3]=parseInt(a[3],10);
            if ((a[0]<1) || (a[1]<1) || (a[3]<0)) continue;
            i.push(a);
        }
        if (i.length>0) { // at least 1 image in the new list
            img=i; // put new list in place
            if (next >= i.length) { // wrap next if next points outside the new list 
                next = 0;
            }
            if (flag) {
                snapNext(); // actually respond to the keypress
            }
        }
    }

    /** handle keydown events for keys selected via keymap
     *
     * @param object event
     */
    function keyHandler(event) {
        var e=event||hwnd.event;
        var c=e.keyCode||e.charCode; // first keyCode, otherwise FF complains about 'meaningless'.
        var h=keymap[c];
        if (h) {
            if (e.preventDefault) e.preventDefault();
            if (e.returnValue) e.returnValue = false;
            h();
            return false;
        } else {
            return true;
        }
    }

    /** close the window and return to the thumbs overview */
    function snapStop() {
        hwnd.close();
    }

    // Initialise 
    if (hwnd.document.addEventListener) {
        hwnd.document.addEventListener("keydown",keyHandler,false);
    } else if (hwnd.attachEvent) {
        hwnd.attachEvent("onkeydown",keyHandler);
    }
    timerFresh=hwnd.setTimeout(snapFreshTimer,1000*refreshTime);
    return {show:snapShow};
} // makeSlideShow()

// eof slideshow.js
